//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//   
// <CTL_AMD_STD>
// </CTL_AMD_STD>
// 
// <DOC_AMD_STD>
// DisplayCplExt.cpp : Implementation of CDisplayCplExt
// </DOC_AMD_STD>
// 
#include "stdafx.h"
#include "atlstr.h"
#include "DisplayPage.h"
#include "DisplayCplExt.h"
#include <cplext.h>
#include "escape.h"
#include "lxshared.h"
#include "stdio.h"
#include "uxtheme.h"

#define MB_INFO MB_OK | MB_ICONINFORMATION
#define MBOX(str) MessageBox(hwnd,str,str,MB_INFO)
#define MBOXN(str) MessageBox(NULL,str,str,MB_INFO)

BOOL CALLBACK NewPageDlgProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
BOOL CALLBACK GammaDlgProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
UINT CALLBACK NewPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp );
BOOL CALLBACK ReplacementPageDlgProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
UINT CALLBACK ReplacementPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp );
BOOL CALLBACK PromptDlgProc (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

void UpdateErrorString (HWND hwnd, int error);
void UpdateDialogFromStruct (HWND hwnd);
void UpdateOutputString (HWND hwnd, int id);
void UpdateOutputInfo (HWND hwnd, int rotId);
void SubClassWindow (HWND hwnd, int id, int index);
void SubClassGammaWindow (HWND hwnd, int id, int index);
void InitializeGammaDialog (HWND hwnd);
void UpdateGammaSliders (HWND hwnd);
void ProcessGammaSliders (HWND hwnd, GAMMA_SLIDER *gammaSlider);
void UpdateGamma(BOOL updateRegistry);
void RestoreGamma(void);
void UpdateSliderValues (HWND hwnd);

REGISTRY_DATA g_regData;
LX_REGISTRY_SETTINGS g_oldSettings;
DEVMODE g_curMode;
DEVMODE g_newMode;
int g_error = 0;
int changed = 0;
int gammaChanged = 0;
int promptThreadRunning = 0;
int promptThreadStop = 0;

char error_string[160];
LX_WND_PROC wndProcs[10];
LX_WND_PROC gammaWndProcs[9];
LONG origRedDispGamma[3];
LONG origGreenDispGamma[3];
LONG origBlueDispGamma[3];
LONG origRedOverGamma[3];
LONG origGreenOverGamma[3];
LONG origBlueOverGamma[3];
LONG curRedDispGamma[3];
LONG curGreenDispGamma[3];
LONG curBlueDispGamma[3];
LONG curRedOverGamma[3];
LONG curGreenOverGamma[3];
LONG curBlueOverGamma[3];
BOOL desktopGamma = TRUE;
BOOL joinGamma = TRUE;
BOOL joinBright = TRUE;
BOOL joinContrast = TRUE;

// Information for automatic processing of all sliders.
//
GAMMA_SLIDER gammaSlidersDisp[] = {
    { IDC_GAMMA_R, IDC_GAMMA_G, IDC_GAMMA_B, &curRedDispGamma[LX_MODE_GAMMA],   &curGreenDispGamma[LX_MODE_GAMMA], &curBlueDispGamma[LX_MODE_GAMMA], LX_GAMMA_MAX, &joinGamma },
    { IDC_GAMMA_G, IDC_GAMMA_R, IDC_GAMMA_B, &curGreenDispGamma[LX_MODE_GAMMA], &curRedDispGamma[LX_MODE_GAMMA],   &curBlueDispGamma[LX_MODE_GAMMA], LX_GAMMA_MAX, &joinGamma },
    { IDC_GAMMA_B, IDC_GAMMA_G, IDC_GAMMA_R, &curBlueDispGamma[LX_MODE_GAMMA],  &curGreenDispGamma[LX_MODE_GAMMA], &curRedDispGamma[LX_MODE_GAMMA],  LX_GAMMA_MAX, &joinGamma },
    { IDC_BRIGHT_R, IDC_BRIGHT_G, IDC_BRIGHT_B, &curRedDispGamma[LX_MODE_BRIGHT],   &curGreenDispGamma[LX_MODE_BRIGHT], &curBlueDispGamma[LX_MODE_BRIGHT], LX_BRIGHT_MAX, &joinBright },
    { IDC_BRIGHT_G, IDC_BRIGHT_R, IDC_BRIGHT_B, &curGreenDispGamma[LX_MODE_BRIGHT], &curRedDispGamma[LX_MODE_BRIGHT],   &curBlueDispGamma[LX_MODE_BRIGHT], LX_BRIGHT_MAX, &joinBright },
    { IDC_BRIGHT_B, IDC_BRIGHT_G, IDC_BRIGHT_R, &curBlueDispGamma[LX_MODE_BRIGHT],  &curGreenDispGamma[LX_MODE_BRIGHT], &curRedDispGamma[LX_MODE_BRIGHT],  LX_BRIGHT_MAX, &joinBright },
    { IDC_CONTRAST_R, IDC_CONTRAST_G, IDC_CONTRAST_B, &curRedDispGamma[LX_MODE_CONTRAST],   &curGreenDispGamma[LX_MODE_CONTRAST], &curBlueDispGamma[LX_MODE_CONTRAST], LX_CONTRAST_MAX, &joinContrast },
    { IDC_CONTRAST_G, IDC_CONTRAST_R, IDC_CONTRAST_B, &curGreenDispGamma[LX_MODE_CONTRAST], &curRedDispGamma[LX_MODE_CONTRAST],   &curBlueDispGamma[LX_MODE_CONTRAST], LX_CONTRAST_MAX, &joinContrast },
    { IDC_CONTRAST_B, IDC_CONTRAST_G, IDC_CONTRAST_R, &curBlueDispGamma[LX_MODE_CONTRAST],  &curGreenDispGamma[LX_MODE_CONTRAST], &curRedDispGamma[LX_MODE_CONTRAST],  LX_CONTRAST_MAX, &joinContrast }
};

GAMMA_SLIDER gammaSlidersOverlay[] = {
    { IDC_GAMMA_R, IDC_GAMMA_G, IDC_GAMMA_B, &curRedOverGamma[LX_MODE_GAMMA],   &curGreenOverGamma[LX_MODE_GAMMA], &curBlueOverGamma[LX_MODE_GAMMA], LX_GAMMA_MAX, &joinGamma },
    { IDC_GAMMA_G, IDC_GAMMA_R, IDC_GAMMA_B, &curGreenOverGamma[LX_MODE_GAMMA], &curRedOverGamma[LX_MODE_GAMMA],   &curBlueOverGamma[LX_MODE_GAMMA], LX_GAMMA_MAX, &joinGamma },
    { IDC_GAMMA_B, IDC_GAMMA_G, IDC_GAMMA_R, &curBlueOverGamma[LX_MODE_GAMMA],  &curGreenOverGamma[LX_MODE_GAMMA], &curRedOverGamma[LX_MODE_GAMMA],  LX_GAMMA_MAX, &joinGamma },
    { IDC_BRIGHT_R, IDC_BRIGHT_G, IDC_BRIGHT_B, &curRedOverGamma[LX_MODE_BRIGHT],   &curGreenOverGamma[LX_MODE_BRIGHT], &curBlueOverGamma[LX_MODE_BRIGHT], LX_BRIGHT_MAX, &joinBright },
    { IDC_BRIGHT_G, IDC_BRIGHT_R, IDC_BRIGHT_B, &curGreenOverGamma[LX_MODE_BRIGHT], &curRedOverGamma[LX_MODE_BRIGHT],   &curBlueOverGamma[LX_MODE_BRIGHT], LX_BRIGHT_MAX, &joinBright },
    { IDC_BRIGHT_B, IDC_BRIGHT_G, IDC_BRIGHT_R, &curBlueOverGamma[LX_MODE_BRIGHT],  &curGreenOverGamma[LX_MODE_BRIGHT], &curRedOverGamma[LX_MODE_BRIGHT],  LX_BRIGHT_MAX, &joinBright },
    { IDC_CONTRAST_R, IDC_CONTRAST_G, IDC_CONTRAST_B, &curRedOverGamma[LX_MODE_CONTRAST],   &curGreenOverGamma[LX_MODE_CONTRAST], &curBlueOverGamma[LX_MODE_CONTRAST], LX_CONTRAST_MAX, &joinContrast },
    { IDC_CONTRAST_G, IDC_CONTRAST_R, IDC_CONTRAST_B, &curGreenOverGamma[LX_MODE_CONTRAST], &curRedOverGamma[LX_MODE_CONTRAST],   &curBlueOverGamma[LX_MODE_CONTRAST], LX_CONTRAST_MAX, &joinContrast },
    { IDC_CONTRAST_B, IDC_CONTRAST_G, IDC_CONTRAST_R, &curBlueOverGamma[LX_MODE_CONTRAST],  &curGreenOverGamma[LX_MODE_CONTRAST], &curRedOverGamma[LX_MODE_CONTRAST],  LX_CONTRAST_MAX, &joinContrast }
};

/////////////////////////////////////////////////////////////////////////////
// EnumChildProc
//
// Routine to look through all child windows for a very particular window class
// chain.  When this chain is identified, the HWND of the second window in
// the chain is sent back to the caller.
//
BOOL CALLBACK EnumChildProc (HWND hwnd, LPARAM lParam)
{
    char className[160];
    FIND_WINDOW *input = (FIND_WINDOW *)lParam;
    int lookingFor = input->lookingFor;

    if (GetClassName (hwnd, className, sizeof(className)))
    {
        if ((lookingFor == 0 && lstrcmp (className, "#32770") == 0) ||
            (lookingFor == 1 && lstrcmp (className, "Static") == 0) ||
            (lookingFor == 2 && lstrcmp (className, "Monitor32") == 0))
        {
            // At this point we will say we are within the right dialog.
            // However, the HWND in this call is not the correct HWND.  The
            // correct HWND is two levels up.
            //
            if (lookingFor == 2)
                ((FIND_WINDOW *)lParam)->found = true;

            else if (lookingFor == 0)
            {
                // This is the correct level for the "Settings" window.  However,
                // as the Window Text could be in a different language, we need
                // to enumerate its children to be sure of a match.
                //
                FIND_WINDOW temp;
                temp.found = false;
                temp.lookingFor = 1;
                EnumChildWindows (hwnd, EnumChildProc, (LPARAM)(&temp));
                if (temp.found)
                {
                    input->hwnd = hwnd;
                    input->found = true;
                }
            }
            else
            {
                FIND_WINDOW temp;
                temp.found = false;
                temp.lookingFor = input->lookingFor + 1;
                EnumChildWindows (hwnd, EnumChildProc, (LPARAM)(&temp));
                if (temp.found)
                    input->found = true;
            }
        }
    }

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// EnumWindowsProc
//
// Callback routine for EnumWindows.  This routine looks for the class code
// #32770 and enumerates all child windows.
//
BOOL CALLBACK EnumWindowsProc (HWND hwnd, LPARAM lParam)
{
    char className[160];
    if (GetClassName (hwnd, className, sizeof(className)) &&
        lstrcmp (className, "#32770") == 0)
    {
        FIND_WINDOW temp;
        temp.lookingFor = 0;
        temp.found = false;

        // Could be a match.  Enumerate the child windows
        //
        EnumChildWindows (hwnd, EnumChildProc, (LPARAM)&temp);

        if (temp.found)
            *((HWND *)lParam) = temp.hwnd;
    }
    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// RefreshDisplayPage
//
// Attempt to send a WM_INITDIALOG message to the Settings dialog.
//
void RefreshDisplayPage()
{
    HWND hwnd = NULL;
    EnumWindows (EnumWindowsProc, (LPARAM)&hwnd);

    if (hwnd)
    {
        SendMessage (hwnd, WM_INITDIALOG, 0, 0);
    }
}

/////////////////////////////////////////////////////////////////////////////
// CDisplayCplExt
//
STDMETHODIMP CDisplayCplExt::Initialize (LPCITEMIDLIST pidlFolder,
                                         IDataObject*  lpdobj, 
                                         HKEY          hkeyProgID)
{
    // Do any initialization here.  pidlFolder and lpdobj are both NULL, so you
    // can ignore them.
    
    return S_OK;
}

STDMETHODIMP CDisplayCplExt::AddPages (LPFNADDPROPSHEETPAGE lpfnAddPage,
                                        LPARAM lParam)
{
    // This method adds a new page to the Display property sheet.

    PROPSHEETPAGE  psp;
    HPROPSHEETPAGE hPage;
    
    // Set up the PROPSHEETPAGE struct.

    ZeroMemory ( &psp, sizeof(PROPSHEETPAGE) );

    psp.dwSize      = sizeof(PROPSHEETPAGE);
    psp.dwFlags     = PSP_USEREFPARENT | PSP_DEFAULT | PSP_USECALLBACK | PSP_USEHICON;
    psp.hInstance   = _Module.GetResourceInstance();
    psp.pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_MEDIUM);
    psp.pfnDlgProc  = NewPageDlgProc;
    psp.pfnCallback = NewPageCallbackProc;
    psp.pcRefParent = (UINT*) &_Module.m_nLockCnt;
    psp.hIcon       = LoadIcon(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDI_AMD_ICON));

    // Create the page & get a handle to it.

    hPage = CreatePropertySheetPage (&psp);

    if (hPage)
    {
        // Call the shell's callback function, so it adds the page to
        // the property sheet.

        if (!lpfnAddPage ( hPage, lParam ))
        {
            DestroyPropertySheetPage ( hPage );
        }
    }

    ZeroMemory ( &psp, sizeof(PROPSHEETPAGE) );

    psp.dwSize      = sizeof(PROPSHEETPAGE);
    psp.dwFlags     = PSP_USEREFPARENT | PSP_DEFAULT | PSP_USECALLBACK | PSP_USEHICON;
    psp.hInstance   = _Module.GetResourceInstance();
    psp.pszTemplate = MAKEINTRESOURCE(IDD_GAMMAPAGE);
    psp.pfnDlgProc  = GammaDlgProc;
    psp.pfnCallback = NewPageCallbackProc;
    psp.pcRefParent = (UINT*) &_Module.m_nLockCnt;
    psp.hIcon       = LoadIcon(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDI_AMD_ICON));

    // Create the page & get a handle to it.

    hPage = CreatePropertySheetPage (&psp);

    if (hPage)
    {
        // Call the shell's callback function, so it adds the page to
        // the property sheet.

        if (!lpfnAddPage ( hPage, lParam ))
        {
            DestroyPropertySheetPage ( hPage );
        }
    }

    return S_OK;
}

STDMETHODIMP CDisplayCplExt::ReplacePage (UINT uPageID, 
                                          LPFNADDPROPSHEETPAGE lpfnReplaceWith,
                                          LPARAM lParam )
{
    PROPSHEETPAGE  psp;
    HPROPSHEETPAGE hPage;

    // Set up the PROPSHEETPAGE struct.

    ZeroMemory ( &psp, sizeof(PROPSHEETPAGE) );

    psp.dwSize      = sizeof(PROPSHEETPAGE);
    psp.dwFlags     = PSP_USEREFPARENT | PSP_DEFAULT | PSP_USECALLBACK | PSP_USEHICON;
    psp.hInstance   = _Module.GetResourceInstance();
    psp.pszTemplate = MAKEINTRESOURCE(IDD_PROPPAGE_MEDIUM);
    psp.pfnDlgProc  = ReplacementPageDlgProc;
    psp.pfnCallback = ReplacementPageCallbackProc;
    psp.pcRefParent = (UINT*) &_Module.m_nLockCnt;
    psp.hIcon = LoadIcon(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDI_AMD_ICON));

    // Create the page & get a handle to it.
    //
    hPage = CreatePropertySheetPage ( &psp );

    if (hPage)
    {
        // Call the shell's callback function, so it adds the page to
        // the property sheet.
        //        
        if (!lpfnReplaceWith (hPage, lParam))
            DestroyPropertySheetPage (hPage);
    }

    // Set up the PROPSHEETPAGE struct.

    ZeroMemory ( &psp, sizeof(PROPSHEETPAGE) );

    psp.dwSize      = sizeof(PROPSHEETPAGE);
    psp.dwFlags     = PSP_USEREFPARENT | PSP_DEFAULT | PSP_USECALLBACK | PSP_USEHICON;
    psp.hInstance   = _Module.GetResourceInstance();
    psp.pszTemplate = MAKEINTRESOURCE(IDD_GAMMAPAGE);
    psp.pfnDlgProc  = ReplacementPageDlgProc;
    psp.pfnCallback = ReplacementPageCallbackProc;
    psp.pcRefParent = (UINT*) &_Module.m_nLockCnt;
    psp.hIcon = LoadIcon(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDI_AMD_ICON));

    // Create the page & get a handle to it.
    //
    hPage = CreatePropertySheetPage ( &psp );

    if (hPage)
    {
        // Call the shell's callback function, so it adds the page to
        // the property sheet.
        //        
        if (!lpfnReplaceWith (hPage, lParam))
            DestroyPropertySheetPage (hPage);
    }
    
    return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// Functions for the page we add to the property sheet
//
BOOL CALLBACK NewPageDlgProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    PAINTSTRUCT ps;
    HBITMAP hbmp;
    BOOL bRet = FALSE;
    NMHDR *phdr;
    LONG rc =0;
    int retval;
    HDC dc, memdc;
    static HWND dlgHWND;
    BOOL userOK = FALSE;

    switch (uMsg)
    {
        case WM_INITDIALOG:

            // Initialize the dialog for themed display.  This allows the tab to
            // look like everybody else's tabs.
            //
            EnableThemeDialogTexture (hwnd, ETDT_ENABLETAB);

            // Query the current driver settings.
            //
            dc = GetDC (NULL);
            g_regData.secretWord = GEODELX_SECRET_WORD;
            retval = ExtEscape (dc, GEODELX_GET_REGISTRY_SETTINGS, sizeof(g_regData), (LPCSTR)&g_regData, sizeof(g_regData), (LPSTR)&g_regData);
            ReleaseDC (NULL, dc);

            // Save a copy of the old registry settings, such that we can identify what changed.
            //
            g_oldSettings = g_regData.driverStruct;
            
            // Initialize the dialog based on the registry settings.
            //
            UpdateErrorString (hwnd, retval);
            UpdateDialogFromStruct (hwnd);

            // Subclass all controls to allow us to capture mouse move messages
            //
            SubClassWindow (hwnd, IDC_COMPRESSION_OFF,      0);
            SubClassWindow (hwnd, IDC_COMPRESSION_DYNAMIC,  1);
            SubClassWindow (hwnd, IDC_DISPLAY_CRT,          2);
            SubClassWindow (hwnd, IDC_DISPLAY_TFT,          3);
            SubClassWindow (hwnd, IDC_DISPLAY_SIMULTANEOUS, 4);
            SubClassWindow (hwnd, IDC_ENABLE_STRETCHED,     5);
            SubClassWindow (hwnd, IDC_ROTATE0,              6);
            SubClassWindow (hwnd, IDC_ROTATE90,             7);
            SubClassWindow (hwnd, IDC_ROTATE180,            8);
            SubClassWindow (hwnd, IDC_ROTATE270,            9);

            if (!g_error)
            {
                SetDlgItemText (hwnd, IDC_HELP_STRING, "Mouse over options for more information.\n");
            }
            break;

        case WM_PAINT:

            // Add the AMD logo bitmap for a touch of flair
            //
            dc = BeginPaint (hwnd, &ps);
            memdc = CreateCompatibleDC (dc);
            if (memdc && dc)
            {
                hbmp = (HBITMAP)LoadImage (_Module.GetResourceInstance(), MAKEINTRESOURCE (IDB_AMD_LOGO), IMAGE_BITMAP,
                    0, 0, 0);

                if (hbmp)
                {
                    SelectObject (memdc, hbmp);
                    TransparentBlt (dc, 100, 10, 210, 72, memdc, 0, 0, 210, 72, 0xd8e9ec);
                    DeleteObject (hbmp);
                }                
            }
            
            // Add an error or info icon for mouse over info and error status.
            //
            if (g_error)
            {
                hbmp = (HBITMAP)LoadImage (_Module.GetResourceInstance(), MAKEINTRESOURCE (IDB_ERROR_ICON), IMAGE_BITMAP,
                    0, 0, 0);
            }
            else
            {
                hbmp = (HBITMAP)LoadImage (_Module.GetResourceInstance(), MAKEINTRESOURCE (IDB_HELP_ICON), IMAGE_BITMAP,
                    0, 0, 0);
            }

            RECT help;
            POINT helpClient;
            GetWindowRect (GetDlgItem (hwnd, IDC_HELP_STRING), &help);
            helpClient.x = help.left;
            helpClient.y = help.top;
            ScreenToClient (hwnd, &helpClient);
            if (hbmp)
            {
                SelectObject (memdc, hbmp);
                TransparentBlt (dc, helpClient.x - 42, helpClient.y, 32, 32, memdc, 0, 0, 32, 32, 0xC8D0D4);
                DeleteObject (hbmp);
            }

            if (memdc)
                DeleteDC (memdc);
            EndPaint(hwnd, &ps); 
            break;


        case WM_NOTIFY :
            
            phdr = (NMHDR *) lParam;
            switch(phdr->code)
            {
                case PSN_APPLY: //The Apply (or OK) button has been clicked
                {
                    if (changed)
                    {
                        // Read the current resolution.  It's important that we do this
                        // before applying any registry settings, as changing the registry
                        // settings can affect the list of available modes.
                        //
                        EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &g_curMode);

                        // Keep the original around in case of a failed modeset, as we will
                        // want to put the original mode back into the registry.  Copy the
                        // contents into a new buffer that will contain the updated mode.
                        //
                        g_newMode = g_curMode;
                        if (g_oldSettings.RotationAngle == 90 ||
                            g_oldSettings.RotationAngle == 270)
                        {
                            unsigned long temp = g_newMode.dmPelsWidth;
                            g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
                            g_newMode.dmPelsHeight = temp;
                        }

                        // Update the registry settings.  This will potentially alter the list
                        // of available modes.
                        //
                        dc = GetDC(NULL);
                        g_regData.secretWord = GEODELX_SECRET_WORD;
                        retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_regData), (LPCSTR)&g_regData, sizeof(g_regData), (LPSTR)&g_regData);
                        ReleaseDC (NULL, dc);

                        UpdateErrorString (hwnd, retval);

                        // Identify the mode we want to set.  If the output path is not changing,
                        // then we simply need to set the existing mode, possibly with a change
                        // in orientation.  Otherwise, we will set the last mode programmed by the
                        // user in the new output format (TFT or CRT).
                        //
                        if (g_oldSettings.PanelEnable != g_regData.driverStruct.PanelEnable)
                        {
                            ULONG lastMode[6];
                            ULONG width, height;
                            dc = GetDC(NULL);
                            lastMode[0] = GEODELX_SECRET_WORD;
                            retval = ExtEscape (dc, GEODELX_GET_LAST_MODE, sizeof(lastMode), (LPCSTR)&lastMode[0], 
                                sizeof(lastMode), (LPSTR)&lastMode[0]);
                            ReleaseDC (NULL, dc);

                            // Values returned are as follows:
                            // lastMode[2] = last set width
                            // lastMode[3] = last set height
                            // lastMode[4] = last set refresh
                            // lastMode[5] = last set rotation
                            //
                            if (retval)
                            {
                                // Extract the unrotated dimensions of the last mode.
                                //
                                if (lastMode[5] == 90 || lastMode[5] == 270)
                                {
                                    width  = lastMode[3];
                                    height = lastMode[2];
                                }
                                else
                                {
                                    width = lastMode[2];
                                    height = lastMode[3];
                                }

                                g_newMode.dmPelsWidth = width;
                                g_newMode.dmPelsHeight = height;
                                g_newMode.dmDisplayFrequency = lastMode[4];
                            }
                        }
        
                        // Adjust for rotation
                        //
                        if (g_regData.driverStruct.RotationAngle == 90 ||
                            g_regData.driverStruct.RotationAngle == 270)
                        {
                            long temp = g_newMode.dmPelsWidth;
                            g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
                            g_newMode.dmPelsHeight = temp;
                        }

                        switch (g_regData.driverStruct.RotationAngle)
                        {
                            case 90:  g_newMode.dmDisplayOrientation = DMDO_90;      break;
                            case 180: g_newMode.dmDisplayOrientation = DMDO_180;     break;
                            case 270: g_newMode.dmDisplayOrientation = DMDO_270;     break;
                            default:  g_newMode.dmDisplayOrientation = DMDO_DEFAULT; break;
                        }

                        // Enumerate the modes.  This forces a call to DrvGetModes, which allows the call 
                        // to set the rotated mode to work.
                        //
                        int mode = 0;
                        DEVMODE dm;
                        while (1)
                        {
                            if (!EnumDisplaySettings (NULL, mode++, &dm))
                                break;
                        }

                        LONG modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
                        modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);

                        if (modeset != DISP_CHANGE_SUCCESSFUL)
                        {
                            // Try again with default values.  One possible scenario here is that
                            // the framebuffer size has changed or that the last set TFT mode was
                            // a fake mode that was set coming out of a resume.
                            //
                            if (g_regData.driverStruct.PanelEnable)
                            {
                                g_newMode.dmPelsWidth  = g_regData.driverStruct.PanelWidth;
                                g_newMode.dmPelsHeight = g_regData.driverStruct.PanelHeight;
                                g_newMode.dmDisplayFrequency = g_regData.driverStruct.PanelRefresh;
                            }
                            else
                            {
                                g_newMode.dmPelsWidth  = 800;
                                g_newMode.dmPelsHeight = 600;
                                g_newMode.dmDisplayFrequency = 60;
                            }

                            // Adjust for rotation
                            //
                            if (g_regData.driverStruct.RotationAngle == 90 ||
                                g_regData.driverStruct.RotationAngle == 270)
                            {
                                long temp = g_newMode.dmPelsWidth;
                                g_newMode.dmPelsWidth = g_newMode.dmPelsHeight;
                                g_newMode.dmPelsHeight = temp;
                            }

                            modeset = ChangeDisplaySettings (&g_newMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
                            modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);
                        }

                        if (modeset != DISP_CHANGE_SUCCESSFUL)
                        {
                            g_error = 1;
                            switch (modeset)
                            {
                                case DISP_CHANGE_BADFLAGS:      strcpy (error_string, "Error: DISP_CHANGE_BADFLAGS");    break;
                                case DISP_CHANGE_BADMODE:       strcpy (error_string, "Error: DISP_CHANGE_BADMODE");     break;
                                case DISP_CHANGE_BADPARAM:      strcpy (error_string, "Error: DISP_CHANGE_BADPARAM");    break;
                                case DISP_CHANGE_FAILED:        strcpy (error_string, "Error: DISP_CHANGE_FAILED");      break;
                                case DISP_CHANGE_NOTUPDATED:    strcpy (error_string, "Error: DISP_CHANGE_NOTUPDATED");  break;
                                case DISP_CHANGE_RESTART:       strcpy (error_string, "Error: DISP_CHANGE_RESTART");     break;
                                default:                        strcpy (error_string, "Uknown Error"); break;
                            }
                            SetDlgItemText (hwnd, IDC_HELP_STRING, (LPCTSTR)error_string);
                        }
                        else
                        {
                            // Prompt the user to confirm that the new mode is acceptable.
                            //
                            userOK = DialogBox (_Module.GetResourceInstance(),
                                MAKEINTRESOURCE (IDD_CHANGE_PROMPT), hwnd, PromptDlgProc) == IDYES;
                        }

                        if (!userOK)
                        {
                            // Put things back the way they were (or try to).  This implies that we 
                            // revert the driver registry settings and change the mode registry settings.
                            //
                            g_regData.driverStruct = g_oldSettings;
                            dc = GetDC(NULL);
                            g_regData.secretWord = GEODELX_SECRET_WORD;
                            retval = ExtEscape (dc, GEODELX_SET_REGISTRY_SETTINGS, sizeof(g_regData), 
                                (LPCSTR)&g_regData, sizeof(g_regData), (LPSTR)&g_regData);
                            ReleaseDC (NULL, dc);

                            mode = 0;
                            DEVMODE dm;
                            while (1)
                            {
                                if (!EnumDisplaySettings (NULL, mode++, &dm))
                                    break;
                            }

                            ChangeDisplaySettings (&g_curMode, CDS_UPDATEREGISTRY | CDS_GLOBAL);
                            modeset = ChangeDisplaySettings (&g_newMode, CDS_RESET);

                            // Repaint the screen to show the error icon.
                            //
                            RECT r;
                            GetClientRect (hwnd, &r);
                            InvalidateRect (hwnd, &r, TRUE);
                        }
                        else
                        {
                            // The new becomes the old...
                            //
                            g_oldSettings = g_regData.driverStruct;
                        }

                        // Try to refresh the parent's dialog page.
                        //
                        RefreshDisplayPage();
                        UpdateDialogFromStruct(hwnd);
                        changed = 0;
                        break;
                    }
                }

                case PSN_RESET: 
                case PSN_SETACTIVE:
                case PSN_KILLACTIVE:
                default:
                    break;
            }

            // Send the message that all went well
            //
            SetWindowLong(hwnd, DWL_MSGRESULT, PSNRET_NOERROR);
            break;

        case WM_COMMAND :
            
            //User has changed some of the settings
            //
            switch (LOWORD (wParam))
            {
                case IDC_COMPRESSION_OFF:

                    changed = 1;
                    g_regData.driverStruct.Compression = 0;
                    UpdateOutputString (hwnd, IDC_COMPRESSION_OFF);
                    break;

                case IDC_COMPRESSION_DYNAMIC:

                    changed = 1;
                    g_regData.driverStruct.Compression = 1;
                    UpdateOutputString (hwnd, IDC_COMPRESSION_DYNAMIC);
                    break;

                case IDC_DISPLAY_CRT:

                    changed = 1;
                    g_regData.driverStruct.PanelEnable = 0;
                    UpdateOutputInfo (hwnd, IDC_DISPLAY_CRT);
                    UpdateOutputString (hwnd, IDC_DISPLAY_CRT);
                    break;

                case IDC_DISPLAY_TFT:

                    changed = 1;
                    g_regData.driverStruct.PanelEnable = 1;
                    g_regData.driverStruct.SimultaneousCRT = 0;
                    UpdateOutputInfo (hwnd, IDC_DISPLAY_TFT);
                    UpdateOutputString (hwnd, IDC_DISPLAY_TFT);
                    break;

                case IDC_DISPLAY_SIMULTANEOUS:

                    changed = 1;
                    g_regData.driverStruct.PanelEnable = 1;
                    g_regData.driverStruct.SimultaneousCRT = 1;
                    UpdateOutputInfo (hwnd, IDC_DISPLAY_SIMULTANEOUS);
                    UpdateOutputString (hwnd, IDC_DISPLAY_SIMULTANEOUS);
                    break;

                case IDC_ENABLE_STRETCHED:
                    
                    changed = 1;
                    g_regData.driverStruct.EnablePanelScale = (int)IsDlgButtonChecked(hwnd, IDC_ENABLE_STRETCHED);
                    UpdateOutputString (hwnd, IDC_ENABLE_STRETCHED);
                    break;

                case IDC_ROTATE0:
                    
                    changed = 1;
                    g_regData.driverStruct.RotationAngle = 0;
                    UpdateOutputString (hwnd, IDC_ROTATE0);
                    break;

                case IDC_ROTATE90:

                    changed = 1;
                    g_regData.driverStruct.RotationAngle = 90;
                    UpdateOutputString (hwnd, IDC_ROTATE90);
                    break;

                case IDC_ROTATE180:

                    changed = 1;
                    g_regData.driverStruct.RotationAngle = 180;
                    UpdateOutputString (hwnd, IDC_ROTATE180);
                    break;

                case IDC_ROTATE270:

                    changed = 1;
                    g_regData.driverStruct.RotationAngle = 270;
                    UpdateOutputString (hwnd, IDC_ROTATE270);
                    break;
            }

            // Notify parent things have changed and enable "Apply" button.
            //
            if (changed)
                SendMessage(GetParent(hwnd), PSM_CHANGED, (WPARAM) hwnd, 0);

            break;
        
        default:
            return FALSE;
    }
    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Gamma page dialog proc
//
BOOL CALLBACK GammaDlgProc (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    HBITMAP hbmp;
    HDC dc, memdc;
    NMHDR *phdr;

    switch (uMsg)
    {
        case WM_INITDIALOG:

            // Initialize the dialog for themed display.  This allows the tab to
            // look like everybody else's tabs.
            //
            EnableThemeDialogTexture (hwnd, ETDT_ENABLETAB);

            // Initialize the dialog
            //
            InitializeGammaDialog (hwnd);

            // Subclass all controls to allow us to capture mouse move messages
            //
            SubClassGammaWindow (hwnd, IDC_BRIGHT_R,    0);
            SubClassGammaWindow (hwnd, IDC_BRIGHT_G,    1);
            SubClassGammaWindow (hwnd, IDC_BRIGHT_B,    2);
            SubClassGammaWindow (hwnd, IDC_GAMMA_R,     3);
            SubClassGammaWindow (hwnd, IDC_GAMMA_G,     4);
            SubClassGammaWindow (hwnd, IDC_GAMMA_B,     5);
            SubClassGammaWindow (hwnd, IDC_CONTRAST_R,  6);
            SubClassGammaWindow (hwnd, IDC_CONTRAST_G,  7);
            SubClassGammaWindow (hwnd, IDC_CONTRAST_B,  8);

            // Set initial slider values
            //
            UpdateSliderValues(hwnd);

            // Sliders initially joined.
            //
            joinGamma = joinBright = joinContrast = TRUE;            
            break;

        case WM_PAINT:

            // Add the AMD logo bitmap for a touch of flair
            //
            dc = BeginPaint (hwnd, &ps);
            memdc = CreateCompatibleDC (dc);
            if (memdc && dc)
            {
                hbmp = (HBITMAP)LoadImage (_Module.GetResourceInstance(), MAKEINTRESOURCE (IDB_AMD_LOGO), IMAGE_BITMAP,
                    0, 0, 0);

                if (hbmp)
                {
                    SelectObject (memdc, hbmp);
                    TransparentBlt (dc, 100, 10, 210, 72, memdc, 0, 0, 210, 72, 0xd8e9ec);
                    DeleteObject (hbmp);
                }                
            }
            
            if (memdc)
                DeleteDC (memdc);
            EndPaint(hwnd, &ps); 
            break;

        case WM_COMMAND:

            //User has changed some of the settings
            //
            switch (LOWORD (wParam))
            {
                case IDC_GAMMA_SELECT:

                    if (HIWORD (wParam) == CBN_SELCHANGE)
                    {
                        desktopGamma = (int)SendDlgItemMessage (hwnd, IDC_GAMMA_SELECT, 
                            CB_GETCURSEL, 0, 0) == 0;
                        UpdateGammaSliders(hwnd);
                        UpdateSliderValues(hwnd);
                    }
                    break;

                case IDC_JOIN_GAMMA:

                    joinGamma = (BOOL)IsDlgButtonChecked (hwnd, IDC_JOIN_GAMMA);
                    break;

                case IDC_JOIN_BRIGHT:

                    joinBright = (BOOL)IsDlgButtonChecked (hwnd, IDC_JOIN_BRIGHT);
                    break;

                case IDC_JOIN_CONTRAST:

                    joinContrast = (BOOL)IsDlgButtonChecked (hwnd, IDC_JOIN_CONTRAST);
                    break;

                case IDC_RESET_DEFAULT:

                    long *red, *green, *blue;
                    if (desktopGamma)
                    {
                        red = &curRedDispGamma[0];
                        green = &curGreenDispGamma[0];
                        blue = &curBlueDispGamma[0];
                    }
                    else
                    {
                        red = &curRedOverGamma[0];
                        green = &curGreenOverGamma[0];
                        blue = &curBlueOverGamma[0];
                    }
                    red[LX_MODE_GAMMA]      = LX_GAMMA_DEFAULT;
                    green[LX_MODE_GAMMA]    = LX_GAMMA_DEFAULT;
                    blue[LX_MODE_GAMMA]     = LX_GAMMA_DEFAULT;
                    red[LX_MODE_BRIGHT]     = LX_BRIGHT_DEFAULT;
                    green[LX_MODE_BRIGHT]   = LX_BRIGHT_DEFAULT;
                    blue[LX_MODE_BRIGHT]    = LX_BRIGHT_DEFAULT;
                    red[LX_MODE_CONTRAST]   = LX_CONTRAST_DEFAULT;
                    green[LX_MODE_CONTRAST] = LX_CONTRAST_DEFAULT;
                    blue[LX_MODE_CONTRAST]  = LX_CONTRAST_DEFAULT;
                    UpdateGamma(FALSE);
                    UpdateGammaSliders(hwnd);
                    UpdateSliderValues(hwnd);
                    break;

            }
            gammaChanged = 1;
            break;

        case WM_HSCROLL:

            GAMMA_SLIDER *gammaSlider;
            if (desktopGamma)
                gammaSlider = gammaSlidersDisp;
            else
                gammaSlider = gammaSlidersOverlay;

            for (int i = 0; i < sizeof(gammaSlidersDisp)/sizeof(GAMMA_SLIDER); i++)
            {
                if (GetDlgItem (hwnd, gammaSlider[i].id) == (HWND)lParam)
                {
                    ProcessGammaSliders (hwnd, &gammaSlider[i]);
                    break;
                }
            }
            gammaChanged = 1;
            break;

        case WM_NOTIFY :
            
            phdr = (NMHDR *) lParam;
            switch(phdr->code)
            {
                case PSN_APPLY: //The Apply (or OK) button has been clicked
                {
                    if (gammaChanged)
                    {
                        // Save the settings to the registry.
                        //
                        BOOL savedDesk = desktopGamma;
                        desktopGamma = TRUE;
                        UpdateGamma (TRUE);
                        desktopGamma = FALSE;
                        UpdateGamma (TRUE);
                        desktopGamma = savedDesk;
                        gammaChanged = 0;

                        // What was new becomes old...
                        //
                        for (int i = 0; i < 3; i++)
                        {
                            origRedDispGamma[i]   = curRedDispGamma[i];
                            origGreenDispGamma[i] = curGreenDispGamma[i];
                            origBlueDispGamma[i]  = curBlueDispGamma[i];
                            origRedOverGamma[i]   = curRedOverGamma[i];
                            origGreenOverGamma[i] = curGreenOverGamma[i];
                            origBlueOverGamma[i]  = curBlueOverGamma[i];
                        }
                    }
                    break;
                }

                case PSN_QUERYCANCEL:
                {
                    // the cancel button has been clicked
                    //
                    RestoreGamma();
                    break;
                }
            }
            break;
        
        default:
            return FALSE;
    }
    if (gammaChanged)
        SendMessage(GetParent(hwnd), PSM_CHANGED, (WPARAM) hwnd, 0);

    return TRUE;
}

UINT CALLBACK NewPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp )
{
    switch (uMsg)
    {
        case PSPCB_CREATE:
            // The page is being created.  Return nonzero to let it be created,
            // zero to destroy it.
            return 1;
            
        case PSPCB_RELEASE:
            // The page is being destroyed.  Put any cleanup code (eg, free up
            // memory or resources) here.  The return value is ignored.
            return 0;
    }

    return 0;
}


/////////////////////////////////////////////////////////////////////////////
// Functions for the replacement background page

BOOL CALLBACK ReplacementPageDlgProc ( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{    
    // Put your normal dialog proc code here...
    return FALSE;
}

UINT CALLBACK ReplacementPageCallbackProc ( HWND hwnd, UINT uMsg, LPPROPSHEETPAGE ppsp )
{
    return FALSE;
}

void UpdateErrorString (HWND hwnd, int success)
{
    if (success <= 0)
    {
        memset (&g_regData.driverStruct, 0, sizeof(g_regData.driverStruct));
        strcpy (error_string, "Display Driver Escapes are disabled.  Enable hardware acceleration in the troubleshoot dialog.");
        SetDlgItemText (hwnd, IDC_HELP_STRING, (LPCTSTR)error_string);
    }
    else if (g_error)
    {
        // Clear an error string if one existed.  Otherwise, do nothing as we
        // would clear out the tooltip text.
        //
        strcpy (error_string, " ");
        SetDlgItemText (hwnd, IDC_HELP_STRING, (LPCTSTR)error_string);
    }

    g_error = (success <= 0);
}

void UpdateOutputString (HWND hwnd, int id)
{
    char helpString[300];
    strcpy (helpString, "");
    
    if (!g_error)
    {
        switch (id)
        {
            case IDC_DISPLAY_CRT:
                strcpy (helpString, "Display only to an attached CRT.  Enabling this option enables all CRT display modes supported by the attached CRT.");
                break;

            case IDC_DISPLAY_TFT:
                strcpy (helpString, "Display only to the TFT.  The CRT output for the system is disabled.");
                break;

            case IDC_DISPLAY_SIMULTANEOUS:
                strcpy (helpString, "Enable simultaneous TFT/CRT display.  The TFT display is passed through to the CRT.  Note that TFT modes may not display correctly on all CRTs.");
                break;

            case IDC_COMPRESSION_OFF:
                strcpy (helpString, "Force display compression to be off.  This potentially saves as much as 750K of video memory, but may increase the bandwidth needs of the active display.");
                break;

            case IDC_COMPRESSION_DYNAMIC:
                strcpy (helpString, "Compression is enabled whenever possible, easing the bandwidth needs of the display.  Enabling compression requires a maximum of 750K of video memory to be allocated.");
                break;

            case IDC_ENABLE_STRETCHED:
                strcpy (helpString, "Enable modes that are smaller or larger than the native TFT resolution to be scaled to fit the display.");
                break;

            case IDC_ROTATE0:
                strcpy (helpString, "Disable display rotation");
                break;

            case IDC_ROTATE90:
                strcpy (helpString, "Portrait mode display with 90 degree display rotation");
                break;

            case IDC_ROTATE180:
                strcpy (helpString, "Inverted landscape display with 180 degree display rotation");
                break;

            case IDC_ROTATE270:
                strcpy (helpString, "Inverted portrait mode display with 270 degree display rotation");
                break;
        }

        SetDlgItemText (hwnd, IDC_HELP_STRING, helpString);
    }
}

void UpdateOutputInfo (HWND hwnd, int rotId)
{
    if (rotId != IDC_DISPLAY_CRT)
        EnableWindow (GetDlgItem (hwnd, IDC_ENABLE_STRETCHED), TRUE);
    else
        EnableWindow (GetDlgItem (hwnd, IDC_ENABLE_STRETCHED), FALSE);
}

void UpdateDialogFromStruct (HWND hwnd)
{
    int compSetting;
    if (g_regData.driverStruct.Compression == 0)
        compSetting = IDC_COMPRESSION_OFF;
    else
        compSetting = IDC_COMPRESSION_DYNAMIC;
    CheckRadioButton (hwnd, IDC_COMPRESSION_OFF, IDC_COMPRESSION_DYNAMIC, compSetting);

    int rotId;
    switch (g_regData.driverStruct.RotationAngle)
    {
        case 90:  rotId = IDC_ROTATE90;  break;
        case 180: rotId = IDC_ROTATE180; break;
        case 270: rotId = IDC_ROTATE270; break;
        default:  rotId = IDC_ROTATE0;   break;
    }
    CheckRadioButton (hwnd, IDC_ROTATE0, IDC_ROTATE270, rotId);

    if (g_regData.driverStruct.BIOSDetectOutput && !g_regData.driverStruct.CustomPanelTimings &&
       !g_regData.driverStruct.PanelInSystem)
    {
        EnableWindow (GetDlgItem (hwnd, IDC_DISPLAY_CRT), FALSE);
        EnableWindow (GetDlgItem (hwnd, IDC_DISPLAY_TFT), FALSE);
        EnableWindow (GetDlgItem (hwnd, IDC_DISPLAY_SIMULTANEOUS), FALSE);
        EnableWindow (GetDlgItem (hwnd, IDC_ENABLE_STRETCHED), FALSE);
    }
    else
    {
        int panelId;
        if (!g_regData.driverStruct.PanelEnable)
            panelId = IDC_DISPLAY_CRT;
        else if (g_regData.driverStruct.SimultaneousCRT)
            panelId = IDC_DISPLAY_SIMULTANEOUS;
        else
            panelId = IDC_DISPLAY_TFT;
        CheckRadioButton (hwnd, IDC_DISPLAY_CRT, IDC_DISPLAY_SIMULTANEOUS, panelId);
        UpdateOutputInfo (hwnd, panelId);
    }

    CheckDlgButton (hwnd, IDC_ENABLE_STRETCHED, g_regData.driverStruct.EnablePanelScale ? BST_CHECKED : BST_UNCHECKED);
}

LRESULT CALLBACK LXWindowProc (HWND hwnd, UINT message, WPARAM wParam, 
    LPARAM lParam)
{
    // First find the control in question
    //
    for (int i = 0; i < sizeof(wndProcs)/sizeof(LX_WND_PROC); i++)
    {
        if (hwnd == wndProcs[i].hwnd)
            break;
    }
    if (i == sizeof(wndProcs)/sizeof(LX_WND_PROC))
        return FALSE;

    // Only care about mouse movememnt
    //
    if (message == WM_MOUSEMOVE)
    {
        UpdateOutputString (wndProcs[i].parent, wndProcs[i].id);
    }
            
    return CallWindowProc (wndProcs[i].wndProc, hwnd, message, wParam, lParam);
} 

LRESULT CALLBACK LXGammaWindowProc (HWND hwnd, UINT message, WPARAM wParam, 
    LPARAM lParam)
{
    // First find the control in question
    //
    for (int i = 0; i < sizeof(gammaWndProcs)/sizeof(LX_WND_PROC); i++)
    {
        if (hwnd == gammaWndProcs[i].hwnd)
            break;
    }
    if (i == sizeof(gammaWndProcs)/sizeof(LX_WND_PROC))
        return FALSE;

    // Only care about mouse movememnt
    //
    if (message == WM_MOUSEMOVE)
    {
        CString valueString;
        double value;
        int stringId;

        double pos = (double)SendMessage(hwnd, TBM_GETPOS, 0, 0);
        if (gammaWndProcs[i].id == IDC_GAMMA_R || 
            gammaWndProcs[i].id == IDC_GAMMA_G ||
            gammaWndProcs[i].id == IDC_GAMMA_B)
        {
            value = 0.5 + (pos * 0.05);
            stringId = IDC_GAMMA_STRING;
        }
        else if (gammaWndProcs[i].id == IDC_BRIGHT_R || 
                 gammaWndProcs[i].id == IDC_BRIGHT_G ||
                 gammaWndProcs[i].id == IDC_BRIGHT_B)
        {
            value = (pos / 256.0) - 0.25;
            stringId = IDC_BRIGHT_STRING;
        }
        else
        {
            value = 0.5 + (0.005 * pos);
            stringId = IDC_CONTRAST_STRING;
        }

        CString neg = "";
        if (value < 0)
        {
            neg = "-";
            value *= -1.0;
        }
        valueString.Format ("%i.%2.2i", (int)value, (int)((double)(value - (int)value) * 100.0));
        valueString = neg + valueString;
        SetDlgItemText (gammaWndProcs[i].parent, stringId, valueString.GetBuffer(1));
    }
            
    return CallWindowProc (gammaWndProcs[i].wndProc, hwnd, message, wParam, lParam);
} 

void SubClassWindow (HWND hwnd, int id, int index)
{
    HWND hControl = GetDlgItem (hwnd, id);
    WNDPROC oldProc = (WNDPROC)SetWindowLong (hControl, GWL_WNDPROC, (LONG)&LXWindowProc);

    wndProcs[index].hwnd = hControl;
    wndProcs[index].id = id;
    wndProcs[index].wndProc = oldProc;
    wndProcs[index].parent = hwnd;
}

void SubClassGammaWindow (HWND hwnd, int id, int index)
{
    HWND hControl = GetDlgItem (hwnd, id);
    WNDPROC oldProc = (WNDPROC)SetWindowLong (hControl, GWL_WNDPROC, (LONG)&LXGammaWindowProc);

    gammaWndProcs[index].hwnd = hControl;
    gammaWndProcs[index].id = id;
    gammaWndProcs[index].wndProc = oldProc;
    gammaWndProcs[index].parent = hwnd;
}

void InitializeGammaDialog (HWND hwnd)
{
    HDC dc;
    ULONG gamma[11];
    BOOL retval;
    int i;
    
    // Read the current gamma settings
    //
    for (i = 0; i < 9; i++)
        gamma[i] = 0;

    dc = GetDC (NULL);
    gamma[0] = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_GET_DISPLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
        (LPSTR)&gamma[0]);

    for (i = 0; i < 3; i++)
    {
        origRedDispGamma[i]   = curRedDispGamma[i]   = gamma[(i * 3) + 2];
        origGreenDispGamma[i] = curGreenDispGamma[i] = gamma[(i * 3) + 3];
        origBlueDispGamma[i]  = curBlueDispGamma[i]  = gamma[(i * 3) + 4];
    }

    gamma[0] = GEODELX_SECRET_WORD;
    retval = ExtEscape (dc, GEODELX_GET_OVERLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
        (LPSTR)&gamma[0]);

    for (i = 0; i < 3; i++)
    {
        origRedOverGamma[i]   = curRedOverGamma[i]   = gamma[(i * 3) + 2];
        origGreenOverGamma[i] = curGreenOverGamma[i] = gamma[(i * 3) + 3];
        origBlueOverGamma[i]  = curBlueOverGamma[i]  = gamma[(i * 3) + 4];
    }
    ReleaseDC (NULL, dc);

    SendDlgItemMessage (hwnd, IDC_GAMMA_SELECT, CB_ADDSTRING, 0, (LPARAM)"Desktop");
    SendDlgItemMessage (hwnd, IDC_GAMMA_SELECT, CB_ADDSTRING, 0, (LPARAM)"Video");
    SendDlgItemMessage (hwnd, IDC_GAMMA_SELECT, CB_SETCURSEL, 0, 0);

    CheckDlgButton(hwnd, IDC_JOIN_GAMMA, BST_CHECKED);
    CheckDlgButton(hwnd, IDC_JOIN_BRIGHT, BST_CHECKED);
    CheckDlgButton(hwnd, IDC_JOIN_CONTRAST, BST_CHECKED);

    SendDlgItemMessage(hwnd, IDC_BRIGHT_R, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_BRIGHT_MAX));
    SendDlgItemMessage(hwnd, IDC_BRIGHT_G, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_BRIGHT_MAX));
    SendDlgItemMessage(hwnd, IDC_BRIGHT_B, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_BRIGHT_MAX));
    SendDlgItemMessage(hwnd, IDC_CONTRAST_R, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_CONTRAST_MAX));
    SendDlgItemMessage(hwnd, IDC_CONTRAST_G, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_CONTRAST_MAX));
    SendDlgItemMessage(hwnd, IDC_CONTRAST_B, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_CONTRAST_MAX));
    SendDlgItemMessage(hwnd, IDC_GAMMA_R, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_GAMMA_MAX));
    SendDlgItemMessage(hwnd, IDC_GAMMA_G, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_GAMMA_MAX));
    SendDlgItemMessage(hwnd, IDC_GAMMA_B, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(0, LX_GAMMA_MAX));

    desktopGamma = TRUE;
    UpdateGammaSliders(hwnd);
}

void UpdateGammaSliders (HWND hwnd)
{
    long *red, *green, *blue;
    if (desktopGamma)
    {
        red   = &curRedDispGamma[0];
        green = &curGreenDispGamma[0];
        blue  = &curBlueDispGamma[0];
    }
    else
    {
        red   = &curRedOverGamma[0];
        green = &curGreenOverGamma[0];
        blue  = &curBlueOverGamma[0];
    }

    SendDlgItemMessage(hwnd, IDC_BRIGHT_R, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)red[LX_MODE_BRIGHT]);
    SendDlgItemMessage(hwnd, IDC_BRIGHT_G, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)green[LX_MODE_BRIGHT]);
    SendDlgItemMessage(hwnd, IDC_BRIGHT_B, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)blue[LX_MODE_BRIGHT]);
    SendDlgItemMessage(hwnd, IDC_CONTRAST_R, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)red[LX_MODE_CONTRAST]);
    SendDlgItemMessage(hwnd, IDC_CONTRAST_G, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)green[LX_MODE_CONTRAST]);
    SendDlgItemMessage(hwnd, IDC_CONTRAST_B, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)blue[LX_MODE_CONTRAST]);
    SendDlgItemMessage(hwnd, IDC_GAMMA_R, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)red[LX_MODE_GAMMA]);
    SendDlgItemMessage(hwnd, IDC_GAMMA_G, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)green[LX_MODE_GAMMA]);
    SendDlgItemMessage(hwnd, IDC_GAMMA_B, TBM_SETPOS, (WPARAM) TRUE, (LPARAM)blue[LX_MODE_GAMMA]);
}

void ProcessGammaSliders (HWND hwnd, GAMMA_SLIDER *gammaSlider)
{
    // Get the new position for control 0.
    //
    long pos = (long)SendMessage(GetDlgItem (hwnd, gammaSlider->id), TBM_GETPOS, 0, 0);

    if (*gammaSlider->join)
    {        
        long diff = pos - *(gammaSlider->pos);
        
        if ((*(gammaSlider->pos) + diff) < 0) diff += -(*(gammaSlider->pos) + diff);
        if ((*(gammaSlider->otherPos1) + diff) < 0) diff += -(*(gammaSlider->otherPos1) + diff);
        if ((*(gammaSlider->otherPos2) + diff) < 0) diff += -(*(gammaSlider->otherPos2) + diff);
        if ((*(gammaSlider->pos) + diff) > gammaSlider->max) diff -= (*(gammaSlider->pos) + diff) - gammaSlider->max;
        if ((*(gammaSlider->otherPos1) + diff) > gammaSlider->max) diff -= (*(gammaSlider->otherPos1) + diff) - gammaSlider->max;
        if ((*(gammaSlider->otherPos2) + diff) > gammaSlider->max) diff -= (*(gammaSlider->otherPos2) + diff) - gammaSlider->max;

        *(gammaSlider->pos) += diff;
        *(gammaSlider->otherPos1) += diff;
        *(gammaSlider->otherPos2) += diff;

        SendDlgItemMessage (hwnd, gammaSlider->id, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)*(gammaSlider->pos));
        SendDlgItemMessage (hwnd, gammaSlider->otherId1, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)*(gammaSlider->otherPos1));
        SendDlgItemMessage (hwnd, gammaSlider->otherId2, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)*(gammaSlider->otherPos2));
    }
    else
        *(gammaSlider->pos) = pos;

    UpdateGamma(FALSE);
}

void UpdateGamma(BOOL updateRegistry)
{
    HDC dc = GetDC (NULL);
    ULONG gamma[11];
    int i;
    BOOL retval;
    
    // Read the current gamma settings
    //
    if (desktopGamma)
    {
        for (i = 0; i < 3; i++)
        {
            gamma[(i * 3) + 2] = curRedDispGamma[i];
            gamma[(i * 3) + 3] = curGreenDispGamma[i];
            gamma[(i * 3) + 4] = curBlueDispGamma[i];
        }
        gamma[0] = GEODELX_SECRET_WORD;
        if (updateRegistry)
        {
            retval = ExtEscape (dc, GEODELX_SET_DISPLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
                (LPSTR)&gamma[0]);
        }
        else
        {
            retval = ExtEscape (dc, GEODELX_SET_DISPLAY_GAMMA, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
                (LPSTR)&gamma[0]);
        }
    }
    else
    {
        for (i = 0; i < 3; i++)
        {
            gamma[(i * 3) + 2] = curRedOverGamma[i];
            gamma[(i * 3) + 3] = curGreenOverGamma[i];
            gamma[(i * 3) + 4] = curBlueOverGamma[i];
        }
        gamma[0] = GEODELX_SECRET_WORD;
        if (updateRegistry)
        {
            retval = ExtEscape (dc, GEODELX_SET_OVERLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
                (LPSTR)&gamma[0]);
        }
        else
        {
            retval = ExtEscape (dc, GEODELX_SET_OVERLAY_GAMMA, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
                (LPSTR)&gamma[0]);
        }
    }

    ReleaseDC (NULL, dc);
}

void RestoreGamma(void)
{
    HDC dc = GetDC (NULL);
    ULONG gamma[11];
    int i;
    BOOL retval;
    
    // Read the current gamma settings
    //
    if (desktopGamma)
    {
        for (i = 0; i < 3; i++)
        {
            gamma[(i * 3) + 2] = origRedDispGamma[i];
            gamma[(i * 3) + 3] = origGreenDispGamma[i];
            gamma[(i * 3) + 4] = origBlueDispGamma[i];
        }
        gamma[0] = GEODELX_SECRET_WORD;
        retval = ExtEscape (dc, GEODELX_SET_DISPLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
            (LPSTR)&gamma[0]);
    }
    else
    {
        for (i = 0; i < 3; i++)
        {
            gamma[(i * 3) + 2] = origRedOverGamma[i];
            gamma[(i * 3) + 3] = origGreenOverGamma[i];
            gamma[(i * 3) + 4] = origBlueOverGamma[i];
        }
        gamma[0] = GEODELX_SECRET_WORD;
        retval = ExtEscape (dc, GEODELX_SET_OVERLAY_GAMMA_REGISTRY, sizeof(gamma), (LPCSTR)&gamma[0], sizeof(gamma), 
            (LPSTR)&gamma[0]);
    }

    ReleaseDC (NULL, dc);
}

BOOL CALLBACK FindConfirmWindow (HWND hwnd, LPARAM lParam)
{
    char className[160];
    if (GetWindowText (hwnd, className, sizeof(className)) &&
        lstrcmp (className, "AMD Confirm Settings") == 0)
    {
        *((HWND *)lParam) = hwnd;
    }
    return TRUE;
}

DWORD WINAPI PromptThread ( LPVOID lpParam ) 
{
    LARGE_INTEGER startTime;
    LARGE_INTEGER currentTicks;
    LARGE_INTEGER countsPerSecond;
    HWND hwnd = NULL;
    QueryPerformanceFrequency (&countsPerSecond);
    EnumWindows (FindConfirmWindow, (LPARAM)&hwnd);
    int previousCount = 0;
    if (!hwnd)
    {
        promptThreadRunning = 0;
        return 0;
    }

    // Read the current performance counter value
    //
    QueryPerformanceCounter (&startTime);

    while (1)
    {
        if (promptThreadStop)
            break;

        Sleep (10);
        QueryPerformanceCounter (&currentTicks);

        currentTicks.QuadPart -= startTime.QuadPart;
        currentTicks.QuadPart /= countsPerSecond.QuadPart;

        if (currentTicks.QuadPart >= 15)
        {
            PostMessage (hwnd, WM_COMMAND, (WPARAM)IDCANCEL, NULL);
            promptThreadRunning = 0;
            return 0;
        }

        if ((int)(currentTicks.QuadPart) != previousCount)
        {
            CString newText;
            previousCount = (int)(currentTicks.QuadPart);

            newText.Format ("Reverting in %i seconds", 15 - previousCount);
            SetDlgItemText (hwnd, IDC_REVERT_STRING, newText.GetBuffer(1));
        }
    }
    promptThreadRunning = 0;
    return 0;
}

BOOL CALLBACK PromptDlgProc (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    static HANDLE hThread = NULL;
    DWORD dwThrdParam, dwThreadId;
    switch (uMsg)
    {
        case WM_INITDIALOG:
            // Start a new thread to update the dialog box and revert if
            // no user input.
            //
            if (!promptThreadRunning)
            {
                dwThrdParam = (DWORD)hwnd;
                promptThreadRunning = 1;
                promptThreadStop = 0;
                hThread = CreateThread( 
                    NULL,                        // default security attributes 
                    0,                           // use default stack size  
                    PromptThread,                // thread function 
                    &dwThrdParam,                // argument to thread function 
                    0,                           // use default creation flags 
                    &dwThreadId);                // returns the thread identifier 

                if (hThread == NULL)
                    break;
            }

        case WM_COMMAND:
            
            switch (LOWORD (wParam))
            {
                case IDOK:
                    if (promptThreadRunning && hThread)
                    {
                        // Signal for the prompt thread to terminate.  1.5 seconds max wait.  
                        // TerminateThread is not used because it can have dangerous consequences
                        // if the execution state of the thread is not known.
                        //
                        int timeout = 150;
                        promptThreadStop = 1;
                        while (promptThreadRunning && timeout--)
                            Sleep(10);

                        CloseHandle (hThread);
                        EndDialog (hwnd, IDYES);
                    }
                    break;

                case IDCANCEL:
                    if (promptThreadRunning && hThread)
                    {
                        // Signal for the prompt thread to terminate.  1.5 seconds max wait.
                        // TerminateThread is not used because it can have dangerous consequences
                        // if the execution state of the thread is not known.
                        //
                        int timeout = 150;
                        promptThreadStop = 1;
                        while (promptThreadRunning && timeout--)
                            Sleep(10);

                        CloseHandle (hThread);
                        EndDialog (hwnd, IDNO);
                    }
                    break;
            }
            break;

        case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC dc = BeginPaint (hwnd, &ps);
            HDC memdc = CreateCompatibleDC (dc);
            HBITMAP hbmp = (HBITMAP)LoadImage (_Module.GetResourceInstance(), 
                MAKEINTRESOURCE (IDB_HELP_ICON), IMAGE_BITMAP, 0, 0, 0);

            RECT help;
            POINT helpClient;
            GetWindowRect (GetDlgItem (hwnd, IDC_HELP_STRING), &help);
            helpClient.x = help.left;
            helpClient.y = help.top;
            ScreenToClient (hwnd, &helpClient);
            if (hbmp)
            {
                SelectObject (memdc, hbmp);
                TransparentBlt (dc, helpClient.x - 52, helpClient.y, 32, 32, memdc, 0, 0, 32, 32, 0xC8D0D4);
                DeleteObject (hbmp);
            }

            if (memdc)
                DeleteDC (memdc);
            EndPaint(hwnd, &ps); 
            break;
        }

        default:
            return FALSE;
    }
    return true;
}

void UpdateSliderValues (HWND hwnd)
{
    CString valueString;
    CString neg;
    double value;
    double pos;
    
    pos = (double)SendMessage(GetDlgItem (hwnd, IDC_GAMMA_R), TBM_GETPOS, 0, 0);
    value = 0.5 + (pos * 0.05);
    valueString.Format ("%i.%2.2i", (int)value, (int)((double)(value - (int)value) * 100.0));
    SetDlgItemText (hwnd, IDC_GAMMA_STRING, valueString.GetBuffer(1));

    pos = (double)SendMessage(GetDlgItem (hwnd, IDC_BRIGHT_R), TBM_GETPOS, 0, 0);
    value = (pos / 256.0) - 0.25;
    neg = "";
    if (value < 0)
    {
        neg = "-";
        value *= -1.0;
    }
    valueString.Format ("%i.%2.2i", (int)value, (int)((double)(value - (int)value) * 100.0));
    valueString = neg + valueString;
    SetDlgItemText (hwnd, IDC_BRIGHT_STRING, valueString.GetBuffer(1));

    pos = (double)SendMessage(GetDlgItem (hwnd, IDC_CONTRAST_R), TBM_GETPOS, 0, 0);
    value = 0.5 + (0.005 * pos);
    valueString.Format ("%i.%2.2i", (int)value, (int)((double)(value - (int)value) * 100.0));
    SetDlgItemText (hwnd, IDC_CONTRAST_STRING, valueString.GetBuffer(1));    
}