//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// DrvBitBlt
// DrvCopyBits
// DrvSynchronize
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "rop4.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"

/////////////////////////////////////////////////////////////////////////
// DrvBitBlt
//
// BitBlt is responsible for a wide variety for rectangular fills and
// memory to screen blits.  All screen to memory BLTs go through DrvCopyBits.
//
BOOL DrvBitBlt(
    SURFOBJ     *psoTrg,
    SURFOBJ     *psoSrc,
    SURFOBJ     *psoMask,
    CLIPOBJ     *pco,
    XLATEOBJ    *pxlo,
    RECTL       *prclDst,
    POINTL      *pptlSrc,
    POINTL      *pptlMask,
    BRUSHOBJ    *pbo,
    POINTL      *pptlBrush,
    ROP4        rop4)
{
    PDEV*       ppdev = NULL;
    PBRUSH*     gBrushPtr;
    RECTL       destRegion;
    int         more, i, solidBrush = 1;
    CLIPENUM    EnumRects8;
    RECTL*      clippedTargetPtr;
    ULONG       destoffset;
    ULONG       index, fromFB;
    ULONG       dstSurfOffset, srcSurfOffset;
    long        dstSurfPitch,  srcSurfPitch;
    USHORT      dhsurfTrg;
    ULONG       bltFlags = 0;
    int         colorConversion;
    ULONG       bitfields[3];
    ULONG       srcFormat;
    ULONG       masked = 0;
    long        x, y;
    int         rotatingDst;
    RECTL       rotDestRegion;
    RECTL       *rotDestPtr = NULL;
    RECTL       *rotSrcPtr;

    DISPDBG ((ppdev, 1, "DrvBitBlt Entry\n"));

    ACQUIRE_PRIMITIVE_SEMAPHORE;

    if (!psoTrg->dhsurf)
    {
        ppdev = (PDEV *)psoSrc->dhpdev;
        goto reversePuntIt;
    }
    ppdev = (PDEV *)psoTrg->dhpdev;

    if (g_puntAllUntilModeSet)
        goto puntIt;

    // Extract dhsurf for the Target surface.  This is needed for rotation
    // cases.
    //
    dhsurfTrg = (USHORT)psoTrg->dhsurf;
    rotatingDst = (dhsurfTrg == 1) && (ppdev->regSettings.RotationAngle != 0);

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

    if ((rop4 & 0xFF) != ((USHORT)rop4 >> 8))
    {
#if DBG
        if (g_puntMask & LX_PUNT_MASKEDBLT)
        {
            goto puntIt;
        }
#endif

        // CHECK PUNT CASES
        // We will punt BLTs that are not a pure masked copy and monochrome
        // masked BLTs.
        //
        if (rop4 == 0xCCAA)
            bltFlags = CIMGP_BLTFLAGS_INVERTMONO;
        else if (rop4 != 0xAACC)
            goto puntIt;

        if (psoSrc->iBitmapFormat == BMF_1BPP || !psoMask)
            goto puntIt;

        masked = 1;

        goto SourceInBlt;
    }

    // CHECK FOR PATTERN FILL CASE
    // Pattern fills are much simpler than anything that involves source, as
    // well as more common...
    //
    if (!(rop4Flags[(UCHAR)rop4] & LX_ROP_SRC))
    {
        index = dhsurf_array[dhsurfTrg];
        if (index & CACHE_FLAG_SYSTEM)
            goto puntSystem;

        dstSurfOffset = bitmap_heap[(USHORT)index].heap_offset;
        dstSurfPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

        if (rop4Flags[(UCHAR)rop4] & LX_ROP_PAT)
        {
            // CHECK FOR SOLID COLOR
            // Color != 0xFFFFFFFF implies solid color instead of pattern.
            //
            if(pbo->iSolidColor != 0xffffffff)
            {
#if DBG
                if (g_puntMask & LX_PUNT_SOLIDFILL)
                {
                    goto puntIt;
                }
#endif

                // SET SOLID COLOR
                //
                gp_declare_blt (0);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_solid_pattern (pbo->iSolidColor);
            }
            else
            {
                // CHECK BRUSH TYPE

                if (pbo->pvRbrush != NULL)
                    gBrushPtr = pbo->pvRbrush;
                else
                {
                    gBrushPtr = BRUSHOBJ_pvGetRbrush (pbo);

                    DISPDBG((ppdev, 1, "gBrushPtr returned 0x%x\n", gBrushPtr));

                    if (gBrushPtr == NULL)
                    {
                        goto puntIt;
                    }
                }

#if DBG
                if (((g_puntMask & LX_PUNT_MONOFILL)  && gBrushPtr->patFormat == 0xFF) ||
                    ((g_puntMask & LX_PUNT_COLORFILL) && gBrushPtr->patFormat != 0xFF))
                {
                    goto puntIt;
                }
#endif

                solidBrush = 0;
                if (gBrushPtr->patFormat == 0xFF)
                {
                    gp_declare_blt (0);
                    gp_set_raster_operation ((UCHAR)rop4);

                    // Rotate the pattern if BLTing to the primary display with
                    // rotation enabled.
                    //
                    if (rotatingDst)
                    {
                        unsigned long data0, data1;
                        unsigned long patX, patY;

                        if (ppdev->regSettings.RotationAngle == 90)
                        {
                            patX = 4000 + prclDst->top - pptlBrush->y;
                            patY = 4000 + pptlBrush->x - prclDst->right;
                        }
                        else if (ppdev->regSettings.RotationAngle == 270)
                        {
                            patY = 4000 + prclDst->left - pptlBrush->x;
                            patX = 4000 + pptlBrush->y - prclDst->bottom;
                        }
                        else
                        {
                            patX = 4000 - prclDst->right + pptlBrush->x;
                            patY = 4000 - prclDst->bottom + pptlBrush->y;
                        }

                        RotateMonoPattern (ppdev, gBrushPtr->Pattern[0], gBrushPtr->Pattern[1],
                            &data0, &data1);
                        gp_set_mono_pattern (gBrushPtr->monoColor0, gBrushPtr->monoColor1,
                            data0, data1, 0, patX, patY);
                    }
                    else
                    {
                        gp_set_mono_pattern (gBrushPtr->monoColor0, gBrushPtr->monoColor1,
                            gBrushPtr->Pattern[0], gBrushPtr->Pattern[1], 0,
                            4000 - pptlBrush->x + prclDst->left,
                            4000 - pptlBrush->y + prclDst->top);
                    }
                }
                else
                {
                    // Rotate the pattern if BLTing to the primary display with
                    // rotation enabled.
                    //
                    if (rotatingDst)
                    {
                        unsigned long data0, data1;
                        unsigned long patX, patY;
                        unsigned long rotPattern[64];

                        if (ppdev->regSettings.RotationAngle == 90)
                        {
                            patX = 4000 + prclDst->top - pptlBrush->y;
                            patY = 4000 + pptlBrush->x - prclDst->right;
                        }
                        else if (ppdev->regSettings.RotationAngle == 270)
                        {
                            patY = 4000 + prclDst->left - pptlBrush->x;
                            patX = 4000 + pptlBrush->y - prclDst->bottom;
                        }
                        else
                        {
                            patX = 4000 - prclDst->right + pptlBrush->x;
                            patY = 4000 - prclDst->bottom + pptlBrush->y;
                        }

                        RotateColorPattern (ppdev, (ULONG *)&(gBrushPtr->Pattern[0]), &rotPattern[0],
                            gBrushPtr->patFormat);
                        gp_set_color_pattern (&rotPattern[0], gBrushPtr->patFormat, patX, patY);
                    }
                    else
                    {
                        gp_set_color_pattern (&(gBrushPtr->Pattern[0]), gBrushPtr->patFormat,
                            4000 - pptlBrush->x + prclDst->left,
                            4000 - pptlBrush->y + prclDst->top);
                    }
                    gp_declare_blt (0);
                    gp_set_raster_operation ((UCHAR)rop4);
                }
            }
        }
        else
        {
            // DESTINATION ONLY, WHITENESS OR BLACKNESS
            // Just write ROP.

            gp_declare_blt (0);
            gp_set_raster_operation ((UCHAR)rop4);
        }

        // SET THE STRIDES
        //
        if (rotatingDst)
            gp_set_strides (mode_pitch, 0);
        else
            gp_set_strides (dstSurfPitch, 0);

        // BRUSH AND ROP SETUP COMPLETE
        // Handle clipping

        if (!pco || pco->iDComplexity == DC_TRIVIAL)
        {
            if (rotatingDst)
            {
                gp_pattern_fill (ROT_OFFSET (ppdev, prclDst->, ppdev->regSettings.RotationAngle),
                    ROT_WIDTH (prclDst->, ppdev->regSettings.RotationAngle),
                    ROT_HEIGHT (prclDst->, ppdev->regSettings.RotationAngle));
            }
            else
            {
                destoffset  = (dstSurfPitch * prclDst->top +
                          (prclDst->left << ppdev->cPelSize));
                destoffset += dstSurfOffset;
                gp_pattern_fill (destoffset,
                    prclDst->right - prclDst->left,
                    prclDst->bottom - prclDst->top);
            }
        }
        else if (pco->iDComplexity == DC_RECT)
        {
            if (!FindIntersection (prclDst, &pco->rclBounds, &destRegion))
                goto returnTrueNoRotate;

            // ADJUST PATTERN ORIGIN

            if (!solidBrush)
            {
                if (rotatingDst)
                {
                    unsigned long patX, patY;
                    if (ppdev->regSettings.RotationAngle == 90)
                    {
                        patX = 4000 + destRegion.top - pptlBrush->y;
                        patY = 4000 + pptlBrush->x - destRegion.right;
                    }
                    else if (ppdev->regSettings.RotationAngle == 270)
                    {
                        patY = 4000 + destRegion.left - pptlBrush->x;
                        patX = 4000 + pptlBrush->y - destRegion.bottom;
                    }
                    else
                    {
                        patX = 4000 - destRegion.right + pptlBrush->x;
                        patY = 4000 - destRegion.bottom + pptlBrush->y;
                    }
                    gp_set_pattern_origin (patX, patY);
                }
                else
                {
                    gp_set_pattern_origin (
                        4000 - pptlBrush->x + destRegion.left,
                        4000 - pptlBrush->y + destRegion.top);
                }
            }

            if (rotatingDst)
            {
                gp_pattern_fill (ROT_OFFSET (ppdev, destRegion., ppdev->regSettings.RotationAngle),
                    ROT_WIDTH (destRegion., ppdev->regSettings.RotationAngle),
                    ROT_HEIGHT (destRegion., ppdev->regSettings.RotationAngle));
            }
            else
            {
                destoffset = (dstSurfPitch * destRegion.top) +
                            (destRegion.left << ppdev->cPelSize);
                destoffset += dstSurfOffset;

                gp_pattern_fill (destoffset,
                    destRegion.right  - destRegion.left,
                    destRegion.bottom - destRegion.top);
            }
        }
        else
        {
            // COMPLEX CLIPPING

            gp_write_parameters();
            CLIPOBJ_cEnumStart(pco, FALSE, CT_RECTANGLES, CD_ANY, 0);

            while (1)
            {
                more = CLIPOBJ_bEnum(pco, sizeof (CLIPENUM), (PULONG) &EnumRects8);

                // ITERATE THROUGH ALL RECTANGLES IN THE CURRENT BATCH

                for (i = 0; i < EnumRects8.c; i++)
                {
                    // INTERSECT WITH TARGET BOUNDS

                    if (!FindIntersection (prclDst, &EnumRects8.arcl[i], &destRegion))
                        continue;

                    if (rotatingDst)
                    {
                        // ADJUST PATTERN ORIGIN
                        //
                        if (!solidBrush)
                        {
                            unsigned long patX, patY;
                            if (ppdev->regSettings.RotationAngle == 90)
                            {
                                patX = 4000 + destRegion.top - pptlBrush->y;
                                patY = 4000 + pptlBrush->x - destRegion.right;
                            }
                            else if (ppdev->regSettings.RotationAngle == 270)
                            {
                                patY = 4000 + destRegion.left - pptlBrush->x;
                                patX = 4000 + pptlBrush->y - destRegion.bottom;
                            }
                            else
                            {
                                patX = 4000 - destRegion.right + pptlBrush->x;
                                patY = 4000 - destRegion.bottom + pptlBrush->y;
                            }
                            gp_set_pattern_origin (patX, patY);
                        }
                        gp_declare_blt (0);
                        gp_pattern_fill (ROT_OFFSET (ppdev, destRegion., ppdev->regSettings.RotationAngle),
                            ROT_WIDTH (destRegion., ppdev->regSettings.RotationAngle),
                            ROT_HEIGHT (destRegion., ppdev->regSettings.RotationAngle));
                    }
                    else
                    {
                        // ADJUST PATTERN ORIGIN
                        //
                        if (!solidBrush)
                        {
                            gp_set_pattern_origin (
                                4000 - pptlBrush->x + destRegion.left,
                                4000 - pptlBrush->y + destRegion.top);
                        }
                        destoffset = (dstSurfPitch * destRegion.top) +
                            (destRegion.left << ppdev->cPelSize);
                        destoffset += dstSurfOffset;

                        gp_declare_blt (0);
                        gp_pattern_fill (destoffset,
                            destRegion.right  - destRegion.left,
                            destRegion.bottom - destRegion.top);
                    }
                }

                if (!more)
                    break;
            }
        }
        goto returnTrueNoRotate;
    }

SourceInBlt:

    // PERFORM THE BLT IN TWO-THREE PASSES FOR ROTATION
    // If the BLT requires source or destination data and one of these is
    // the primary surface, then reverse rotate before the BLT begins.  It
    // is possible for both surfaces to be the primary, in which case we
    // could accomplish the BLT in a single pass.  However, experience has
    // shown that this doesn't happen.  Also, note that we rotate around the
    // whole enchilada, instead of around each BLT.  This saves a lot of
    // extra work when performing complex clipping.  However, it implies that
    // we must do a little clipping of our own.
    //
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1)
    {
        RECTL srcRect;
        srcRect.left   = pptlSrc->x;
        srcRect.top    = pptlSrc->y;
        srcRect.right  = pptlSrc->x + prclDst->right - prclDst->left;
        srcRect.bottom = pptlSrc->y + prclDst->bottom - prclDst->top;

        rotSrcPtr = &srcRect;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotSrcPtr = &rotDestRegion;
            if (!FindIntersection (&srcRect, &ppdev->scrnBounds, rotSrcPtr))
                goto returnTrueNoRotate;
        }

        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotSrcPtr->top) + (rotSrcPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }
    if (rotatingDst && ((rop4Flags[(UCHAR)rop4] & LX_ROP_DST) || masked || pco))
    {
        rotDestPtr = prclDst;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclDst, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    // DECODE SOURCE FORMAT
    //
    if ((psoSrc->iBitmapFormat != psoTrg->iBitmapFormat) || (pxlo && !(pxlo->flXlate & XO_TRIVIAL)))
    {
        if (ppdev->cBitsPerPel == 8 && psoSrc->iBitmapFormat != BMF_1BPP)
            goto puntIt;

        colorConversion = 1;

        switch (psoSrc->iBitmapFormat)
        {
            case BMF_1BPP:

                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_mono_source (pxlo->pulXlate[0], pxlo->pulXlate[1], 0);
                break;

            case BMF_4BPP:

                bltFlags |= CIMGP_BLTFLAGS_PRES_LUT;
                gp_program_lut (pxlo->pulXlate, 0);
                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_source_format (CIMGP_SOURCE_FMT_4BPP_INDEXED);
                break;

            case BMF_8BPP:

                bltFlags |= CIMGP_BLTFLAGS_PRES_LUT;
                gp_program_lut (pxlo->pulXlate, 1);
                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_source_format (CIMGP_SOURCE_FMT_8BPP_INDEXED);
                break;

            case BMF_16BPP:

                XLATEOBJ_cGetPalette (pxlo, XO_SRCBITFIELDS, 3, bitfields);
                if      (bitfields[0] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_0_5_6_5;
                else if (bitfields[0] == 0x7C00) srcFormat = CIMGP_SOURCE_FMT_1_5_5_5;
                else if (bitfields[2] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_16BPP_BGR;
                else                             srcFormat = CIMGP_SOURCE_FMT_15BPP_BGR;
                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_source_format (srcFormat);
                break;

            case BMF_24BPP:

                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_source_format (CIMGP_SOURCE_FMT_24BPP);
                break;

            case BMF_32BPP:

                gp_declare_blt (bltFlags);
                gp_set_raster_operation ((UCHAR)rop4);
                gp_set_source_format (CIMGP_SOURCE_FMT_8_8_8_8);
                break;

            default:

                DISPDBG ((ppdev, 1, "Unsupported source format.\n"));
                goto puntIt;
        }
    }
    else
    {
        // SET DEFAULT SOURCE FORMAT.
        // This is used when not color converting a masked BLT.
        //
        if (masked)
            gp_set_source_format (ppdev->cimFormat);

        colorConversion = 0;
        gp_declare_blt (bltFlags);
        gp_set_raster_operation ((UCHAR)rop4);
    }

    // READ DESTINATION AND SOURCE SURFACES
    // Needed because all surfaces are opaque.
    //
    index = dhsurf_array[(USHORT)psoTrg->dhsurf];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntSystem;

    dstSurfOffset = bitmap_heap[(USHORT)index].heap_offset;
    dstSurfPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

    if (psoSrc->dhsurf)
    {
        index = dhsurf_array[(USHORT)psoSrc->dhsurf];
        if (index & CACHE_FLAG_SYSTEM)
        {
            fromFB = 0;
            srcSurfOffset = system_heap[(USHORT)index].heap_offset;
            srcSurfPitch  = system_heap[(USHORT)index].flags_and_pitch;
        }
        else
        {
            fromFB = 1;
            srcSurfOffset = bitmap_heap[(USHORT)index].heap_offset;
            srcSurfPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

            // CHECK CORNER CASE
            // In rare circumstances, an offscreen surface can be passed in with a XlateObj
            // and treated as a color conversion.  In these cases, the srcSurfOffset will 
            // be an offset into the framebuffer and not a true pointer, causing a failure.
            //
            if (colorConversion)
                srcSurfOffset += (ULONG)cim_fb_ptr;
        }
    }
    else
    {
        fromFB = 0;
        srcSurfOffset = (ULONG)psoSrc->pvScan0;
        srcSurfPitch  = (ULONG)psoSrc->lDelta;
    }

    // SET BLT STRIDES
    //
    gp_set_strides (dstSurfPitch, srcSurfPitch);

    // CHECK FOR PATTERN
    //
    if (rop4Flags[(UCHAR)rop4] & LX_ROP_PAT)
    {
        // CHECK FOR SOLID COLOR
        // Color != 0xFFFFFFFF implies solid color instead of pattern.

        if(pbo->iSolidColor != 0xffffffff)
        {
#if DBG
            if (g_puntMask & LX_PUNT_SRCANDSOLIDPAT)
                goto puntIt;
#endif

            // SET SOLID COLOR

            gp_set_solid_pattern (pbo->iSolidColor);
        }
        else
        {
            // CHECK BRUSH TYPE

            if (pbo->pvRbrush != NULL)
                gBrushPtr = pbo->pvRbrush;
            else
            {
                gBrushPtr = BRUSHOBJ_pvGetRbrush (pbo);

                DISPDBG((ppdev, 1, "gBrushPtr returned 0x%x\n", gBrushPtr));

                if (gBrushPtr == NULL)
                    goto puntIt;
            }

#if DBG
            if (((g_puntMask & LX_PUNT_SRCANDMONOPAT)  && gBrushPtr->patFormat == 0xFF) ||
                ((g_puntMask & LX_PUNT_SRCANDCOLORPAT) && gBrushPtr->patFormat != 0xFF))
            {
                goto puntIt;
            }
#endif

            solidBrush = 0;
            if (gBrushPtr->patFormat == 0xFF)
            {
                gp_set_mono_pattern (gBrushPtr->monoColor0, gBrushPtr->monoColor1,
                    gBrushPtr->Pattern[0], gBrushPtr->Pattern[1], 0,
                    4000 - pptlBrush->x + prclDst->left,
                    4000 - pptlBrush->y + prclDst->top);
            }
            else
            {
                // CHECK PUNT CASE
                // GeodeLX cannot do color patterns and color conversion in one BLT.
                //
                if (colorConversion && psoSrc->iBitmapFormat != BMF_1BPP)
                {
                    DISPDBG ((ppdev, 1, "Unable to combine a color pattern with color conversion\n"));
                    goto puntIt;
                }

                bltFlags = CIMGP_BLTFLAGS_PRES_COLOR_PAT;
                gp_write_parameters();
                gp_set_color_pattern (&(gBrushPtr->Pattern[0]), gBrushPtr->patFormat,
                    4000 - pptlBrush->x + prclDst->left,
                    4000 - pptlBrush->y + prclDst->top);
                gp_declare_blt (bltFlags);
            }
        }
    }

#if DBG
    if (colorConversion)
    {
        if (((psoSrc->iBitmapFormat == BMF_1BPP)  && (g_puntMask & LX_PUNT_CONVERTMONO)) ||
            ((psoSrc->iBitmapFormat == BMF_4BPP)  && (g_puntMask & LX_PUNT_CONVERT4BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_8BPP)  && (g_puntMask & LX_PUNT_CONVERT8BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_16BPP) && (g_puntMask & LX_PUNT_CONVERT16BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_24BPP) && (g_puntMask & LX_PUNT_CONVERT24BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_32BPP) && (g_puntMask & LX_PUNT_CONVERT32BPP)))
        {
            DISPDBG ((ppdev, 1, "Punting BLT from %i\n", psoSrc->iBitmapFormat));
            goto puntIt;
        }
    }
    else
    {
        if (g_puntMask & LX_PUNT_NOCONVERT)
        {
            goto puntIt;
        }
    }
#endif

    // CLIPPING
    // Complex-clipped source BLTs default to function pointers.

    x = pptlSrc->x;
    y = pptlSrc->y;

    if (!pco || pco->iDComplexity == DC_TRIVIAL || pco->iDComplexity == DC_RECT)
    {
#if DBG
        if (g_puntMask & LX_PUNT_BLTSIMPLE)
        {
            goto puntIt;
        }
#endif

        if (pco && pco->iDComplexity == DC_RECT)
        {
            clippedTargetPtr = &destRegion;
            if (!FindIntersection (prclDst, &pco->rclBounds, &destRegion))
                goto returnTrueNoRotate;

            // ADJUST PATTERN ORIGIN

            if (!solidBrush)
            {
                gp_set_pattern_origin (
                    4000 - pptlBrush->x + destRegion.left,
                    4000 - pptlBrush->y + destRegion.top);
            }

            x += (destRegion.left - prclDst->left);
            y += (destRegion.top  - prclDst->top);
        }
        else
            clippedTargetPtr = prclDst;

        if (psoSrc->iBitmapFormat == BMF_1BPP)
        {
            gp_mono_bitmap_to_screen_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }
        else if (masked)
        {
            if (psoSrc->dhsurf)
            {
                gp_screen_to_screen_masked (
                    (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                    (y * srcSurfPitch) + (x << ppdev->cPelSize) + srcSurfOffset,
                    clippedTargetPtr->right  - clippedTargetPtr->left,
                    clippedTargetPtr->bottom - clippedTargetPtr->top,
                    pptlMask->x + x - pptlSrc->x,
                    (unsigned char *)psoMask->pvScan0 + ((pptlMask->y + y - pptlSrc->y) * psoMask->lDelta),
                    psoMask->lDelta);
            }
            else
            {
                gp_masked_blt (
                    (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                    clippedTargetPtr->right  - clippedTargetPtr->left,
                    clippedTargetPtr->bottom - clippedTargetPtr->top,
                    pptlMask->x + x - pptlSrc->x,
                    x,
                    (unsigned char *)psoMask->pvScan0 + ((pptlMask->y + y - pptlSrc->y) * psoMask->lDelta),
                    (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                    psoMask->lDelta,
                    srcSurfPitch);
            }
        }
        else if (colorConversion)
        {
            gp_color_convert_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }
        else if (fromFB)
        {
            ULONG flags = 0;
            if (srcSurfOffset == dstSurfOffset)
            {
                if (prclDst->top > y)
                    flags = CIMGP_NEGYDIR;
                else if ((prclDst->top == y) && (prclDst->left > x))
                    flags = CIMGP_NEGXDIR;
            }

            gp_screen_to_screen_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                (y * srcSurfPitch) + (x << ppdev->cPelSize) + srcSurfOffset,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                flags);
        }
        else
        {
            gp_color_bitmap_to_screen_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }
    }
    else
    {
        // COMPLEX CLIPPING

        ULONG enumOrder = CD_LEFTDOWN;
        ULONG flags = 0;
        ULONG tempx, tempy;

#if DBG
        if (g_puntMask & LX_PUNT_BLTCOMPLEX)
        {
            goto puntIt;
        }
#endif

        gp_write_parameters();

        if (psoSrc->dhsurf && srcSurfOffset == dstSurfOffset)
        {
            enumOrder = 0;
            if (prclDst->top > y)
            {
                enumOrder |= CD_UPWARDS;
                flags     |= CIMGP_NEGYDIR;
            }
            if (prclDst->left > x)
            {
                enumOrder |= CD_LEFTWARDS;
                flags     |= CIMGP_NEGXDIR;
            }
        }

        // COMPLEX CLIPPING

        CLIPOBJ_cEnumStart(pco, FALSE, CT_RECTANGLES, enumOrder, 0);

        while (1)
        {
            more = CLIPOBJ_bEnum(pco, sizeof (CLIPENUM), (PULONG) &EnumRects8);

            // ITERATE THROUGH ALL RECTANGLES IN THE CURRENT BATCH

            for (i = 0; i < EnumRects8.c; i++)
            {
                // INTERSECT WITH TARGET BOUNDS

                if (!FindIntersection (prclDst, &EnumRects8.arcl[i], &destRegion))
                    continue;

                // ADJUST PATTERN ORIGIN

                if (!solidBrush)
                {
                    gp_set_pattern_origin (
                        4000 - pptlBrush->x + destRegion.left,
                        4000 - pptlBrush->y + destRegion.top);
                }

                tempx = x + destRegion.left - prclDst->left;
                tempy = y + destRegion.top  - prclDst->top;

                gp_declare_blt (bltFlags);
                if (psoSrc->iBitmapFormat == BMF_1BPP)
                {
                    gp_mono_bitmap_to_screen_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }
                else if (masked)
                {
                    if (psoSrc->dhsurf)
                    {
                        gp_screen_to_screen_masked (
                            (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                            (tempy * srcSurfPitch) + (tempx << ppdev->cPelSize) + srcSurfOffset,
                            destRegion.right  - destRegion.left,
                            destRegion.bottom - destRegion.top,
                            pptlMask->x + tempx - pptlSrc->x,
                            (unsigned char *)psoMask->pvScan0 + ((pptlMask->y + tempy - pptlSrc->y) * psoMask->lDelta),
                            psoMask->lDelta);
                    }
                    else
                    {
                        gp_masked_blt (
                            (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                            destRegion.right  - destRegion.left,
                            destRegion.bottom - destRegion.top,
                            pptlMask->x + tempx - pptlSrc->x,
                            tempx,
                            (unsigned char *)psoMask->pvScan0 + ((pptlMask->y + tempy - pptlSrc->y) * psoMask->lDelta),
                            (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                            psoMask->lDelta,
                            srcSurfPitch);
                    }
                }
                else if (colorConversion)
                {
                    gp_color_convert_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }
                else if (fromFB)
                {
                    gp_screen_to_screen_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        (tempy * srcSurfPitch) + (tempx << ppdev->cPelSize) + srcSurfOffset,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        flags);
                }
                else
                {
                    gp_color_bitmap_to_screen_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }
            }

            if (!more)
                break;
        }
    }

returnTrue:

    // Rotate the result back.  Destination region will already have been specified
    // unless clipping object is not specified and the ROP did not involve destination
    // data.
    //
    if (rotatingDst)
    {
        if (!rotDestPtr)
            rotDestPtr = prclDst;
            
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            rotDestPtr->right - rotDestPtr->left,
            rotDestPtr->bottom - rotDestPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

returnTrueNoRotate:

    // Same as returnTrue, except no need to check rotation
    //
    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG((ppdev, 1, "DrvBitBlt: return\n"));
    return TRUE;

puntSystem:

    DISPDBG ((ppdev, 1, "Punting System Memory BLT\n"));

    // This code assumes 'index' has been setup immediately before jumping to
    // this location.
    //
    index &= 0xFFFF;
    if (!system_heap[index].lockedSurf)
    {
        if (!AddLockedSurfForSystemBitmap (ppdev, index))
            goto returnTrue;
    }
    psoTrg = system_heap[index].lockedSurf;

    if (psoSrc && psoSrc->dhsurf)
    {
        MAKE_PUNTABLE(psoSrc);
    }

    gp_wait_until_idle();

    EngBitBlt (psoTrg, psoSrc, psoMask, pco, pxlo, prclDst, pptlSrc,
        pptlMask, pbo, pptlBrush, rop4);

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;

reversePuntIt:

    DISPDBG ((ppdev, 1, "DrvBitBlt Reverse Punt\n"));

    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1 &&
        !g_puntAllUntilModeSet)
    {
        RECTL tempRegion;     
        tempRegion.left   = pptlSrc->x;
        tempRegion.top    = pptlSrc->y;
        tempRegion.right  = pptlSrc->x + prclDst->right  - prclDst->left;
        tempRegion.bottom = pptlSrc->y + prclDst->bottom - prclDst->top;

        rotSrcPtr = &tempRegion;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotSrcPtr = &rotDestRegion;
            if (!FindIntersection (&tempRegion, &ppdev->scrnBounds, rotSrcPtr))
                goto returnTrueNoRotate;
        }

        // Reverse rotate the screen before punting to software
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotSrcPtr->top) + (rotSrcPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    // RETRIEVE LOCKED SURFACE
    //
    MAKE_PUNTABLE(psoSrc);

    gp_wait_until_idle();

    EngBitBlt (psoTrg, psoSrc, psoMask, pco, pxlo, prclDst, pptlSrc,
        pptlMask, pbo, pptlBrush, rop4);

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;

puntIt:

    DISPDBG ((ppdev, 1, "DrvBitBlt Punt\n"));

    if (ppdev->regSettings.RotationAngle != 0 && psoSrc && (USHORT)psoSrc->dhsurf == 1 &&
        !g_puntAllUntilModeSet)
    {
        RECTL tempRegion;     
        tempRegion.left   = pptlSrc->x;
        tempRegion.top    = pptlSrc->y;
        tempRegion.right  = pptlSrc->x + prclDst->right  - prclDst->left;
        tempRegion.bottom = pptlSrc->y + prclDst->bottom - prclDst->top;

        rotSrcPtr = &tempRegion;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotSrcPtr = &rotDestRegion;
            if (!FindIntersection (&tempRegion, &ppdev->scrnBounds, rotSrcPtr))
                goto returnTrueNoRotate;
        }

        // Reverse rotate the screen before punting to software
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotSrcPtr->top) + (rotSrcPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoTrg->dhsurf == 1 &&
        !g_puntAllUntilModeSet)
    {
        rotDestPtr = prclDst;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclDst, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }

        // Reverse rotate the destination before punting to software.  This only
        // needs to be done if the ROP involves destination.  However, this case
        // should be rare, and a blanket rotate saves the worry about missing
        // some cases, like quarternary ROPS.
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    dhsurfTrg = (USHORT)psoTrg->dhsurf;
    MAKE_PUNTABLE(psoTrg);

    if (psoSrc && psoSrc->dhsurf)
    {
        MAKE_PUNTABLE(psoSrc);
    }

    gp_wait_until_idle();

    // CALL ENGBITBLT

    EngBitBlt (psoTrg, psoSrc, psoMask, pco, pxlo, prclDst, pptlSrc,
        pptlMask, pbo, pptlBrush, rop4);

    if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 &&
        !g_puntAllUntilModeSet)
    {
        // Rotate the screen after punting to software.  The destination region
        // was calculated before we punted.
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            rotDestPtr->right - rotDestPtr->left,
            rotDestPtr->bottom - rotDestPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// DrvCopyBits
//
// This function handles a variety of data copies from memory to screen,
// screen to memory, and screen to screen.
//
BOOL DrvCopyBits(
    SURFOBJ *psoTrg,  // target surface
    SURFOBJ *psoSrc,  // source surface
    CLIPOBJ *pco,     // clip object
    XLATEOBJ *pxlo,   // translation object - color translation tables
    RECTL *prclDst,   // target rectangle
    POINTL *pptlSrc)  // origin of source
{
    PDEV       *ppdev = NULL;
    RECTL      *clippedTargetPtr;
    CLIPENUM    EnumRects8;
    RECTL       destRegion;
    int         i, more;
    long        x, y;
    ULONG       srcFormat;
    ULONG       colorConversion;
    ULONG       dstSurfOffset, srcSurfOffset;
    long        dstSurfPitch,  srcSurfPitch;
    ULONG       index, fromFB;
    ULONG       bltFlags;
    ULONG       bitfields[3];
    int         rotatingDst;
    USHORT      dhsurfTrg;
    USHORT      dhsurfSrc;

    DISPDBG ((ppdev, 1, "DrvCopyBits Entry\n"));

    ACQUIRE_PRIMITIVE_SEMAPHORE;

    if (!psoTrg->dhsurf)
    {
        ppdev = (PDEV *)psoSrc->dhpdev;
        goto reversePuntIt;
    }
    ppdev = (PDEV *)psoTrg->dhpdev;

    if (g_puntAllUntilModeSet)
        goto puntIt;

    // Check for rotation
    //
    dhsurfTrg = (USHORT)psoTrg->dhsurf;
    rotatingDst = (dhsurfTrg == 1) && (ppdev->regSettings.RotationAngle != 0);

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

    // DECODE SOURCE FORMAT
    //
    bltFlags = 0;
    if ((psoSrc->iBitmapFormat != psoTrg->iBitmapFormat) || (pxlo && !(pxlo->flXlate & XO_TRIVIAL)))
    {
        if (ppdev->cBitsPerPel == 8 && psoSrc->iBitmapFormat != BMF_1BPP)
            goto puntIt;

        colorConversion = 1;

        // Configure the source format.  Note that we will inevitably have to reprogram the
        // source format when rotating and performing a complex BLT.
        //
        if (!ConfigureLookup (psoSrc, pxlo, &bltFlags, 0xCC))
            goto puntIt;
    }
    else
    {
        colorConversion = 0;
        gp_declare_blt (0);
        gp_set_raster_operation (0xCC);
    }

    // READ DESTINATION AND SOURCE INFORMATION
    // Needed because all device surfaces are opaque.
    //
    index = dhsurf_array[dhsurfTrg];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntSystem;

    dstSurfOffset = bitmap_heap[(USHORT)index].heap_offset;
    dstSurfPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

    if (psoSrc->dhsurf)
    {
        index = dhsurf_array[(USHORT)psoSrc->dhsurf];
        if (index & CACHE_FLAG_SYSTEM)
        {
            fromFB = 0;
            srcSurfOffset = system_heap[(USHORT)index].heap_offset;
            srcSurfPitch  = system_heap[(USHORT)index].flags_and_pitch;
        }
        else
        {
            fromFB = 1;
            srcSurfOffset = bitmap_heap[(USHORT)index].heap_offset;
            srcSurfPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

            // CHECK CORNER CASE
            // In rare circumstances, an offscreen surface can be passed in with a XlateObj
            // and treated as a color conversion.  In these cases, the srcSurfOffset will 
            // be an offset into the framebuffer and not a true pointer, causing a failure.
            //
            if (colorConversion)
                srcSurfOffset += (ULONG)cim_fb_ptr;
        }
    }
    else
    {
        fromFB = 0;
        srcSurfOffset = (ULONG)psoSrc->pvScan0;
        srcSurfPitch  = (ULONG)psoSrc->lDelta;
    }

    // SET BLT STRIDES
    //
    gp_set_strides (dstSurfPitch, srcSurfPitch);

#if DBG
    if (colorConversion)
    {
        if (((psoSrc->iBitmapFormat == BMF_1BPP)  && (g_puntMask & LX_PUNT_COPYMONO)) ||
            ((psoSrc->iBitmapFormat == BMF_4BPP)  && (g_puntMask & LX_PUNT_COPY4BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_8BPP)  && (g_puntMask & LX_PUNT_COPY8BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_16BPP) && (g_puntMask & LX_PUNT_COPY16BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_24BPP) && (g_puntMask & LX_PUNT_COPY24BPP)) ||
            ((psoSrc->iBitmapFormat == BMF_32BPP) && (g_puntMask & LX_PUNT_COPY32BPP)))
        {
            goto puntIt;
        }
    }
    else
    {
        if (g_puntMask & LX_PUNT_COPYSIMPLE)
            goto puntIt;
    }
#endif

    // CLIPPING

    x = pptlSrc->x;
    y = pptlSrc->y;

    if (!pco || pco->iDComplexity == DC_TRIVIAL || pco->iDComplexity == DC_RECT)
    {
        // Requires a rotation after the main BLT
        //
        int twoPass = 0;
#if DBG
        if (g_puntMask & LX_PUNT_BLTSIMPLE)
            goto puntIt;
#endif

        if (pco && pco->iDComplexity == DC_RECT)
        {
            clippedTargetPtr = &destRegion;
            if (!FindIntersection (prclDst, &pco->rclBounds, &destRegion))
                goto returnTrue;

            x += (destRegion.left - prclDst->left);
            y += (destRegion.top  - prclDst->top);
        }
        else
            clippedTargetPtr = prclDst;

        if (psoSrc->iBitmapFormat == BMF_1BPP)
        {
            if (rotatingDst)
                twoPass = 1;

            gp_mono_bitmap_to_screen_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }
        else if (colorConversion)
        {
            if (rotatingDst)
                twoPass = 1;

            gp_color_convert_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }
        else if (fromFB)
        {
            // FINALLY SOMETHING WE CAN ROTATE!
            // Operation has both source and destination in the frame buffer.
            // There are three scenarios here:
            //   1. Dest is primary display - rotate to dest
            //   2. Src is primary display - rotate from src
            //   3. Both are primary display - adjust coordinates and flags
            //
            if (rotatingDst || ((int)psoSrc->dhsurf == 1 && ppdev->regSettings.RotationAngle != 0))
            {
                if (srcSurfOffset == dstSurfOffset)
                {
                    ULONG flags = 0;
                    ULONG width, height, sx, sy, dx, dy;

                    // BLT is from x,y to clippedTargetPtr->left,clippedTargetPtr->top
                    //
                    if (ppdev->regSettings.RotationAngle == 90)
                    {
                        sx = y;
                        dx = clippedTargetPtr->top;
                        sy = ppdev->cyScreen - (x + clippedTargetPtr->right - clippedTargetPtr->left);
                        dy = ppdev->cyScreen - clippedTargetPtr->right;
                        width  = clippedTargetPtr->bottom - clippedTargetPtr->top;
                        height = clippedTargetPtr->right - clippedTargetPtr->left;
                    }
                    else if (ppdev->regSettings.RotationAngle == 270)
                    {
                        sy = x;
                        dy = clippedTargetPtr->left;
                        sx = ppdev->cxScreen - (y + clippedTargetPtr->bottom - clippedTargetPtr->top);
                        dx = ppdev->cxScreen - clippedTargetPtr->bottom;
                        width  = clippedTargetPtr->bottom - clippedTargetPtr->top;
                        height = clippedTargetPtr->right - clippedTargetPtr->left;
                    }
                    else
                    {
                        dx = ppdev->cxScreen - clippedTargetPtr->right;
                        dy = ppdev->cyScreen - clippedTargetPtr->bottom;
                        sx = ppdev->cxScreen - (x + clippedTargetPtr->right  - clippedTargetPtr->left);
                        sy = ppdev->cyScreen - (y + clippedTargetPtr->bottom - clippedTargetPtr->top);
                        width  = clippedTargetPtr->right - clippedTargetPtr->left;
                        height = clippedTargetPtr->bottom - clippedTargetPtr->top;
                    }
                    if (dy > sy)
                        flags = CIMGP_NEGYDIR;
                    else if ((dy == sy) && (dx > sx))
                        flags = CIMGP_NEGXDIR;
                    gp_set_strides (mode_pitch, mode_pitch);
                    gp_screen_to_screen_blt (
                        (dy * mode_pitch) + (dx << ppdev->cPelSize),
                        (sy * mode_pitch) + (sx << ppdev->cPelSize),
                        width, height, flags);
                }
                else if (rotatingDst)
                {
                    gp_set_strides (mode_pitch, srcSurfPitch);
                    gp_set_source_format (ppdev->cimFormat);
                    gp_rotate_blt (ROT_OFFSET(ppdev, clippedTargetPtr->, ppdev->regSettings.RotationAngle),
                        (y * srcSurfPitch) + (x << ppdev->cPelSize) + srcSurfOffset,
                        clippedTargetPtr->right  - clippedTargetPtr->left,
                        clippedTargetPtr->bottom - clippedTargetPtr->top,
                        360 - ppdev->regSettings.RotationAngle);
                }
                else
                {
                    RECTL src;
                    src.left   = x;
                    src.top    = y;
                    src.right  = x + clippedTargetPtr->right - clippedTargetPtr->left;
                    src.bottom = y + clippedTargetPtr->bottom - clippedTargetPtr->top;
                    gp_set_strides (dstSurfPitch, mode_pitch);
                    gp_set_source_format (ppdev->cimFormat);
                    gp_rotate_blt (
                        (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                        ROT_OFFSET(ppdev, src., ppdev->regSettings.RotationAngle),
                        ROT_WIDTH (clippedTargetPtr->, ppdev->regSettings.RotationAngle),
                        ROT_HEIGHT (clippedTargetPtr->, ppdev->regSettings.RotationAngle),
                        ppdev->regSettings.RotationAngle);
                }
            }
            else
            {
                ULONG flags = 0;
                if (srcSurfOffset == dstSurfOffset)
                {
                    if (prclDst->top > y)
                        flags = CIMGP_NEGYDIR;
                    else if ((prclDst->top == y) && (prclDst->left > x))
                        flags = CIMGP_NEGXDIR;
                }

                gp_screen_to_screen_blt (
                    (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                    (y * srcSurfPitch) + (x << ppdev->cPelSize) + srcSurfOffset,
                    clippedTargetPtr->right  - clippedTargetPtr->left,
                    clippedTargetPtr->bottom - clippedTargetPtr->top,
                    flags);
            }
        }
        else
        {
            if (rotatingDst)
                twoPass = 1;

            gp_color_bitmap_to_screen_blt (
                (clippedTargetPtr->top * dstSurfPitch) + (clippedTargetPtr->left << ppdev->cPelSize) + dstSurfOffset,
                x,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                (unsigned char *)srcSurfOffset + (y * srcSurfPitch),
                srcSurfPitch);
        }

        if (twoPass)
        {
            // Rotate the result
            //
            gp_declare_blt (0);
            gp_set_source_format (ppdev->cimFormat);
            gp_set_strides (mode_pitch, rot_pitch);
            gp_rotate_blt (ROT_OFFSET(ppdev, clippedTargetPtr->, ppdev->regSettings.RotationAngle),
                (clippedTargetPtr->left << ppdev->cPelSize) + (clippedTargetPtr->top * rot_pitch) + rot_offset,
                clippedTargetPtr->right  - clippedTargetPtr->left,
                clippedTargetPtr->bottom - clippedTargetPtr->top,
                360 - ppdev->regSettings.RotationAngle);
        }
    }
    else
    {
        // COMPLEX CLIPPING

        ULONG enumOrder = CD_LEFTDOWN;
        ULONG flags = 0;
        ULONG tempx, tempy;

#if DBG
        if (g_puntMask & LX_PUNT_BLTCOMPLEX)
            goto puntIt;
#endif

        gp_write_parameters();

        if (fromFB && srcSurfOffset == dstSurfOffset)
        {
            enumOrder = 0;
            if (prclDst->top > y)
            {
                enumOrder |= CD_UPWARDS;
                flags     |= CIMGP_NEGYDIR;
            }
            if (prclDst->left > x)
            {
                enumOrder |= CD_LEFTWARDS;
                flags     |= CIMGP_NEGXDIR;
            }
        }

        // COMPLEX CLIPPING

        CLIPOBJ_cEnumStart(pco, FALSE, CT_RECTANGLES, enumOrder, 0);

        while (1)
        {
            // Requires an extra copy after the BLT
            //
            int twoPass = 0;

            more = CLIPOBJ_bEnum(pco, sizeof (CLIPENUM), (PULONG) &EnumRects8);

            // ITERATE THROUGH ALL RECTANGLES IN THE CURRENT BATCH

            for (i = 0; i < EnumRects8.c; i++)
            {
                // INTERSECT WITH TARGET BOUNDS

                if (!FindIntersection (prclDst, &EnumRects8.arcl[i], &destRegion))
                    continue;

                tempx = x + destRegion.left - prclDst->left;
                tempy = y + destRegion.top  - prclDst->top;

                gp_declare_blt (bltFlags);
                if (psoSrc->iBitmapFormat == BMF_1BPP)
                {
                    if (rotatingDst)
                    {
                        ConfigureLookup (psoSrc, pxlo, &bltFlags, 0xCC);
                        gp_set_strides (dstSurfPitch, 0);
                        twoPass = 1;
                    }
                    gp_mono_bitmap_to_screen_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }
                else if (colorConversion)
                {
                    if (rotatingDst)
                    {
                        ConfigureLookup (psoSrc, pxlo, &bltFlags, 0xCC);
                        gp_set_strides (dstSurfPitch, 0);
                        twoPass = 1;
                    }

                    gp_color_convert_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }
                else if (fromFB)
                {
                    if (rotatingDst || ((int)psoSrc->dhsurf == 1 && ppdev->regSettings.RotationAngle != 0))
                    {
                        if (srcSurfOffset == dstSurfOffset)
                        {
                            ULONG flags = 0;
                            ULONG width, height, sx, sy, dx, dy;

                            // BLT is from x,y to clippedTargetPtr->left,clippedTargetPtr->top
                            //
                            if (ppdev->regSettings.RotationAngle == 90)
                            {
                                sx = tempy;
                                dx = destRegion.top;
                                sy = ppdev->cyScreen - (tempx + destRegion.right - destRegion.left);
                                dy = ppdev->cyScreen - destRegion.right;
                                width  = destRegion.bottom - destRegion.top;
                                height = destRegion.right - destRegion.left;
                            }
                            else if (ppdev->regSettings.RotationAngle == 270)
                            {
                                sy = tempx;
                                dy = destRegion.left;
                                sx = ppdev->cxScreen - (tempy + destRegion.bottom - destRegion.top);
                                dx = ppdev->cxScreen - destRegion.bottom;
                                width  = destRegion.bottom - destRegion.top;
                                height = destRegion.right - destRegion.left;
                            }
                            else
                            {
                                dx = ppdev->cxScreen - destRegion.right;
                                dy = ppdev->cyScreen - destRegion.bottom;
                                sx = ppdev->cxScreen - (tempx + destRegion.right  - destRegion.left);
                                sy = ppdev->cyScreen - (tempy + destRegion.bottom - destRegion.top);
                                width  = destRegion.right - destRegion.left;
                                height = destRegion.bottom - destRegion.top;
                            }
                            if (dy > sy)
                                flags = CIMGP_NEGYDIR;
                            else if ((dy == sy) && (dx > sx))
                                flags = CIMGP_NEGXDIR;
                            gp_set_strides (mode_pitch, mode_pitch);
                            gp_screen_to_screen_blt (
                                (dy * mode_pitch) + (dx << ppdev->cPelSize),
                                (sy * mode_pitch) + (sx << ppdev->cPelSize),
                                width, height, flags);
                        }
                        else if (rotatingDst)
                        {
                            gp_set_strides (mode_pitch, srcSurfPitch);
                            gp_set_source_format (ppdev->cimFormat);
                            gp_rotate_blt (ROT_OFFSET(ppdev, destRegion., ppdev->regSettings.RotationAngle),
                                (tempy * srcSurfPitch) + (tempx << ppdev->cPelSize) + srcSurfOffset,
                                destRegion.right  - destRegion.left,
                                destRegion.bottom - destRegion.top,
                                360 - ppdev->regSettings.RotationAngle);
                        }
                        else
                        {
                            RECTL src;
                            src.left   = tempx;
                            src.top    = tempy;
                            src.right  = tempx + destRegion.right - destRegion.left;
                            src.bottom = tempy + destRegion.bottom - destRegion.top;
                            gp_set_strides (dstSurfPitch, mode_pitch);
                            gp_set_source_format (ppdev->cimFormat);
                            gp_rotate_blt (
                                (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                                ROT_OFFSET(ppdev, src., ppdev->regSettings.RotationAngle),
                                ROT_WIDTH (destRegion., ppdev->regSettings.RotationAngle),
                                ROT_HEIGHT (destRegion., ppdev->regSettings.RotationAngle),
                                ppdev->regSettings.RotationAngle);
                        }
                    }
                    else
                    {
                        gp_screen_to_screen_blt (
                            (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                            (tempy * srcSurfPitch) + (tempx << ppdev->cPelSize) + srcSurfOffset,
                            destRegion.right  - destRegion.left,
                            destRegion.bottom - destRegion.top,
                            flags);
                    }
                }
                else
                {
                    if (rotatingDst)
                    {
                        gp_set_strides (dstSurfPitch, 0);
                        twoPass = 1;
                    }
                    gp_color_bitmap_to_screen_blt (
                        (destRegion.top * dstSurfPitch) + (destRegion.left << ppdev->cPelSize) + dstSurfOffset,
                        tempx,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        (unsigned char *)srcSurfOffset + (tempy * srcSurfPitch),
                        srcSurfPitch);
                }

                if (twoPass)
                {
                    // Rotate the result
                    //
                    gp_declare_blt (0);
                    gp_set_source_format (ppdev->cimFormat);
                    gp_set_strides (mode_pitch, rot_pitch);
                    gp_rotate_blt (ROT_OFFSET(ppdev, destRegion., ppdev->regSettings.RotationAngle),
                        (destRegion.left << ppdev->cPelSize) + (destRegion.top * rot_pitch) + rot_offset,
                        destRegion.right  - destRegion.left,
                        destRegion.bottom - destRegion.top,
                        360 - ppdev->regSettings.RotationAngle);
                }
            }

            if (!more)
                break;
        }
    }

returnTrue:

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 1, "DrvCopyBits Exit\n"));
    return TRUE;

puntSystem:

    DISPDBG ((ppdev, 1, "Punting System Memory BLT\n"));

    // This code assumes 'index' has been setup immediately before jumping to
    // this location.
    //
    index &= 0xFFFF;
    if (!system_heap[index].lockedSurf)
    {
        if (!AddLockedSurfForSystemBitmap (ppdev, index))
            goto returnTrue;
    }
    psoTrg = system_heap[index].lockedSurf;

    if (psoSrc->dhsurf)
    {
        MAKE_PUNTABLE(psoSrc);
    }

    gp_wait_until_idle();

    EngCopyBits (psoTrg, psoSrc, pco, pxlo,
        prclDst, pptlSrc);

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;

puntIt:

    DISPDBG ((ppdev, 1, "DrvCopyBits Punt\n"));

    dhsurfTrg = (USHORT)psoTrg->dhsurf;
    MAKE_PUNTABLE(psoTrg);

    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1 &&
        !g_puntAllUntilModeSet)
    {
        RECTL tempRegion;
        RECTL rotDestRegion;
        RECTL *rotSrcPtr;        
        tempRegion.left   = pptlSrc->x;
        tempRegion.top    = pptlSrc->y;
        tempRegion.right  = pptlSrc->x + prclDst->right  - prclDst->left;
        tempRegion.bottom = pptlSrc->y + prclDst->bottom - prclDst->top;

        rotSrcPtr = &tempRegion;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotSrcPtr = &rotDestRegion;
            if (!FindIntersection (&tempRegion, &ppdev->scrnBounds, rotSrcPtr))
                goto returnTrue;
        }

        // Reverse rotate the screen before punting to software
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotSrcPtr->top) + (rotSrcPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    if (psoSrc->dhsurf)
    {
        MAKE_PUNTABLE(psoSrc);
    }

    gp_wait_until_idle();

    // CALL ENGCOPYBITS

    EngCopyBits (psoTrg, psoSrc, pco, pxlo,
        prclDst, pptlSrc);

    if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 && !g_puntAllUntilModeSet)
    {
        RECTL *rotDestPtr = prclDst;
        RECTL  rotDestRegion;

        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclDst, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrue;
        }

        // Rotate the screen after punting to software
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            rotDestPtr->right - rotDestPtr->left,
            rotDestPtr->bottom - rotDestPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;

reversePuntIt:

    DISPDBG ((ppdev, 1, "DrvCopyBits Reverse Punt\n"));

    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1 &&
        !g_puntAllUntilModeSet)
    {
        RECTL tempRegion;
        RECTL rotDestRegion;
        RECTL *rotSrcPtr;        
        tempRegion.left   = pptlSrc->x;
        tempRegion.top    = pptlSrc->y;
        tempRegion.right  = pptlSrc->x + prclDst->right  - prclDst->left;
        tempRegion.bottom = pptlSrc->y + prclDst->bottom - prclDst->top;

        rotSrcPtr = &tempRegion;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotSrcPtr = &rotDestRegion;
            if (!FindIntersection (&tempRegion, &ppdev->scrnBounds, rotSrcPtr))
                goto returnTrue;
        }

        // Reverse rotate the screen before punting to software
        //
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotSrcPtr->top) + (rotSrcPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotSrcPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    // RETRIEVE LOCKED SURFACE
    //
    MAKE_PUNTABLE(psoSrc);

    gp_wait_until_idle();

    EngCopyBits(psoTrg, psoSrc, pco, pxlo, prclDst, pptlSrc);

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// RotateMonoPattern
//
// Rotates a mono pattern before it is programmed into the GP.
//
void RotateMonoPattern (PDEV *ppdev, ULONG data0_in, ULONG data1_in,
    ULONG *data0_out, ULONG *data1_out)
{
    unsigned char bytes_out[8];
    unsigned long datain[2];
    unsigned char *bytes_in;
    int i, j;

    datain[0] = data0_in;
    datain[1] = data1_in;
    bytes_in = (unsigned char *)&datain[0];

    if (ppdev->regSettings.RotationAngle == 90)
    {
        for (i = 0 ; i < 8; i++)
        {
            bytes_out[i] = 0;
            for (j = 0; j < 8; j++)
            {
                bytes_out[i] <<= 1;
                bytes_out[i] |= bytes_in[j] & 1;
            }
            // Shift the dword shifts all bytes together
            //
            datain[0] >>= 1;
            datain[1] >>= 1;
        }
    }
    else if (ppdev->regSettings.RotationAngle == 270)
    {
        for (i = 0 ; i < 8; i++)
        {
            bytes_out[i] = 0;
            for (j = 0; j < 8; j++)
            {
                bytes_out[i] >>= 1;
                bytes_out[i] |= bytes_in[j] & 0x80;
            }
            // Shift the dword shifts all bytes together
            //
            datain[0] <<= 1;
            datain[1] <<= 1;
        }
    }
    else
    {
        for (i = 0 ; i < 8; i++)
        {
            bytes_out[i] = 0;
            for (j = 0; j < 8; j++)
            {
                bytes_out[i] |= ((bytes_in[7 - i] >> j) & 1) << (7 - j);
            }
        }
    }

    *data0_out = *((unsigned long *)(&bytes_out[0]));
    *data1_out = *((unsigned long *)(&bytes_out[0]) + 1);
}

/////////////////////////////////////////////////////////////////////////
// RotateColorPattern
//
// Rotates a color pattern before it is programmed into the GP.
//
void RotateColorPattern (PDEV *ppdev, ULONG *src, ULONG *dst, ULONG format)
{
    if (format == CIMGP_SOURCE_FMT_8_8_8_8)
    {
        int x, y;
        unsigned long *tmp_src;
        switch (ppdev->regSettings.RotationAngle)
        {
            case 270:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = &src[y];
                    for (x = 0; x < 8; x++)
                        dst[7 - x] = tmp_src[x << 3];
                    dst += 8;
                }
                break;

            case 180:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = &src[(7 - y) << 3];
                    for (x = 0; x < 8; x++)
                        dst[x] = tmp_src[7 - x];
                    dst += 8;
                }
                break;

            case 90:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = &src[7 - y];
                    for (x = 0; x < 8; x++)
                        dst[x] = tmp_src[x << 3];
                    dst += 8;
                }
                break;
        }
    }
    else
    {
        int x, y;
        unsigned short *tmp_src;
        unsigned short *tmp_dst = (unsigned short *)dst;
        switch (ppdev->regSettings.RotationAngle)
        {
            case 270:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = (unsigned short *)src + y;
                    for (x = 0; x < 8; x++)
                        tmp_dst[7 - x] = tmp_src[x << 3];
                    tmp_dst += 8;
                }
                break;

            case 180:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = (unsigned short *)src + ((7 - y) << 3);
                    for (x = 0; x < 8; x++)
                        tmp_dst[x] = tmp_src[7 - x];
                    tmp_dst += 8;
                }
                break;

            case 90:

                for (y = 0; y < 8; y++)
                {
                    tmp_src = (unsigned short *)src + 7 - y;
                    for (x = 0; x < 8; x++)
                        tmp_dst[x] = tmp_src[x << 3];
                    tmp_dst += 8;
                }
                break;
        }
    }
}

/////////////////////////////////////////////////////////////////////////
// ConfigureLookup
//
// Prepares for a color conversion operation.
//
int ConfigureLookup (SURFOBJ *psoSrc, XLATEOBJ *pxlo, ULONG *bltFlags, UCHAR rop)
{
    ULONG bitfields[3];
    ULONG srcFormat;
    PDEV *ppdev = NULL;

    switch (psoSrc->iBitmapFormat)
    {
        case BMF_1BPP:

            gp_declare_blt (0);
            gp_set_raster_operation (rop);
            gp_set_mono_source (pxlo->pulXlate[0], pxlo->pulXlate[1], 0);
            break;

        case BMF_4BPP:

            *bltFlags = CIMGP_BLTFLAGS_PRES_LUT;
            gp_program_lut (pxlo->pulXlate, 0);
            gp_declare_blt (CIMGP_BLTFLAGS_PRES_LUT);
            gp_set_raster_operation (rop);
            gp_set_source_format (CIMGP_SOURCE_FMT_4BPP_INDEXED);
            break;

        case BMF_8BPP:

            *bltFlags = CIMGP_BLTFLAGS_PRES_LUT;
            gp_program_lut (pxlo->pulXlate, 1);
            gp_declare_blt (CIMGP_BLTFLAGS_PRES_LUT);
            gp_set_raster_operation (rop);
            gp_set_source_format (CIMGP_SOURCE_FMT_8BPP_INDEXED);
            break;

        case BMF_16BPP:

            XLATEOBJ_cGetPalette (pxlo, XO_SRCBITFIELDS, 3, bitfields);
            if      (bitfields[0] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_0_5_6_5;
            else if (bitfields[0] == 0x7C00) srcFormat = CIMGP_SOURCE_FMT_1_5_5_5;
            else if (bitfields[2] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_16BPP_BGR;
            else                             srcFormat = CIMGP_SOURCE_FMT_15BPP_BGR;
            gp_declare_blt (0);
            gp_set_raster_operation (rop);
            gp_set_source_format (srcFormat);
            break;

        case BMF_24BPP:

            gp_declare_blt (0);
            gp_set_raster_operation (rop);
            gp_set_source_format (CIMGP_SOURCE_FMT_24BPP);
            break;

        case BMF_32BPP:

            gp_declare_blt (0);
            gp_set_raster_operation (rop);
            gp_set_source_format (CIMGP_SOURCE_FMT_8_8_8_8);
            break;

        default:

            DISPDBG ((ppdev, 1, "Unsupported source format.\n"));
            return 0;
    }

    return 1;
}
