//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// DirectDraw base routines.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"

extern unsigned long ddraw_invalid;
extern unsigned long ddraw_count;

/////////////////////////////////////////////////////////////////////////
// DrvGetDirectDrawInfo
//
// Will be called twice before DrvEnableDirectDraw is called.
//
BOOL DrvGetDirectDrawInfo(
	DHPDEV			dhpdev,
	DD_HALINFO* 	pHalInfo,
	DWORD*			pdwNumHeaps,
	VIDEOMEMORY*	pvmList,			// Will be NULL on first call
	DWORD*			pdwNumFourCC,
	DWORD*			pdwFourCC)			// Will be NULL on first call
{
	PDEV *ppdev = NULL;
	
	ppdev = (PDEV*) dhpdev;

	DISPDBG((ppdev, 300, "DrvGetDirectDrawInfo\n"));

	pHalInfo->dwSize = sizeof(*pHalInfo);

	// FILL IN HALINFO
	// Current primary surface attributes.	Since HalInfo is zero-initialized
	// by GDI, we only have to fill in the fields which should be non-zero:

	pHalInfo->vmiData.pvPrimary 	  = cim_fb_ptr;
	pHalInfo->vmiData.dwDisplayWidth  = ppdev->cxScreen;
	pHalInfo->vmiData.dwDisplayHeight = ppdev->cyScreen;
	pHalInfo->vmiData.lDisplayPitch   = vg_get_display_pitch();

	pHalInfo->vmiData.ddpfDisplay.dwSize  = sizeof(DDPIXELFORMAT);
	pHalInfo->vmiData.ddpfDisplay.dwFlags = DDPF_RGB;

	pHalInfo->vmiData.ddpfDisplay.dwRGBBitCount = ppdev->cBitsPerPel;

	if (ppdev->iBitmapFormat == BMF_8BPP)
		pHalInfo->vmiData.ddpfDisplay.dwFlags |= DDPF_PALETTEINDEXED8;

	// RGB MASKS
	// These masks will be zero at 8 BPP

	pHalInfo->vmiData.ddpfDisplay.dwRBitMask = ppdev->flRed;
	pHalInfo->vmiData.ddpfDisplay.dwGBitMask = ppdev->flGreen;
	pHalInfo->vmiData.ddpfDisplay.dwBBitMask = ppdev->flBlue;

    // FORCE DWORD ALIGNMENT FOR OFFSCREEN SURFACES

	pHalInfo->vmiData.dwOffscreenAlign = 4;
    pHalInfo->vmiData.dwOverlayAlign   = 32;

	// SELF-MANAGED HEAP FOR OFFSCREEN MEMORY

	*pdwNumHeaps = 0;

    if (pvmList)
    {
        DISPDBG ((ppdev, 3000, "Heap Pointer is non-zero even though heap count is zero!!!\n"));
        pvmList->dwFlags           = VIDMEM_ISLINEAR;
        pvmList->fpStart           = heap_start;
        pvmList->fpEnd             = heap_end - 1;
        pvmList->ddsCaps.dwCaps    = 0;
        pvmList->ddsCapsAlt.dwCaps = 0;
    }

	// --- CORE CAPABILITIES ---

	pHalInfo->ddCaps.dwCaps =

           // GRAPHICS

           DDCAPS_BLT
         | DDCAPS_BLTCOLORFILL
         | DDCAPS_BLTQUEUE
	     | DDCAPS_COLORKEY
         | DDCAPS_COLORKEYHWASSIST
         | DDCAPS_READSCANLINE
         | DDCAPS_VBI

         // VIDEO

         | DDCAPS_OVERLAY
         | DDCAPS_OVERLAYSTRETCH
         | DDCAPS_OVERLAYFOURCC
         | DDCAPS_OVERLAYCANTCLIP
         | DDCAPS_COLORKEYHWASSIST;

    // --- COLOR KEY CAPABILITIES ---

	pHalInfo->ddCaps.dwCKeyCaps = DDCKEYCAPS_SRCBLT
                                //| DDCKEYCAPS_DESTBLT
                                | DDCKEYCAPS_DESTOVERLAY
                                | DDCKEYCAPS_DESTOVERLAYONEACTIVE
                                | DDCKEYCAPS_NOCOSTOVERLAY
                                | DDCKEYCAPS_SRCOVERLAY
                                | DDCKEYCAPS_SRCOVERLAYONEACTIVE
                                | DDCKEYCAPS_SRCOVERLAYYUV;

    // --- SECONDARY CAPABILITIES ---

    pHalInfo->ddCaps.dwCaps2 =

           // GRAPHICS

           DDCAPS2_COPYFOURCC;

    if (g_vipHardware)
    {
        pHalInfo->ddCaps.dwCaps2 |=
            
            // VIDEOPORT

             DDCAPS2_CANBOBINTERLEAVED
           | DDCAPS2_CANBOBNONINTERLEAVED
           | DDCAPS2_CANFLIPODDEVEN
           | DDCAPS2_VIDEOPORT;
    }           

    // --- SURFACE CAPABILITIES ---

	pHalInfo->ddCaps.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN
									| DDSCAPS_PRIMARYSURFACE
									| DDSCAPS_FLIP
                                    | DDSCAPS_OVERLAY
									| DDSCAPS_COMPLEX;

    if (g_vipHardware)
    {
        pHalInfo->ddCaps.ddsCaps.dwCaps |= DDSCAPS_VIDEOPORT;
    }

    // --- STRETCHING CAPABILITIES ---

    pHalInfo->ddCaps.dwFXCaps = DDFXCAPS_OVERLAYARITHSTRETCHY
                              | DDFXCAPS_OVERLAYSTRETCHX
                              | DDFXCAPS_OVERLAYSTRETCHY
                              | DDFXCAPS_OVERLAYSHRINKX
                              | DDFXCAPS_OVERLAYSHRINKY;

	// --- FOURCC INFORMATION ---
	
	*pdwNumFourCC = 6;
		
	if (pdwFourCC)
	{
		DISPDBG((ppdev, 200, "Setting fourcc types\n"));
		pdwFourCC[0] = FOURCC_I420;
		pdwFourCC[1] = FOURCC_IYUV;
		pdwFourCC[2] = FOURCC_YV12;
		pdwFourCC[3] = FOURCC_YUY2;
		pdwFourCC[4] = FOURCC_UYVY;
		pdwFourCC[5] = FOURCC_YVYU;
	}
	
    // --- ADVANCED INFO CALLBACK

    pHalInfo->GetDriverInfo = DdGetDriverInfo;
    pHalInfo->dwFlags = DDHALINFO_GETDRIVERINFOSET;

    // --- OVERLAY CAPABILITIES ---

    if (g_vipHardware)
        pHalInfo->ddCaps.dwMaxVideoPorts = 1;
    else
        pHalInfo->ddCaps.dwMaxVideoPorts = 0;
    pHalInfo->ddCaps.dwMaxVisibleOverlays  = 1;
    pHalInfo->ddCaps.dwCurrVisibleOverlays = 0;
    pHalInfo->ddCaps.dwCurrVideoPorts      = 0;

    // SUPPORT 16:1 STRETCHING AND 1:4 DOWNSCALING
    // Stretch factors are related to 1000.  A value of 8000
    // indicates an 8:1 stretch while a value of 500 indicates
    // a 1:2 downscale.

	pHalInfo->ddCaps.dwMinOverlayStretch   = 250;
	pHalInfo->ddCaps.dwMinLiveVideoStretch = 250;
	pHalInfo->ddCaps.dwMinHwCodecStretch   = 250;

	pHalInfo->ddCaps.dwMaxOverlayStretch   = 16000;
	pHalInfo->ddCaps.dwMaxLiveVideoStretch = 16000;
	pHalInfo->ddCaps.dwMaxHwCodecStretch   = 16000;

	DISPDBG((ppdev, 300, "DrvGetDirectDrawInfo leaves\n"));
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// DrvEnableDirectDraw
//
// This function is called by GDI to enable DirectDraw when a DirectDraw
// program is started and DirectDraw is not already active.
//
BOOL DrvEnableDirectDraw(
	DHPDEV					dhpdev,
	DD_CALLBACKS*			pCallBacks,
	DD_SURFACECALLBACKS*	pSurfaceCallBacks,
	DD_PALETTECALLBACKS*	pPaletteCallBacks)
{
	PDEV	*ppdev;

	ppdev = (PDEV*) dhpdev;

	DISPDBG((ppdev, 300, "DrvEnableDirectDraw\n"));

	// GENERAL CALLBACKS

	pCallBacks->WaitForVerticalBlank = DdWaitForVerticalBlank;
	pCallBacks->MapMemory			 = DdMapMemory;
    pCallBacks->CanCreateSurface	 = DdCanCreateSurface;
	pCallBacks->CreateSurface		 = DdCreateSurface;
    pCallBacks->GetScanLine          = DdGetScanLine;
	pCallBacks->dwFlags 			 = DDHAL_CB32_WAITFORVERTICALBLANK
									 | DDHAL_CB32_MAPMEMORY
                                     | DDHAL_CB32_CANCREATESURFACE
	  								 | DDHAL_CB32_CREATESURFACE
                                     | DDHAL_CB32_GETSCANLINE;

    // SURFACE CALLBACKS

    pSurfaceCallBacks->DestroySurface		= DdDestroySurface;
    pSurfaceCallBacks->Flip 		        = DdFlip;
    pSurfaceCallBacks->Lock 		        = DdLock;
    pSurfaceCallBacks->Blt			        = DdBlt;
    pSurfaceCallBacks->GetBltStatus         = DdGetBltStatus;
    pSurfaceCallBacks->GetFlipStatus        = DdGetFlipStatus;
    pSurfaceCallBacks->UpdateOverlay		= DdUpdateOverlay;
	pSurfaceCallBacks->SetOverlayPosition	= DdSetOverlayPosition;
	
	pSurfaceCallBacks->dwFlags		        = DDHAL_SURFCB32_DESTROYSURFACE
                                            | DDHAL_SURFCB32_FLIP
                                            | DDHAL_SURFCB32_LOCK
                                            | DDHAL_SURFCB32_BLT
                                            | DDHAL_SURFCB32_GETBLTSTATUS
                                            | DDHAL_SURFCB32_GETFLIPSTATUS
                                            | DDHAL_SURFCB32_UPDATEOVERLAY
										    | DDHAL_SURFCB32_SETOVERLAYPOSITION;									

	DISPDBG((ppdev, 300, "DrvEnableDirectDraw leaves\n"));

	ppdev->VideoFlipRecord.bFlipFlag    = FALSE;
	ppdev->GraphicsFlipRecord.bFlipFlag = FALSE;
    ppdev->VPEFlipRecord.bFlipFlag      = FALSE;

    // INITIALIZE THE GP BUFFER LIMIT
    // DirectDraw will only perform screen to screen copies and pattern
    // fills.  Consequently, software can get very far ahead of the GP
    // hardware.  We limit the allowable software lead for DirectDraw to
    // a reasonable value to prevent watchdog errors.  We limit the
    // command buffer lead to 500 BLTs.
    //
    gp_set_limit_on_buffer_lead (GP3_BLT_COMMAND_SIZE * 500);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// DrvDisableDirectDraw
//
// This function is called by GDI when the last active DirectDraw program
// is quit and DirectDraw will no longer be active.
//
VOID DrvDisableDirectDraw (DHPDEV dhpdev)
{
	PDEV *ppdev = (PDEV*) dhpdev;

	DISPDBG((ppdev, 300, "DrvDisableDirectDraw\n"));
}

/////////////////////////////////////////////////////////////////////////
// UpdateVideoFlipStatus
//
// Checks and sees if the most recent video flip has occurred.  This routine
// is only called for video flips.
//
HRESULT UpdateVideoFlipStatus(
	PDEV	*ppdev,
	FLATPTR fpVidMem)
{
	HRESULT ret = DD_OK;

	if ((ppdev->VideoFlipRecord.bFlipFlag) &&
        ((fpVidMem == ppdev->VideoFlipRecord.fpFlipFrom) || (fpVidMem == ppdev->VideoFlipRecord.fpFlipTo)))
	{
		int vid_enable;
		unsigned long flags;
		df_get_video_enable(&vid_enable, &flags);
		if (!vid_enable || df_test_video_flip_status())
		{
			// FLIP FINISHED
            // 'Flip To' surface is now the 'Flip From' surface.
            //
			ppdev->VideoFlipRecord.bFlipFlag  = FALSE;
            ppdev->VideoFlipRecord.fpFlipFrom = ppdev->VideoFlipRecord.fpFlipTo;
		}
		else
		{
			// FLIP STILL IN PROGRESS

			ret = DDERR_WASSTILLDRAWING;
		}
	}
	
	DISPDBG((ppdev, 200, "UpdateVideoFlipStatus leaves\n"));

	return ret;
}

/////////////////////////////////////////////////////////////////////////
// drvalUpdateFlipStatus
//
// Checks and sees if the most recent display flip has occurred.  This routine
// is not called for video flips.
//
HRESULT ddrvalUpdateFlipStatus(
	PDEV	*ppdev,
	FLATPTR fpVidMem)
{
	HRESULT ret = DD_OK;

	DISPDBG((ppdev, 200, "ddrvalUpdateFlipStatus\n"));

	if ((ppdev->GraphicsFlipRecord.bFlipFlag) && (fpVidMem == ppdev->GraphicsFlipRecord.fpFlipFrom))
	{
		if (vg_test_flip_status())
		{
			// FLIP FINISHED

			ppdev->GraphicsFlipRecord.bFlipFlag = FALSE;	
		}
		else
		{
			// FLIP STILL IN PROGRESS

			ret = DDERR_WASSTILLDRAWING;
		}
	}
	else
	{
		// NO WORK

		DISPDBG((ppdev, 200, "no status, flip %d, fpVidMem %x, flipFrom %x\n",
			 ppdev->GraphicsFlipRecord.bFlipFlag, fpVidMem, ppdev->GraphicsFlipRecord.fpFlipFrom));
	}

	DISPDBG((ppdev, 200, "ddrvalUpdateFlipStatus leaves\n"));

	return ret;
}

/////////////////////////////////////////////////////////////////////////
// DdBlt
//
// Generic BitBlt routine.
//
DWORD DdBlt (PDD_BLTDATA lpBlt)
{
	PDD_SURFACE_GLOBAL	srcSurf;
	PDD_SURFACE_GLOBAL	dstSurf;
	PDEV*				ppdev = NULL;
	HRESULT 			ddrval, ddrval2;
	DWORD				dstX;
	DWORD				dstY;
	DWORD				dwFlags;
	DWORD				dstWidth;
	DWORD				dstHeight;
	DWORD				srcWidth;
	DWORD				srcHeight;
	DWORD				srcX;
	DWORD				srcY;
	ULONG               dstOffset, srcOffset;
    ULONG               bltFlags;

	ppdev	 = (PDEV*) lpBlt->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdBlt\n"));

	dstSurf = lpBlt->lpDDDestSurface->lpGbl;
	
	DISPDBG((ppdev, 200, "rops: rop %x, ddrop %x\n", lpBlt->bltFX.dwROP, lpBlt->bltFX.dwDDROP));

	// IS A FLIP IN PROGRESS?
    // Check both graphics and video as we may be BLTing to a video surface.
    //	
	ddrval  = ddrvalUpdateFlipStatus (ppdev, dstSurf->fpVidMem);
    ddrval2 = UpdateVideoFlipStatus (ppdev, dstSurf->fpVidMem);

	if (ddrval != DD_OK || ddrval2 != DD_OK)
	{
		lpBlt->ddRVal = DDERR_WASSTILLDRAWING;
		DISPDBG((ppdev, 200, "DdBlt leaves, flip in progress\n"));
		return DDHAL_DRIVER_HANDLED;
	}

	dwFlags = lpBlt->dwFlags;
	DISPDBG((ppdev, 200, "dwFlags %x\n", dwFlags));
	
	// DEST RECT

	dstX	  = lpBlt->rDest.left;
	dstY	  = lpBlt->rDest.top;
	dstWidth  = lpBlt->rDest.right  - lpBlt->rDest.left;
	dstHeight = lpBlt->rDest.bottom - lpBlt->rDest.top;

	DISPDBG((ppdev, 200, "dest org (%d,%d), size (%d,%d)\n",
				dstX, dstY, dstWidth, dstHeight));

    // PREPARE FOR THE BLT

    gp_set_bpp (ppdev->cBitsPerPel);
    gp_declare_blt (CIMGP_BLTFLAGS_LIMITBUFFER);

	if (dwFlags & DDBLT_COLORFILL)
	{
		// CHECK PIXEL FORMAT
		// The LX can't do color fills for off-screen surfaces that
		// are a different pixel format than that of the primary display:

		if (dstSurf->dwReserved1 != 0)
		{
            DISPDBG((ppdev, 2000, "Can't do colorfill to odd pixel format\n"));
			return DDHAL_DRIVER_NOTHANDLED;
		}
		else
		{
			// SOLID FILL

			calculateOffset (dstOffset, dstX, dstY, dstSurf);

            gp_set_raster_operation (0xF0);
            gp_set_solid_pattern (lpBlt->bltFX.dwFillColor);
            gp_set_strides (dstSurf->lPitch, 0);
            gp_pattern_fill (dstOffset, dstWidth, dstHeight);
			
			lpBlt->ddRVal = DD_OK;
			return DDHAL_DRIVER_HANDLED;
		}
	}

	// CHECK REMAINING COMMAND
	// We specified with Our ddCaps.dwCaps that we handle a limited number
	// of commands, and by this point in our routine we've handled everything
	// except DDBLT_ROP.  DirectDraw and GDI shouldn't pass us anything
	// else; we'll assert on debug builds to prove this:

	ASSERTVDD((dwFlags & DDBLT_ROP) && (lpBlt->lpDDSrcSurface),
		"Expected dwFlags commands of only DDBLT_ASYNC and DDBLT_COLORFILL");

	// GET OFFSET, WIDTH AND HEIGHT FOR SOURCE

	srcSurf 	 = lpBlt->lpDDSrcSurface->lpGbl;
	srcX		 = lpBlt->rSrc.left;
	srcY		 = lpBlt->rSrc.top;
	srcWidth	 = lpBlt->rSrc.right - lpBlt->rSrc.left;
	srcHeight	 = lpBlt->rSrc.bottom - lpBlt->rSrc.top;
	
	if ((srcWidth  == dstWidth)  &&
		(srcHeight == dstHeight) &&
		(srcSurf->dwReserved1 == dstSurf->dwReserved1))
	{
	    DISPDBG((ppdev, 200, "Screen to screen BLT\n"));

		// SCREEN TO SCREEN BLT

        if (srcSurf->dwReserved1 != 0)
        {
            if ((srcSurf->dwReserved1 & 0xF) >= DF_VIDFMT_Y0Y1Y2Y3 &&
                (srcSurf->dwReserved1 & 0xF) <= DF_VIDFMT_Y1Y2Y3Y0)
            {
                ULONG dstSurfHeight;
                ULONG srcSurfHeight;

                dstSurfHeight = lpBlt->lpDDDestSurface->lpGbl->wHeight;
                srcSurfHeight = lpBlt->lpDDSrcSurface->lpGbl->wHeight;
                gp_set_bpp (8);

                bltFlags = 0;
		        if (dstSurf == srcSurf)
		        {
                    if (dstY > srcY)
                        bltFlags = CIMGP_NEGYDIR;
                    else if (dstY == srcY && dstX > srcX)
                        bltFlags = CIMGP_NEGXDIR;
                }

                // 4:2:0 Video - Y, U and V are stored in separate buffers.
                //
                dstOffset  = (dstY * dstSurf->lPitch) + dstSurf->fpVidMem;
	            dstOffset +=  dstX;
                srcOffset  = (srcY * srcSurf->lPitch) + srcSurf->fpVidMem;
	            srcOffset +=  srcX;

                // Copy Y data
                //
                gp_set_raster_operation (0xCC);
                gp_set_strides (dstSurf->lPitch, srcSurf->lPitch);
                gp_screen_to_screen_blt (dstOffset, srcOffset, dstWidth, dstHeight, bltFlags);

                // Copy U Data
                //
                dstOffset  = dstSurf->fpVidMem + (dstSurfHeight * dstSurf->lPitch);
                srcOffset  = srcSurf->fpVidMem + (srcSurfHeight * srcSurf->lPitch);
                dstOffset += (dstY * (dstSurf->lPitch >> 1)) + (dstX >> 1);
                srcOffset += (srcY * (srcSurf->lPitch >> 1)) + (srcX >> 1);
                gp_declare_blt (CIMGP_BLTFLAGS_LIMITBUFFER);
                gp_set_strides (dstSurf->lPitch >> 1, srcSurf->lPitch >> 1);
                gp_screen_to_screen_blt (dstOffset, srcOffset, dstWidth >> 1, dstHeight >> 1, bltFlags);

                // Copy V Data
                //
                dstOffset  = dstSurf->fpVidMem + (dstSurfHeight * dstSurf->lPitch) + ((dstSurfHeight >> 1) * (dstSurf->lPitch >> 1));
                srcOffset  = srcSurf->fpVidMem + (srcSurfHeight * srcSurf->lPitch) + ((srcSurfHeight >> 1) * (srcSurf->lPitch >> 1));
                dstOffset += (dstY * (dstSurf->lPitch >> 1)) + (dstX >> 1);
                srcOffset += (srcY * (srcSurf->lPitch >> 1)) + (srcX >> 1);
                gp_declare_blt (CIMGP_BLTFLAGS_LIMITBUFFER);
                gp_screen_to_screen_blt (dstOffset, srcOffset, dstWidth >> 1, dstHeight >> 1, bltFlags);

                lpBlt->ddRVal = DD_OK;

	            DISPDBG((ppdev, 200, "DdBlt leaves\n"));

	            return DDHAL_DRIVER_HANDLED;
            }

            // FOURCC COPY
            // Data is 16-bit instead of native display BPP

            dstOffset  = (dstY * dstSurf->lPitch) + dstSurf->fpVidMem;
	        dstOffset += (dstX << 1);
            srcOffset  = (srcY * srcSurf->lPitch) + srcSurf->fpVidMem;
	        srcOffset += (srcX << 1);

            if      (ppdev->cPelSize == 2) dstWidth >>= 1;
            else if (ppdev->cPelSize == 0) dstWidth <<= 1;
        }
        else
        {
		    calculateOffset (dstOffset, dstX, dstY, dstSurf);
		    calculateOffset (srcOffset, srcX, srcY, srcSurf);
        }
			
        gp_set_raster_operation (0xCC);
        gp_set_strides (dstSurf->lPitch, srcSurf->lPitch);
		if (dwFlags & DDBLT_KEYSRCOVERRIDE)
            gp_set_source_transparency (lpBlt->bltFX.ddckSrcColorkey.dwColorSpaceLowValue, 0xFFFFFF);
        else if (dwFlags & DDBLT_KEYDESTOVERRIDE)
            DISPDBG ((ppdev, 5000, "Destination override\n"));
		
        bltFlags = 0;
		if (dstSurf == srcSurf)
		{
            if (dstY > srcY)
                bltFlags = CIMGP_NEGYDIR;
            else if (dstY == srcY && dstX > srcX)
                bltFlags = CIMGP_NEGXDIR;
        }
        gp_screen_to_screen_blt (dstOffset, srcOffset, dstWidth, dstHeight, bltFlags);
    }
	else
	{
		DISPDBG((ppdev, 300, "Unhandled BLT, dwFlags %x\n", dwFlags));
	}

	lpBlt->ddRVal = DD_OK;

	DISPDBG((ppdev, 200, "DdBlt leaves\n"));

	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DWORD DdFlip
//
// DirectDraw Surface Flip Callback.
//
DWORD DdFlip (PDD_FLIPDATA lpFlip)
{
	PDEV               *ppdev;
	HRESULT             ddrval;
    PDD_SURFACE_GLOBAL  dstSurf;
    ULONG               uOffset;
    ULONG               vOffset;
	
	ppdev = (PDEV*) lpFlip->lpDD->dhpdev;

	DISPDBG((ppdev, 250, "DdFlip\n"));

    dstSurf = lpFlip->lpSurfTarg->lpGbl;

	if (lpFlip->lpSurfCurr->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
	{
		DISPDBG((ppdev, 200, "Overlay flip\n"));

		// CHECK FOR PREVIOUS VIDEO FLIP COMPLETED
		
		if (ppdev->VideoFlipRecord.bFlipFlag)
		{
            if (!df_test_video_flip_status())
		    {
                lpFlip->ddRVal = DDERR_WASSTILLDRAWING;
				return DDHAL_DRIVER_HANDLED;
            }

            ppdev->VideoFlipRecord.fpFlipFrom = ppdev->VideoFlipRecord.fpFlipTo;
        }

        // REMEMBER THE FLIP
		//
		ppdev->VideoFlipRecord.bFlipFlag  = TRUE;
		ppdev->VideoFlipRecord.fpFlipTo = dstSurf->fpVidMem;

        // HANDLE INTERLACED OUTPUT
        // If flipping on an interlaced display, we first check if the content
        // is already interlaced.  If so, we simply update the appropriate odd
        // or even pointers.
        //
        uOffset = ppdev->videoSource0.u_offset - ppdev->videoSource0.y_offset;
        vOffset = ppdev->videoSource0.v_offset - ppdev->videoSource0.y_offset;

        if (ppdev->interlacedOut)
        {
            // ODD FIELD ONLY
            // DDOVER_BOB flag is implied??
            //
            if (lpFlip->dwFlags & DDFLIP_ODD)
            {
                if ((lpFlip->lpSurfTarg->lpSurfMore->dwOverlayFlags & (DDOVER_BOB | DDOVER_INTERLEAVED)) == (DDOVER_BOB | DDOVER_INTERLEAVED))
                {
                    ppdev->videoSource1.y_offset = dstSurf->fpVidMem + (ppdev->videoSource0.y_pitch >> 1);
                    ppdev->videoSource1.u_offset = dstSurf->fpVidMem + uOffset + (ppdev->videoSource0.uv_pitch >> 1);
                    ppdev->videoSource1.v_offset = dstSurf->fpVidMem + vOffset + (ppdev->videoSource0.uv_pitch >> 1);
                }
                else
                {
                    ppdev->videoSource1.y_offset = dstSurf->fpVidMem;
                    ppdev->videoSource1.u_offset = dstSurf->fpVidMem + uOffset;
                    ppdev->videoSource1.v_offset = dstSurf->fpVidMem + vOffset;
                }
                df_set_video_offsets (1,
                    ppdev->videoSource1.y_offset + ppdev->yBufferOffset,
                    ppdev->videoSource1.u_offset + ppdev->uvBufferOffset,
                    ppdev->videoSource1.v_offset + ppdev->uvBufferOffset);
            }
            // EVEN FIELD ONLY
            // DDOVER_BOB flag is implied??
            //
            else if (lpFlip->dwFlags & DDFLIP_EVEN)
            {
                ppdev->videoSource0.y_offset = dstSurf->fpVidMem;
                ppdev->videoSource0.u_offset = dstSurf->fpVidMem + uOffset;
                ppdev->videoSource0.v_offset = dstSurf->fpVidMem + vOffset;
                df_set_video_offsets (0,
                    ppdev->videoSource0.y_offset + ppdev->yBufferOffset,
                    ppdev->videoSource0.u_offset + ppdev->uvBufferOffset,
                    ppdev->videoSource0.v_offset + ppdev->uvBufferOffset);
            }
            else
            {
                ppdev->videoSource0.y_offset = dstSurf->fpVidMem;
                ppdev->videoSource0.u_offset = dstSurf->fpVidMem + uOffset;
                ppdev->videoSource0.v_offset = dstSurf->fpVidMem + vOffset;

                if (ppdev->customMode->line_double_video_on_interlaced)
                {
                    ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset;
                    ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset;
                    ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset;
                }
                else
                {
                    ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset + (ppdev->videoSource0.y_pitch >> 1);
                    ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset + (ppdev->videoSource0.uv_pitch >> 1);
                    ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset + (ppdev->videoSource0.uv_pitch >> 1);
                }

                df_set_video_offsets (0,
                    ppdev->videoSource0.y_offset + ppdev->yBufferOffset,
                    ppdev->videoSource0.u_offset + ppdev->uvBufferOffset,
                    ppdev->videoSource0.v_offset + ppdev->uvBufferOffset);

                df_set_video_offsets (1,
                    ppdev->videoSource1.y_offset + ppdev->yBufferOffset,
                    ppdev->videoSource1.u_offset + ppdev->uvBufferOffset,
                    ppdev->videoSource1.v_offset + ppdev->uvBufferOffset);
            }
        }
        else
        {
            ppdev->videoSource0.y_offset = dstSurf->fpVidMem;
            ppdev->videoSource0.u_offset = dstSurf->fpVidMem + uOffset;
            ppdev->videoSource0.v_offset = dstSurf->fpVidMem + vOffset;

            if (lpFlip->dwFlags & DDFLIP_ODD)
            {
                WRITE_VID32 (DF_VIDEO_Y_POS, ppdev->videoY + 1);

                // ADJUST THE OFFSETS FOR INTERLEAVED BOB
                //
                if ((lpFlip->lpSurfTarg->lpSurfMore->dwOverlayFlags & (DDOVER_BOB | DDOVER_INTERLEAVED)) == (DDOVER_BOB | DDOVER_INTERLEAVED))
                {
                    df_set_video_offsets (0,
                        ppdev->videoSource0.y_offset + ppdev->yBufferOffset  + (ppdev->videoSource0.y_pitch >> 1),
                        ppdev->videoSource0.u_offset + ppdev->uvBufferOffset + (ppdev->videoSource0.uv_pitch >> 1),
                        ppdev->videoSource0.v_offset + ppdev->uvBufferOffset + (ppdev->videoSource0.uv_pitch >> 1));
                }
                else
                {
                    df_set_video_offsets (0,
                        ppdev->videoSource0.y_offset + ppdev->yBufferOffset,
                        ppdev->videoSource0.u_offset + ppdev->uvBufferOffset,
                        ppdev->videoSource0.v_offset + ppdev->uvBufferOffset);
                }
            }
            else
            {
                if (lpFlip->dwFlags & DDFLIP_EVEN)
                    WRITE_VID32 (DF_VIDEO_Y_POS, ppdev->videoY);

                df_set_video_offsets (0,
                    ppdev->videoSource0.y_offset + ppdev->yBufferOffset,
                    ppdev->videoSource0.u_offset + ppdev->uvBufferOffset,
                    ppdev->videoSource0.v_offset + ppdev->uvBufferOffset);
            }
        }

	}
	else
	{
		// WAIT FOR PREVIOUS FLIP

		if (ppdev->GraphicsFlipRecord.bFlipFlag)
		{
			if (!(vg_test_flip_status()))
			{
				lpFlip->ddRVal = DDERR_WASSTILLDRAWING;
				return DDHAL_DRIVER_HANDLED;
			}
		}
		
		DISPDBG((ppdev, 200, "Ddraw flip\n"));

        // REMEMBER THE FLIP

		ppdev->GraphicsFlipRecord.bFlipFlag  = TRUE;
		ppdev->GraphicsFlipRecord.fpFlipFrom = vg_get_display_offset();

        // PROGRAM THE DISPLAY OFFSET
        // Compression should have been disabled in a call to DdSetExclusiveMode
        // If not, Cimarron will disable compression as a fallback.

		vg_set_display_offset (dstSurf->fpVidMem);
	}

	lpFlip->ddRVal = DD_OK;

	DISPDBG((ppdev, 250, "DdFlip leaves\n"));

	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdLock
//
DWORD DdLock(PDD_LOCKDATA lpLock)
{
	PDEV	*ppdev = NULL;
	HRESULT ddrval;

	ppdev = (PDEV*) lpLock->lpDD->dhpdev;
	DISPDBG((ppdev, 200, "DdLock\n"));

	// CHECK FLIP STATUS
	// Check to see if any pending physical flip has occurred.	Don't allow
	// a lock if a flip is in progress:

    if (lpLock->lpDDSurface->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
    {
        ddrval = UpdateVideoFlipStatus (ppdev, lpLock->lpDDSurface->lpGbl->fpVidMem);

        if (ddrval != DD_OK)
        {
            lpLock->ddRVal = ddrval;
		    return DDHAL_DRIVER_HANDLED;
        }
    }
    else
    {
	    ddrval = ddrvalUpdateFlipStatus(ppdev, lpLock->lpDDSurface->lpGbl->fpVidMem);
	    if (ddrval != DD_OK)
	    {
		    lpLock->ddRVal = DDERR_WASSTILLDRAWING;
		    return DDHAL_DRIVER_HANDLED;
	    }
    }

	// Here's one of the places where the Windows 95 and Windows NT DirectDraw
	// implementations differ: on Windows NT, you should watch for
	// DDLOCK_WAIT and loop in the driver while the accelerator is busy.
	// On Windows 95, it doesn't really matter.
	//
	// (The reason is that Windows NT allows applications to draw directly
	// to the frame buffer even while the accelerator is running, and does
	// not synchronize everything on the Win16Lock.  Note that on Windows NT,
	// it is even possible for multiple threads to be holding different
	// DirectDraw surface locks at the same time.)

	if (lpLock->dwFlags & DDLOCK_WAIT)
	{
		gp_wait_until_idle();
	}
	else if (gp_test_blt_busy())
	{
		lpLock->ddRVal = DDERR_WASSTILLDRAWING;
		return DDHAL_DRIVER_HANDLED;
	}

	DISPDBG((ppdev, 200, "DdLock leaves\n"));

	return DDHAL_DRIVER_NOTHANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdGetBltStatus
//
// Doesn't currently really care what surface is specified, just checks
// and goes.
//
DWORD DdGetBltStatus (PDD_GETBLTSTATUSDATA lpGetBltStatus)
{
	PDEV	*ppdev;
	HRESULT ddRVal;

	ppdev	 = (PDEV*) lpGetBltStatus->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdGetBltStatus\n"));

	ddRVal = DD_OK;

	if (lpGetBltStatus->dwFlags == DDGBS_CANBLT)
	{
		// DDGBS_CANBLT -> CAN WE ADD A BLT

		ddRVal = ddrvalUpdateFlipStatus(ppdev,
						lpGetBltStatus->lpDDSurface->lpGbl->fpVidMem);

		// NO FLIP, SO WE CAN ADD A BLT
        // With the command buffer, we assume there is always
        // room enough to add a BLT
	}
	else
	{
		// DDGBS_ISBLTDONE -> IS A BLT IN PROGRESS?

		if (gp_test_blt_busy())
		{
			ddRVal = DDERR_WASSTILLDRAWING;
		}
	}

	lpGetBltStatus->ddRVal = ddRVal;

	DISPDBG((ppdev, 200, "DdGetBltStatus leaves\n"));

	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdMapMemory
//
// This is a new DDI call specific to Windows NT that is used to map
// or unmap all the application modifiable portions of the frame buffer
// into the specified process's address space.
//
DWORD DdMapMemory(PDD_MAPMEMORYDATA lpMapMemory)
{
	PDEV*							ppdev;
	VIDEO_SHARE_MEMORY				ShareMemory;
	VIDEO_SHARE_MEMORY_INFORMATION	ShareMemoryInformation;
	DWORD							ReturnedDataLength;
	DWORD                           heap_size;

	ppdev = (PDEV*) lpMapMemory->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdMapMemory\n"));
	DISPDBG((ppdev, 200, "hProc %x, fpProc %x, bMap %d\n", lpMapMemory->hProcess,
			lpMapMemory->fpProcess, lpMapMemory->bMap));

	// BMAP == TRUE -> MAP THE MEMORY
	
	if (lpMapMemory->bMap)
	{
		DISPDBG((ppdev, 200, "mapping memory\n"));

		ShareMemory.ProcessHandle = lpMapMemory->hProcess;

		// REQUESTEDVIRTUALADDRESS NOT USED

		ShareMemory.RequestedVirtualAddress = 0;

		// MAP TO TOP OF THE FRAMEBUFFER

		ShareMemory.ViewOffset = 0;

		// MAP THE FRAME BUFFER */
		
		ShareMemory.ViewSize = g_fbSize;

		if (EngDeviceIoControl(ppdev->hDriver,
							   IOCTL_VIDEO_SHARE_VIDEO_MEMORY,
							   &ShareMemory,
							   sizeof(VIDEO_SHARE_MEMORY),
							   &ShareMemoryInformation,
							   sizeof(VIDEO_SHARE_MEMORY_INFORMATION),
							   &ReturnedDataLength))
		{
			DISPDBG((ppdev, 300, "Failed IOCTL_VIDEO_SHARE_MEMORY"));

			lpMapMemory->ddRVal = DDERR_GENERIC;
			return DDHAL_DRIVER_HANDLED;
		}

		DISPDBG((ppdev, 200, "Returned address %x\n",
			ShareMemoryInformation.VirtualAddress));

		lpMapMemory->fpProcess = (DWORD) ShareMemoryInformation.VirtualAddress;
	}

	// ELSE, UNMAP

	else
	{
		DISPDBG((ppdev, 300, "unmap\n"));

		ShareMemory.ProcessHandle			= lpMapMemory->hProcess;
		ShareMemory.ViewOffset				= 0;
		ShareMemory.ViewSize				= 0;
		ShareMemory.RequestedVirtualAddress = (VOID*) lpMapMemory->fpProcess;

		if (EngDeviceIoControl(ppdev->hDriver,
							   IOCTL_VIDEO_UNSHARE_VIDEO_MEMORY,
							   &ShareMemory,
							   sizeof(VIDEO_SHARE_MEMORY),
							   NULL,
							   0,
							   &ReturnedDataLength))
		{
			DISPDBG ((ppdev, 5000, "Failed IOCTL_VIDEO_UNSHARE_MEMORY\n"));
		}
	}

	lpMapMemory->ddRVal = DD_OK;

	DISPDBG((ppdev, 200, "DdMapMemory leaves\n"));

	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdGetFlipStatus
//
// If the display has gone through one refresh cycle since the flip
// occurred, we return DD_OK.  If it has not gone through one refresh
// cycle we return DDERR_WASSTILLDRAWING to indicate that this surface
// is still busy "drawing" the flipped page.   We also return
// DDERR_WASSTILLDRAWING if the bltter is busy and the caller wanted
// to know if they could flip yet.
//
DWORD DdGetFlipStatus(PDD_GETFLIPSTATUSDATA lpGetFlipStatus)
{
	PDEV *ppdev;
    HRESULT ddrval;

	ppdev = (PDEV*) lpGetFlipStatus->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdGetFlipStatus\n"));

	// HANDLE OVERLAY FLIPPING SEPARATE FROM GRAPHICS FLIPPING

	if (lpGetFlipStatus->lpDDSurface->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
	{
		// TEST FOR FLIP PENDING
        // If DirectDraw is asking for 'Can Flip', we must wait for the GP to be idle and
        // any pending flips to have occured.
        //
        if (lpGetFlipStatus->dwFlags == DDGFS_CANFLIP)
        {
            lpGetFlipStatus->ddRVal = UpdateVideoFlipStatus (ppdev, ppdev->VideoFlipRecord.fpFlipFrom);
            if (gp_test_blt_busy())
                lpGetFlipStatus->ddRVal = DDERR_WASSTILLDRAWING;
        }

        // Otherwise we simply need to query whether a particular surface is currently involved
        // in a flip.
        //
        else
        {
            lpGetFlipStatus->ddRVal = UpdateVideoFlipStatus (ppdev, lpGetFlipStatus->lpDDSurface->lpGbl->fpVidMem);
        }

		return DDHAL_DRIVER_HANDLED;
	}

	// CHECK LAST FLIP STATUS

	lpGetFlipStatus->ddRVal = ddrvalUpdateFlipStatus(ppdev, ppdev->GraphicsFlipRecord.fpFlipFrom);

	// CHECK BLT BUSY

	if (lpGetFlipStatus->dwFlags == DDGFS_CANFLIP)
	{
		if ((lpGetFlipStatus->ddRVal == DD_OK) && gp_test_blt_busy())
		{
			lpGetFlipStatus->ddRVal = DDERR_WASSTILLDRAWING;
		}
	}

	DISPDBG((ppdev, 200, "DdGetFlipStatus leaves\n"));

	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdInVerticalBlank
//
// Returns TRUE if vertical blank is active, otherwise false.
//
BOOL DdInVerticalBlank(PDEV *ppdev)
{
	return (!vg_test_vertical_active());
}

/////////////////////////////////////////////////////////////////////////
// DdWaitForVerticalBlank
//
DWORD DdWaitForVerticalBlank(PDD_WAITFORVERTICALBLANKDATA lpWaitForVerticalBlank)
{
	PDEV	*ppdev=NULL;

	ppdev = (PDEV*) lpWaitForVerticalBlank->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdWaitForVerticalBlank\n"));

	switch (lpWaitForVerticalBlank->dwFlags)
	{
		case DDWAITVB_I_TESTVB:

			// TESTVB -> CURRENT VBLANK STATUS
			
			if (DdInVerticalBlank(ppdev))
				lpWaitForVerticalBlank->bIsInVB = TRUE;
			else
				lpWaitForVerticalBlank->bIsInVB = FALSE;

			lpWaitForVerticalBlank->ddRVal = DD_OK;
			return DDHAL_DRIVER_HANDLED;

		case DDWAITVB_BLOCKBEGIN:

			// BLOCKBEGIN -> WAIT FOR VBLANK AND THEN DISPLAY BEGIN
			
			while (DdInVerticalBlank(ppdev))
				;
			while (!DdInVerticalBlank(ppdev))
				;

			lpWaitForVerticalBlank->ddRVal = DD_OK;
			return DDHAL_DRIVER_HANDLED;

		case DDWAITVB_BLOCKEND:

			// BLOCKEND -> WAIT FOR VBLANK TO END

			while (!DdInVerticalBlank(ppdev))
				;
			while (DdInVerticalBlank(ppdev))
				;

			lpWaitForVerticalBlank->ddRVal = DD_OK;
			return DDHAL_DRIVER_HANDLED;
	}

	DISPDBG((ppdev, 200, "DdWaitForVerticalBlank leaves\n"));

	return DDHAL_DRIVER_NOTHANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdCanCreateSurface
//
DWORD DdCanCreateSurface (PDD_CANCREATESURFACEDATA lpCanCreateSurface)
{
	PDEV           *ppdev;
	DWORD			dwRet;
	LPDDSURFACEDESC lpSurfaceDesc;

	ppdev = (PDEV*) lpCanCreateSurface->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdCanCreateSurface\n"));

	lpSurfaceDesc = lpCanCreateSurface->lpDDSurfaceDesc;

	dwRet = DDHAL_DRIVER_NOTHANDLED;

	if (lpCanCreateSurface->bIsDifferentPixelFormat)
	{
		DISPDBG((ppdev, 200, "Different pixel format\n"));
		
		if(lpSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
			DISPDBG((ppdev, 200, "overlay surface\n"));

		if (lpSurfaceDesc->ddpfPixelFormat.dwFlags & DDPF_FOURCC)
		{
			// CHECK FOR SUPPORTED FOURCC TYPE

			DISPDBG((ppdev, 200, "FourCC format 0x%x\n",
				lpSurfaceDesc->ddpfPixelFormat.dwFourCC));

			switch(lpSurfaceDesc->ddpfPixelFormat.dwFourCC)
			{
				case FOURCC_YUY2:
					lpSurfaceDesc->ddpfPixelFormat.dwYUVBitCount = 16;
					dwRet = DDHAL_DRIVER_HANDLED;
					DISPDBG((ppdev, 200, "Handled yuy2\n"));
					break;

				case FOURCC_YVYU:
					lpSurfaceDesc->ddpfPixelFormat.dwYUVBitCount = 16;
					dwRet = DDHAL_DRIVER_HANDLED;
					DISPDBG((ppdev, 200, "Handled yvyu\n"));
					break;

				case FOURCC_UYVY:
					lpSurfaceDesc->ddpfPixelFormat.dwYUVBitCount = 16;
					dwRet = DDHAL_DRIVER_HANDLED;
					DISPDBG((ppdev, 200, "Handled uyvy\n"));
					break;

				case FOURCC_YV12:
				case FOURCC_I420:
				case FOURCC_IYUV:
					lpSurfaceDesc->ddpfPixelFormat.dwYUVBitCount = 12;
					dwRet = DDHAL_DRIVER_HANDLED;
					DISPDBG((ppdev, 200, "Handled yv12\n"));
					break;

				default:
					DISPDBG((ppdev, 200, "Punted\n"));
					break;
			}
		}
		else
		{
            if(lpSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
            {
                // ONLY ALLOW RGB SURFACES DIFFERENT FROM THE PRIMARY FOR OVERLAYS

                if(lpSurfaceDesc->ddpfPixelFormat.dwFlags & DDPF_RGB)
			    {
				    if(lpSurfaceDesc->ddpfPixelFormat.dwRGBBitCount == 16 &&
                      (lpSurfaceDesc->ddpfPixelFormat.dwRBitMask & 0x8000))
                    {    					
					    DISPDBG((ppdev, 200, "Handled 5:6:5 RGB format\n"));
					    dwRet = DDHAL_DRIVER_HANDLED;
				    }
			    }
            }
		}
	}
	else
	{
		DISPDBG((ppdev, 200, "trivial surface\n"));
		
		// TRIVIAL SURFACE
		// It's trivially easy to create plain surfaces that are the same
		// type as the primary surface:

		dwRet = DDHAL_DRIVER_HANDLED;
	}

	// PRINT ERROR MESSAGE IF SURFACE WAS NOT CREATED

	if (dwRet == DDHAL_DRIVER_NOTHANDLED)
	{
		DISPDBG((ppdev, 200, "non trivial surface\n"));

		if (lpSurfaceDesc->ddpfPixelFormat.dwFlags & DDPF_RGB)
		{
			DISPDBG((ppdev, 300, "Failed creation of %libpp RGB surface %lx %lx %lx\n",
				lpSurfaceDesc->ddpfPixelFormat.dwRGBBitCount,
				lpSurfaceDesc->ddpfPixelFormat.dwRBitMask,
				lpSurfaceDesc->ddpfPixelFormat.dwGBitMask,
				lpSurfaceDesc->ddpfPixelFormat.dwBBitMask));
		}
		else
		{
			DISPDBG((ppdev, 300, "Failed creation of type 0x%lx YUV 0x%lx surface\n",
				lpSurfaceDesc->ddpfPixelFormat.dwFlags,
				lpSurfaceDesc->ddpfPixelFormat.dwFourCC));
		}

        lpCanCreateSurface->ddRVal = DDERR_INVALIDPIXELFORMAT;
        return DDHAL_DRIVER_HANDLED;
	}

	lpCanCreateSurface->ddRVal = DD_OK;

	DISPDBG((ppdev, 200, "DdCanCreateSurface leaves\n"));
	return dwRet;
}

/////////////////////////////////////////////////////////////////////////
// DdCreateSurface
//
DWORD DdCreateSurface (PDD_CREATESURFACEDATA lpCreateSurface)
{
	PDEV               *ppdev;
	DD_SURFACE_LOCAL   *lpSurfaceLocal;
	DD_SURFACE_GLOBAL  *lpSurfaceGlobal;
	LPDDSURFACEDESC     lpSurfaceDesc;
	LONG                lLinearPitch;
	ULONG               surfaceSize;
    ULONG               Format = DD_RESERVED_VIDEO_SURFACE;
	int                 ufirst = 0;

	ppdev = (PDEV*) lpCreateSurface->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdCreateSurface\n"));

	lpCreateSurface->ddRVal = DD_OK;

	// On Windows XP, dwSCnt will always be 1, so there will only ever
	// be one entry in the 'lplpSList' array:

	lpSurfaceLocal	= lpCreateSurface->lplpSList[0];
	lpSurfaceGlobal = lpSurfaceLocal->lpGbl;
	lpSurfaceDesc	= lpCreateSurface->lpDDSurfaceDesc;

	DISPDBG((ppdev, 200, "Surface local 0x%x, global 0x%x, desc 0x%x\n",
			lpSurfaceLocal, lpSurfaceGlobal, lpSurfaceDesc));

	// ALLOCATE A SURFACE
	// The DDraw heap grows from the end of memory backward.  If a surface needs
	// more memory than is available, we try to free up memory by packing the GDI
	// heap or by relocating GDI bitmaps to system memory.  If we still don't
	// have room, we fail the call.

	// CHECK IF THERE IS A HEAP

	if (!heap_start)
	{
		lpCreateSurface->ddRVal = DDERR_OUTOFVIDEOMEMORY;
		return DDHAL_DRIVER_HANDLED;
	}

	// CHECK FOURCC CODES

	if (lpSurfaceGlobal->ddpfSurface.dwFlags & DDPF_FOURCC)
	{
		DISPDBG((ppdev, 200, "Deterining FourCC surface type\n"));

		switch(lpSurfaceGlobal->ddpfSurface.dwFourCC)
		{
			case FOURCC_YUY2: Format |= DF_VIDFMT_YUYV; break;
			case FOURCC_YVYU: Format |= DF_VIDFMT_YVYU; break;
			case FOURCC_UYVY: Format |= DF_VIDFMT_UYVY; break;
            case FOURCC_YV12: Format |= DF_VIDFMT_Y0Y1Y2Y3; break;
			case FOURCC_IYUV:
            case FOURCC_I420: Format |= DF_VIDFMT_Y0Y1Y2Y3;
                              ufirst = 1;
                              break;

			default:
				lpCreateSurface->ddRVal = DDERR_INVALIDPIXELFORMAT;
				DISPDBG((ppdev, 3000, "Unexpected FourCC surface %x\n",
					lpSurfaceGlobal->ddpfSurface.dwFourCC));
				return DDHAL_DRIVER_HANDLED;
		}

		DISPDBG((ppdev, 200, "Created YUV: %li x %li\n",
			lpSurfaceGlobal->wWidth, lpSurfaceGlobal->wHeight));

        if ((Format & 0x1F) == DF_VIDFMT_Y0Y1Y2Y3)
		{
            lpSurfaceGlobal->ddpfSurface.dwYUVBitCount = 12;

            if (lpSurfaceLocal->ddsCaps.dwCaps & DDSCAPS_VIDEOPORT)
            {
                // Planar videoport surfaces must have a y pitch that
                // is 64-byte aligned.
                //
                lLinearPitch =  (lpSurfaceGlobal->wWidth + 63) & ~0x3F;
			    surfaceSize  = ((lpSurfaceGlobal->wHeight) * lLinearPitch) +
					        ((lpSurfaceGlobal->wHeight) * (lLinearPitch >> 1));

            }
            else
            {
			    lLinearPitch =  (lpSurfaceGlobal->wWidth + 15) & ~0xF;
			    surfaceSize  = ((lpSurfaceGlobal->wHeight) * lLinearPitch) +
					        ((lpSurfaceGlobal->wHeight) * (lLinearPitch >> 1));
            }
		}
		else
		{
            lpSurfaceGlobal->ddpfSurface.dwYUVBitCount = 16;

            if (lpSurfaceLocal->ddsCaps.dwCaps & DDSCAPS_VIDEOPORT)
            {
                // Videoport surfaces must have a 32-byte aligned pitch.
                //
                lLinearPitch = (lpSurfaceGlobal->wWidth * 2 + 31) & ~0x1F;
			    surfaceSize  = (lpSurfaceGlobal->wHeight * lLinearPitch);
            }
            else
            {
                lLinearPitch = (lpSurfaceGlobal->wWidth * 2 + 7) & ~7;
			    surfaceSize  = (lpSurfaceGlobal->wHeight * lLinearPitch);
            }
		}
	}
	else
	{
        // RGB SURFACE
        // Surface may still be an overlay surface
        //

		DISPDBG((ppdev, 200, "RGB surface type\n"));
		
        if (lpSurfaceLocal->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
        {
		    // 16BPP ONLY

		    if ((lpSurfaceGlobal->ddpfSurface.dwRGBBitCount >> 3) != 2)
		    {
			    DISPDBG((ppdev, 300, "Invalid byte count\n"));
			    lpCreateSurface->ddRVal = DDERR_INVALIDPIXELFORMAT;
			    return DDHAL_DRIVER_HANDLED;
		    }

		    Format |= DF_VIDFMT_RGB;

            if (lpSurfaceLocal->ddsCaps.dwCaps & DDSCAPS_VIDEOPORT)
            {
                // Videoport surfaces must have a 32-byte aligned pitch.
                //
                lLinearPitch = (lpSurfaceGlobal->wWidth * 2 + 31) & ~0x1F;
			    surfaceSize  = (lpSurfaceGlobal->wHeight * lLinearPitch);
            }
            else
            {
                lLinearPitch = (lpSurfaceGlobal->wWidth * 2 + 7) & ~7;
			    surfaceSize  = (lpSurfaceGlobal->wHeight * lLinearPitch);
            }
        }
        else
        {
            if (lpSurfaceGlobal->ddpfSurface.dwRGBBitCount != ppdev->cBitsPerPel)
            {
                DISPDBG ((ppdev, 3000, "Failing RGB Surface creation different from primary.\n"));
                lpCreateSurface->ddRVal = DDERR_INVALIDPIXELFORMAT;
				return(DDHAL_DRIVER_HANDLED);
            }

            // TRIVIAL SURFACE
            // Format is that of the primary.
            //
            if ((lpSurfaceDesc->dwFlags & DDSD_CAPS) &&
			    (lpSurfaceDesc->ddsCaps.dwCaps & DDSCAPS_PRIMARYSURFACE))
		    {
			    lLinearPitch = mode_pitch;
		    }
		    else
		    {
			    lLinearPitch = ((lpSurfaceGlobal->wWidth << ppdev->cPelSize) + 3) & ~3l;
		    }
		    surfaceSize = lLinearPitch * lpSurfaceGlobal->wHeight;

            Format = 0;
        }
	}

    DISPDBG ((ppdev, 30, "Surface Size and pitch are 0x%X, 0x%X\n", surfaceSize, lLinearPitch));

    // CALCULATE THE MEMORY NEEDED

	if (lpSurfaceLocal->ddsCaps.dwCaps & DDSCAPS_OVERLAY)
	{
		DISPDBG((ppdev, 200, "an overlay surface\n"));

		// DO NOT ALLOCATE AN OVERLAY SURFACE WHEN AN OVERLAY IS VISIBLE
		// This makes the assumption that all applications will create all
		// overlay surfaces in one batch (one buffer with multiple backbuffers)
		// and then enable video.

		if (ppdev->OverlayActive)
		{
			DISPDBG ((ppdev, 3000, "Trying to create an overlay surface with overlay active\n"));
			lpCreateSurface->ddRVal = DDERR_OUTOFCAPS;
			return DDHAL_DRIVER_HANDLED;
		}
    }

	// CHECK FOR AVAILABLE ROOM

	if (!AddDDrawBitmap (ppdev, surfaceSize, lLinearPitch))
	{
		DISPDBG ((ppdev, 3000, "Ran out of memory!\n"));
		lpCreateSurface->ddRVal = DDERR_OUTOFVIDEOMEMORY;
		return DDHAL_DRIVER_HANDLED;
	}

	// FILL IN FPVIDMEM POINTER

	lpSurfaceGlobal->fpVidMem = ddraw_offset;
	lpSurfaceGlobal->lPitch   = lLinearPitch;
	
	lpSurfaceDesc->lPitch	= lLinearPitch;
	lpSurfaceDesc->dwFlags |= DDSD_PITCH;

    // FILL IN RESERVED INFORMATION
    // We use the dwReserved1 member to hold the decoded surface format.
    // If the surface is a plain offscreen surface, we set the value to
    // 0.  Otherwise, we set it to the Cimarron video format that
    // corresponds to the FOURCC code passed in.  Finally, if the
    // video format is 4:2:0, we set bit 32 if the buffer ordering is
    // Y then U then V.  Otherwise, the buffers are assumed to order Y
    // then V then U

    if (ufirst) lpSurfaceGlobal->dwReserved1 = Format | DD_RESERVED_420_UFIRST;
    else        lpSurfaceGlobal->dwReserved1 = Format;

	lpCreateSurface->ddRVal = DD_OK;
	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdDestroySurface
//
// Destroys a surface created by DdCreateSurface.
//
DWORD DdDestroySurface (PDD_DESTROYSURFACEDATA  lpDestroySurface)
{
	PDEV *ppdev;

	ppdev = (PDEV*) lpDestroySurface->lpDD->dhpdev;
	DISPDBG((ppdev, 200, "DdDestroySurface\n"));

	// CLEAR SURFACE POINTER

	lpDestroySurface->lpDDSurface->lpGbl->fpVidMem = 0;
    lpDestroySurface->lpDDSurface->lpGbl->dwReserved1 = 0;

	// DECREMENT ACTIVE SURFACE COUNT
	// Reset the heap if this is the last surface.  Flag an error if this
	// call happened for invalid surfaces.

	ddraw_invalid++;

	ASSERTVDD(ddraw_count >= ddraw_invalid, "Too many bitmaps deleted");

	if (ddraw_invalid == ddraw_count)
	{
		DISPDBG ((ppdev, 3000, "Resetting ddraw heap...\n"));
		ResetDDrawHeap();
	}
			
	lpDestroySurface->ddRVal = DD_OK;
	return DDHAL_DRIVER_NOTHANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DWORD DdUpdateOverlay
//
DWORD DdUpdateOverlay(
	PDD_UPDATEOVERLAYDATA lpUpdateOverlay)
{
	PDEV               *ppdev;
	DD_SURFACE_GLOBAL*	lpSource;
	DWORD               DstOffset;
	DWORD               DstWidth;
	DWORD               DstHeight;
	DWORD               SrcWidth;
	DWORD               SrcHeight;
	DWORD               ColorKey;
	BOOL                dstColorKey;
	BOOL                srcColorKey;
	DWORD               i;
	RECT                rDst;
	int                 xFilter, yFilter;

	ppdev = (PDEV*) lpUpdateOverlay->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdUpdateOverlay\n"));
	DISPDBG((ppdev, 200, "Visible Overlay 0x%x\n", ppdev->OverlayActive));

    // Updating an overlay should fail when rotating.  This will allow software
    // rendering, albeit much more slowly.
    //
    if (ppdev->regSettings.RotationAngle != 0)
    {
        DISPDBG ((ppdev, 3000, "Refusing UpdateOverlay call during rotation\n"));
		lpUpdateOverlay->ddRVal = DDERR_OUTOFCAPS;
		return DDHAL_DRIVER_HANDLED;
    }

	// READ POINTERS
	// 'Source' is the overlay surface, 'destination' is the surface to be
	// overlayed.

	lpSource = lpUpdateOverlay->lpDDSrcSurface->lpGbl;
	DISPDBG((ppdev, 200, "source video pointer 0x%x\n",
			lpSource->fpVidMem));

	if (lpUpdateOverlay->dwFlags & DDOVER_HIDE)
	{
		DISPDBG((ppdev, 200, "Turn off overlay\n"));
		
		// CHECK IF WE ARE DISABLING THE CORRECT OVERLAY
        //
		if (ppdev->videoSource0.y_offset != lpSource->fpVidMem && ppdev->OverlayActive)
		{
			DISPDBG ((ppdev, 3000, "Trying to hide invalid overlay\n"));
			lpUpdateOverlay->ddRVal = DDERR_OUTOFCAPS;
		}

		// DISABLE OVERLAY
		
		else
		{		
			df_set_video_enable (0, 0);
			ppdev->OverlayActive = FALSE;
			lpUpdateOverlay->ddRVal = DD_OK;
		}		
	}
	else
	{
		// CHECK FOR OVERLAY MATCH

		if (ppdev->videoSource0.y_offset != lpSource->fpVidMem && ppdev->OverlayActive)
		{
			DISPDBG ((ppdev, 3000, "Trying to update invalid overlay\n"));
			lpUpdateOverlay->ddRVal = DDERR_OUTOFCAPS;
			return DDHAL_DRIVER_HANDLED;
		}

		DISPDBG((ppdev, 200, "Something else\n"));
		
		if (lpUpdateOverlay->dwFlags & DDOVER_SHOW)
		{
            // ONLY CONFIGURE ON DDOVER_SHOW
            // When DDOVER_SHOW is not set, we only configure video color keys.

            // STORE THE CURRENT SOURCE CONFIGURATION
            // These values are not passed directly to Cimarron.  Instead, they
            // are modified by another routine based on clipping and panning.
            // They merely represent how the overlay would be configured
            // completely unclipped.
            //
            ppdev->videoSource0.video_format = lpSource->dwReserved1 & 0x3FFFFFFF;
            ppdev->videoSource0.width        = lpSource->wWidth;
            ppdev->videoSource0.height       = lpSource->wHeight;
            ppdev->videoSource0.flags        = DF_SOURCEFLAG_IMPLICITSCALING;	
            ppdev->videoSource0.y_offset     = lpSource->fpVidMem;
            ppdev->videoSource0.y_pitch      = lpSource->lPitch;
            ppdev->videoSource0.uv_pitch     = lpSource->lPitch >> 1;

            if (ppdev->videoSource0.video_format == DF_VIDFMT_Y0Y1Y2Y3)
            {
                unsigned long offset1, offset2;
			    						
			    offset1 = lpSource->wHeight * ppdev->videoSource0.y_pitch;
			    offset2 = offset1 + (lpSource->wHeight >> 1) * (ppdev->videoSource0.y_pitch >> 1);

			    if (lpSource->dwReserved1 & DD_RESERVED_420_UFIRST)
			    {
				    ppdev->videoSource0.u_offset = offset1 + lpSource->fpVidMem;
				    ppdev->videoSource0.v_offset = offset2 + lpSource->fpVidMem;
			    }
			    else
			    {
				    ppdev->videoSource0.v_offset = offset1 + lpSource->fpVidMem;
				    ppdev->videoSource0.u_offset = offset2 + lpSource->fpVidMem;
			    }
            }

            DstWidth  = lpUpdateOverlay->rDest.right  - lpUpdateOverlay->rDest.left;
		    DstHeight = lpUpdateOverlay->rDest.bottom - lpUpdateOverlay->rDest.top;
		    SrcWidth  = lpUpdateOverlay->rSrc.right   - lpUpdateOverlay->rSrc.left;
		    SrcHeight = lpUpdateOverlay->rSrc.bottom  - lpUpdateOverlay->rSrc.top;

            // HANDLE INTERLACED OUTPUT
            // Adding support for interlaced displays is tricky... On GeodeLX, the easiest
            // way to display progressive video content on an interlaced display is
            // to fetch every other line of the video buffer, i.e. to 'de-weave'.
            // However, due to the GeodeLX architecture, this implies that the even and odd
            // video fields will be scaled and filtered independently and then shown
            // together on an interlaced display.  This can yield to ghastly flicker.
            // The solution we are implementing is as follows:
            //
            // 1. When the overlay is being upscaled, we downscale the entire video buffer
            //    by an additional factor of 2 and display the same scaled data twice.
            // 2. When the overlay is being downscaled or is unscaled, we simply display
            //    one scaled field of the video window twice.
            // 3. The above two options are enabled via the registry when a user selects a
            //    custom mode.  The default driver behavior is to simply interlace the
            //    content, possibly causing flicker.
            //
            // The solution is further complicated in that some content is already interlaced.
            // For this content, we simply apply the even field data to the even output and the
            // odd field data to the odd field output.  The only mechanism that we currently
            // have to detect interlaced data is through the DDOVER_BOB and DDOVER_INTERLEAVED
            // flags.
            //
            if (ppdev->interlacedOut)
            {
                ppdev->downscaleVideo = 0;

                if (lpUpdateOverlay->dwFlags & DDOVER_INTERLEAVED)
                {
                    ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset + ppdev->videoSource0.y_pitch;
                    ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset + ppdev->videoSource0.uv_pitch;
                    ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset + ppdev->videoSource0.uv_pitch;

                    ppdev->videoSource0.height   >>= 1;
                    ppdev->videoSource0.y_pitch  <<= 1;
                    ppdev->videoSource0.uv_pitch <<= 1;
                }
                else if (lpUpdateOverlay->dwFlags & DDOVER_BOB)
                {
                    SrcHeight <<= 1;
                    ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset;
                    ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset;
                    ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset;
                }
                else
                {
                    if (ppdev->customMode->line_double_video_on_interlaced)
                    {
                        ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset;
                        ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset;
                        ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset;

                        // Apply the downscale only if the video is upscaled more than 1.5X.
                        // In these cases, applying a vertical downscale actually achieves a
                        // smaller scale factor.
                        //
                        if (DstHeight > SrcHeight && ((DstHeight - SrcHeight) > (SrcHeight >> 1)))
                        {
                            SrcHeight <<= 1;
                            ppdev->downscaleVideo = 1;
                        }
                        else
                        {
                            ppdev->videoSource0.height   >>= 1;
                            ppdev->videoSource0.y_pitch  <<= 1;
                            ppdev->videoSource0.uv_pitch <<= 1;
                        }
                    }
                    else
                    {
                        ppdev->videoSource1.y_offset = ppdev->videoSource0.y_offset + ppdev->videoSource0.y_pitch;
                        ppdev->videoSource1.u_offset = ppdev->videoSource0.u_offset + ppdev->videoSource0.uv_pitch;
                        ppdev->videoSource1.v_offset = ppdev->videoSource0.v_offset + ppdev->videoSource0.uv_pitch;

                        ppdev->videoSource0.height   >>= 1;
                        ppdev->videoSource0.y_pitch  <<= 1;
                        ppdev->videoSource0.uv_pitch <<= 1;
                    }
                }
            }
            else
            {
                // HANDLE INTERLEAVED BOB
                // When Bobbing interleaved data, we must double the pitch
                // (such that we only fetch one field of data.
                //
                if ((lpUpdateOverlay->dwFlags & (DDOVER_BOB | DDOVER_INTERLEAVED)) == (DDOVER_BOB | DDOVER_INTERLEAVED))
                {
                    ppdev->videoSource0.y_pitch  <<= 1;
                    ppdev->videoSource0.uv_pitch <<= 1;
                    SrcHeight >>= 1;
                }
            }

            // STORE THE CURRENT VIDEO POSITION
            // This is similar to storing the source configuration.  It is not passed
            // directly, but updated based on clipping and panning.
            //
            ppdev->scaleSrcWidth           = SrcWidth;
            ppdev->scaleSrcHeight          = SrcHeight;
            ppdev->scaleDstWidth           = DstWidth;
            ppdev->scaleDstHeight          = DstHeight;
            ppdev->videoPosition.x         = lpUpdateOverlay->rDest.left;
            ppdev->videoPosition.y         = lpUpdateOverlay->rDest.top;
            ppdev->videoPosition.width     = DstWidth;
            ppdev->videoPosition.height    = DstHeight;
            ppdev->videoPosition.left_clip = lpUpdateOverlay->rSrc.left;
            ppdev->videoPosition.flags     = 0;
            ppdev->clippedSrcLines         = lpUpdateOverlay->rSrc.top;

            SetVideoPosition (ppdev, 1);

            df_set_video_filter_coefficients (NULL, 1);
            df_set_video_enable (1, 0);
			
			ppdev->OverlayActive  = TRUE;
			ppdev->VideoFlipRecord.fpFlipFrom = lpSource->fpVidMem;
            ppdev->VideoFlipRecord.fpFlipTo   = 0;
		}
						
   		// COLOR KEY?
		
		dstColorKey = FALSE;
		srcColorKey = FALSE;

		if (lpUpdateOverlay->dwFlags & DDOVER_KEYDEST)
		{
			ColorKey  = lpUpdateOverlay->lpDDDestSurface->ddckCKDestOverlay.dwColorSpaceLowValue;
			dstColorKey = TRUE;
		}
		else if (lpUpdateOverlay->dwFlags & DDOVER_KEYDESTOVERRIDE)
		{
			ColorKey  = lpUpdateOverlay->overlayFX.dckDestColorkey.dwColorSpaceLowValue;
			dstColorKey = TRUE;
		}
		if (lpUpdateOverlay->dwFlags & DDOVER_KEYSRC)
		{
			ColorKey  = lpUpdateOverlay->lpDDSrcSurface->ddckCKSrcOverlay.dwColorSpaceLowValue;
			srcColorKey = TRUE;
		}
		else if (lpUpdateOverlay->dwFlags & DDOVER_KEYSRCOVERRIDE)
		{
			ColorKey  = lpUpdateOverlay->overlayFX.dckSrcColorkey.dwColorSpaceLowValue;
			srcColorKey = TRUE;
		}

		if (srcColorKey || dstColorKey)
		{
			if (srcColorKey && dstColorKey)
			{
				// ONLY ONE COLOR KEY IS SUPPORTED
				
				lpUpdateOverlay->ddRVal = DDERR_OUTOFCAPS;
			
				return DDHAL_DRIVER_HANDLED;
			}		
			
			// SOURCE COLOR KEY - CHROMA KEY OFF VIDEO DATA
			// The video data will always be 16BPP

			if (srcColorKey)
			{
				SetChromaKey (ppdev, ColorKey, 16, 0);
			}

			// DESTINATION COLOR KEY - COLOR KEY OFF GRAPHICS DATA
			// The color key is in the current display mode.

			else
			{
				SetChromaKey (ppdev, ColorKey, ppdev->cBitsPerPel, 1);
			}
		}
		else
		{
			// DISABLE CHROMA KEY
			// Video will always be on top

			SetChromaKey(ppdev, 0, 0, 1);
		}

		lpUpdateOverlay->ddRVal = DD_OK;
	}
	DISPDBG((ppdev, 200, "DdUpdateOverlay leaves\n"));
	return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DWORD DdSetOverlayPosition
//
DWORD DdSetOverlayPosition(
	PDD_SETOVERLAYPOSITIONDATA lpSetOverlayPosition)
{
	PDEV	*ppdev;

	ppdev = (PDEV*) lpSetOverlayPosition->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "SET OVERLAY POSITION\n" ));
	DISPDBG((ppdev, 200, "Overlay surface = %08lx\n", lpSetOverlayPosition->lpDDSrcSurface ));
	DISPDBG((ppdev, 200, "(%ld,%ld)\n", lpSetOverlayPosition->lXPos, lpSetOverlayPosition->lYPos ));

	ppdev->videoPosition.x = lpSetOverlayPosition->lXPos;
	ppdev->videoPosition.y = lpSetOverlayPosition->lYPos;
	SetVideoPosition (ppdev, 0);

	lpSetOverlayPosition->ddRVal = DD_OK;
	return DDHAL_DRIVER_HANDLED;

}

/////////////////////////////////////////////////////////////////////////
// DWORD DdGetScanLine
//
DWORD DdGetScanLine (PDD_GETSCANLINEDATA pGetScanLine)
{
	PDEV	*ppdev;

	ppdev = (PDEV*) pGetScanLine->lpDD->dhpdev;

	DISPDBG((ppdev, 200, "DdGetScanLine\n" ));
	
    if (vg_test_timing_active() && vg_test_vertical_active())
    {
        pGetScanLine->dwScanLine = vg_get_current_vline();
        pGetScanLine->ddRVal = DD_OK;
    }
    else
        pGetScanLine->ddRVal = DDERR_VERTICALBLANKINPROGRESS;

    return DDHAL_DRIVER_HANDLED;
}

