//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// DirectDraw video helper routines.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"

/////////////////////////////////////////////////////////////////////////
// bIntersect
//
// Intersects two rectangles and returns the common area.
//
BOOL bIntersect (
    RECTL FAR *  pRcl1,
    RECTL FAR *  pRcl2,
    RECTL FAR *  pRclResult)
{
    pRclResult->left  = max(pRcl1->left,  pRcl2->left);
    pRclResult->right = min(pRcl1->right, pRcl2->right);

    // CHECK FOR HORIZONTAL INTERSECTION

    if (pRclResult->left <= pRclResult->right)
    {
        pRclResult->top    = max(pRcl1->top,    pRcl2->top);
        pRclResult->bottom = min(pRcl1->bottom, pRcl2->bottom);

        if (pRclResult->top <= pRclResult->bottom)
        {
            // AN INTERSECTION HAS BEEN FOUND

            return TRUE;
        }
    }

    return FALSE;
}

/////////////////////////////////////////////////////////////////////////
// SetChromaKey
//
// Loads the chroma key color into the hardware
// Key should be in Bpp specified
//
// If Key and Bpp are 0 then disable color key (video always shown)
//
BOOL SetChromaKey (PDEV *ppdev, DWORD Key, DWORD Bpp, int graphics)
{
	DWORD R,G,B;
	DWORD data32;
    ULONG lock;
    ULONG palB, palG, palR;

    // Save away the key information for later
    //
    ppdev->colorKey = Key;
    ppdev->graphics = graphics;
    ppdev->colorBPP = Bpp;
    
	if ((Key == 0) && (Bpp == 0))
	{
		df_set_video_color_key (0,0,1);
	}
	else
	{
		if (Bpp == 16)
		{
            // EXPAND 16BPP COLORS
            // The VG pads the colors with 0, while the DF copies the MSBs to the LSBs.
            //
            if (graphics)
            {
                B =  (Key & 0x1F) << 3;
			    G = ((Key >> 5) & 0x3F) << 2;
			    R = ((Key >> 11) & 0x1F) << 3;			
            }
            else
            {
                B = ((Key & 0x001F) << 3) | ((Key & 0x001F) >> 2);
			    G = ((Key & 0x07E0) >> 3) | ((Key & 0x07E0) >> 9);
			    R = ((Key & 0xF800) >> 8) | ((Key & 0xF800) >> 13);
            }
		}
		else if (Bpp == 8)
		{
			// READ 8BPP PALETTE ENTRY
			
			vg_get_display_palette_entry (Key, &data32);
			R = (data32 >> 16) & 0xFF;
			G = (data32 >> 8) & 0xFF;
			B =  data32 & 0xFF;
		}
		else
		{
			R = (Key & 0xFF0000) >> 16;
			G = (Key & 0x00FF00) >> 8;
			B = (Key & 0x0000FF);
		}	

        if (ppdev->cPelSize != 0 && graphics)
        {
            // Run through display palette.  We assume that the display palette is always
            // enabled.
            //
            vg_get_display_palette_entry (R, &palR);
            vg_get_display_palette_entry (G, &palG);
            vg_get_display_palette_entry (B, &palB);

            R = (palR >> 16) & 0xFF;
            G = (palG >> 8)  & 0xFF;
            B =  palB & 0xFF;
        }
		df_set_video_color_key ((B | (G << 8) | (R << 16)),
			0x00FFFFFF, graphics);

        // SPECIAL CASE FOR CHROMA KEYING
        // This driver configures and uses an alpha window to allow the cursor to
        // be alpha-blended with video data (instead of the underlying color key data).
        // However, the alpha blending hardware will blend graphics and video on the
        // event of a chroma key mismatch.  In order for video to appear properly, we
        // must hardwire the color key hardware to be enabled while chroma keying.  Note
        // that this implementation assumes no per-pixel alpha blending, so it will need to
        // be updated to support that feature.
        //
        if (!graphics)
        {
            lock = READ_REG32 (DC3_UNLOCK);
            WRITE_REG32 (DC3_UNLOCK, DC3_UNLOCK_VALUE);
            WRITE_REG32 (DC3_COLOR_KEY, DC3_CLR_KEY_ENABLE);
		    WRITE_REG32 (DC3_COLOR_MASK, 0);
            WRITE_REG32 (DC3_UNLOCK, lock);
        }
    }

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// SetVideoPosition
//
// Sets the position and size of the overlay window
//
BOOL SetVideoPosition (PDEV *ppdev, int scale)
{
    DF_VIDEO_SOURCE_PARAMS  videoSource0;
    DF_VIDEO_SOURCE_PARAMS  videoSource1;
    DF_VIDEO_POSITION       videoPosition;
	LX_REGISTRY_SETTINGS    *pSettings;
	RECTL display, ovly, result;
	long xstart, xend, ystart, yend, xclip = 0;
	unsigned long temp, lines, offset;
	unsigned long xcrop, vcfg;

    // START WITH SAVED VIDEO SETTINGS

    videoSource0  = ppdev->videoSource0;
    videoSource1  = ppdev->videoSource1;
    videoPosition = ppdev->videoPosition;
    videoPosition.flags = DF_POSFLAG_DIRECTCLIP;
    videoPosition.dst_clip = 0;

    // CHECK FOR INTERSECTION BETWEEN VIDEO AND DISPLAY
    // This is only necessary when panning.
	
    pSettings = &ppdev->regSettings;
    if (pSettings->CustomPanelTimings ||
       (pSettings->ManualTimingEnable && (ppdev->customMode->manual_timings.flags & VG_MODEFLAG_PANELOUT)) ||
      (!pSettings->ManualTimingEnable && pSettings->PanelEnable))
    {
        display.left   = ppdev->panning.start_x;
        display.top    = ppdev->panning.start_y;

        if (pSettings->ManualTimingEnable || pSettings->CustomPanelTimings)
        {
            display.right  = display.left + ppdev->customMode->manual_timings.panel_width;
            display.bottom = display.top  + ppdev->customMode->manual_timings.panel_height;
        }
        else
        {
            display.right  = display.left + pSettings->PanelWidth;
            display.bottom = display.top  + pSettings->PanelHeight;
        }

        ovly.left   = videoPosition.x;
        ovly.top    = videoPosition.y;
        ovly.right  = videoPosition.x + videoPosition.width;
        ovly.bottom = videoPosition.y + videoPosition.height;

		if (bIntersect(&display, &ovly, &result))
		{
			videoPosition.x -= ppdev->panning.start_x;
            videoPosition.y -= ppdev->panning.start_y;
		}
        else
        {
            videoPosition.x = videoPosition.y = 0;
            videoPosition.width = 0;
        }
    }

	// ACCOUNT FOR TOP AND LEFT CLIPPING
	// Clipping from Windows Media Player is in terms of source pixels.
    // Clipping from panning is in terms of destination pixels.  We
    // must therefore keep track of the scale and calculate the exact
    // number of bytes needed to combine the two clipping values.

	lines = ppdev->clippedSrcLines;

	// CHECK FOR TOP CLIPPING FROM PANNING OR NEGATIVE COORDINATES

	if (videoPosition.y < 0)
	{
		lines  += (-(videoPosition.y) * ppdev->scaleSrcHeight) / ppdev->scaleDstHeight;
		videoPosition.y = 0;
	}
	if (videoPosition.x < 0)
	{
		videoPosition.left_clip += (-(videoPosition.x) * ppdev->scaleSrcWidth) / ppdev->scaleDstWidth;
        videoPosition.dst_clip   = ((videoPosition.left_clip & 3) * ppdev->scaleDstWidth) / ppdev->scaleSrcWidth;
        videoPosition.left_clip &= 0xFFFC;
		videoPosition.x = 0;
	}

    // SAVE BUFFER OFFSETS
    // These offsets are precalculated and used later when flipping
    // the video buffer.

    if (ppdev->interlacedOut && !ppdev->downscaleVideo)
        lines >>= 1;

    ppdev->yBufferOffset  =  lines * videoSource0.y_pitch;
    ppdev->uvBufferOffset = (lines >> 1) * videoSource0.uv_pitch;

    videoSource0.y_offset += ppdev->yBufferOffset;
    videoSource0.u_offset += ppdev->uvBufferOffset;
    videoSource0.v_offset += ppdev->uvBufferOffset;

    if (ppdev->interlacedOut)
    {
        videoSource1.y_offset += ppdev->yBufferOffset;
        videoSource1.u_offset += ppdev->uvBufferOffset;
        videoSource1.v_offset += ppdev->uvBufferOffset;
    }

    // CONFIGURE AN ALPHA WINDOW
    // This will allow a color cursor to blend with video data or allow
    // user-specified alpha blending parameters.
    //
    ppdev->videoAlpha.x           = videoPosition.x;
    ppdev->videoAlpha.y           = videoPosition.y;
    ppdev->videoAlpha.width       = videoPosition.width;
    ppdev->videoAlpha.height      = videoPosition.height;
    ppdev->videoAlpha.priority    = 2;
    ppdev->videoAlpha.delta       = 0;

    if (!(ppdev->blendFlags & LX_BLEND_ENABLE))
    {
        ppdev->videoAlpha.alpha_value = 0xFF;
        ppdev->videoAlpha.flags       = DF_ALPHAFLAG_PERPIXELENABLED;
    }

    if (scale)
    {
        // DO NOT UPDATE DURING VBLANK
        // The DF latches at the end of VSync while the VG latches at the
        // beginning.  If we update the configuration in the middle of VSync
        // we will thus see a frame of tearing.
        //
        while (!vg_test_vertical_active())
            ;

        df_set_video_scale (ppdev->scaleSrcWidth, ppdev->scaleSrcHeight,
            ppdev->scaleDstWidth, ppdev->scaleDstHeight,
            DF_SCALEFLAG_CHANGEX | DF_SCALEFLAG_CHANGEY);
    }

    df_configure_video_source (&videoSource0, &videoSource1);
    df_configure_alpha_window (0, &(ppdev->videoAlpha));

    df_set_alpha_window_enable (0, 1);
    df_set_video_position (&videoPosition);

    // Save the alpha window positions.
    // Save the y position locally and in the miniport.
    //
    ppdev->videoY = READ_VID32 (DF_VIDEO_Y_POS);
    if (EngDeviceIoControl (ppdev->hDriver,
                            IOCTL_SAVE_VIDEO_Y,
                            NULL,
                            0,
                            NULL,
                            0,
                            &temp))
    {
        DISPDBG((ppdev, 3000, "Failed IOCTL_SAVE_VIDEO_Y\n"));
    }

	return TRUE;
}

