//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// DirectDraw VPE routines.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"

extern unsigned long exclusive_mode;

// PREDEFINED MISCELLANEOUS CALLBACKS
//
const DD_MISCELLANEOUSCALLBACKS GeodeLXMiscellaneousCallbacks =
{
    sizeof (DD_MISCELLANEOUSCALLBACKS),
    DDHAL_MISCCB32_GETAVAILDRIVERMEMORY,
    DdGetAvailDriverMemory
};

// PREDEFINED VIDEOPORT CALLBACKS
// All callback routines are supported.


const DD_VIDEOPORTCALLBACKS GeodeLXVideoPortCallbacks =
{
    sizeof(DD_VIDEOPORTCALLBACKS),

    DDHAL_VPORT32_CANCREATEVIDEOPORT     |
    DDHAL_VPORT32_CREATEVIDEOPORT        |
    DDHAL_VPORT32_FLIP                   |
    DDHAL_VPORT32_GETBANDWIDTH           |
    DDHAL_VPORT32_GETINPUTFORMATS        |
    DDHAL_VPORT32_GETOUTPUTFORMATS       |
    DDHAL_VPORT32_GETFIELD               |
    DDHAL_VPORT32_GETLINE                |
    DDHAL_VPORT32_GETCONNECT             |
    DDHAL_VPORT32_DESTROY                |
    DDHAL_VPORT32_GETFLIPSTATUS          |
    DDHAL_VPORT32_UPDATE                 |
    DDHAL_VPORT32_WAITFORSYNC            |
    DDHAL_VPORT32_GETSIGNALSTATUS        |
    DDHAL_VPORT32_COLORCONTROL,

    DdVideoPortCanCreate,
    DdVideoPortCreate,
    DdVideoPortFlip,
    DdVideoPortGetBandwidth,
    DdVideoPortGetInputFormats,
    DdVideoPortGetOutputFormats,
    NULL ,
    DdVideoPortGetField,
    DdVideoPortGetLine,
    DdVideoPortGetConnectInfo,
    DdVideoPortDestroy,
    DdVideoPortGetFlipStatus,
    DdVideoPortUpdate,
    DdVideoPortWaitForSync,
    DdVideoPortGetSignalStatus,
    NULL
};

const DDVIDEOPORTCAPS GeodeLXVideoPortCaps =
{
    sizeof(DDVIDEOPORTCAPS),

    // VIDEOPORT FLAGS

    DDVPD_WIDTH             |
    DDVPD_HEIGHT            |
    DDVPD_ID                |
    DDVPD_CAPS              |
    DDVPD_FX                |
    DDVPD_ALIGN,

    4096,                       // Max video width
    4096,                       // Max VBI width
    4096,                       // Max video height
    0,                          // VP ID

    // VIDEOPORT CAPS

    DDVPCAPS_INTERLACED     |
    DDVPCAPS_NONINTERLACED  |
    DDVPCAPS_OVERSAMPLEDVBI |
    DDVPCAPS_READBACKFIELD  |
    DDVPCAPS_READBACKLINE   |
    DDVPCAPS_SYNCMASTER     |
    DDVPCAPS_VBISURFACE,

    // VIDEOPORT FX

    DDVPFX_INTERLEAVE       |
    DDVPFX_VBICONVERT       |
    DDVPFX_VBINOSCALE       |
    0,
    VIP_MAX_BUFFERS,            // Number of autoflippable surfaces
    4,                          // Byte restriction of placement within surface
    4,                          // Byte restriction of width after prescaling
    2,                          // Byte restriction of left cropping
    2,                          // Byte restriction of cropping width
    0,                          // shrink in steps of original width
    0,                          // shrink in steps of original height
    VIP_MAX_BUFFERS,            // Num VBI AutoFlip surfaces
    3,                          // Preferred Autoflip
    4,                          // Filter taps
    4
};

const DD_KERNELCALLBACKS GeodeLXKernelCallbacks =
{
    sizeof(DD_KERNELCALLBACKS),
    DDHAL_KERNEL_SYNCSURFACEDATA    |
    DDHAL_KERNEL_SYNCVIDEOPORTDATA,
    DdSyncSurfaceData,
    DdSyncVideoPortData
};

const DDKERNELCAPS GeodeLXKernelCaps =
{
    sizeof (DDKERNELCAPS),

    DDKERNELCAPS_AUTOFLIP                  |
    DDKERNELCAPS_FIELDPOLARITY             |
    DDKERNELCAPS_FLIPOVERLAY               |
    DDKERNELCAPS_FLIPVIDEOPORT             |
    DDKERNELCAPS_LOCK                      |
    DDKERNELCAPS_SETSTATE                  |
    0,

    DDIRQ_DISPLAY_VSYNC |
    DDIRQ_VPORT0_LINE   |
    DDIRQ_VPORT0_VSYNC
};

const DD_NTCALLBACKS GeodeLXNTCallbacks =
{
    sizeof (DD_NTCALLBACKS),
    DDHAL_NTCB32_SETEXCLUSIVEMODE,
    NULL,
    DdSetExclusiveMode,
    NULL
};

const DWORD dwConnectionCaps =
    DDVPCONNECT_DISCARDSVREFDATA    |    
    DDVPCONNECT_INTERLACED          |
    DDVPCONNECT_INVERTPOLARITY;

const GUID *guids[] =
{
    &DDVPTYPE_CCIR656,
    &DDVPTYPE_E_HREFH_VREFH,
    &DDVPTYPE_E_HREFH_VREFL,
    &DDVPTYPE_E_HREFL_VREFH,
    &DDVPTYPE_E_HREFL_VREFL
};

#define MAX_VP_TYPES (sizeof(guids)/sizeof(GUID *))

#define mmioFOURCC( ch0, ch1, ch2, ch3 )                                \
        ((DWORD)(BYTE)(ch0) | ( (DWORD)(BYTE)(ch1) << 8) |              \
        ((DWORD)(BYTE)(ch2) << 16 ) | ( (DWORD)(BYTE)(ch3) << 24 ))

const DDPIXELFORMAT pf422[] =
{
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('Y','U','Y','2'),16,(DWORD)-1,(DWORD)-1,(DWORD)-1},
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('Y','V','Y','U'),16,(DWORD)-1,(DWORD)-1,(DWORD)-1},
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('U','Y','V','Y'),16,(DWORD)-1,(DWORD)-1,(DWORD)-1},
    {sizeof(DDPIXELFORMAT),DDPF_RGB, 0,16,0xF800,0x07E0,0x001F}
};

#define NUM_FORMATS_422 (sizeof(pf422) / sizeof(DDPIXELFORMAT))

const DDPIXELFORMAT pf420[] =
{
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('Y','V','1','2'),12,(DWORD)-1,(DWORD)-1,(DWORD)-1},
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('I','Y','U','V'),12,(DWORD)-1,(DWORD)-1,(DWORD)-1},
    {sizeof(DDPIXELFORMAT),DDPF_FOURCC, mmioFOURCC('I','4','2','0'),12,(DWORD)-1,(DWORD)-1,(DWORD)-1},
};

#define NUM_FORMATS_420 (sizeof(pf420) / sizeof(DDPIXELFORMAT))

/////////////////////////////////////////////////////////////////////////
// DWORD DDGetDriverInfo
//
DWORD DdGetDriverInfo(PDD_GETDRIVERINFODATA lpData)
{
    PDEV *ppdev = NULL;

    ppdev = (PDEV*) lpData->dhpdev;
    lpData->ddRVal = DDERR_CURRENTLYNOTAVAIL;

    DISPDBG((ppdev,200,"DdGetDriverInfo Begins\n"));

    if (IsEqualIID(&(lpData->guidInfo), &GUID_MiscellaneousCallbacks) )
    {
        DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DD_MISCELLANEOUSCALLBACKS));

        lpData->dwActualSize = sizeof(DD_MISCELLANEOUSCALLBACKS);

        DISPDBG ((ppdev,2000, "GetDriverInfo: Miscellaneous Callbacks\n"));

        memcpy (lpData->lpvData, &GeodeLXMiscellaneousCallbacks, dwSize);
        lpData->ddRVal = DD_OK;
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_VideoPortCallbacks) )
    {
        if (g_vipHardware)
        {
            // VIP Functionality is optional in graphics device.
            //
            DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DD_VIDEOPORTCALLBACKS));

            lpData->dwActualSize = sizeof(DD_VIDEOPORTCALLBACKS);

            DISPDBG ((ppdev,200, "GetDriverInfo: VideoPortCallbacks\n"));

            memcpy (lpData->lpvData, &GeodeLXVideoPortCallbacks, dwSize);
            lpData->ddRVal = DD_OK;
        }
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_VideoPortCaps) )
    {
        if (g_vipHardware)
        {
            // VIP Functionality is optional in graphics device.
            //
            DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DDVIDEOPORTCAPS));
            
            lpData->dwActualSize = sizeof(DDVIDEOPORTCAPS);

            DISPDBG ((ppdev, 2000, "GetDriverInfo: VideoPort Caps\n"));
            DISPDBG ((ppdev, 2000, "dwSize is %i\n", dwSize));

            memcpy (lpData->lpvData, &GeodeLXVideoPortCaps, dwSize);
            lpData->ddRVal = DD_OK;  
        }
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_NTCallbacks) )
    {
        DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DD_NTCALLBACKS));
        
        lpData->dwActualSize = sizeof(DD_NTCALLBACKS);

        DISPDBG ((ppdev, 2000, "GetDriverInfo: NT Callbacks\n"));

        memcpy (lpData->lpvData, &GeodeLXNTCallbacks, dwSize);
        lpData->ddRVal = DD_OK;        
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_KernelCallbacks) )
    {
        if (g_vipHardware)
        {
            // VIP Functionality is optional in graphics device.
            //
            DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DD_KERNELCALLBACKS));
            
            lpData->dwActualSize = sizeof(DD_KERNELCALLBACKS);

            DISPDBG ((ppdev, 2000, "GetDriverInfo: Kernel Callbacks\n"));

            memcpy (lpData->lpvData, &GeodeLXKernelCallbacks, dwSize);
            lpData->ddRVal = DD_OK;
        }
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_KernelCaps) )
    {
        if (g_vipHardware)
        {
            // VIP Functionality is optional in graphics device.
            //
            DWORD dwSize = min(lpData->dwExpectedSize, sizeof(DDKERNELCAPS));
            
            lpData->dwActualSize = sizeof(DDKERNELCAPS);

            DISPDBG ((ppdev, 2000, "GetDriverInfo: Kernel Caps\n"));

            memcpy (lpData->lpvData, &GeodeLXKernelCaps, dwSize);
            lpData->ddRVal = DD_OK;
        }
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_ColorControlCallbacks) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_ColorControlCallbacks\n"));
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_ZPixelFormats) )
    {
         DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_ZPixelFormats\n"));
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_D3DExtendedCaps) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_D3DExtendedCaps\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_D3DCallbacks) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_D3DCallbacks\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_D3DCallbacks3) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_D3DCallbacks3\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_GetHeapAlignment) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_GetHeapAlignment\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_Miscellaneous2Callbacks) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_Miscellaneous2Callbacks\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_MotionCompCallbacks) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_MotionCompCallbacks\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_NonLocalVidMemCaps) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_UpdateNonLocalHeap\n"));    
    }
    if (IsEqualIID(&(lpData->guidInfo), &GUID_UpdateNonLocalHeap) )
    {
        DISPDBG((ppdev,200,"DdGetDriverInfo: Checking for GUID_NonLocalVidMemCaps\n"));    
    }

    DISPDBG((ppdev,200," \nDdGetDriverInfo Leaves\n"));
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortCanCreate
//
// This function is required
//
DWORD DdVideoPortCanCreate (PDD_CANCREATEVPORTDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;
    int i;

    DISPDBG ((ppdev, 3000, "Inside DdVideoPortCanCreate.\n"));
    
    lpInput->ddRVal = DD_OK;

    if (lpInput->lpDDVideoPortDesc->dwVideoPortID != 0)
        lpInput->ddRVal = DDERR_UNSUPPORTED;

    for (i = 0; i < MAX_VP_TYPES; i++)
    {
        if (IsEqualIID (&(lpInput->lpDDVideoPortDesc->VideoPortType.guidTypeID), guids[i]))
            break;
    }

    if (i == MAX_VP_TYPES)
        lpInput->ddRVal = DDERR_UNSUPPORTED;
    
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortCreate
//
DWORD DdVideoPortCreate(PDD_CREATEVPORTDATA lpInput)
{    
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DISPDBG ((ppdev, 3000, "Inside DdVideoPortCreate.\n"));

    // No work in this routine for the moment.  All VIP setup can be
    // accomplished in DdVideoPortUpdate
    //
    lpInput->ddRVal = 0;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortFlip
//
DWORD DdVideoPortFlip (PDD_FLIPVPORTDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;
    ULONG diff;

    DISPDBG ((ppdev, 3000, "Inside DdVideoPortFlip.\n"));

    if (ppdev->VPEFlipRecord.bFlipFlag)
    {
        if (!vip_is_buffer_update_latched ())
        {
            lpInput->ddRVal = DDERR_WASSTILLDRAWING;
            return DDHAL_DRIVER_HANDLED;
        }
    }

    ppdev->VPEFlipRecord.bFlipFlag = TRUE;

    // This routine is called whenever an application wants to manually control
    // the videoport stream.  Why it would do that when we support autoflipping
    // we may never know.
    //
    diff = ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A].even_base[0] -
           ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A].odd_base[0];

    ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A].odd_base[0] =
        lpInput->lpSurfTarg->lpGbl->fpVidMem + g_baseOffsets.framebuffer_base;
    ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A].even_base[0] =
        ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A].odd_base[0] + diff;

    vip_toggle_video_offsets (VIP_BUFFER_A, &ppdev->vipBuffer);
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetFlipStatus
//
// This required function allows DDRAW to restrict access to a surface
// until the physical flip has occurred, allowing doubled buffered capture.
//
DWORD DdVideoPortGetFlipStatus (PDD_GETVPORTFLIPSTATUSDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DISPDBG ((ppdev, 3000, "Inside DdVideoPortGetFlipStatus.\n"));

    if (!ppdev->VPEFlipRecord.bFlipFlag || vip_is_buffer_update_latched())
    {
        ppdev->VPEFlipRecord.bFlipFlag = FALSE;
        lpInput->ddRVal = DD_OK;
        return DDHAL_DRIVER_HANDLED;
    }

    lpInput->ddRVal = DDERR_WASSTILLDRAWING;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetBandwidth
//
DWORD DdVideoPortGetBandwidth (PDD_GETVPORTBANDWIDTHDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DDVIDEOPORTBANDWIDTH *lpOutput = lpInput->lpBandwidth;
    DDVIDEOPORTINFO *pInfo = &(lpInput->lpVideoPort->ddvpInfo);
    DDVIDEOPORTDESC *pDesc = &(lpInput->lpVideoPort->ddvpDesc);
    
    DISPDBG ((ppdev, 3000, "Inside DdVideoPortGetBandwidth.\n"));

    if (lpInput->dwFlags & DDVPB_TYPE)
    {
        lpOutput->dwSize = sizeof(DDVIDEOPORTBANDWIDTH);
        lpOutput->dwCaps = DDVPBCAPS_DESTINATION;
    }
    if (lpInput->dwFlags & DDVPB_VIDEOPORT)
    {    
        // VALUES OF 1000 INDICATE ZOOM FACTOR OF 1X
            
        lpOutput->dwOverlay            = 500;
        lpOutput->dwColorkey           = 500;
        lpOutput->dwYInterpolate       = 500;
        lpOutput->dwYInterpAndColorkey = 500;
    }
    if (lpInput->dwFlags & DDVPB_OVERLAY)
    {
        lpInput->ddRVal = DDERR_INVALIDPARAMS;
        return DDHAL_DRIVER_HANDLED;
    }
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetInputFormats
//
DWORD DdVideoPortGetInputFormats (PDD_GETVPORTINPUTFORMATDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DISPDBG ((ppdev, 2000, "Inside DdVideoPortGetInputFormats.\n"));

    lpInput->dwNumFormats = NUM_FORMATS_422;
    if (lpInput->lpddpfFormat != NULL)
    {
        memcpy (lpInput->lpddpfFormat, pf422, lpInput->dwNumFormats * sizeof (DDPIXELFORMAT));
    }
    
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetOutputFormats
//
DWORD DdVideoPortGetOutputFormats (PDD_GETVPORTOUTPUTFORMATDATA lpOutput)
{
    PDEV *ppdev = lpOutput->lpDD->lpGbl->dhpdev;

    DISPDBG ((ppdev, 2000, "Inside DdVideoPortGetOutputFormats.\n"));

    // ASSUME 1:1 TRANSLATION
    // We only support one output type for each input type.  The VIP is not
    // capable of any color conversion.  The VIP can, however, convert UYVY
    // to planar format.
    //
    if ((lpOutput->lpddpfInputFormat->dwFlags & DDPF_FOURCC) &&
         lpOutput->lpddpfInputFormat->dwFourCC == FOURCC_UYVY)
    {
        lpOutput->dwNumFormats = 1 + NUM_FORMATS_420;
        if (lpOutput->lpddpfOutputFormats != NULL)
        {
            memcpy (lpOutput->lpddpfOutputFormats,     lpOutput->lpddpfInputFormat, sizeof (DDPIXELFORMAT));
            memcpy (&lpOutput->lpddpfOutputFormats[1], pf420, NUM_FORMATS_420 * sizeof (DDPIXELFORMAT));
        }
    }
    else
    {
        lpOutput->dwNumFormats = 1;
        if (lpOutput->lpddpfOutputFormats != NULL)
        {
            memcpy (lpOutput->lpddpfOutputFormats, lpOutput->lpddpfInputFormat, sizeof (DDPIXELFORMAT));
        }
    }    

    lpOutput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetConnectInfo
//
// Returns the GUIDs describing the connections that can be supported.
// Think of a GUID simply as a 128 bit FOURCC that you don't have to
// register.  This function is required.
//
DWORD DdVideoPortGetConnectInfo (PDD_GETVPORTCONNECTDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;
    DWORD j;
    LPDDVIDEOPORTCONNECT pEntry;

    DISPDBG((ppdev, 3000, "In GetVideoPortConnectInfo32\n"));

    // RETURN GUIDS SUPPORTED
    //
    // If lpInput->lpGUIDs is NULL, simply return the number of GUIDS
    // supported.  Note that we double the value to indicate support for
    // 16 bit modes.
    //
    lpInput->dwNumEntries = MAX_VP_TYPES << 1;

    if (&lpInput->lpConnect[0] != NULL)
    {
        // FILL THE BUFFER
        // Otherwise, copy the supported GUIDs into the specifed buffer.
        // The HAL guarantees that the buffer is large enough to hold the
        // GUIDs.
    
        for (j = 0; j < (MAX_VP_TYPES << 1); j += 2)
        {
            pEntry = &(lpInput->lpConnect[j]);
            memcpy(&(pEntry->guidTypeID), guids[j >> 1], sizeof(GUID));
            pEntry->dwPortWidth = 8;
            pEntry->dwFlags = dwConnectionCaps;
            pEntry->dwSize = sizeof(DDVIDEOPORTCONNECT);
            pEntry = &(lpInput->lpConnect[j + 1]);
            memcpy(&(pEntry->guidTypeID), guids[j >> 1], sizeof(GUID));
            pEntry->dwPortWidth = 16;
            pEntry->dwFlags = dwConnectionCaps;
            pEntry->dwSize = sizeof(DDVIDEOPORTCONNECT);
        }
    }
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortDestroy
//
// This optional function notifies the HAL when the video port
// has been destroyed.
//
DWORD DdVideoPortDestroy(PDD_DESTROYVPORTDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;
    
    DISPDBG((ppdev, 2000, "In DdVideoPortDestroy\n"));

    // Currently no work as there was no work in DdVideoPortCreate
    //
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_NOTHANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortUpdate
//
// This required function sets up the video port
//
DWORD DdVideoPortUpdate (PDD_UPDATEVPORTDATA lpInput)
{
    PDEV *ppdev = lpInput->lpDD->lpGbl->dhpdev;
    LPDDPIXELFORMAT     outputFormat;
    VIPSUBWINDOWBUFFER  subwindowBuffer;
    VIPGENLOCKBUFFER    genlockBuffer;
    VIPINPUTBUFFER_ADDR *taskAOffsets;
    DWORD               offset, flags, size;
    ULONG               height;
    ULONG               pitch;
    ULONG               stopHeight;
    ULONG               i, autoFlip[VIP_MAX_BUFFERS + 2];
    ULONG               ReturnedDataLength;

    DISPDBG ((ppdev, 2000, "Inside DdVideoPortUpdate.\n"));
    DISPDBG ((ppdev, 2000, "Flags = %i\n", lpInput->dwFlags));

    lpInput->ddRVal = DD_OK;

    if (lpInput->dwFlags == DDRAWI_VPORTSTART ||
        lpInput->dwFlags == DDRAWI_VPORTUPDATE)
    {
        // DECLINE ANY AUTOFLIP UPDATES
        // If we do not fail this call, DirectShow will in spite of our caps
        // assume that we support hardware autoflipping without the intervention
        // of a DxApi interrupt scheme.  However, if we fail this call, DirectDraw
        // will immediately recall this routine with the autoflip flag disabled and
        // fallback on an ISR-driven autoflip scheme.  The whole process should be
        // transparent to the application.
        //
        if (lpInput->lpVideoInfo->dwVPFlags & DDVP_AUTOFLIP)
        {
            lpInput->ddRVal = DDERR_UNSUPPORTED;
            return DDHAL_DRIVER_HANDLED;
        }

        if (!lpInput->lplpDDSurface)
        {
            DISPDBG ((ppdev, 3000, "  No surfaces in the call to DdVideoPortUpdate\n"));
            lpInput->ddRVal = DDERR_INVALIDPARAMS;
            return DDHAL_DRIVER_HANDLED;
        }

        // We only allow 10 autoflip surfaces
        //
        if (lpInput->dwNumAutoflip > VIP_MAX_BUFFERS)
        {
            lpInput->ddRVal = DDERR_INVALIDPARAMS;
            return DDHAL_DRIVER_HANDLED;
        }

        // Inform the miniport driver of the autoflip surfaces.  This allows the
        // miniport a working knowledge of the autoflip chain.
        //
        flags = lpInput->lpVideoPort->ddvpDesc.VideoPortType.dwFlags;
        size  = lpInput->lpVideoPort->ddvpDesc.VideoPortType.dwPortWidth;
        autoFlip[0] = flags & DDVPCONNECT_INTERLACED;
        autoFlip[1] = lpInput->dwNumAutoflip;
        for (i = 0; i < lpInput->dwNumAutoflip; i++)
            autoFlip[i + 2] = lpInput->lplpDDSurface[i]->lpLcl->lpGbl->fpVidMem;

        if (EngDeviceIoControl (ppdev->hDriver,
                            IOCTL_SET_AUTOFLIP_SURFACES,
                            &autoFlip[0],
                            (lpInput->dwNumAutoflip + 2) << 2,
                            &autoFlip[0],
                            (lpInput->dwNumAutoflip + 2) << 2,
                            &ReturnedDataLength))
        {
            DISPDBG((ppdev, 3000, "DdVideoPortUpdate - Failed IOCTL_SET_AUTOFLIP_SURFACES\n"));
            lpInput->ddRVal = DDERR_INVALIDPARAMS;
            return DDHAL_DRIVER_HANDLED;
        }

        // At last the heart of the code.  This routine is the most important of all
        // the VPE routines.  It is responsible for starting, updating or stopping the
        // flow of data through the videoport and for properly configuring the hardware.
        //
        ppdev->vipBuffer.current_buffer  = 0;
        ppdev->vipBuffer.flags           = 0;
        taskAOffsets = &(ppdev->vipBuffer.offsets[VIP_BUFFER_TASK_A]);

        // Start with the appropriate VIP format
        //
        memset (&ppdev->vipMode, 0, sizeof (ppdev->vipMode));
        ppdev->vipMode.stream_enables = VIP_ENABLE_TASKA;

        // CCIR656
        //
        if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_CCIR656))
        {
            if (size == 16)
                ppdev->vipMode.operating_mode = VIP_MODE_VIP2_16BIT;
            else if (size == 8 && !(flags & DDVPCONNECT_INTERLACED))
                ppdev->vipMode.operating_mode = VIP_MODE_VIP2_8BIT;
            else if (size == 8)
                ppdev->vipMode.operating_mode = VIP_MODE_VIP1_8BIT;
            else
            {
                DISPDBG ((ppdev, 3000, "Unknown videoport size!\n"));
                lpInput->ddRVal = DDERR_INVALIDPARAMS;
                return DDHAL_DRIVER_HANDLED;
            }
        }

        // VIP 601
        //
        else
        {
            if (size == 16)
                ppdev->vipMode.operating_mode = VIP_MODE_16BIT601;
            else if (size == 8)
                ppdev->vipMode.operating_mode = VIP_MODE_8BIT601;
            else
            {
                DISPDBG ((ppdev, 3000, "Unknown videoport size!\n"));
                lpInput->ddRVal = DDERR_INVALIDPARAMS;
                return DDHAL_DRIVER_HANDLED;
            }

            // The polarity of the input signals is specified in the GUID.
            //
            if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_E_HREFH_VREFH))
                ppdev->vipMode.vip601_settings.flags = VIP_MODEFLAG_VSYNCACTIVEHIGH | VIP_MODEFLAG_HSYNCACTIVEHIGH;
            else if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_E_HREFH_VREFL))
                ppdev->vipMode.vip601_settings.flags = VIP_MODEFLAG_HSYNCACTIVEHIGH;
            else if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_E_HREFL_VREFH))
                ppdev->vipMode.vip601_settings.flags = VIP_MODEFLAG_VSYNCACTIVEHIGH;
            else if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_E_HREFL_VREFL))
                ppdev->vipMode.vip601_settings.flags = 0;
            else
            {
                DISPDBG ((ppdev, 3000, "Invalid GUID!\n"));
                lpInput->ddRVal = DDERR_INVALIDPARAMS;
                return DDHAL_DRIVER_HANDLED;
            }

            // Cropping rectangle specifies the video origins.  One very important thing to
            // note here is that the cropping rectangle is used even though cropping support
            // is disabled.  The driver is working on the assumption that the DirectDraw
            // middleware does not touch or even check the cropping parameters when the 
            // cropping flag is not set.  As such, it provides a custom 'back-door' method
            // for specifying VIP 601 origins.  In the event that this rectangle becomes
            // eventually filtered by the middleware, then a driver escape will be added to
            // supplement this feature.
            //
            ppdev->vipMode.vip601_settings.horz_start      = lpInput->lpVideoInfo->rCrop.left;
            ppdev->vipMode.vip601_settings.vert_start_odd  = lpInput->lpVideoInfo->rCrop.top;
            ppdev->vipMode.vip601_settings.vert_start_even = lpInput->lpVideoInfo->rCrop.bottom;          
            ppdev->vipMode.vip601_settings.width           = lpInput->lpVideoPort->ddvpDesc.dwFieldWidth;
            ppdev->vipMode.vip601_settings.odd_height      = lpInput->lpVideoPort->ddvpDesc.dwFieldHeight;
            ppdev->vipMode.vip601_settings.even_height     = lpInput->lpVideoPort->ddvpDesc.dwFieldHeight;
        }

        // Parse connection flags
        //
        if (flags & DDVPCONNECT_INVERTPOLARITY)
            ppdev->vipMode.flags |= VIP_MODEFLAG_INVERTPOLARITY;
        if (!(flags & DDVPCONNECT_INTERLACED))
            ppdev->vipMode.flags |= VIP_MODEFLAG_PROGRESSIVE;

        pitch  = lpInput->lplpDDSurface[0]->lpLcl->lpGbl->lPitch;
        height = lpInput->lplpDDSurface[0]->lpLcl->lpGbl->wHeight;
        offset = lpInput->lplpDDSurface[0]->lpLcl->lpGbl->fpVidMem;

        // Pitch is always the surface pitch.  Our DdCreateSurface routine will set
        // the pitch for YV12 to the width of the Y buffer and the width of the whole
        // buffer for 4:2:2.  This matches VIP's expectations.
        //
        taskAOffsets->y_pitch  = pitch;
        taskAOffsets->uv_pitch = pitch >> 1;

        if ((flags & DDVPCONNECT_INTERLACED) && (lpInput->lpVideoInfo->dwVPFlags & DDVP_INTERLEAVE))
        {
            // Set even offset to point to the second line.
            //
            taskAOffsets->odd_base[0]  = offset + g_baseOffsets.framebuffer_base;
            taskAOffsets->even_base[0] = offset + g_baseOffsets.framebuffer_base + pitch;

            // Pitch is 2X when interleaving in memory
            //
            taskAOffsets->y_pitch  <<= 1;
            taskAOffsets->uv_pitch <<= 1;

            // Limit capture to 1/2 the surface
            //
            stopHeight = height >> 1;
        }
        else
        {
            // Bob or progressive.  Program the hardware to toggle each field.
            //
            ppdev->vipMode.flags |= VIP_MODEFLAG_TOGGLEEACHFIELD;

            // Set even and odd offsets to the same buffer.
            //
            taskAOffsets->odd_base[0]  = offset + g_baseOffsets.framebuffer_base;
            taskAOffsets->even_base[0] = offset + g_baseOffsets.framebuffer_base;

            // Limit capture to the entire surface
            //
            stopHeight = height;
        }

        // Parse input fourCC
        //
        outputFormat = &lpInput->lplpDDSurface[0]->lpLcl->lpGbl->ddpfSurface;
        if ((outputFormat->dwFlags & DDPF_FOURCC) &&
            (outputFormat->dwFourCC == FOURCC_YV12 ||
             outputFormat->dwFourCC == FOURCC_IYUV ||
             outputFormat->dwFourCC == FOURCC_I420))
        {
            // 4:2:0 planar capture
            //
            ppdev->vipMode.flags   |= VIP_MODEFLAG_PLANARCAPTURE;
            ppdev->vipBuffer.flags |= VIP_INPUTFLAG_PLANAR;
            ppdev->vipMode.planar_capture = VIP_420CAPTURE_ALTERNATINGLINES;

            // Program the U and V offsets for weave deinterlacing.
            //
            if ((flags & DDVPCONNECT_INTERLACED) && (lpInput->lpVideoInfo->dwVPFlags & DDVP_INTERLEAVE))
            {
                if (outputFormat->dwFourCC == FOURCC_YV12)
                {
                    // V Before U
                    //
                    taskAOffsets->odd_voffset = pitch * height;
                    taskAOffsets->odd_uoffset = taskAOffsets->odd_voffset + ((height >> 1) * (pitch >> 1));

                    taskAOffsets->even_voffset = taskAOffsets->odd_voffset - (pitch >> 1);
                    taskAOffsets->even_uoffset = taskAOffsets->even_voffset + ((height >> 1) * (pitch >> 1));
                }
                else
                {
                    // U Before V
                    //
                    taskAOffsets->odd_uoffset = pitch * height;
                    taskAOffsets->odd_voffset = taskAOffsets->odd_uoffset + ((height >> 1) * (pitch >> 1));

                    taskAOffsets->even_uoffset = taskAOffsets->odd_uoffset - (pitch >> 1);
                    taskAOffsets->even_voffset = taskAOffsets->even_uoffset + ((height >> 1) * (pitch >> 1));
                }
            }
            else
            {
                if (outputFormat->dwFourCC == FOURCC_YV12)
                {
                    // V Before U
                    //
                    taskAOffsets->odd_voffset = taskAOffsets->even_voffset = pitch * height;
                    taskAOffsets->odd_uoffset = taskAOffsets->even_uoffset =
                        taskAOffsets->odd_voffset + ((height >> 1) * (pitch >> 1));
                }
                else
                {
                    // U Before V
                    //
                    taskAOffsets->odd_uoffset = taskAOffsets->even_uoffset = pitch * height;
                    taskAOffsets->odd_voffset = taskAOffsets->even_voffset =
                        taskAOffsets->odd_uoffset + ((height >> 1) * (pitch >> 1));
                }
            }
        }

        // Initialize the mode.  Note that the default threshold values are
        // programmed inside the initialization routine.
        //
        vip_initialize (&ppdev->vipMode);

        // Enable error detection (except frame error).  Why, you may ask.  The
        // answer is that the video OK signal between the VIP and the VG can be
        // used as a live signal status indicator for the VPE routines.  The
        // criteriae by which the video OK signal is deasserted are specified
        // in the genlock code.
        //
        memset (&genlockBuffer, 0, sizeof (genlockBuffer));
        genlockBuffer.vip_signal_loss =
            VIP_VDE_RUNAWAY_LINE    |
            VIP_VDE_CLOCKS_PER_LINE |
            VIP_VDE_LOST_CLOCK;

        vip_configure_genlock (&genlockBuffer);

        // Configure subwindow capture and start VIP capture.  Subwindow capture only
        // applies to CCIR656 input.  The subwindow is specified implicitly with VIP 
        // 601 capture.
        //
        if (IsEqualIID (&(lpInput->lpVideoPort->ddvpDesc.VideoPortType.guidTypeID), &DDVPTYPE_CCIR656))
        {
            subwindowBuffer.enable = 1;
            subwindowBuffer.start  = 0;
            subwindowBuffer.stop   = stopHeight;
            vip_set_subwindow_enable (&subwindowBuffer);
        }

        vip_configure_capture_buffers (VIP_BUFFER_A, &ppdev->vipBuffer);

        // Configure the max address.  A very important thing to note here is that
        // the VIP engine will generate a comparison when incrementing the line
        // counter at the end of the line, even when no writes are generated.  This
        // implies that the max address must be programmed at least one pitch value
        // past the end of the buffer for weaved capture.  We are making the assumption
        // that the pitch is never more than 4K.
        //
        vip_max_address_enable (g_baseOffsets.framebuffer_base + heap_end + 0x1000, 1);
        vip_set_capture_state (VIP_STARTCAPTUREATNEXTFRAME);
    }
    else
    {
        vip_terminate();
    }
        
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortWaitForSync
//
DWORD DdVideoPortWaitForSync (PDD_WAITFORVPORTSYNCDATA lpWaitForSync)
{
    PDEV *ppdev = lpWaitForSync->lpDD->lpGbl->dhpdev;
    unsigned long timeout;

    DISPDBG ((ppdev, 2000, "Inside DdVideoPortWaitForSync.\n"));

    // Assume 25000 reads per ms
    //
    timeout = 25000 * lpWaitForSync->dwTimeOut;

    if (lpWaitForSync->dwFlags == DDVPWAIT_BEGIN ||
        lpWaitForSync->dwFlags == DDVPWAIT_END)
    {
        while (timeout)
        {
            if (!(READ_VIP32 (VIP_STATUS) & VIP_STATUS_VBLANK))
                break;

            timeout--;
        }

        while (timeout)
        {
            if (READ_VIP32 (VIP_STATUS) & VIP_STATUS_VBLANK)
                break;

            timeout--;
        }

        if (!timeout)
        {
            lpWaitForSync->ddRVal = DDERR_VIDEONOTACTIVE;
            return DDHAL_DRIVER_HANDLED;
        }
    }
    else
    {
        while (timeout)
        {
            if (vip_get_current_line() == lpWaitForSync->dwLine)
                break;

            timeout--;
        }

        if (!timeout)
        {
            lpWaitForSync->ddRVal = DDERR_VIDEONOTACTIVE;
            return DDHAL_DRIVER_HANDLED;
        }
    }

    lpWaitForSync->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetSignalStatus
//
DWORD DdVideoPortGetSignalStatus(PDD_GETVPORTSIGNALDATA lpInput)
{
    PDEV *ppdev = NULL;

    ppdev = lpInput->lpDD->lpGbl->dhpdev;
    
    DISPDBG((ppdev, 2000, "In DdVideoPortGetSignalStatus\n"));

    if (vip_test_signal_status())
        lpInput->dwStatus = DDVPSQ_SIGNALOK;
    else
        lpInput->dwStatus = DDVPSQ_NOSIGNAL;

    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetField
//
// This function is only required if readback of the current
// field is supported.
//
DWORD DdVideoPortGetField(PDD_GETVPORTFIELDDATA lpInput)
{
    DWORD dwStatus;
    PDEV *ppdev=NULL;
    ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DISPDBG((ppdev,2000, "In DdVideoPortGetField\n"));

    // CHECK VIDEO PORT ENABLE STATUS
    // Make sure the video port is ON.  If not, set
    // lpInput->ddRVal to DDERR_VIDEONOTACTIVE and return.
    //
    if (vip_get_capture_state() < VIP_STARTCAPTUREATNEXTLINE)
    {
        lpInput->ddRVal = DDERR_VIDEONOTACTIVE;
        return DDHAL_DRIVER_HANDLED;
    }

    if (vip_get_current_field() == VIP_EVEN_FIELD)
        lpInput->bField = TRUE;
    else
        lpInput->bField = FALSE
        ;
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdVideoPortGetLine
//
// This function is only required if readback of the current
// video line number (0 relative) is supported.
//
DWORD DdVideoPortGetLine(PDD_GETVPORTLINEDATA lpInput)
{
    DWORD dwStatus;
    PDEV *ppdev = NULL;

    ppdev = lpInput->lpDD->lpGbl->dhpdev;

    DISPDBG((ppdev, 2000, "In DdVideoPortGetLine\n"));

    // CHECK VIDEO PORT ENABLE STATUS
    // Make sure the video port is ON.  If not, set
    // lpInput->ddRVal to DDERR_VIDEONOTACTIVE and return.
    //
    if (vip_get_capture_state() < VIP_STARTCAPTUREATNEXTLINE)
    {
        lpInput->ddRVal = DDERR_VIDEONOTACTIVE;
        return DDHAL_DRIVER_HANDLED;
    }

    lpInput->dwLine = vip_get_current_line();
    lpInput->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdSyncSurfaceData
//
DWORD DdSyncSurfaceData(PDD_SYNCSURFACEDATA  lpSyncSurface)
{
    DD_SURFACE_GLOBAL*    lpSource;
    unsigned long offset1, offset2;
    unsigned long pitch;

    PDEV *ppdev = (PDEV *)(lpSyncSurface->lpDD->lpGbl->dhpdev);

    DISPDBG ((ppdev, 3000, "DdSyncSurfaceData\n"));

    lpSource = lpSyncSurface->lpDDSurface->lpGbl;

    lpSyncSurface->dwSurfaceOffset = lpSource->fpVidMem;
    lpSyncSurface->dwOverlayOffset = lpSource->fpVidMem;
    lpSyncSurface->ddRVal = DD_OK;

    // Save U and V offsets to reduce flip latency.  Note that the offsets are adjusted if we
    // are Bobbing and interleaving.
    //
    pitch = lpSource->lPitch;
    offset1  = lpSource->wHeight * pitch;
    offset2  = offset1 + (lpSource->wHeight >> 1) * (pitch >> 1);
    lpSyncSurface->dwDriverReserved3 = 0;

    if ((lpSyncSurface->lpDDSurface->lpSurfMore->dwOverlayFlags & (DDOVER_BOB | DDOVER_INTERLEAVED)) == (DDOVER_BOB | DDOVER_INTERLEAVED))
        pitch <<= 1;

    if (lpSource->dwReserved1 & DD_RESERVED_420_UFIRST)
    {
        lpSyncSurface->dwDriverReserved1 = offset1 + lpSource->fpVidMem;
        lpSyncSurface->dwDriverReserved2 = offset2 + lpSource->fpVidMem;
    }
    else
    {
        lpSyncSurface->dwDriverReserved1 = offset2 + lpSource->fpVidMem;
        lpSyncSurface->dwDriverReserved2 = offset1 + lpSource->fpVidMem;
    }

    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdSyncVideoPortData
//
DWORD DdSyncVideoPortData(PDD_SYNCVIDEOPORTDATA  lpSyncVideoPort)
{
    PDEV *ppdev = (PDEV *)(lpSyncVideoPort->lpDD->lpGbl->dhpdev);

    DISPDBG ((ppdev, 3000, "DdSyncVideoPortData\n"));

    lpSyncVideoPort->dwOriginOffset = lpSyncVideoPort->lpVideoPort->lpSurface->lpLcl->lpGbl->fpVidMem;
    lpSyncVideoPort->ddRVal = DD_OK;

    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdSetExclusiveMode
//
// This routine is called when DirectDraw is entering or exiting exclusive mode.
//
DWORD DdSetExclusiveMode (PDD_SETEXCLUSIVEMODEDATA lpSetExclusiveMode)
{
    PDEV *ppdev = (PDEV *)(lpSetExclusiveMode->lpDD->dhpdev);

    if (lpSetExclusiveMode->dwEnterExcl)
    {
        DISPDBG ((ppdev, 3000, "Entering Exclusive Mode\n"));

        // FLAG EXCLUSIVE MODE
        // Exclusive mode is generally used when DirectDraw sets a custom mode.
        // These modes generally involve display flipping.  We thus flag exclusive
        // mode such that the next mode set sets a linear pitch and does not enable
        // compression, as a power of 2 pitch and compression do not benefit
        // applications that perform full screen animation.

        exclusive_mode = 1;
    }
    else
    {
        DISPDBG ((ppdev, 3000, "Exiting Exclusive Mode\n"));
        exclusive_mode = 0;
    }

    lpSetExclusiveMode->ddRVal = DD_OK;
    return DDHAL_DRIVER_HANDLED;
}

/////////////////////////////////////////////////////////////////////////
// DdGetAvailDriverMemory
//
// This routine is called to retrieve the current state of the DirectDraw heap.
//
DWORD DdGetAvailDriverMemory (PDD_GETAVAILDRIVERMEMORYDATA lpGetAvailDriverMemory)
{
    PDEV *ppdev = NULL;
    DISPDBG ((ppdev, 200, "DdGetAvailDriverMemory\n"));
    lpGetAvailDriverMemory->ddRVal  = DD_OK;
    lpGetAvailDriverMemory->dwTotal = heap_end - heap_start;
    lpGetAvailDriverMemory->dwFree  = ddraw_offset - heap_start;

    return DDHAL_DRIVER_HANDLED;
}

