//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Display driver prototypes.
// </DOC_AMD_STD>
//

#ifndef _DRIVER_H
#define _DRIVER_H

#include "cim_parm.h"
#include "escape.h"
#include "lxshared.h"

//---------------------------------------------------------
// GLOBAL DRIVER DEFINITIONS
//---------------------------------------------------------

// DEBUG DEFINITIONS

#define DEBUG_LOG_SIZE          0x50000
#define DEBUG_MODE_NT           0x001
#define DEBUG_MODE_LOG          0x002

// RANGE MACROS

#ifndef MAX
#define MAX(a,b)    (((a) >= (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b)    (((a) <= (b)) ? (a) : (b))
#endif

// MISCELLANEOUS

#define DLL_NAME                L"geodelx"        
#define STANDARD_DEBUG_PREFIX   "GEODELX: "     
                                            
// DRIVER EXTRA INFORMATION
// Currently not used
//
#define DRIVER_EXTRA_SIZE    0

// PALETTE IS 256 ENTRIES
//
#define MAX_CLUT_SIZE (sizeof(VIDEO_CLUT) + (sizeof(ULONG) * 256))

// ROUNDING MACRO
//
#define ROUND_UP_TO_64K(x)  (((ULONG)(x) + 0x10000 - 1) & ~(0x10000 - 1))

// DRIVER MEMORY ALLOCATION
// This macro hardcodes calls to EngAllocMem with our heap ID, 'Dlx '
//
#define LX_ALLOC_MEM(flags, size)  EngAllocMem((flags), (size), 0x20786C44)

// DRIVER PUNT MACRO
// This macro is used to check if a surface is actually currently being stored
// in system memory due to a GDI surface eviction caused by a new DirectDraw
// surface.
//
#define MAKE_PUNTABLE(surf)                                                  \
{                                                                            \
    ULONG index = dhsurf_array[(USHORT)surf->dhsurf];                        \
    if (!(index & CACHE_FLAG_SYSTEM))                                        \
    {                                                                        \
        index &= 0xFFFF;                                                     \
        if (!bitmap_heap[index].lockedSurf)                                  \
        {                                                                    \
            if (!AddLockedSurf (ppdev, index))                               \
                goto returnTrue;                                             \
        }                                                                    \
        surf = bitmap_heap[index].lockedSurf;                                \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        index &= 0xFFFF;                                                     \
        if (!system_heap[index].lockedSurf)                                  \
        {                                                                    \
            if (!AddLockedSurfForSystemBitmap (ppdev, index))                \
                goto returnTrue;                                             \
        }                                                                    \
        surf = system_heap[index].lockedSurf;                                \
    }                                                                        \
}

#define ROT_OFFSET(pdev_ptr, region, angle) \
    ((angle == 90)  ? ((region##top << (pdev_ptr)->cPelSize) + (((pdev_ptr)->cyScreen - region##right) * mode_pitch))   : \
     (angle == 270) ? ((((pdev_ptr)->cxScreen - region##bottom) << (pdev_ptr)->cPelSize) + (region##left * mode_pitch)) : \
      ((((pdev_ptr)->cxScreen - region##right) << (pdev_ptr)->cPelSize) + (((pdev_ptr)->cyScreen - region##bottom) * mode_pitch)))

#define ROT_WIDTH(region, angle) \
    ((angle != 180) ? (region##bottom - region##top) : (region##right - region##left))

#define ROT_HEIGHT(region, angle) \
    ((angle != 180) ? (region##right - region##left) : (region##bottom - region##top))

///////////////////////////////////////
//           I2C DEFINITIONS         //
///////////////////////////////////////

#define SMBUS_BASE 0x6000

#define SMBUS_SMBSDA                    (SMBUS_BASE + 0x00)
#define SMBUS_SMBST                     (SMBUS_BASE + 0x01)
#define SMBST_SLVSTP_SET                (1 << 7)
#define SMBST_SDAST_SET                 (1 << 6)
#define SMBST_BER_SET                   (1 << 5)
#define SMBST_NEGACK_SET                (1 << 4)
#define SMBST_STASTR_SET                (1 << 3)
#define SMBST_NMATCH_SET                (1 << 2)
#define SMBST_MASTER_SET                (1 << 1)
#define SMBST_XMIT_SET                  (1 << 0)
#define SMBUS_SMBCST                    (SMBUS_BASE + 0x02)
#define SMBCST_TGSCL_SET                (1 << 5)
#define SMBCST_TSDA_SET                 (1 << 4)
#define SMBCST_GCMTCH_SET               (1 << 3)
#define SMBCST_MATCH_SET                (1 << 2)
#define SMBCST_BB_SET                   (1 << 1)
#define SMBCST_BUSY_SET                 (1 << 0)
#define SMBUS_SMBCTL1                   (SMBUS_BASE + 0x03)
#define SMBCTL1_STASTRE_SET             (1 << 7)
#define SMBCTL1_NMINTE_SET              (1 << 6)
#define SMBCTL1_GCMEN_SET               (1 << 5)
#define SMBCTL1_RECACK_SET              (1 << 4)
#define SMBCTL1_DMAEN_SET               (1 << 3)
#define SMBCTL1_INTEN_SET               (1 << 2)
#define SMBCTL1_STOP_SET                (1 << 1)
#define SMBCTL1_START_SET               (1 << 0)
#define SMBUS_SMBADDR                   (SMBUS_BASE + 0x04)
#define SMBADDR_SAEN_SET                (1 << 7)
#define SMBUS_SMBCTL2                   (SMBUS_BASE + 0x05)
#define SMBCTL2_SCLFRQ_SHIFT            (1 << 1)
#define SMBCTL2_ENABLE_SET              (1 << 0)

int SMBusWrite (unsigned long data, unsigned long base, unsigned long address);

// FRAME BUFFER ALLOCATION DEFINITIONS
// We allocate 2MB of space for the command buffer and 1MB of space for
// a temporary buffer, for bug fixes, two pass operations, etc.  These definitions
// are mirrored in lxshared.h
//
#define LX_CMD_BUFFER_SIZE       0x200000
#define LX_CIM_SCRATCH_SIZE      0x100000
#define LX_RESERVED_SPACE        (LX_CMD_BUFFER_SIZE + LX_CIM_SCRATCH_SIZE)

// CRUDE SEMAPHORE
// The following macros are used to prevent re-entrant or multithreaded code from executing
// two routines that touch the GP at once.
//
#define ACQUIRE_PRIMITIVE_SEMAPHORE                     \
{                                                       \
    while (*(volatile unsigned long *)(&g_blt))         \
        WRITE_REG32 (DC3_UNLOCK, DC3_UNLOCK_VALUE);     \
    g_blt = 1;                                          \
}
#define RELEASE_PRIMITIVE_SEMAPHORE   { g_blt = 0; }

// GEODELX VG VIRTUAL REGISTER DEFINITIONS
// These defines are used to interact with the video BIOS.
//
#define VR_INDEX                0xAC1C
#define VR_DATA                 0xAC1E
#define VR_UNLOCK               0xFC53
#define VRC_VG                  0x0002
#define VG_MEM_SIZE             0x0000
#define FP_DETECT_MASK          0x0100
#define FP_SIMULTANEOUS_CRT     0x0400

// PUNT DEFINITIONS
// These defines are used to disable one or more driver features without restarting
// the system.
//
#define LX_PUNT_LINETO              0x00000001
#define LX_PUNT_STROKEPATH          0x00000002
#define LX_PUNT_MONOTEXT            0x00000004
#define LX_PUNT_ANTIALIASED         0x00000008
#define LX_PUNT_PREMULTIPLIED       0x00000010
#define LX_PUNT_CONSTANTALPHA       0x00000020
#define LX_PUNT_TRANSPARENTCONVERT  0x00000040
#define LX_PUNT_TRANSPARENTCOPY     0x00000080
#define LX_PUNT_MASKEDBLT           0x00000100
#define LX_PUNT_SOLIDFILL           0x00000200
#define LX_PUNT_MONOFILL            0x00000400
#define LX_PUNT_COLORFILL           0x00000800
#define LX_PUNT_SRCANDSOLIDPAT      0x00001000
#define LX_PUNT_SRCANDMONOPAT       0x00002000
#define LX_PUNT_SRCANDCOLORPAT      0x00004000
#define LX_PUNT_CONVERTMONO         0x00008000
#define LX_PUNT_CONVERT4BPP         0x00010000
#define LX_PUNT_CONVERT8BPP         0x00020000
#define LX_PUNT_CONVERT16BPP        0x00040000
#define LX_PUNT_CONVERT24BPP        0x00080000
#define LX_PUNT_CONVERT32BPP        0x00100000
#define LX_PUNT_NOCONVERT           0x00200000
#define LX_PUNT_BLTCOMPLEX          0x00400000
#define LX_PUNT_BLTSIMPLE           0x00800000
#define LX_PUNT_COPYMONO            0x01000000
#define LX_PUNT_COPY4BPP            0x02000000
#define LX_PUNT_COPY8BPP            0x04000000
#define LX_PUNT_COPY16BPP           0x08000000
#define LX_PUNT_COPY24BPP           0x10000000
#define LX_PUNT_COPY32BPP           0x20000000
#define LX_PUNT_COPYSIMPLE          0x40000000
#define LX_PUNT_ALPHACURSOR         0x80000000
#define LX_PUNT_ALL                 0xFFFFFFFF

// BLEND DEFINITIONS
// These definitions are used when the user manually enables
// an alpha window.
//
#define LX_BLEND_ENABLE             0x00000001
#define LX_BLEND_PER_PIXEL          0x00000002

//---------------------------------------------------------
// GLOBAL DRIVER STRUCTURE DEFINITIONS
//---------------------------------------------------------

// CLIPPING STRUCTURE
// The CLIPOBJ_xxx routines take a pointer to an ENUMRECTS structure.  However
// the definition for that structure only defines an array of one RECTL
// structure.  We create our own compatible definition to facilitate stack
// allocation.  This definitions allocates room for 50 clipping rectangles,
// taking a total of 804 bytes of stack space.

#define CLIP_LIMIT      50    

typedef struct _CLIPENUM
{
    LONG    c;
    RECTL    arcl[CLIP_LIMIT];    

} CLIPENUM;

// FLIPPING STRUCTURE
// Used to hold information about the offset we last flipped from and if we
// are currently waiting for a flip to be latched.
//
typedef struct _FLIPRECORD
{
    DWORD       fpFlipFrom;             // Surface we last flipped from
    DWORD       fpFlipTo;               // Surface we are flipping to
    BOOL        bFlipFlag;              // True if we think a flip is
                                        //   still pending
} FLIPRECORD;

// DRIVER PDEV
// The all-important omnipresent driver data structure.  This structure
// holds all relevant global information and is passed to virtually every
// driver routine.

typedef struct    _PDEV
{
    LONG        lDelta;                 // Distance from one scan to the next.
    LONG        cPelSize;               // 0 if 8bpp, 1 if 16bpp, 2 if 32bpp
    ULONG       iBitmapFormat;          // BMF_8BPP or BMF_16BPP or BMF_32BPP
                                        //     (our current colour depth)    
    ULONG       cimFormat;              // Cimarron source format for gp_set_source_format

    HANDLE      hDriver;                // Handle to \Device\Screen
    HDEV        hdevEng;                // Engine's handle to PDEV
    HSURF       hsurfScreen;            // Engine's handle to screen surface
    SURFOBJ*    pdsurfScreen;           // Our private DSURF for the screen

    ULONG       cxScreen;               // Visible screen width
    ULONG       cyScreen;               // Visible screen height
    ULONG       cBitsPerPel;            // Bits per pel (8, 15, 16, 24 or 32)
    LONG        refreshRate;            // Vertical refresh rate of the screen
    ULONG       screenStride;           // Screen Stride
    ULONG       ulMode;                 // Mode the mini-port driver is in.

    FLONG       flHooks;                // What we're hooking from GDI

    ////////// Palette stuff:

    PALETTEENTRY* pPal;                 // The palette if palette managed
    HPALETTE    hpalDefault;            // GDI handle to the default palette.
    FLONG       flRed;                  // Red mask for 16/32bpp bitfields
    FLONG       flGreen;                // Green mask for 16/32bpp bitfields
    FLONG       flBlue;                 // Blue mask for 16/32bpp bitfields
    ULONG       cPaletteShift;          // number of bits the 8-8-8 palette must
                                        // be shifted by to fit in the hardware
                                        // palette.

      ///////// Device Surfaces

    RECTL       scrnBounds;             // Dimensions of screen
    SURFOBJ    *pLockedSurf;            // Pre-locked surface around ppdev->LXBitmap for GDI punts
    HBITMAP     LXBitmap;               // Created bitmap for frame buffer access.
    HSURF       hsurf;                  // Device managed surface handle

    ///////// Cursor Settings

    VG_PANNING_COORDINATES  panning;
    BOOL        cursorEnabled;          // TRUE if HW cursor is on
    ULONG       startX;
    ULONG       startY;                 // Starting coordinates for panning
    ULONG       xHotSpot;               // Cursor hotspots
    ULONG       yHotSpot;
    ULONG       cursorOffset;           // Cursor buffer offset
    ULONG       activeBuffer;
    ULONG      *pCursorBuffer;
    BOOL        bSimulatedCursor;
    BOOL        bColorCursor;

    ///////// Miniport Settings

    LX_REGISTRY_SETTINGS    regSettings;
    GEODELX_CUSTOM_MODE *   customMode;
    ULONG                   interlacedOut;
    ULONG                   downscaleVideo;
    
    ///////// VPE Settings

    VIPINPUTBUFFER          vipBuffer;
    VIPSETMODEBUFFER        vipMode;
    FLIPRECORD              VPEFlipRecord;

    ///////// Gamma Settings

    PVOID                   fpState;
    ULONG                   redDisplayGamma[3];
    ULONG                   greenDisplayGamma[3];
    ULONG                   blueDisplayGamma[3];
    ULONG                   redOverlayGamma[3];
    ULONG                   greenOverlayGamma[3];
    ULONG                   blueOverlayGamma[3];

    ///////// Video Settings

    DF_VIDEO_SOURCE_PARAMS  videoSource0;
    DF_VIDEO_SOURCE_PARAMS  videoSource1;
    DF_VIDEO_POSITION       videoPosition;
    ULONG                   scaleSrcWidth;
    ULONG                   scaleSrcHeight;
    ULONG                   scaleDstWidth;
    ULONG                   scaleDstHeight;
    ULONG                   clippedSrcLines;
    ULONG                   yBufferOffset;
    ULONG                   uvBufferOffset;

    ///////// Alpha Blend Settings

    unsigned long           blendFlags;
    DF_ALPHA_REGION_PARAMS  videoAlpha;
    unsigned long           bpp32;
    
    ///////// DDraw Flip Settings

    FLIPRECORD              VideoFlipRecord;
    FLIPRECORD              GraphicsFlipRecord;
    BOOL                    OverlayActive;
    ULONG                   videoY;

    ULONG                   colorKey;
    ULONG                   colorBPP;
    int                     graphics;

        
#ifdef DBG
    UCHAR debugLog[DEBUG_LOG_SIZE];
    ULONG debugBegin;            
    ULONG debugEnd;             
    ULONG debugMode;            
#endif

} PDEV;

// BRUSH STRUCTURE
// The structure holds all information for created brushes.
//
#define PBRUSH_MASK 0x01
#define PBRUSH_2COLOR 0x02

typedef struct tagPBrush
{
    ULONG patFormat;
    ULONG monoColor0;
    ULONG monoColor1;
    ULONG Pattern[1];

} PBRUSH, *PPBRUSH;

//---------------------------------------------------------
// GLOBAL DRIVER ROUTINES
//---------------------------------------------------------

BOOL bEnablePalette(PDEV*);
BOOL bInitializePalette(PDEV*, DEVINFO*);
VOID vUninitializePalette(PDEV*);
BOOL bInitializeModeFields(PDEV*, GDIINFO*, DEVINFO*, DEVMODEW*);
BOOL bEnableHardware(PDEV*);
VOID vDisableHardware(PDEV*);
BOOL bAssertModeHardware(PDEV*, BOOL);
BOOL FindIntersection(RECTL *rect1, RECTL *rect2, RECTL *rectResult);
int ConfigureLookup (SURFOBJ *psoSrc, XLATEOBJ *pxlo, ULONG *bltFlags, UCHAR rop);
BOOL ConfigureLockedPrimary (PDEV *ppdev);
void RotateMonoPattern (PDEV *ppdev, ULONG data0_in, ULONG data1_in,
    ULONG *data0_out, ULONG *data1_out);
void RotateColorPattern (PDEV *ppdev, ULONG *src, ULONG *dst, ULONG format);
void UpdateBIOSOutputMode (PDEV *ppdev);
void cim_outw (unsigned short port, unsigned short address);
unsigned short cim_inw (unsigned short port);
void SetPalette (PDEV *ppdev, BOOL video);

//---------------------------------------------------------
// DRIVER GLOBALS
//---------------------------------------------------------

#ifdef INCLUDE_DRIVER_GLOBALS
#define DRIVER_GLOBAL
#else
#define DRIVER_GLOBAL  extern
#endif

DRIVER_GLOBAL INIT_BASE_ADDRESSES g_baseOffsets;
DRIVER_GLOBAL unsigned long compression_enabled;
DRIVER_GLOBAL unsigned long g_fbSize;
DRIVER_GLOBAL unsigned long heap_offset;
DRIVER_GLOBAL unsigned long heap_start;
DRIVER_GLOBAL unsigned long heap_end;
DRIVER_GLOBAL unsigned long ddraw_offset;
DRIVER_GLOBAL unsigned long mode_pitch;
DRIVER_GLOBAL unsigned long rot_pitch;
DRIVER_GLOBAL unsigned long rot_offset;
DRIVER_GLOBAL unsigned long rot_size;
DRIVER_GLOBAL unsigned long exclusive_mode;
DRIVER_GLOBAL unsigned long exclusive_set;
DRIVER_GLOBAL int g_vipHardware;
DRIVER_GLOBAL int g_blt;
DRIVER_GLOBAL PDEV *activePDEV;
DRIVER_GLOBAL GEODELX_CUSTOM_MODE *g_customModes;
DRIVER_GLOBAL unsigned long g_modeCount;
DRIVER_GLOBAL int g_puntAllUntilModeSet;
#if DBG
DRIVER_GLOBAL unsigned long g_puntMask;
#endif

#ifndef INCLUDE_DRIVER_GLOBALS
extern unsigned long gaussian_htaps[][5];
extern unsigned long gaussian_vtaps[][3];
#endif

#endif // _DRIVER_H
