//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//   
// <CTL_AMD_STD>
// </CTL_AMD_STD>
// 
// <DOC_AMD_STD>
// Pre-defined driver settings.
// </DOC_AMD_STD>
//

/////////////////////////////////////////////////////////////////////////
// GDIINFO ggdiDefault
//
// This structure contains the default GDIINFO fields that are passed back to GDI
// during DrvEnablePDEV.
//
// NOTE: This structure defaults to values for an 8bpp palette device.
//		 Some fields are overwritten for different color depths.
//
GDIINFO ggdiDefault = {
	GDI_DRIVER_VERSION,		// ulVersion (3.51, 4.0, etc.     
	DT_RASDISPLAY,			// ulTechnology                   
	0,						// ulHorzSize (filled in later)   
	0,						// ulVertSize (filled in later)   
	0,						// ulHorzRes (filled in later)    
	0,						// ulVertRes (filled in later)    
	0,						// cBitsPixel (filled in later)   
	0,						// cPlanes (filled in later)      
	20, 					// ulNumColors (palette managed)  
	0,						// flRaster (DDI reserved field)  

	0,						// ulLogPixelsX (filled in later) 
	0,						// ulLogPixelsY (filled in later) 

	// flTextCaps -- If we had wanted console windows to scroll by repainting 
	// the entire window, instead of doing a screen-to-screen blt, we         
	// would have set TC_SCROLLBLT (yes, the flag is bass-ackwards).          
	
	TC_RA_ABLE,

	0,						// ulDACRed   (filled in later)             
	0,						// ulDACGreen (filled in later)             
	0,						// ulDACBlue  (filled in later)             

	0x0024, 				// ulAspectX                                
	0x0024, 				// ulAspectY                                
	0x0033, 				// ulAspectXY (one-to-one aspect ratio)     

	1,						// xStyleStep                               
	1,						// yStyleStep                               
	3,						// denStyleStep -- Styles have a one-to-one 
	                        //                 aspect ratio, and every  
							//                 'dot' is 3 pixels long.  

	{ 0, 0 },				// ptlPhysOffset                            
	{ 0, 0 },				// szlPhysSize                              

	256,					// ulNumPalReg                              

	// These fields are for halftone initialization.  The actual values are 
	// a bit magic, but seem to work well on our display.                   
	 
	{						// ciDevice            
	   { 6700, 3300, 0 },	//		Red            
	   { 2100, 7100, 0 },	//		Green          
	   { 1400,	800, 0 },	//		Blue           
	   { 1750, 3950, 0 },	//		Cyan           
	   { 4050, 2050, 0 },	//		Magenta        
	   { 4400, 5200, 0 },	//		Yellow         
	   { 3127, 3290, 0 },	//		AlignmentWhite 
	   20000,				//		RedGamma       
	   20000,				//		GreenGamma     
	   20000,				//		BlueGamma      
	   0, 0, 0, 0, 0, 0 	//		No dye correction for raster displays 
	},

	0,						 // ulDevicePelsDPI (for printers only) 
	PRIMARY_ORDER_CBA,		 // ulPrimaryOrder                      
	HT_PATSIZE_4x4_M,		 // ulHTPatternSize                     
	HT_FORMAT_8BPP, 		 // ulHTOutputFormat                    
	HT_FLAG_ADDITIVE_PRIMS,  // flHTFlags                           
	0,						 // ulVRefresh                          
	0,						 // ulBltAlignment                      
	0,						 // ulPanningHorzRes                    
	0,						 // ulPanningVertRes                    
};

// SYSTEM FONTS 

#define SYSTM_LOGFONT {16,7,0,0,700,0,0,0,ANSI_CHARSET,OUT_DEFAULT_PRECIS,\
					   CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,\
					   VARIABLE_PITCH | FF_DONTCARE,L"System"}
#define HELVE_LOGFONT {12,9,0,0,400,0,0,0,ANSI_CHARSET,OUT_DEFAULT_PRECIS,\
					   CLIP_STROKE_PRECIS,PROOF_QUALITY,\
					   VARIABLE_PITCH | FF_DONTCARE,L"MS Sans Serif"}
#define COURI_LOGFONT {12,9,0,0,400,0,0,0,ANSI_CHARSET,OUT_DEFAULT_PRECIS,\
					   CLIP_STROKE_PRECIS,PROOF_QUALITY,\
					   FIXED_PITCH | FF_DONTCARE, L"Courier"}

/////////////////////////////////////////////////////////////////////////
// DEVINFO gdevinfoDefault
//
// This contains the default DEVINFO fields that are passed back to GDI
// during DrvEnablePDEV.
//
// NOTE: This structure defaults to values for an 8bpp palette device.
//		Some fields are overwritten for different color depths.
//
DEVINFO gdevinfoDefault = {
	(GCAPS_OPAQUERECT		|
	 GCAPS_MONO_DITHER      |
	 GCAPS_GRAY16),
	SYSTM_LOGFONT,								// lfDefaultFont 
	HELVE_LOGFONT,								// lfAnsiVarFont 
	COURI_LOGFONT,								// lfAnsiFixFont 
	0,											// cFonts 
	BMF_8BPP,									// iDitherFormat 
	0,											// cxDither 
	0,											// cyDither 
	0											// hpalDefault (filled in later) 
};

/////////////////////////////////////////////////////////////////////////
// DFVFN gadrvfn[]
//
// Build the driver function table gadrvfn with function index/address
// pairs.  This table tells GDI which DDI calls we support, and their
// location (GDI does an indirect call through this table to call us).
//
DRVFN gadrvfn[] = {
	{	INDEX_DrvEnablePDEV,			(PFN) DrvEnablePDEV 		    },
	{	INDEX_DrvCompletePDEV,			(PFN) DrvCompletePDEV		    },
	{	INDEX_DrvDisablePDEV,			(PFN) DrvDisablePDEV		    },
	{	INDEX_DrvEnableSurface, 		(PFN) DrvEnableSurface		    },
	{	INDEX_DrvDisableSurface,		(PFN) DrvDisableSurface 	    },
	{	INDEX_DrvCreateDeviceBitmap,	(PFN) DrvCreateDeviceBitmap     },
	{	INDEX_DrvDeleteDeviceBitmap,	(PFN) DrvDeleteDeviceBitmap     },
	{	INDEX_DrvAssertMode,			(PFN) DrvAssertMode 		    },
	{	INDEX_DrvMovePointer,			(PFN) DrvMovePointer		    },
	{	INDEX_DrvSetPointerShape,		(PFN) DrvSetPointerShape	    },
	{	INDEX_DrvSetPalette,			(PFN) DrvSetPalette 		    },
	{	INDEX_DrvCopyBits,				(PFN) DrvCopyBits			    },
	{	INDEX_DrvStrokePath,			(PFN) DrvStrokePath			    },
	{	INDEX_DrvTextOut,				(PFN) DrvTextOut			    },
	{	INDEX_DrvBitBlt,				(PFN) DrvBitBlt 			    },
	{	INDEX_DrvGetModes,				(PFN) DrvGetModes			    },
	{	INDEX_DrvRealizeBrush,			(PFN) DrvRealizeBrush		    },
	{	INDEX_DrvEscape,				(PFN) DrvEscape 			    },
    {   INDEX_DrvGetDirectDrawInfo,     (PFN) DrvGetDirectDrawInfo      },
    {   INDEX_DrvEnableDirectDraw,      (PFN) DrvEnableDirectDraw       },
    {   INDEX_DrvDisableDirectDraw,     (PFN) DrvDisableDirectDraw      },
	{   INDEX_DrvAlphaBlend,            (PFN) DrvAlphaBlend             },
	{   INDEX_DrvTransparentBlt,        (PFN) DrvTransparentBlt         },
    {   INDEX_DrvDisableDriver,         (PFN) DrvDisableDriver          },
    {   INDEX_DrvLineTo,                (PFN) DrvLineTo                 },
    {   INDEX_DrvIcmSetDeviceGammaRamp, (PFN) DrvIcmSetDeviceGammaRamp  },
};

#define NUM_DRIVER_ROUTINES (sizeof(gadrvfn) / sizeof(DRVFN))