//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// GeodeLX Driver escapes.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "escape.h"
#include "cim_rtns.h"

/////////////////////////////////////////////////////////////////////////
// DrvEscape
//
// Driver escape entry point.  This function should return TRUE for any
// supported escapes in response to QUERYESCSUPPORT, and FALSE for any
// others.  All supported escapes are called from this routine.
//
ULONG DrvEscape(
    SURFOBJ*    pso,
    ULONG        iEsc,
    ULONG        cjIn,
    VOID*        pvIn,
    ULONG        cjOut,
    VOID*        pvOut)
{
    PDEV*    ppdev = NULL;
    ULONG    iQuery, temp;
    LONG    lRet;
    int     enable;
    DWORD ReturnedDataLength;
    int i;

    ppdev = (PDEV*) pso->dhpdev;

    lRet = 0;  // Assume failure, or not suported

    switch (iEsc)
    {
        case QUERYESCSUPPORT:

            // Note:  we don't need to check cjIn for this case since
            // GDI validates this for us.

            iQuery = *(ULONG *)pvIn;

            switch (iQuery)
            {
                case GEODELX_GET_REGISTRY_SETTINGS:
                case GEODELX_SET_REGISTRY_SETTINGS:
                case GEODELX_CONFIGURE_ALPHA_WINDOW:
                case GEODELX_GET_LAST_MODE:
    
                    lRet = TRUE;
                    break;
            }
            break;

        case GEODELX_GET_REGISTRY_SETTINGS :
    
            // VERIFY THAT THE SIZE IS RIGHT
            //
            if (cjIn  == (sizeof (LX_REGISTRY_SETTINGS) + 8) &&
                cjOut == (sizeof (LX_REGISTRY_SETTINGS) + 8))
            {
                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    if (EngDeviceIoControl(ppdev->hDriver, IOCTL_GET_MINIPORT_REGISTRY,
                        (PVOID)((ULONG)pvIn + 8), cjIn - 8, (PVOID)((ULONG)pvOut + 8), cjOut - 8, &ReturnedDataLength))
                    {
                        DISPDBG((ppdev, 99999, "Registry reading failed\n"));
                        break;
                    }

                    lRet = TRUE;
                }
            }
    
            break;

        case GEODELX_SET_REGISTRY_SETTINGS :

            // VERIFY THAT THE SIZE IS RIGHT
            //
            if (cjIn  == (sizeof (LX_REGISTRY_SETTINGS) + 8) &&
                cjOut == (sizeof (LX_REGISTRY_SETTINGS) + 8))
            {
                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    if (EngDeviceIoControl(ppdev->hDriver, IOCTL_SET_LX_REGISTRY,
                        (PVOID)((ULONG)pvIn + 8), cjIn - 8, (PVOID)((ULONG)pvOut + 8), cjOut - 8, &ReturnedDataLength))
                    {
                        DISPDBG((ppdev, 99999, "Registry writing failed\n"));
                        break;
                    }

                    lRet = TRUE;
                }
            }
            break;

        case GEODELX_GET_DISPLAY_GAMMA_REGISTRY:
        case GEODELX_SET_DISPLAY_GAMMA_REGISTRY:
        case GEODELX_GET_OVERLAY_GAMMA_REGISTRY:
        case GEODELX_SET_OVERLAY_GAMMA_REGISTRY:
            
            if (cjIn  == (11 * sizeof(DWORD)) &&
                cjOut == (11 * sizeof(DWORD)))
            {
                DWORD ioCode = 0;

                if      (iEsc == GEODELX_GET_DISPLAY_GAMMA_REGISTRY) ioCode = IOCTL_GET_DISPLAY_GAMMA;
                else if (iEsc == GEODELX_SET_DISPLAY_GAMMA_REGISTRY) ioCode = IOCTL_SET_DISPLAY_GAMMA;
                else if (iEsc == GEODELX_GET_OVERLAY_GAMMA_REGISTRY) ioCode = IOCTL_GET_OVERLAY_GAMMA;
                else if (iEsc == GEODELX_SET_OVERLAY_GAMMA_REGISTRY) ioCode = IOCTL_SET_OVERLAY_GAMMA;

                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    if (EngDeviceIoControl(ppdev->hDriver, ioCode, (PVOID)((ULONG)pvIn + 8), 
                        cjIn - 8, (PVOID)((ULONG)pvOut + 8), cjOut - 8, &ReturnedDataLength))
                    {
                        DISPDBG((ppdev, 99999, "Gamma IOCTL failed: %i\n", (int)ioCode));
                        break;
                    }

                    if (iEsc == GEODELX_SET_DISPLAY_GAMMA_REGISTRY)
                    {
                        ULONG *gamma = (ULONG *)pvIn + 2;
                        for (i = 0; i < 3; i++)
                        {
                            ppdev->redDisplayGamma[i]   = gamma[i * 3];
                            ppdev->greenDisplayGamma[i] = gamma[(i * 3) + 1];
                            ppdev->blueDisplayGamma[i]  = gamma[(i * 3) + 2];
                        }
                        SetPalette (ppdev, FALSE);
                    }
                    else if (iEsc == GEODELX_SET_OVERLAY_GAMMA_REGISTRY)
                    {
                        ULONG *gamma = (ULONG *)pvIn + 2;
                        for (i = 0; i < 3; i++)
                        {
                            ppdev->redOverlayGamma[i]   = gamma[i * 3];
                            ppdev->greenOverlayGamma[i] = gamma[(i * 3) + 1];
                            ppdev->blueOverlayGamma[i]  = gamma[(i * 3) + 2];
                        }
                        SetPalette (ppdev, TRUE);
                    }

                    lRet = TRUE;
                }
            }
            break;

        case GEODELX_SET_DISPLAY_GAMMA:
        case GEODELX_SET_OVERLAY_GAMMA:
            
            if (cjIn  == (11 * sizeof(DWORD)) &&
                cjOut == (11 * sizeof(DWORD)))
            {
                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    if (iEsc == GEODELX_SET_DISPLAY_GAMMA)
                    {
                        ULONG *gamma = (ULONG *)pvIn + 2;
                        for (i = 0; i < 3; i++)
                        {
                            ppdev->redDisplayGamma[i]   = gamma[i * 3];
                            ppdev->greenDisplayGamma[i] = gamma[(i * 3) + 1];
                            ppdev->blueDisplayGamma[i]  = gamma[(i * 3) + 2];
                        }
                        SetPalette (ppdev, FALSE);
                    }
                    else if (iEsc == GEODELX_SET_OVERLAY_GAMMA)
                    {
                        ULONG *gamma = (ULONG *)pvIn + 2;
                        for (i = 0; i < 3; i++)
                        {
                            ppdev->redOverlayGamma[i]   = gamma[i * 3];
                            ppdev->greenOverlayGamma[i] = gamma[(i * 3) + 1];
                            ppdev->blueOverlayGamma[i]  = gamma[(i * 3) + 2];
                        }
                        SetPalette (ppdev, TRUE);
                    }

                    lRet = TRUE;
                }
            }
            break;


        case GEODELX_GET_LAST_MODE:
            
            if (cjIn  == (6 * sizeof(DWORD)) &&
                cjOut == (6 * sizeof(DWORD)))
            {
                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    if (EngDeviceIoControl(ppdev->hDriver, IOCTL_GET_LAST_MODE,
                        (PVOID)((ULONG)pvIn + 8), cjIn - 8, (PVOID)((ULONG)pvOut + 8), cjOut - 8, &ReturnedDataLength))
                    {
                        DISPDBG((ppdev, 99999, "Get Last Mode failed\n"));
                        break;
                    }

                    lRet = TRUE;
                }
            }
            break;

        case GEODELX_CONFIGURE_ALPHA_WINDOW:

            // VERIFY THAT THE SIZE IS RIGHT
            //
            if (cjIn  == (sizeof (LX_VIDEO_BLEND) + 8))
            {
                // CHECK FOR SECRET APPLICATION KEYWORD
                //
                if (*(unsigned long *)pvIn == GEODELX_SECRET_WORD)
                {
                    LX_VIDEO_BLEND *pBlend = (LX_VIDEO_BLEND *)((ULONG)pvIn + 8);

                    ppdev->blendFlags = pBlend->enable ? LX_BLEND_ENABLE : 0;

                    if (ppdev->blendFlags)
                    {
                        if (pBlend->per_pixel)
                        {
                            ppdev->blendFlags |= LX_BLEND_PER_PIXEL;
                            ppdev->videoAlpha.alpha_value = 0xFF;
                            ppdev->videoAlpha.flags = DF_ALPHAFLAG_PERPIXELENABLED;

                            if (ppdev->cBitsPerPel >= 24)
                                vg_set_display_bpp (32);
                            else
                            {
                                // Per-pixel blending is only available in 32BPP
                                //
                                return FALSE;
                            }
                        }
                        else
                        {
                            ppdev->videoAlpha.alpha_value = pBlend->fixed_alpha;
                            ppdev->videoAlpha.flags = 0;

                            if (ppdev->cBitsPerPel >= 24)
                            {
                                // Restore the BPP back to the user specified BPP
                                //
                                if (ppdev->bpp32)
                                    vg_set_display_bpp (32);
                                else
                                    vg_set_display_bpp (24);
                            }
                        }

                        // Apply the new alpha blending settings immediately.  This is
                        // benign if the video window is not currently enabled.
                        //
                        df_configure_alpha_window (0, &(ppdev->videoAlpha));
                        df_set_alpha_window_enable (0, 1);

                        // Disable color keying to allow exclusive alpha blending.
                        //
                        df_get_video_enable (&enable, &temp);
                        df_set_video_enable (enable, DF_ENABLEFLAG_NOCOLORKEY);
                    }
                    else
                    {
                        // Revert to per-pixel blending to allow the cursor to be blended
                        //
                        ppdev->videoAlpha.alpha_value = 0xFF;
                        ppdev->videoAlpha.flags = DF_ALPHAFLAG_PERPIXELENABLED;
                        df_configure_alpha_window (0, &(ppdev->videoAlpha));
                        df_set_alpha_window_enable (0, 1);

                        // Restore color/chroma keying.
                        //
                        df_get_video_enable (&enable, &temp);
                        df_set_video_enable (enable, 0);

                        if (ppdev->cBitsPerPel >= 24)
                        {
                            if (ppdev->bpp32)
                                vg_set_display_bpp (32);
                            else
                                vg_set_display_bpp (24);
                        }
                    }

                    lRet = TRUE;
                }
            }
            break;
#if DBG

        case GEODELX_SET_PUNT_FLAGS:

            // VERIFY THAT THE SIZE IS RIGHT
            //
            if (cjIn != 8)
            {
                lRet = FALSE;
                break;
            }

            // VERIFY THAT THE FIRST DWORD IS OUR SIGNATURE DWORD - 0xABADDOOD
            //

            if (*((ULONG *)pvIn) != GEODELX_SECRET_WORD)
            {
                lRet = FALSE;
                break;
            }

            // OTHERWISE SET THE NEW GLOBAL FLAGS
            //
            g_puntMask = *((ULONG *)((ULONG)pvIn + 4));
            lRet = TRUE;
            break;
#endif
    }

    return ((ULONG) lRet);
}
