//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// GDI DrvLineTo support.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_regs.h"
#include "cim_defs.h"

// ROP LOOKUP TABLE FOR PAINT ROUTINES
// Line routines use one of 16 basic ROPs, labeled
// a "mix".
//
USHORT MixToRop[] = {
	0xff,	// 1 - Used by paint routines
	0,		// 0
	0x05,	// DPon
	0x0a,	// DPna
	0x0f,	// Pn
	0x50,	// PDna
	0x55,	// Dn
	0x5a,	// DPx
	0x5f,	// DPan
	0xa0,	// DPa
	0xa5,	// DPxn
	0xaa,	// D
	0xaf,	// DPno
	0xf0,	// P
	0xf5,	// PDno
	0xfa,	// DPo
	0xff,	// 1
};

//////////////////////////////////////////////////////////////////////
// DrvLineTo
//
// Draws a single solid integer-only cosmetic line.
//
BOOL DrvLineTo(
	SURFOBJ*	pso,
	CLIPOBJ*	pco,
	BRUSHOBJ*	pbo,
	LONG		x0,
	LONG		y0,
	LONG		x1,
	LONG		y1,
	RECTL*		prclBounds,
	MIX 		mix)
{
    PDEV    *ppdev = NULL;
    BOOL    ret;
    ULONG   ROP;
    ULONG   VecMode;
    ULONG   dstOffset, dstPitch;
    ULONG   index;
    long    axialerr, diagerr, initerr;
    long    dx, dy, dmaj, dmin;
    USHORT  dhsurfTrg;
    RECTL   destRegion;
    RECTL  *clippedPtr;

    ppdev = (PDEV *)pso->dhpdev;

    DISPDBG ((ppdev, 1, "DrvLineTo Entry\n"));

    ACQUIRE_PRIMITIVE_SEMAPHORE;
    if (g_puntAllUntilModeSet)
        goto puntIt;

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

#if DBG
    if (g_puntMask & LX_PUNT_LINETO)
        goto puntIt;
#endif

    // EXTRACT DESTINATION OFFSET AND PITCH
    //
    index = dhsurf_array[(USHORT)pso->dhsurf];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntIt;

    dstOffset = bitmap_heap[(USHORT)index].heap_offset;
    dstPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

    // CALCULATE ROP AND DESTINATION REQUIRED

    ROP = (ULONG)MixToRop[mix & 0xF];

    if ((ROP & 0x55) ^ ((ROP >> 1) & 0x55))
        VecMode = GP3_VM_DST_REQ;
    else
        VecMode = 0;

    if (pco == NULL || pco->iDComplexity  == DC_TRIVIAL)
    {
        // Adjust the coordinates for rotation
        //
        if (ppdev->regSettings.RotationAngle != 0 && (USHORT)pso->dhsurf == 1)
        {
            dstOffset = 0;
            dstPitch  = mode_pitch;
            switch (ppdev->regSettings.RotationAngle)
            {
                case 90:
                {
                    long tx = x0;
                    x0 = y0;
                    y0 = (ppdev->cyScreen - (tx + 1));
                    tx = x1;
                    x1 = y1;
                    y1 = (ppdev->cyScreen - (tx + 1));
                    break;
                }
                case 180:
                {
                    x0 = (ppdev->cxScreen - (x0 + 1));
                    y0 = (ppdev->cyScreen - (y0 + 1));
                    x1 = (ppdev->cxScreen - (x1 + 1));
                    y1 = (ppdev->cyScreen - (y1 + 1));
                    break;
                }
                case 270:
                {
                    long tx = x0;
                    x0 = (ppdev->cxScreen - (y0 + 1));
                    y0 = tx;
                    tx = x1;
                    x1 = (ppdev->cxScreen - (y1 + 1));
                    y1 = tx;
                    break;
                }
            }
        }

        dx = x1 - x0;
        if (dx < 0) dx = -dx;
        dy = y1 - y0;
        if (dy < 0) dy = -dy;

        if (dx >= dy)
        {
            dmaj = dx;
            dmin = dy;
            if (x1 > x0) VecMode |= 2;
            if (y1 > y0) VecMode |= 4;
        }
        else
        {
            dmaj = dy;
            dmin = dx;
            VecMode |= 1;
            if (x1 > x0) VecMode |= 4;
            if (y1 > y0) VecMode |= 2;
        }
        axialerr = dmin << 1;
        diagerr  = (dmin - dmaj) << 1;
        initerr  = (dmin << 1) - dmaj;
	    if (!(VecMode & 4)) initerr--;

        gp_declare_vector(0);
        gp_set_raster_operation ((unsigned char)ROP);
        gp_set_strides (dstPitch, 0);
        gp_set_solid_pattern (pbo->iSolidColor);
        gp_bresenham_line (
            (y0 * dstPitch) + (x0 << ppdev->cPelSize) + dstOffset,
            (unsigned short)dmaj,
            (unsigned short)initerr,
            (unsigned short)axialerr,
            (unsigned short)diagerr,
            VecMode);

        RELEASE_PRIMITIVE_SEMAPHORE;

        DISPDBG ((ppdev, 1, "DrvLineTo Exit\n"));
        return TRUE;
    }

puntIt:

    // Pre-rotate the screen
    //
    dhsurfTrg = (USHORT)pso->dhsurf;
    if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 && !g_puntAllUntilModeSet)
    {
        // Clip to the screen if necessary
        //
        clippedPtr = prclBounds;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            clippedPtr = &destRegion;
            if (!FindIntersection (prclBounds, &ppdev->scrnBounds, clippedPtr))
                goto returnTrue;
        }

        gp_declare_blt (0);
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (clippedPtr->left << ppdev->cPelSize) + (clippedPtr->top * rot_pitch) + rot_offset,
            ROT_OFFSET (ppdev, clippedPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (clippedPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (clippedPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    MAKE_PUNTABLE(pso);

    gp_wait_until_idle();
    ret = EngLineTo (pso, pco, pbo, x0, y0, x1, y1, prclBounds, mix);

    if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 && !g_puntAllUntilModeSet)
    {
        // Clipped region was initialized during the pre-punt rotate
        //
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, clippedPtr->, ppdev->regSettings.RotationAngle),
            (clippedPtr->left << ppdev->cPelSize) + (clippedPtr->top * rot_pitch) + rot_offset, 
            clippedPtr->right  - clippedPtr->left,
            clippedPtr->bottom - clippedPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

    RELEASE_PRIMITIVE_SEMAPHORE;

    return ret;

returnTrue:

    RELEASE_PRIMITIVE_SEMAPHORE;

    return TRUE;
}

