//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Shared Driver definitions.
// </DOC_AMD_STD>
//

#ifndef __LXSHARED_H_
#define __LXSHARED_H_

#include "cim_parm.h"

#define DRIVER_FLAG_LINEAR_MODE 0x80000000l
#define DRIVER_FLAG_COMPRESSION 0x40000000l
#define DRIVER_FLAG_32BPP       0x20000000l

// SMBUS STRUCTURE
// The following structure holds all information for a single SMBus
// write, including device and register base and the data.
//
typedef struct tagSMBusWrite
{
    unsigned long base;
    unsigned long addr;
    unsigned long data;
} SMBUS_WRITE;

// CUSTOM MODE FILE SIGNATURE (LXXP_BIN)
//
#define LX_MODE_SIGNATURE1 0x5058584C
#define LX_MODE_SIGNATURE2 0x4E49425F

// CUSTOM MODE STRUCTURE
// The following structure includes all information needed to
// set a user-specified mode.
//
typedef struct tagGeodeLXCustomMode
{
    VG_DISPLAY_MODE manual_timings;
    VOPCONFIGURATIONBUFFER vop_config;
    unsigned long configure_vop;
    unsigned long bpp_mask;
    unsigned long output_path;
    unsigned long output_space;
    unsigned long flicker_strength;
    unsigned long line_double_video_on_interlaced;
    unsigned long border_color;
    unsigned long premode_smbus_count;
    unsigned long postmode_smbus_count;
    unsigned long *hcoeff;
    unsigned long *vcoeff;
    SMBUS_WRITE *premode_smbus;
    SMBUS_WRITE *postmode_smbus;

} GEODELX_CUSTOM_MODE;

// GLOBAL REGISTRY STRUCTURE
// This structure holds all user-modifiable registry
// information, such as panel enable, custom timings, etc.
//
typedef struct tagLXRegistrySettings
{
    unsigned long BIOSDetectOutput;     // Use the BIOS to detect the output device.
    unsigned long Compression;          // Compression Settings (enabled, auto, etc.)
    unsigned long FBLinear;             // Linear frame buffer setting
    unsigned long DDCEnabled;           // Enable DDC communication to a monitor
    unsigned long DDC_GPIOData;         // GPIO for DDC data line.
    unsigned long DDC_GPIOClock;        // GPIO for DDC clock line.
    unsigned long PanelEnable;          // Dynamic state of panel enable.
    unsigned long PanelWidth;           // Manually specified panel width for supported panel sizes.
    unsigned long PanelHeight;          // Manually specified panel height for supported panel sizes.
    unsigned long XVGAPanel;            // XVGA Panel for supported panel sizes.
    unsigned long SimultaneousCRT;      // Dynamic flag enabling simultaneous CRT display.    
    unsigned long ManualTimingEnable;   // Custom mode timings have been specified.    
    unsigned long EnablePanelScale;     // Allow modes to be stretched to fit the display.
    unsigned long EnableNon60HzPanel;   // Allow non-60Hz CRT modes when displaying on a TFT.
    unsigned long CustomPanelTimings;   // Custom panel timings have been specified.
    unsigned long RotationAngle;        // Angle of desktop rotation.
    unsigned long PanelRefresh;         // Detected refresh rate of the panel
    unsigned long PanelInSystem;        // Panel detected with BIOSDetectOutput or panel
                                        // specifically specified at install time.

} LX_REGISTRY_SETTINGS;

// BACKDOOR ALPHA BLENDING STRUCTURE
// This structure is used inside a driver escape to configure
// an alpha blending window for use with the video window.
//
typedef struct tagLXBlend
{
    int enable;
    int per_pixel;
    unsigned long fixed_alpha;

} LX_VIDEO_BLEND;

#define VIDEO_CUSTOMER_BASE 0x800

#define IOFN_SET_HEAP_INFORMATION   (VIDEO_CUSTOMER_BASE + 101)
#define IOFN_GET_LX_REGISTRY        (VIDEO_CUSTOMER_BASE + 103)
#define IOFN_SET_LX_REGISTRY        (VIDEO_CUSTOMER_BASE + 104)
#define IOFN_GET_MINIPORT_REGISTRY  (VIDEO_CUSTOMER_BASE + 105)
#define IOFN_SET_GP_RESTORE_INFO    (VIDEO_CUSTOMER_BASE + 106)
#define IOFN_SET_AUTOFLIP_SURFACES  (VIDEO_CUSTOMER_BASE + 107)
#define IOFN_SAVE_VIDEO_Y           (VIDEO_CUSTOMER_BASE + 108)
#define IOFN_GET_CUSTOM_MODE_SIZE   (VIDEO_CUSTOMER_BASE + 109)
#define IOFN_GET_CUSTOM_MODES       (VIDEO_CUSTOMER_BASE + 110)
#define IOFN_GET_PANEL_TIMINGS      (VIDEO_CUSTOMER_BASE + 111)
#define IOFN_GET_LAST_MODE          (VIDEO_CUSTOMER_BASE + 112)
#define IOFN_SET_DISPLAY_GAMMA      (VIDEO_CUSTOMER_BASE + 113)
#define IOFN_GET_DISPLAY_GAMMA      (VIDEO_CUSTOMER_BASE + 114)
#define IOFN_SET_OVERLAY_GAMMA      (VIDEO_CUSTOMER_BASE + 115)
#define IOFN_GET_OVERLAY_GAMMA      (VIDEO_CUSTOMER_BASE + 116)

#define IOCTL_SET_HEAP_INFORMATION\
    CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_HEAP_INFORMATION, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_LX_REGISTRY\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_LX_REGISTRY, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SET_LX_REGISTRY\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_LX_REGISTRY, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_MINIPORT_REGISTRY\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_MINIPORT_REGISTRY, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SET_GP_RESTORE_INFO\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_GP_RESTORE_INFO, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SET_AUTOFLIP_SURFACES\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_AUTOFLIP_SURFACES, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SAVE_VIDEO_Y\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SAVE_VIDEO_Y, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_CUSTOM_MODE_SIZE\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_CUSTOM_MODE_SIZE, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_CUSTOM_MODES\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_CUSTOM_MODES, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SAVE_VIDEO_Y\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SAVE_VIDEO_Y, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_PANEL_TIMINGS\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_PANEL_TIMINGS, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_LAST_MODE\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_LAST_MODE, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SET_DISPLAY_GAMMA\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_DISPLAY_GAMMA, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_DISPLAY_GAMMA\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_DISPLAY_GAMMA, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_SET_OVERLAY_GAMMA\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_SET_OVERLAY_GAMMA, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_OVERLAY_GAMMA\
      CTL_CODE(FILE_DEVICE_VIDEO, IOFN_GET_OVERLAY_GAMMA, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define LX_CMD_BUFFER_SIZE       0x200000
#define LX_CIM_SCRATCH_SIZE      0x100000
#define LX_RESERVED_SPACE        (LX_CMD_BUFFER_SIZE + LX_CIM_SCRATCH_SIZE)

#define LX_MODE_GAMMA       0
#define LX_MODE_BRIGHT      1
#define LX_MODE_CONTRAST    2
#define LX_GAMMA_MAX        60
#define LX_GAMMA_DEFAULT    10
#define LX_BRIGHT_MAX       128
#define LX_BRIGHT_DEFAULT   64
#define LX_CONTRAST_MAX     200
#define LX_CONTRAST_DEFAULT 100

#endif

