//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Driver Palette support.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"

//////////////////////////////////////////////////////////////////////
// GLOBAL WINDOWS PALETTE
// Windows reserves the first and last 10 palette entries for system
// colors when the display mode is set to 256 colors.
//
PALETTEENTRY g_baseColors[20] =
{
	{ 0x00, 0x00, 0x00, 0 },		// 0
	{ 0x80, 0x00, 0x00, 0 },		// 1
	{ 0x00, 0x80, 0x00, 0 },		// 2
	{ 0x80, 0x80, 0x00, 0 },		// 3
	{ 0x00, 0x00, 0x80, 0 },		// 4
	{ 0x80, 0x00, 0x80, 0 },		// 5
	{ 0x00, 0x80, 0x80, 0 },		// 6
	{ 0xC0, 0xC0, 0xC0, 0 },		// 7
	{ 0xC0, 0xDC, 0xC0, 0 },		// 8
	{ 0xA6, 0xCA, 0xF0, 0 },		// 9
	{ 0xFF, 0xFB, 0xF0, 0 },		// 10
	{ 0xA0, 0xA0, 0xA4, 0 },		// 11
	{ 0x80, 0x80, 0x80, 0 },		// 12
	{ 0xFF, 0x00, 0x00, 0 },		// 13
	{ 0x00, 0xFF, 0x00, 0 },		// 14
	{ 0xFF, 0xFF, 0x00, 0 },		// 15
	{ 0x00, 0x00, 0xFF, 0 },		// 16
	{ 0xFF, 0x00, 0xFF, 0 },		// 17
	{ 0x00, 0xFF, 0xFF, 0 },		// 18
	{ 0xFF, 0xFF, 0xFF, 0 },		// 19
};

/////////////////////////////////////////////////////////////////////////
// BOOL bInitializePalette
//
// Initializes default palette for the PDEV.
//
BOOL bInitializePalette(
	PDEV*	 ppdev,
	DEVINFO* pdi)
{
	PALETTEENTRY*	ppal;
	PALETTEENTRY*	ppalTmp;
	ULONG			ulLoop;
	BYTE			jRed;
	BYTE			jGre;
	BYTE			jBlu;
	HPALETTE		hpal;

	DISPDBG((ppdev, 1, "bInitializePalette Entry\n"));

	if (ppdev->iBitmapFormat == BMF_8BPP)
	{
		// Allocate our palette:

		ppal = (PALETTEENTRY*)LX_ALLOC_MEM(FL_ZERO_MEMORY, (sizeof(PALETTEENTRY) * 256));
		if (ppal == NULL)
			goto ReturnFalse;

		ppdev->pPal = ppal;

		// Generate 256 (8*4*4) RGB combinations to fill the palette

		jRed = 0;
		jGre = 0;
		jBlu = 0;

		ppalTmp = ppal;

		for (ulLoop = 256; ulLoop != 0; ulLoop--)
		{
			ppalTmp->peRed	 = jRed;
			ppalTmp->peGreen = jGre;
			ppalTmp->peBlue  = jBlu;
			ppalTmp->peFlags = 0;

			ppalTmp++;

			if (!(jRed += 32))
				if (!(jGre += 32))
					jBlu += 64;
		}

		// Fill in Windows reserved colours.  The Window Manager reserved the first
        // and last 10 colours for painting windows borders and for non-palette
        // managed applications.
        //
		for (ulLoop = 0; ulLoop < 10; ulLoop++)
		{
			// First 10

			ppal[ulLoop]	   = g_baseColors[ulLoop];

			// Last 10

			ppal[246 + ulLoop] = g_baseColors[ulLoop + 10];
		}

		// Create handle for palette.

		hpal = EngCreatePalette(PAL_INDEXED, 256, (ULONG*) ppal, 0, 0, 0);
	}
	else
	{
		ASSERTVDD((ppdev->iBitmapFormat == BMF_16BPP) ||
				  (ppdev->iBitmapFormat == BMF_32BPP),
				 "This case handles only 16 or 32bpp");

		hpal = EngCreatePalette(PAL_BITFIELDS, 0, NULL,
            ppdev->flRed, ppdev->flGreen, ppdev->flBlue);
	}

	ppdev->hpalDefault = hpal;
	pdi->hpalDefault   = hpal;

	if (hpal == 0)
		goto ReturnFalse;

    DISPDBG ((ppdev, 1, "bInitializePalette Exit\n"));
	return TRUE;

ReturnFalse:

	DISPDBG((ppdev, 0, "Failed bInitializePalette"));
	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// VOID vUninitializePalette
//
// Frees resources allocated by bInitializePalette.
// Note: In an error case, this may be called before bInitializePalette.
//
VOID vUninitializePalette(PDEV* ppdev)
{
	DISPDBG((ppdev, 1, "vUnititializePalette Entry\n"));

	// Delete the default palette if we created one:

	if (ppdev->hpalDefault)
		EngDeletePalette(ppdev->hpalDefault);

	if (ppdev->pPal)
		EngFreeMem (ppdev->pPal);

    DISPDBG((ppdev, 1, "vUnititializePalette Exit\n"));
}

//////////////////////////////////////////////////////////////////////
// BOOL bEnablePalette
//
// Initialize the hardware's 8bpp palette registers.
//
BOOL bEnablePalette(PDEV* ppdev)
{
	PVIDEO_CLUT pScreenClut;
	ULONG		ulReturnedDataLength;
	ULONG		cColors;
	PVIDEO_CLUTDATA pScreenClutData;

	DISPDBG((ppdev, 1, "bEnablePalette Entry\n"));

	if (ppdev->iBitmapFormat == BMF_8BPP)
	{
		// Allocate temporary space for the LUT data.  This is a dynamic
        // allocation instead of a stack allocation because large objects
        // on the stack can in rare cases lead to stack overflow
        //
		pScreenClut = (PVIDEO_CLUT) LX_ALLOC_MEM(0, MAX_CLUT_SIZE);
        if (!pScreenClut)
            goto ReturnFalse;

		pScreenClut->NumEntries = 256;
		pScreenClut->FirstEntry = 0;

		// Copy colours in:

		cColors = 256;
		pScreenClutData = (PVIDEO_CLUTDATA) (&(pScreenClut->LookupTable[0]));

		while(cColors--)
		{
			pScreenClutData[cColors].Red =	  ppdev->pPal[cColors].peRed >>
											  ppdev->cPaletteShift;
			pScreenClutData[cColors].Green =  ppdev->pPal[cColors].peGreen >>
											  ppdev->cPaletteShift;
			pScreenClutData[cColors].Blue =   ppdev->pPal[cColors].peBlue >>
											  ppdev->cPaletteShift;
			pScreenClutData[cColors].Unused = 0;
		}

		// SET PALETTE REGISTERS ON THE HARDWARE
        // As the miniport has to implement support for this IOCTL, we might
        // as well reuse that code.
        //
		if (EngDeviceIoControl(ppdev->hDriver,
							 IOCTL_VIDEO_SET_COLOR_REGISTERS,
							 pScreenClut,
							 MAX_CLUT_SIZE,
							 NULL,
							 0,
							 &ulReturnedDataLength))
		{
			DISPDBG((ppdev, 300, "Failed bEnablePalette"));
			goto ReturnFalse;
		}

         EngFreeMem (pScreenClut);
	}

	DISPDBG((ppdev, 1, "bEnablePalette Exit\n"));

	return TRUE;

ReturnFalse:

    if (pScreenClut)
        EngFreeMem (pScreenClut);
    return FALSE;
}

/////////////////////////////////////////////////////////////////////////
// BOOL DrvSetPalette
//
// DDI entry point for manipulating the palette.
//
BOOL DrvSetPalette(
	DHPDEV	dhpdev,
	PALOBJ* ppalo,
	FLONG	fl,
	ULONG	iStart,
	ULONG	cColors)
{
	PVIDEO_CLUT 	pScreenClut;
	PVIDEO_CLUTDATA pScreenClutData;
	PDEV*			ppdev;

	ppdev = (PDEV*) dhpdev;

    // Allocate temporary space for the LUT data.  This is a dynamic
    // allocation instead of a stack allocation because large objects
    // on the stack can in rare cases lead to stack overflow
    //
	pScreenClut = (PVIDEO_CLUT) LX_ALLOC_MEM(0, MAX_CLUT_SIZE);
    if (!pScreenClut)
        goto ReturnFalse;

	// Fill in pScreenClut header info:
    //
	pScreenClut->NumEntries = (USHORT) cColors;
	pScreenClut->FirstEntry = (USHORT) iStart;

	pScreenClutData = (PVIDEO_CLUTDATA) (&(pScreenClut->LookupTable[0]));

	if (cColors != PALOBJ_cGetColors(ppalo, iStart, cColors, (ULONG*) pScreenClutData))
	{
		DISPDBG((ppdev, 0, "DrvSetPalette failed PALOBJ_cGetColors\n"));
		goto ReturnFalse;
	}

	// Set the high reserved byte in each palette entry to 0.
	// Do the appropriate palette shifting to fit in the DAC.
    //
	if (ppdev->cPaletteShift)
	{
		while(cColors--)
		{
			pScreenClutData[cColors].Red   >>= ppdev->cPaletteShift;
			pScreenClutData[cColors].Green >>= ppdev->cPaletteShift;
			pScreenClutData[cColors].Blue  >>= ppdev->cPaletteShift;
			pScreenClutData[cColors].Unused = 0;
		}
	}
	else
	{
		while(cColors--)
		{
			pScreenClutData[cColors].Unused = 0;
		}
	}

	// Set palette registers
    //
	if (EngDeviceIoControl(ppdev->hDriver,
						 IOCTL_VIDEO_SET_COLOR_REGISTERS,
						 pScreenClut,
						 MAX_CLUT_SIZE,
						 NULL,
						 0,
						 &cColors))
	{
		DISPDBG((ppdev, 0, "DrvSetPalette failed EngDeviceIoControl\n"));
		goto ReturnFalse;
	}

    EngFreeMem (pScreenClut);

    DISPDBG ((ppdev, 1, "DrvSetPalette Exit\n"));

	return TRUE;

ReturnFalse:

    if (pScreenClut)
        EngFreeMem (pScreenClut);

    return FALSE;
}

/////////////////////////////////////////////////////////////////////////
// BOOL DrvIcmSetDeviceGammaRamp
//
// This routine programs the gamma RAM for the primary surface.
//
BOOL DrvIcmSetDeviceGammaRamp (
    DHPDEV  dhpdev,
    ULONG   iFormat,
    LPVOID  lpRamp)
{
    GAMMARAMP *pGamma = (GAMMARAMP *)lpRamp;
    PDEV *ppdev = (PDEV *)dhpdev;
    unsigned long red, green, blue;
    int i;

    DISPDBG ((ppdev, 3000, "DrvIcmSetDeviceGammaRamp\n"));

    if (iFormat != IGRF_RGB_256WORDS)
    {
        DISPDBG ((ppdev, 3000, "Invalid Format for gamma ramp %i\n", iFormat));
        return FALSE;
    }

    if (ppdev->cPelSize == 0)
    {
        DISPDBG ((ppdev, 3000, "Unable to set gamma ramp in 8-bit modes.\n"));
        return FALSE;
    }

    for (i = 0; i < 256; i++)
    {
        red   = (ULONG)pGamma->Red[i]   >> 8;
        green = (ULONG)pGamma->Green[i] >> 8;
        blue  = (ULONG)pGamma->Blue[i]  >> 8;

        vg_set_display_palette_entry (i,
                (red << 16) | (green << 8) | blue);
    }

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// SetPalette
//
// Routine to calculate and program gamma settings for the display or
// overlay.
//
void SetPalette (PDEV *ppdev, BOOL video)
{
    ULONG fpSize;
    long *redGamma, *blueGamma, *greenGamma;
    double ImageRed, ImageGreen, ImageBlue, ImageIn;
    double InvRG, InvGG, InvBG;   
    double RedSlope, GreenSlope, BlueSlope;
    double RedG, GreenG, BlueG;
    double Red, Green, Blue;
    double RedShift, GreenShift, BlueShift;
    int i;

    if (video)
    {
        redGamma   = &ppdev->redOverlayGamma[0];
        blueGamma  = &ppdev->blueOverlayGamma[0];
        greenGamma = &ppdev->greenOverlayGamma[0];
    }
    else
    {
        redGamma   = &ppdev->redDisplayGamma[0];
        blueGamma  = &ppdev->blueDisplayGamma[0];
        greenGamma = &ppdev->greenDisplayGamma[0];
    }

    // Save and restore the floating point state
    //
    fpSize = EngSaveFloatingPointState (NULL, 0);
    ppdev->fpState = LX_ALLOC_MEM (0, fpSize);
    EngSaveFloatingPointState (ppdev->fpState, fpSize);
    if (!ppdev->fpState)
    {
        DISPDBG ((ppdev, 3000, "Unable to allocate floating point state.\n"));
        return;
    }

    // GAMMA CORRECTION
    // Range is 0.5 to 3.5 in 0.05 increments.
    //
    if (redGamma[LX_MODE_GAMMA] > LX_GAMMA_MAX)
        redGamma[LX_MODE_GAMMA] = LX_GAMMA_MAX;
    if (greenGamma[LX_MODE_GAMMA] > LX_GAMMA_MAX)
        greenGamma[LX_MODE_GAMMA] = LX_GAMMA_MAX;
    if (blueGamma[LX_MODE_GAMMA] > LX_GAMMA_MAX)
        blueGamma[LX_MODE_GAMMA] = LX_GAMMA_MAX;

    RedG   = ((double)redGamma[LX_MODE_GAMMA] * .05) + 0.5;
    GreenG = ((double)greenGamma[LX_MODE_GAMMA] * .05) + 0.5;
    BlueG  = ((double)blueGamma[LX_MODE_GAMMA] * .05) + 0.5;

    InvRG = 1.0 / RedG;
    InvGG = 1.0 / GreenG;
    InvBG = 1.0 / BlueG;

    // BRIGHTNESS CORRECTION
    // Range is plus or minus 1/4 of the color range.
    //
    if (redGamma[LX_MODE_BRIGHT] > LX_BRIGHT_MAX)
        redGamma[LX_MODE_BRIGHT] = LX_BRIGHT_MAX;
    if (greenGamma[LX_MODE_BRIGHT] > LX_BRIGHT_MAX)
        greenGamma[LX_MODE_BRIGHT] = LX_BRIGHT_MAX;
    if (blueGamma[LX_MODE_BRIGHT] > LX_BRIGHT_MAX)
        blueGamma[LX_MODE_BRIGHT] = LX_BRIGHT_MAX;

    RedShift   = (double)(redGamma[LX_MODE_BRIGHT] - (LX_BRIGHT_MAX >> 1)) / 256.0;
    GreenShift = (double)(greenGamma[LX_MODE_BRIGHT] - (LX_BRIGHT_MAX >> 1)) / 256.0;
    BlueShift  = (double)(blueGamma[LX_MODE_BRIGHT] - (LX_BRIGHT_MAX >> 1)) / 256.0;
    
    // CONTRAST CORRECTION
    // Range is 0.5 to 1.5 in slope.
    //
    if (redGamma[LX_MODE_CONTRAST] > LX_CONTRAST_MAX)
        redGamma[LX_MODE_CONTRAST] = LX_CONTRAST_MAX;
    if (greenGamma[LX_MODE_CONTRAST] > LX_CONTRAST_MAX)
        greenGamma[LX_MODE_CONTRAST] = LX_CONTRAST_MAX;
    if (blueGamma[LX_MODE_CONTRAST] > LX_CONTRAST_MAX)
        blueGamma[LX_MODE_CONTRAST] = LX_CONTRAST_MAX;

    RedSlope   = 0.5 + (.005 * (double)redGamma[LX_MODE_CONTRAST]);
    GreenSlope = 0.5 + (.005 * (double)greenGamma[LX_MODE_CONTRAST]);
    BlueSlope  = 0.5 + (.005 * (double)blueGamma[LX_MODE_CONTRAST]);
        
	// PROGRAM THE VALUES IN THE HARDWARE
	//
    for(i = 0; i < 256; i++)
    {
        ImageIn = (double)i / 255.0;
        Red   = pow(ImageIn,InvRG);
        Green = pow(ImageIn,InvGG);
        Blue  = pow(ImageIn,InvBG);

        Red   *= RedSlope;
        Green *= GreenSlope;
        Blue  *= BlueSlope;
          
        Red   += RedShift;
        Green += GreenShift;
        Blue  += BlueShift;

        // Add 0.5 for rounding
        //
        Red   = (Red * 255.0) + 0.5;
        Green = (Green * 255.0) + 0.5;
        Blue  = (Blue * 255.0) + 0.5;

        if(Red < 0)   Red   = 0; if(Red > 255)   Red   = 255;
        if(Green < 0) Green = 0; if(Green > 255) Green = 255;    
        if(Blue < 0)  Blue  = 0; if(Blue > 255)  Blue  = 255;

        if (video)
        {
            df_set_video_palette_entry (i,
                ((DWORD)Blue) | ((DWORD)Green << 8) | ((DWORD)Red << 16));
        }
        else
        {
		    vg_set_display_palette_entry (i,
                ((DWORD)Blue) | ((DWORD)Green << 8) | ((DWORD)Red << 16));
        }
    }

    // Program an updated color key in case of video playback.
    //
    if (ppdev->graphics)
        SetChromaKey (ppdev, ppdev->colorKey, ppdev->colorBPP, ppdev->graphics);

    EngRestoreFloatingPointState (ppdev->fpState);
    EngFreeMem (ppdev->fpState);
}

