//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Hardware cursor support.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "pointer.h"

/////////////////////////////////////////////////////////////////////////
// LXSetCursorPos
//
// Private routine to adjust the current pointer position.  If panning
// occurs and a video window is active, the current video position is also
// updated.
//
void LXSetCursorPos(PDEV *ppdev, long x, long y)
{
    if (ppdev->regSettings.RotationAngle != 0)
    {
        switch (ppdev->regSettings.RotationAngle)
        {
            case 90:
            {
                long tx = x;
                x = y;
                y = (ppdev->cyScreen - (tx + 1));                
                break;
            }
            case 180:
            {
                x = (ppdev->cxScreen - (x + 1));
                y = (ppdev->cyScreen - (y + 1));
                break;
            }
            case 270:
            {
                long tx = x;
                x = (ppdev->cxScreen - (y + 1));
                y = tx;
                break;
            }
        }
    }
    vg_set_cursor_position (x, y, &ppdev->panning);
    
    if (ppdev->panning.start_updated)
    {
        // Re-enable compression if the offset is back to 0.
        //
        if (x == 0 && y == 0 && compression_enabled)
            vg_set_compression_enable (1);
    }

    if (ppdev->OverlayActive &&
         (ppdev->regSettings.CustomPanelTimings ||
         (ppdev->regSettings.ManualTimingEnable && (ppdev->customMode->manual_timings.flags && VG_MODEFLAG_PANELOUT)) ||
        (!ppdev->regSettings.ManualTimingEnable && ppdev->regSettings.PanelEnable)))
    {
        if (ppdev->panning.start_updated)
        {
            SetVideoPosition (ppdev, 0);
        }
    }
}

/////////////////////////////////////////////////////////////////////////
// DrvMovePointer
//
// Moves the hardware pointer to a new position.
//
VOID DrvMovePointer(
    SURFOBJ * pso,
    LONG x,
    LONG y,
    RECTL * prcl)
{
    PDEV *ppdev = (PDEV *) pso->dhpdev;
    
    // CHECK IF CURSOR SHOULD BE DISABLED
    // A new x position of -1 means hide the pointer.
    //
    if (x == -1)
    {
        vg_set_cursor_enable (0);
        ppdev->cursorEnabled = FALSE;
    }
    else
    {   
        // CHANGE POINTER POSITION
        //
        if (y < 0)
        {
            // When y is less than 0, this is a callback enabling us to pan
            // a display with a simulated cursor.  If the cursor is not
            // simulated, we will have already received a normal call to
            // DrvMovePointer by this point.
            //
            if (ppdev->bSimulatedCursor)
            {
                if (ppdev->regSettings.RotationAngle == 90 || 
                    ppdev->regSettings.RotationAngle == 270)
                {
                    LXSetCursorPos (ppdev, x, y + ppdev->cxScreen);
                }
                else
                    LXSetCursorPos (ppdev, x, y + ppdev->cyScreen);
            }
        }
        else
        {
            if (!ppdev->cursorEnabled)
            {
                ppdev->cursorEnabled = TRUE;
                vg_set_cursor_enable (1);
            }
            LXSetCursorPos (ppdev, x, y);
        }
    }
}

/////////////////////////////////////////////////////////////
// RotateMonoPointer
// Code to rotate cursor data.  The buffers are expected to contain
// data for a 64x64 mono cursor.
//
void RotateMonoPointer (PDEV *ppdev, ULONG *pAnd_In, ULONG *pXor_In, ULONG *pAnd_Out, ULONG *pXor_Out)
{
    int x, y;
    switch (ppdev->regSettings.RotationAngle)
    {
        case 90:
            for (x = 0; x < 32; x++)
            {
                ULONG andValue = *pAnd_In++;
                ULONG xorValue = *pXor_In++;

                for (y = 0; y < 64; y++)
                {
                    if (y == 32)
                    {
                        andValue = *pAnd_In++;
                        xorValue = *pXor_In++;
                    }

                    pAnd_Out[(63 - y) << 1] = (pAnd_Out[(63 - y) << 1] << 1) | (andValue >> 31);
                    pXor_Out[(63 - y) << 1] = (pXor_Out[(63 - y) << 1] << 1) | (xorValue >> 31);
                    andValue <<= 1;
                    xorValue <<= 1;
                }
            }
            for (x = 0; x < 32; x++)
            {
                ULONG andValue = *pAnd_In++;
                ULONG xorValue = *pXor_In++;

                for (y = 0; y < 64; y++)
                {
                    if (y == 32)
                    {
                        andValue = *pAnd_In++;
                        xorValue = *pXor_In++;
                    }

                    pAnd_Out[((63 - y) << 1) + 1] = (pAnd_Out[((63 - y) << 1) + 1] << 1) | (andValue >> 31);
                    pXor_Out[((63 - y) << 1) + 1] = (pXor_Out[((63 - y) << 1) + 1] << 1) | (xorValue >> 31);
                    andValue <<= 1;
                    xorValue <<= 1;
                }
            }
            break;
        
        case 270:
            for (x = 0; x < 32; x++)
            {
                ULONG andValue = *pAnd_In++;
                ULONG xorValue = *pXor_In++;

                for (y = 0; y < 64; y++)
                {
                    if (y == 32)
                    {
                        andValue = *pAnd_In++;
                        xorValue = *pXor_In++;
                    }

                    pAnd_Out[(y << 1) + 1] = (pAnd_Out[(y << 1) + 1] >> 1) | (andValue & 0x80000000);
                    pXor_Out[(y << 1) + 1] = (pXor_Out[(y << 1) + 1] >> 1) | (xorValue & 0x80000000);
                    andValue <<= 1;
                    xorValue <<= 1;
                }
            }
            for (x = 0; x < 32; x++)
            {
                ULONG andValue = *pAnd_In++;
                ULONG xorValue = *pXor_In++;

                for (y = 0; y < 64; y++)
                {
                    if (y == 32)
                    {
                        andValue = *pAnd_In++;
                        xorValue = *pXor_In++;
                    }

                    pAnd_Out[y << 1] = (pAnd_Out[y << 1] >> 1) | (andValue & 0x80000000);
                    pXor_Out[y << 1] = (pXor_Out[y << 1] >> 1) | (xorValue & 0x80000000);
                    andValue <<= 1;
                    xorValue <<= 1;
                }
            }
            break;
        
        case 180:
            for (y = 0; y < 64; y++)
            {
                ULONG andValue = *pAnd_In++;
                ULONG xorValue = *pXor_In++;
                int index = ((63 - y) << 1) + 1;

                pAnd_Out[((63 - y) << 1) + 1] = pXor_Out[((63 - y) << 1) + 1] =
                    pAnd_Out[(63 - y) << 1] = pXor_Out[(63 - y) << 1] = 0;

                for (x = 0; x < 32; x++)
                {
                    pAnd_Out[index] |= ((andValue >> x) & 1) << (31 - x);
                    pXor_Out[index] |= ((xorValue >> x) & 1) << (31 - x);
                }
                andValue = *pAnd_In++;
                xorValue = *pXor_In++;
                index = (63 - y) << 1;
                for (x = 0; x < 32; x++)
                {
                    pAnd_Out[index] |= ((andValue >> x) & 1) << (31 - x);
                    pXor_Out[index] |= ((xorValue >> x) & 1) << (31 - x);
                }
            }
            break;
    }
}

/////////////////////////////////////////////////////////////////////////
// bCopyMonoPointer
//
// Copies two monochrome masks into a buffer of the maximum size handled by the
// hardware, with any extra bits set to 0.  The masks are converted to topdown
// form if they aren't already.  Returns TRUE if we can handle this pointer in
// hardware, FALSE if not.
//
BOOL bCopyMonoPointer(
    PDEV *ppdev,
    SURFOBJ * pso)
{
    ULONG cxSrc, cySrc;
    ULONG x, y, index;
    ULONG srcBytes, maxBytes;
    BYTE andArray[512], xorArray[512];
    PBYTE pSrcAnd, pSrcXor;
    PBYTE pDstAnd, pDstXor;

    // GRAB CURSOR SIZES
    // Cursor height includes both masks
    //
    cxSrc = pso->sizlBitmap.cx;
    cySrc = pso->sizlBitmap.cy >> 1;

    // CHECK SIZE
    // Make sure the new pointer isn't too big to handle
    //
    if ((cxSrc > MAX_GEODELX_CURSOR_WIDTH) ||
        (cySrc > MAX_GEODELX_CURSOR_HEIGHT))
    {
    	return FALSE;
	}

	// FILL IN CUSTOM CURSOR ARRAY
	// We are going to call Cimarron directly from here.  This simplifies
	// the interface, as Cimarron expects an array of XOR and AND data.
    // Unfortunately, however, the arrays are expected to be constant
    // size and the expected byte ordering is reversed.
    //
    srcBytes   = (cxSrc + 7) >> 3;
    maxBytes   = (MAX_GEODELX_CURSOR_WIDTH + 7) >> 3;
    pSrcAnd    = (PBYTE) pso->pvBits;
    pSrcXor    = pSrcAnd + (cySrc * pso->lDelta);
    pDstAnd    = &andArray[0];
    pDstXor    = &xorArray[0];

    for (y = 0; y < cySrc; y++)
    {
        // FILL IN VALID BYTES
        //
        for (x = 0; x < srcBytes; x++)
        {
            index = (x & 0xFFC) + 3 - (x & 3);
            pDstAnd[index] = pSrcAnd[x];
            pDstXor[index] = pSrcXor[x];
        }

        // FILL IN REMAINING BYTES WITH TRANSPARENCY
        //
        for (x = srcBytes; x < maxBytes; x++)
        {
            index = (x & 0xFFC) + 3 - (x & 3);
            pDstAnd[index] = 0xFF;
            pDstXor[index] = 0x00;
        }
        
        // INCREMENT TO NEXT LINE OF DATA
        //
        pSrcAnd += pso->lDelta;
        pSrcXor += pso->lDelta;
        pDstAnd += 8;
        pDstXor += 8;
    }

    // PAD REMAINING LINES WITH TRANSPARENCY
    //
    for (y = cySrc; y < MAX_GEODELX_CURSOR_HEIGHT; y++)
    {
        *((ULONG *)(pDstAnd))     = 0xFFFFFFFF;
        *((ULONG *)(pDstAnd + 4)) = 0xFFFFFFFF;
        *((ULONG *)(pDstXor))     = 0x00000000;
        *((ULONG *)(pDstXor + 4)) = 0x00000000;
        
        pDstAnd += 8;
        pDstXor += 8;
    }

    // CALL CIMARRON TO INITIALIZE CURSOR
    //
    if (ppdev->regSettings.RotationAngle != 0)
    {
        RotateMonoPointer (ppdev, (ULONG *)(&andArray[0]), (ULONG *)(&xorArray[0]),
            (ULONG *)ppdev->pCursorBuffer, (ULONG *)ppdev->pCursorBuffer + 128);
        vg_set_mono_cursor_shape64 (ppdev->cursorOffset + (ppdev->activeBuffer << 14),
            (ULONG *)ppdev->pCursorBuffer, (ULONG *)ppdev->pCursorBuffer + 128,
            ppdev->xHotSpot, ppdev->yHotSpot);
    }
    else
    {
        vg_set_mono_cursor_shape64 (ppdev->cursorOffset + (ppdev->activeBuffer << 14),
            (ULONG *)(&andArray[0]), (ULONG *)(&xorArray[0]),
            ppdev->xHotSpot, ppdev->yHotSpot);
    }

    ppdev->activeBuffer = (ppdev->activeBuffer + 1) & 7;

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// RotateColorPointer
//
// Rotates a color pointer in memory.  This routine assumes that a 48x64
// buffer is being passed in.  It rotates the first 48x48 portion.
//
void RotateColorPointer (PDEV *ppdev, unsigned long *buf_in, unsigned long *buf_out)
{
    int x, y;
    if (ppdev->regSettings.RotationAngle == 90)
    {
        for (y = 0; y < 48; y++)
        {
            for (x = 0; x < 48; x++)
            {
                buf_out[y + ((48 - x - 1) * 48)] = buf_in[x];
            }
            buf_in += 48;
        }
    }
    else if (ppdev->regSettings.RotationAngle == 270)
    {
        for (y = 0; y < 48; y++)
        {
            for (x = 0; x < 48; x++)
            {
                buf_out[48 - y  - 1 + (x * 48)] = buf_in[x];
            }
            buf_in += 48;
        }
    }
    else
    {
        for (y = 0; y < 48; y++)
        {
            for (x = 0; x < 48; x++)
            {
                buf_out[48 - x - 1 + ((48 - y - 1) * 48)] = buf_in[x];
            }
            buf_in += 48;
        }
    }
}

/////////////////////////////////////////////////////////////////////////
// bCopyColorPointer
//
// Copies the mono and color masks into the buffer of maximum size
// handled by the miniport with any extra bits set to 0. Color translation
// is handled at this time. The masks are converted to topdown form if they
// aren't already.  Returns TRUE if we can handle this pointer in  hardware,
// FALSE if not.  Current LX hardware does not support a color cursor.
//
BOOL bCopyColorPointer(
    PDEV *ppdev,
    SURFOBJ *psoMask,
    SURFOBJ *psoColor,
    XLATEOBJ *pxlo)
{
    unsigned long bits;
    unsigned long width, height, pitch, tempwidth, tempheight;
    unsigned long bufptr = (ULONG)ppdev->pCursorBuffer;

#if DBG
    if (g_puntMask & LX_PUNT_ALPHACURSOR)
        return FALSE;
#endif

    // WE DON'T SUPPORT MASKS FOR NOW
    //
    if (psoMask)
    {
        DISPDBG ((ppdev, 5000, "Unable to support masked color cursor.\n"));
        return FALSE;
    }

    width  = psoColor->sizlBitmap.cx;
    height = psoColor->sizlBitmap.cy;
    pitch  = psoColor->lDelta;

    if (ppdev->regSettings.RotationAngle != 0)
    {
        if (width  > MAX_GEODELX_COLOR_CURSOR_WIDTH  ||
            height > MAX_GEODELX_COLOR_CURSOR_WIDTH ||
            psoColor->iBitmapFormat != BMF_32BPP)
        {
            DISPDBG ((ppdev, 5000, "Color cursor could not be enabled in rotation.\n"));
            return FALSE;
        }
    }
    else
    {
        if (width  > MAX_GEODELX_COLOR_CURSOR_WIDTH ||
            height > MAX_GEODELX_COLOR_CURSOR_HEIGHT ||
            psoColor->iBitmapFormat != BMF_32BPP)
        {
            DISPDBG ((ppdev, 5000, "Color cursor not supported by hardware.\n"));
            return FALSE;
        }
    }

    // REVERSE PRE-MULTIPLY
    // Windows pre-multiplies alpha blended cursors, leaving us A + (1 - alpha)B, but the
    // hardware expects alpha*A + (1 - alpha)B.
    //
    bits = (unsigned long )psoColor->pvScan0;

    _asm {

        cld
        mov     edi, bufptr
        mov     eax, height
        mov     tempheight, eax

line_loop:
        mov     esi, bits
        mov     eax, width
        mov     tempwidth, eax

pixel_loop:
        lodsd

        // NO WORK IF ALPHA IS 0xFF OR 0x00
        // 0xFF pre-multiplied by a value is the value.  Likewise, if
        // alpha is 0, the result will always be 0 after the hardware.

        xor     ecx, ecx
        shld    ecx, eax, 8
        setz    bl
        cmp     cl, 0xFF
        setz    bh
        or      bl, bh
        jnz     nextPixel

        mov     ebx, eax

        // CALCULATE BLUE

        and     eax, 0x000000FF
        imul    eax, alphaTable[ecx*4]
        shr     eax, 16
        inc     eax
        shl     eax, 24

        // CALCULATE GREEN

        shld    edx, ebx, 24
        and     edx, 0x000000FF
        imul    edx, alphaTable[ecx*4]
        shr     edx, 16
        inc     edx
        shrd    eax, edx, 8

        // CALCULATE RED

        shr     ebx, 16
        and     ebx, 0x000000FF
        imul    ebx, alphaTable[ecx*4]
        shr     ebx, 16
        inc     ebx
        shrd    eax, ebx, 8

        // RE-ADD ALPHA

        shrd    eax, ecx, 8

nextPixel:

        stosd
        dec     tempwidth
        jnz     pixel_loop

        mov     ecx, MAX_GEODELX_COLOR_CURSOR_WIDTH
        sub     ecx, width
        jz      line_loop

        // Pad out buffer with 0's

        xor     eax, eax
        rep     stosd

        mov     eax, pitch
        add     bits, eax
        dec     tempheight
        jnz     line_loop

        // Pad out remaining lines

        mov     ebx, MAX_GEODELX_COLOR_CURSOR_HEIGHT
        sub     ebx, height
        jz      loop_done

zero_line_loop:

        mov     ecx, MAX_GEODELX_COLOR_CURSOR_WIDTH
        xor     eax, eax
        rep     stosd
        dec     ebx
        jnz     zero_line_loop

loop_done:
    }

    if (ppdev->regSettings.RotationAngle != 0)
    {
        RotateColorPointer (ppdev, (ULONG *)ppdev->pCursorBuffer, (ULONG *)ppdev->pCursorBuffer + (48*64));
        vg_set_color_cursor_shape (ppdev->cursorOffset + (ppdev->activeBuffer << 14),
            (UCHAR *)ppdev->pCursorBuffer + (48*64*4),
            MAX_GEODELX_COLOR_CURSOR_WIDTH, MAX_GEODELX_COLOR_CURSOR_HEIGHT,
            MAX_GEODELX_COLOR_CURSOR_WIDTH * 4, ppdev->xHotSpot, ppdev->yHotSpot);
    }
    else
    {
        vg_set_color_cursor_shape (ppdev->cursorOffset + (ppdev->activeBuffer << 14),
            (UCHAR *)ppdev->pCursorBuffer,
            MAX_GEODELX_COLOR_CURSOR_WIDTH, MAX_GEODELX_COLOR_CURSOR_HEIGHT,
            MAX_GEODELX_COLOR_CURSOR_WIDTH * 4, ppdev->xHotSpot, ppdev->yHotSpot);
    }

    // 8 different buffers in a chain to prevent tearing issues.
    //
    ppdev->activeBuffer = (ppdev->activeBuffer + 1) & 7;

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// bSetHardwarePointerShape
//
// Changes the shape of the Hardware Pointer.
//
// Returns: True if successful, False if Pointer shape can't be hardware.
//
BOOL bSetHardwarePointerShape(
    SURFOBJ *pso,
    SURFOBJ *psoMask,
    SURFOBJ *psoColor,
    XLATEOBJ *pxlo,
    LONG x,
    LONG y,
    FLONG fl)
{
    PDEV *ppdev = (PDEV *) pso->dhpdev;
    DWORD returnedDataLength;

    DISPDBG((ppdev, 0, "bSetHardwarePointerShape: file %s, line %d\n", __FILE__, __LINE__));

    // CHECK COLOR CURSOR
    //
    if (psoColor != (SURFOBJ *) NULL)
    {
        ppdev->bColorCursor = TRUE;
        if (!bCopyColorPointer(ppdev, psoMask, psoColor, pxlo))
        {
            return FALSE;
        }
    }
    else
    {
        ppdev->bColorCursor = FALSE;

        // MONOCHROME CURSOR
        //
        if (!bCopyMonoPointer(ppdev, psoMask))
        {
            return FALSE;
        }
        vg_set_mono_cursor_colors (0x00000000, 0x00FCFEFD);
    }

    // INITIALIZE POINTER ATTRIBUTES AND POSITION
    // We will call Cimarron directly, instead of using an IOCTL service.
    //
    if (x != -1 || y != -1)
    {
        if (ppdev->regSettings.RotationAngle != 0)
        {
            switch (ppdev->regSettings.RotationAngle)
            {
                case 90:
                {
                    long tx = x;
                    x = y;
                    y = (ppdev->cyScreen - (tx + 1));                
                    break;
                }
                case 180:
                {
                    x = (ppdev->cxScreen - (x + 1));
                    y = (ppdev->cyScreen - (y + 1));
                    break;
                }
                case 270:
                {
                    long tx = x;
                    x = (ppdev->cxScreen - (y + 1));
                    y = tx;
                    break;
                }
            }
        }
        vg_set_cursor_position (x, y, &ppdev->panning);
        vg_set_cursor_enable (1);
        ppdev->cursorEnabled = TRUE;
    }
    
    return (TRUE);
}

/////////////////////////////////////////////////////////////////////////
// DrvSetPointerShape
//
// Sets the new pointer shape.
//
ULONG DrvSetPointerShape(
    SURFOBJ * pso,
    SURFOBJ * psoMask,
    SURFOBJ * psoColor,
    XLATEOBJ * pxlo,
    LONG xHot,
    LONG yHot,
    LONG x,
    LONG y,
    RECTL * prcl,
    FLONG fl)
{
    PDEV *ppdev = (PDEV *) pso->dhpdev;
    DWORD returnedDataLength;

    // STORE HOTSPOTS
    //
    ppdev->xHotSpot = xHot;
    ppdev->yHotSpot = yHot;

    if (ppdev->regSettings.RotationAngle != 0)
    {
        ULONG size;
        if (psoColor != NULL)
            size = 48;
        else
            size = 64;
            
        // Rotated color cursor is 48x48
        //
        switch (ppdev->regSettings.RotationAngle)
        {
            case 90:
            {
                int tx = ppdev->xHotSpot;
                ppdev->xHotSpot = ppdev->yHotSpot;
                ppdev->yHotSpot = (size - 1 - tx);
            }
            break;

            case 180:
            {
                ppdev->xHotSpot = (size - 1 - ppdev->xHotSpot);
                ppdev->yHotSpot = (size - 1 - ppdev->yHotSpot);
            }
            break;
            
            case 270:
            {
                int tx = ppdev->xHotSpot;
                ppdev->xHotSpot = (size - 1 - ppdev->yHotSpot);
                ppdev->yHotSpot = tx;
            }
            break;
        }
    }

    // PROGRAM CURSOR
    // Note that we fall back to a software cursor when performing per-pixel alpha
    // blending, as the hardware will override the alpha value with 0 after
    // blending the cursor with the underlying graphics data.
    //
    if ((ppdev->blendFlags & (LX_BLEND_ENABLE | LX_BLEND_PER_PIXEL)) == (LX_BLEND_ENABLE | LX_BLEND_PER_PIXEL) ||
        !bSetHardwarePointerShape(pso, psoMask, psoColor, pxlo, x, y, fl))
    {
        // UNSUPPORTED CURSOR - DISABLE HARDWARE CURSOR

        if (ppdev->cursorEnabled)
        {
            ppdev->cursorEnabled = FALSE;
            vg_set_cursor_enable(0);
        }

        // MINIPORT DECLINES TO REALIZE THIS POINTER

        ppdev->bSimulatedCursor = TRUE;
        return SPS_DECLINE;
    }

    ppdev->bSimulatedCursor = FALSE;

    // SET NOEXCLUDE
    // Noexclude informs Windows that the cursor is independent of all
    // drawing operations.
    //
    return SPS_ACCEPT_NOEXCLUDE;
}

