//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Text Acceleration for GeodeLX
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_regs.h"
#include "cim_defs.h"

/////////////////////////////////////////////////////////////////////////
// DrvTextOut
//
// Render a text string with optional opaque background. Theoretically
// the background could be any brush, practically I haven't seen anything
// other than solid rectangles, and, in fact, will punt anything else.
//
BOOL DrvTextOut(
    SURFOBJ     *surfPtr,
    STROBJ      *stringPtr,
    FONTOBJ     *fontPtr,
    CLIPOBJ     *clipPtr,
    RECTL       *extraPtr,
    RECTL       *opaqueRectPtr,
    BRUSHOBJ    *fgBrushPtr,
    BRUSHOBJ    *opaqueBrushPtr,
    POINTL      *destPntPtr,
    MIX         mix)
{
    BOOL        more;
    BOOL        moreGlyphs;
    PDEV       *ppdev = NULL;
    long        i;
    long        rectCount;
    BYTE        complexity;
    RECTL       clippedTarget;
    RECTL       destRegion;
    CLIPENUM    clipRectEnum;
    int         xClipped;
    ULONG       glyphCount;
    GLYPHBITS  *glyphBitPtr;
    GLYPHPOS   *glyphPosPtr;
    UCHAR      *glyphDataPtr;
    ULONG       dstOffset;
    ULONG       dstPitch;
    ULONG       index;
    int         rotatingDst;
    USHORT      dhsurfTrg;

    ppdev = (PDEV *)surfPtr->dhpdev;

    DISPDBG ((ppdev, 0, "DrvTextOut Entry\n"));

    if (g_puntAllUntilModeSet)
        goto puntIt;

#if DBG
    if ((g_puntMask & LX_PUNT_ANTIALIASED) && (fontPtr->flFontType & FO_GRAY16) ||
        (g_puntMask & LX_PUNT_MONOTEXT) && !(fontPtr->flFontType & FO_GRAY16))
    {
        goto puntIt;
    }
#endif

    // ROTATING?
    //
    dhsurfTrg = (USHORT)surfPtr->dhsurf;
    rotatingDst = (dhsurfTrg == 1) && (ppdev->regSettings.RotationAngle != 0);

    // READ SURFACE INFORMATION

    index = dhsurf_array[dhsurfTrg];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntIt;

    dstOffset = bitmap_heap[(USHORT)index].heap_offset;
    dstPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;

    ACQUIRE_PRIMITIVE_SEMAPHORE;

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

    // SET DESTINATION STRIDE

    gp_declare_blt (0);
    gp_set_strides (dstPitch, 0);
    gp_write_parameters();

    // HANDLE CLIPPING
    // We are using the dreaded "goto" statement to avoid calling a function
    // and/or duplicating code.

    complexity = (clipPtr == NULL) ? DC_TRIVIAL : clipPtr->iDComplexity;

    if (complexity != DC_COMPLEX)
    {
        if (complexity == DC_TRIVIAL)
            clippedTarget = ppdev->scrnBounds;
        else
            clippedTarget = clipPtr->rclBounds;

        rectCount = 0;
        more = FALSE;

        goto SingleFont;
    }

    // COMPLEX CLIPPING

    CLIPOBJ_cEnumStart(clipPtr, FALSE, CT_RECTANGLES, CD_ANY, 0);

    DISPDBG((ppdev, 0, "Complex clipping\n"));

    for(;;)
    {
        more = CLIPOBJ_bEnum(clipPtr, sizeof(clipRectEnum), (ULONG *) &clipRectEnum);

        // SET RECTANGLE COUNT
        // Count can also be set from above.

        rectCount = clipRectEnum.c;
        DISPDBG((ppdev, 0, "Clip rects %d\n", rectCount));

        while (rectCount)
        {
            rectCount--;
            if(!FindIntersection(&ppdev->scrnBounds, &clipRectEnum.arcl[rectCount], &clippedTarget))
                continue;

            // SINGLE FONT CODE
            // The code to execute a single block is the same as the code to execute
            // multiple complex rectangles.

SingleFont:

            // HANDLE OPAQUE RECTANGLES

            if (opaqueRectPtr != NULL)
            {
                if (FindIntersection (opaqueRectPtr, &clippedTarget, &destRegion))
                {
                    if (rotatingDst)
                    {
                        gp_declare_blt (0);
                        gp_set_raster_operation (0xF0);
                        gp_set_strides (mode_pitch, 0);
                        gp_set_solid_pattern (opaqueBrushPtr->iSolidColor);
                        gp_pattern_fill (
                            ROT_OFFSET (ppdev, destRegion., ppdev->regSettings.RotationAngle),
                            ROT_WIDTH (destRegion., ppdev->regSettings.RotationAngle),
                            ROT_HEIGHT (destRegion., ppdev->regSettings.RotationAngle));
                    }
                    else
                    {
                        gp_declare_blt (0);
                        gp_set_raster_operation (0xF0);
                        gp_set_solid_pattern (opaqueBrushPtr->iSolidColor);
                        gp_pattern_fill (
                            (destRegion.left << (ppdev->cPelSize)) + (destRegion.top * dstPitch) + dstOffset,
                            (destRegion.right  - destRegion.left),
                            (destRegion.bottom - destRegion.top));
                    }

                }
            }

            // RETRIEVE THE GLYPH DATA

            if (stringPtr->pgp == NULL)
            {
                STROBJ_vEnumStart(stringPtr);
                moreGlyphs = STROBJ_bEnum(stringPtr, &glyphCount, &glyphPosPtr);
            }
            else
            {
                glyphPosPtr = stringPtr->pgp;
                glyphCount  = stringPtr->cGlyphs;
                moreGlyphs  = FALSE;
            }

            // SET MONOCHROME OR ANTIALIASED PARAMETERS
            //
            if (fontPtr->flFontType & FO_GRAY16)
            {
                gp_declare_blt (0);
                gp_set_solid_source (fgBrushPtr->iSolidColor);
                gp_write_parameters();
            }
            else
            {
                gp_declare_blt (0);
                gp_set_raster_operation (0xCC);
                gp_set_mono_source (0, fgBrushPtr->iSolidColor, 1);
                gp_write_parameters();
            }

            // LOOP THROUGH EACH STRING

            for(;;)
            {
                ULONG g;

                // HANDLE MONO SPACED FONTS

                if (stringPtr->ulCharInc != 0)
                {
                    ULONG i;
                    long x, y;

                    // SET COORDINATES IN ADVANCE
                    // Since this is a mono font, set the y coordinate of all
                    // glyphs to the same as the first, and space out the x
                    // values equally.

                    x = glyphPosPtr[0].ptl.x;
                    y = glyphPosPtr[0].ptl.y;

                    for (i = 1; i < glyphCount; i++)
                    {
                        x += stringPtr->ulCharInc;
                        glyphPosPtr[i].ptl.x = x;
                        glyphPosPtr[i].ptl.y = y;
                    }
                }

                // LOOP THROUGH FOR EACH CHARACTER

                for (g = 0; g < glyphCount; g++, glyphPosPtr++)
                {
                    long glyphWidth, glyphHeight;
                    long glyphDelta;
                    long glyphXOffset = 0;
                    long dx, dy;
                    POINTL destPnt;
                    ULONG surfDst;
                    RECT tempRegion;

                    // GRAB A POINTER TO THE GLYPH

                    glyphBitPtr = glyphPosPtr->pgdf->pgb;

                    // GLYPH DIMENSIONS

                    glyphWidth  = glyphBitPtr->sizlBitmap.cx;
                    glyphHeight = glyphBitPtr->sizlBitmap.cy;

                    // CALCULATE DESTINATION COORDINATES

                    destPnt.x = glyphPosPtr->ptl.x + glyphBitPtr->ptlOrigin.x;
                    destPnt.y = glyphPosPtr->ptl.y + glyphBitPtr->ptlOrigin.y;

                    // GRAB A POINTER TO THE GLYPH DATA

                    glyphDataPtr = glyphBitPtr->aj;

                    // CALCULATE THE BITMAP PITCH
                    // Grayscale fonts are 4BPP

                    if (fontPtr->flFontType & FO_GRAY16)
                        glyphDelta = (glyphWidth + 1) >> 1;
                    else
                        glyphDelta = (glyphWidth + 7) >> 3;

                    // HORIZONTAL CLIPPING
                    // Horizontal clipping is done before vertical clipping for
                    // the following reason:  If x-clipping is not needed, the
                    // font data is contiguous, (byte-packed).  The interface
                    // to Cimarron is thus simpler (no srcx, srcy, pitch, etc.)

                    xClipped = 0;
                    dx = clippedTarget.left - destPnt.x;

                    if (dx > 0)
                    {
                        xClipped = 1;
                        destPnt.x = clippedTarget.left;

                        glyphXOffset = dx;
                        glyphWidth  -= dx;
                    }

                    // RIGHT CLIPPING

                    dx = (destPnt.x + glyphWidth) - clippedTarget.right;

                    if(dx > 0)
                    {
                        xClipped    = 1;
                        glyphWidth -= dx;
                    }

                    // CHECK FOR TRIVIAL CLIPPING

                    if(glyphWidth <= 0)
                        continue;

                    // BOTTOM CLIPPING

                    dy = (destPnt.y + glyphHeight) - clippedTarget.bottom;

                    if(dy > 0)
                        glyphHeight -= dy;

                    // TOP CLIPPING

                    dy = clippedTarget.top - destPnt.y;

                    if(dy > 0)
                    {
                        destPnt.y     = clippedTarget.top;
                        glyphHeight  -= dy;
                        glyphDataPtr += dy * glyphDelta;
                    }

                    // CHECK TRIVIAL CLIP

                    if (glyphHeight <= 0)
                        continue;

                    surfDst = (destPnt.x << ppdev->cPelSize) +
                        (destPnt.y * dstPitch) + dstOffset;

                    if (rotatingDst)
                    {
                        // Collapse all cases into a single pre-rotate
                        //
                        tempRegion.left   = destPnt.x;
                        tempRegion.right  = destPnt.x + glyphWidth;
                        tempRegion.top    = destPnt.y;
                        tempRegion.bottom = destPnt.y + glyphHeight;

                        gp_declare_blt (0);
                        gp_set_strides (rot_pitch, mode_pitch);
                        gp_set_source_format (ppdev->cimFormat);
                        gp_rotate_blt (surfDst,
                            ROT_OFFSET (ppdev, tempRegion., ppdev->regSettings.RotationAngle),
                            ROT_WIDTH (tempRegion., ppdev->regSettings.RotationAngle),
                            ROT_HEIGHT (tempRegion., ppdev->regSettings.RotationAngle),
                            ppdev->regSettings.RotationAngle);
                    }

                    // Blended or pure mono text
                    //
                    if (fontPtr->flFontType & FO_GRAY16)
                    {
                        gp_declare_blt (0);
                        gp_antialiased_text (
                            surfDst,
                            glyphXOffset,
                            glyphWidth,
                            glyphHeight,
                            glyphDataPtr,
                            glyphDelta,
                            1);
                    }
                    else if (xClipped)
                    {
                        gp_declare_blt (0);
                        if (rotatingDst)
                            gp_set_mono_source (0, fgBrushPtr->iSolidColor, 1);
                        gp_mono_bitmap_to_screen_blt (
                            surfDst,
                            glyphXOffset,
                            glyphWidth,
                            glyphHeight,
                            glyphDataPtr,
                            glyphDelta);
                    }
                    else
                    {
                        gp_declare_blt (0);
                        if (rotatingDst)
                            gp_set_mono_source (0, fgBrushPtr->iSolidColor, 1);
                        gp_text_blt (
                             surfDst,
                             glyphWidth,
                             glyphHeight,
                             glyphDataPtr);
                    }

                    if (rotatingDst)
                    {
                        // Collapse all cases into a single post-rotate
                        //
                        gp_declare_blt (0);
                        gp_set_strides (mode_pitch, rot_pitch);
                        gp_rotate_blt (
                            ROT_OFFSET (ppdev, tempRegion., ppdev->regSettings.RotationAngle),
                            surfDst,
                            glyphWidth,
                            glyphHeight,
                            360 - ppdev->regSettings.RotationAngle);
                    }
                }

                // LAST STRING?

                if(!moreGlyphs)
                    break;

                moreGlyphs = STROBJ_bEnum(stringPtr, &glyphCount, &glyphPosPtr);
            }

        } // while (rectCount)

        // LAST RECTANGLE

        if (!more)
            break;

    }

returnTrue:

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 0, "DrvTextOut Exit\n"));
    return (TRUE);

puntIt:

    if (fontPtr->flFontType & FO_GRAY16)
    {
        gp_wait_until_idle();
        return EngTextOut (surfPtr, stringPtr, fontPtr, clipPtr, extraPtr, opaqueRectPtr,
            fgBrushPtr, opaqueBrushPtr, destPntPtr, mix);
    }
    else
    {
        ULONG index;
        BOOL  ret;
        USHORT dhsurfTrg = (USHORT)surfPtr->dhsurf;

        ACQUIRE_PRIMITIVE_SEMAPHORE;

        // Pre-rotate the screen
        // This is a huge thing to do for a simple font copy.  However, we should
        // NEVER get here except when performing a debug punt.  The normal driver always
        // handles this call.
        //
        if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 && !g_puntAllUntilModeSet)
        {
            gp_declare_blt (0);
            gp_set_bpp (ppdev->cBitsPerPel);
            gp_set_strides (rot_pitch, mode_pitch);
            gp_set_source_format (ppdev->cimFormat);
            gp_rotate_blt (
                rot_offset,
                0,
                ppdev->cxScreen, ppdev->cyScreen,
                ppdev->regSettings.RotationAngle);
        }

        MAKE_PUNTABLE(surfPtr);
        gp_wait_until_idle();

        ret = EngTextOut (surfPtr, stringPtr, fontPtr, clipPtr, extraPtr, opaqueRectPtr,
            fgBrushPtr, opaqueBrushPtr, destPntPtr, mix);

        if (ppdev->regSettings.RotationAngle != 0 && dhsurfTrg == 1 && !g_puntAllUntilModeSet)
        {
            gp_declare_blt (0);
            gp_set_strides (mode_pitch, rot_pitch);
            gp_rotate_blt (
                0, rot_offset,
                (ppdev->regSettings.RotationAngle == 180) ? ppdev->cxScreen : ppdev->cyScreen,
                (ppdev->regSettings.RotationAngle == 180) ? ppdev->cyScreen : ppdev->cxScreen,
                360 - ppdev->regSettings.RotationAngle);
        }

        RELEASE_PRIMITIVE_SEMAPHORE;

        return ret;
    }
}
