// Copyright (c) 2006 Advanced Micro Devices ("AMD").
//
// All Rights Reserved.  Unpublished rights reserved under the 
// copyright laws of the United States of America, other countries 
// and international treaties.  The software is provided without 
// fee.  Permission to use, copy, store, modify, disclose, transmit 
// or distribute the software is granted, provided that this 
// copyright notice must appear in any copy, modification, 
// disclosure, transmission or distribution of the software.
//
// AMD retains all ownership, copyright, trade secret and 
// proprietary rights in the software. THIS SOFTWARE HAS BEEN 
// PROVIDED "AS IS," WITHOUT EXPRESS OR IMPLIED WARRANTY INCLUDING, 
// WITHOUT LIMITATION, IMPLIED WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR USE AND NON-INFRINGEMENT.
//
// AMD does not assume or authorize any other person to assume for 
// it any liability in connection with the Software. AMD SHALL NOT 
// BE LIABLE TO COMPANY, OR ANY THIRD PARTY, IN CONTRACT, TORT, 
// WARRANTY, STRICT LIABILITY, OR OTHERWISE FOR ANY DIRECT DAMAGES, 
// OR FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL 
// DAMAGES, INCLUDING BUT NOT LIMITED TO, BUSINESS INTERRUPTION, 
// LOST PROFITS OR GOODWILL, OR LOSS OF INFORMATION EVEN IF AMD IS 
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
#include "stdafx.h"
#include "lx_gtf.h"
#include "lx_gtfDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// Routine to calculate M, N and P settings for an arbitrary frequency
//
unsigned long CalculatePLL (double frequency);

// CLX_GTFDlg dialog

CLX_GTFDlg::CLX_GTFDlg(CWnd* pParent /*=NULL*/)
    : CDialog(CLX_GTFDlg::IDD, pParent)
    , m_panelWidth(800)
    , m_panelHeight(600)
    , m_panelRefresh(60)
    , m_panelBusWidth(0)
    , m_simultaneous(TRUE)
    , m_negHsync(TRUE)
    , m_negVsync(TRUE)
{
    m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CLX_GTFDlg::DoDataExchange(CDataExchange* pDX)
{
    CDialog::DoDataExchange(pDX);
    DDX_Text(pDX, IDC_PANEL_WIDTH, m_panelWidth);
    DDV_MinMaxUInt(pDX, m_panelWidth, 0, 8192);
    DDX_Text(pDX, IDC_PANEL_HEIGHT, m_panelHeight);
    DDV_MinMaxUInt(pDX, m_panelHeight, 0, 8192);
    DDX_Text(pDX, IDC_PANEL_REFRESH, m_panelRefresh);
    DDV_MinMaxUInt(pDX, m_panelRefresh, 0, 200);
    DDX_Radio(pDX, IDC_RADIO1, m_panelBusWidth);
    DDX_Control(pDX, IDC_GENERATE, m_generateButton);
    DDX_Check(pDX, IDC_SIMULTANEOUS, m_simultaneous);
    DDX_Check(pDX, IDC_NEG_HSYNC, m_negHsync);
    DDX_Check(pDX, IDC_NEG_VSYNC, m_negVsync);
}

BEGIN_MESSAGE_MAP(CLX_GTFDlg, CDialog)
    ON_WM_SYSCOMMAND()
    ON_WM_PAINT()
    ON_WM_QUERYDRAGICON()
    //}}AFX_MSG_MAP
    ON_BN_CLICKED(IDC_GENERATE, OnBnClickedGenerate)
END_MESSAGE_MAP()

// CLX_GTFDlg message handlers

BOOL CLX_GTFDlg::OnInitDialog()
{
    CDialog::OnInitDialog();

    // Set the icon for this dialog.  The framework does this automatically
    //  when the application's main window is not a dialog
    //
    SetIcon(m_hIcon, TRUE);         // Set big icon
    SetIcon(m_hIcon, FALSE);        // Set small icon

    m_generateButton.SetFocus();
    return FALSE;  // return TRUE  unless you set the focus to a control
}

void CLX_GTFDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
    CDialog::OnSysCommand(nID, lParam);
}

void CLX_GTFDlg::OnPaint() 
{
    if (IsIconic())
    {
        CPaintDC dc(this); // device context for painting

        SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

        // Center icon in client rectangle
        //
        int cxIcon = GetSystemMetrics(SM_CXICON);
        int cyIcon = GetSystemMetrics(SM_CYICON);
        CRect rect;
        GetClientRect(&rect);
        int x = (rect.Width() - cxIcon + 1) / 2;
        int y = (rect.Height() - cyIcon + 1) / 2;

        // Draw the icon
        //
        dc.DrawIcon(x, y, m_hIcon);
    }
    else
    {
        CDialog::OnPaint();
    }
}

// The system calls this function to obtain the cursor to display while the user drags
// the minimized window.
//
HCURSOR CLX_GTFDlg::OnQueryDragIcon()
{
    return static_cast<HCURSOR>(m_hIcon);
}

void CLX_GTFDlg::OnBnClickedGenerate()
{
    CString temp;
    UpdateData (TRUE);

    // Generate CRT-compatible timings inside a customization file format.
    //
    CDisplayTimings timingDialog;
    timingDialog.m_generatedTimings = "";

    // Add file header
    //
    temp.Format ("binary_file lx_%ix%i_custom.out\r\n", (int)m_panelWidth, (int)m_panelHeight);
    timingDialog.m_generatedTimings += temp;

    temp.Format ("inf_file lx_%ix%i_custom.inf\r\n", (int)m_panelWidth, (int)m_panelHeight);
    timingDialog.m_generatedTimings += temp;

    temp.Format ("disk_id %c%AMD Custom Driver For %ix%i Panel%c%\r\n", '"', (int)m_panelWidth, (int)m_panelHeight, '"');
    timingDialog.m_generatedTimings += temp;    

    timingDialog.m_generatedTimings += "show_crt_modes 0\r\n";
    temp.Format ("simultaneous_CRT %i\r\n\r\n", (int)m_simultaneous);
    timingDialog.m_generatedTimings += temp;

    // Add mode flags based on user-selected settings.
    //
    timingDialog.m_generatedTimings += "flags (VG_MODEFLAG_MANUAL_FREQUENCY |\r\n";
    if (m_negHsync)
        timingDialog.m_generatedTimings += "       VG_MODEFLAG_NEG_HSYNC        |\r\n";
    if (m_negVsync)
        timingDialog.m_generatedTimings += "       VG_MODEFLAG_NEG_VSYNC        |\r\n";  
    if (m_panelWidth > 1280)
        timingDialog.m_generatedTimings += "       VG_MODEFLAG_HIGH_BAND";
    else
        timingDialog.m_generatedTimings += "       VG_MODEFLAG_AVG_BAND";
    
    // No dithering selected, so no need to specify custom panel timings.
    //
    if (m_panelBusWidth == 0)
        timingDialog.m_generatedTimings += ")\r\n";
    else
    {
        timingDialog.m_generatedTimings += "         |\r\n";
        timingDialog.m_generatedTimings += "       VG_MODEFLAG_CUSTOM_PANEL)\r\n";
    }
    
    // Add panel register settings.
    //
    temp.Format ("panel_width            %i\r\n", (int)m_panelWidth);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("panel_height           %i\r\n", (int)m_panelHeight);
    timingDialog.m_generatedTimings += temp;
    if (m_panelBusWidth == 0)
    {
        timingDialog.m_generatedTimings += "panel_tim1             0\r\n";
        timingDialog.m_generatedTimings += "panel_tim2             0\r\n";
        timingDialog.m_generatedTimings += "panel_dither_ctl       0\r\n";
        timingDialog.m_generatedTimings += "panel_pad_sel_low      0\r\n";
        timingDialog.m_generatedTimings += "panel_pad_sel_high     0\r\n";
    }
    else
    {
        timingDialog.m_generatedTimings += "panel_tim1             0x00000000\r\n";
        timingDialog.m_generatedTimings += "panel_tim2             0x08C00000\r\n";
        switch (m_panelBusWidth)
        {
            case 1:  timingDialog.m_generatedTimings += "panel_dither_ctl       0x00000047\r\n"; break;
            case 2:  timingDialog.m_generatedTimings += "panel_dither_ctl       0x00000059\r\n"; break;
            default: timingDialog.m_generatedTimings += "panel_dither_ctl       0x0000006B\r\n"; break;
        }
        timingDialog.m_generatedTimings += "panel_pad_sel_low      0xDFFFFFFF\r\n";
        timingDialog.m_generatedTimings += "panel_pad_sel_high     0x0000003F\r\n";
    }

    // Calculate the timing values according to the default generalized
    // timing formula.  The following calculations are pulled from the generalized
    // timing formula spreadsheet with standard defaults.
    //
    double hperiod = ((1.0 / (double)m_panelRefresh) - TGT_VERT_FLYBACK / 1000000.0);
    hperiod /= (double)m_panelHeight + MIN_V_FRONT_PORCH;
    hperiod *= 1000000.0;
    double sync_lines = (TGT_VERT_FLYBACK / hperiod) - NUM_VSYNC_LINES;\
    double tot_lines = m_panelHeight + MIN_V_FRONT_PORCH + NUM_VSYNC_LINES + (DWORD)sync_lines;
    double est_rate = 1.0 / hperiod / tot_lines * 1000000.0;
    double act_period_us = hperiod / ((double)m_panelRefresh / est_rate);
    double ideal_duty_cycle = ACTUAL_C -(ACTUAL_M * act_period_us / 1000.0);
    double blank_pixels =((double)m_panelWidth * ideal_duty_cycle / (100.0 - ideal_duty_cycle));
    blank_pixels = (double)((DWORD)((blank_pixels / (CHAR_WIDTH * 2.0)) + 0.5));
    blank_pixels *= (CHAR_WIDTH * 2.0);
    double tot_pixels = m_panelWidth + blank_pixels;
    double freq_mhz = tot_pixels / hperiod;
    double hsync_width = (double)((DWORD)((HSYNC_PERCENTAGE * tot_pixels / CHAR_WIDTH) + 0.5) * CHAR_WIDTH);
    double hfront_porch = (blank_pixels / 2.0) - hsync_width;
    double hback_porch = hfront_porch + hsync_width;

    temp.Format ("hactive                %i\r\n", (int)m_panelWidth);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("hblankstart            %i\r\n", (int)m_panelWidth);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("hsyncstart             %i\r\n", (int)(m_panelWidth + hfront_porch));
    timingDialog.m_generatedTimings += temp;
    temp.Format ("hsyncend               %i\r\n", (int)(m_panelWidth + hfront_porch + hsync_width));
    timingDialog.m_generatedTimings += temp;
    temp.Format ("hblankend              %i\r\n", (int)tot_pixels);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("htotal                 %i\r\n", (int)tot_pixels);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vactive                %i\r\n", (int)m_panelHeight);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vblankstart            %i\r\n", (int)m_panelHeight);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vsyncstart             %i\r\n", (int)(m_panelHeight + (DWORD)MIN_V_FRONT_PORCH));
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vsyncend               %i\r\n", 
        (int)(m_panelHeight + (DWORD)MIN_V_FRONT_PORCH + NUM_VSYNC_LINES));
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vblankend              %i\r\n", (int)tot_lines);
    timingDialog.m_generatedTimings += temp;
    temp.Format ("vtotal                 %i\r\n", (int)tot_lines);
    timingDialog.m_generatedTimings += temp;

    // Use a custom PLL setting, as otherwise the setting would need to be in Cimarron's
    // internal PLL tables.
    //
    temp.Format ("frequency              0x%8.8lX\r\n", CalculatePLL (freq_mhz));
    timingDialog.m_generatedTimings += temp;

    timingDialog.DoModal();
}

void CLX_GTFDlg::OnOK()
{
    // Stubbed out routine to prevent the enter key from closing out the dialog.
}

// CDisplayTimings Class
// This class presents the final customization text for the user to cut and
// paste.
//
IMPLEMENT_DYNAMIC(CDisplayTimings, CDialog)
CDisplayTimings::CDisplayTimings(CWnd* pParent /*=NULL*/)
    : CDialog(CDisplayTimings::IDD, pParent)
    , m_generatedTimings(_T(""))
{
    m_fontCreated = false;

    // Attempt to create and use a font that sets each character to the same pixel
    // width.  Slight formatting win.
    //
    if (m_font.CreateFont(
        13,                        // nHeight
        0,                         // nWidth
        0,                         // nEscapement
        0,                         // nOrientation
        FW_NORMAL,                 // nWeight
        FALSE,                     // bItalic
        FALSE,                     // bUnderline
        0,                         // cStrikeOut
        ANSI_CHARSET,              // nCharSet
        OUT_DEFAULT_PRECIS,        // nOutPrecision
        CLIP_DEFAULT_PRECIS,       // nClipPrecision
        DEFAULT_QUALITY,           // nQuality
        DEFAULT_PITCH | FF_SWISS,  // nPitchAndFamily
        "Lucida Console"))         // lpszFacename
    {
        m_fontCreated = true;
    }
}

CDisplayTimings::~CDisplayTimings()
{

}

void CDisplayTimings::DoDataExchange(CDataExchange* pDX)
{
    CDialog::DoDataExchange(pDX);
    DDX_Text(pDX, IDC_TIMINGS_EDIT, m_generatedTimings);
    DDX_Control(pDX, IDC_TIMINGS_EDIT, m_textEdit);
}


BEGIN_MESSAGE_MAP(CDisplayTimings, CDialog)
END_MESSAGE_MAP()


// CDisplayTimings message handlers

BOOL CDisplayTimings::OnInitDialog()
{
    CDialog::OnInitDialog();

    if (m_fontCreated)
    {
        m_textEdit.SetFont (&m_font, FALSE);
    }

    return TRUE;  // return TRUE unless you set the focus to a control
    // EXCEPTION: OCX Property Pages should return FALSE
}

