//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Console application to generate a binary customization file and associated
// .inf install file.
// </DOC_AMD_STD>
//

#include <afx.h>
#include "lxcustom.h"
#include "..\display\lxshared.h"

int g_line = 1;

int main(int argc, char* argv[])
{
    unsigned long length;
    unsigned long extraOffset;
    unsigned char *outBuffer;
    unsigned char **extraArray;
    unsigned long *extraSizes;
    unsigned long size;
    CString bin_file;
    CString filename;
    CString filetext;
    CString token1, token2, token3;
    CString disk_id;
    CFile input, output, inf;
    char *filePtr;
    int i, x, count;
    int extraIndex = 0;
    
    if (argc != 2)
    {
        DisplayUsage();
        return 0;
    }

    filename = argv[1];

    input.Open (filename, CFile::modeRead, NULL); 
    if (input.m_hFile == CFile::hFileNull)
    {
        printf ("Unable to open input file: %s\n", filename);
        return 0;
    }

    // Buffer the entire file contents in a CString.  Don't forget
    // the NULL termination.
    //
    length = (long)input.GetLength();
    filePtr = filetext.GetBuffer (length + 1);
    input.Read (filePtr, length);
    filePtr[length] = 0;

    // Check if the first four digits are a binary signature.
    //
    if (*((unsigned long *)filePtr) == LX_MODE_SIGNATURE1 &&
        *((unsigned long *)filePtr + 1) == LX_MODE_SIGNATURE2)
    {
        fprintf (stderr, "Binary Input File: %s\n", filename.GetBuffer (1));

        // Binary file - Reverse engineer the contents and output to standard out.
        //
        filePtr += 8;

        count = *((unsigned long *)filePtr);
        filePtr += 4;

        printf ("binary_file %s\n", filename.GetBuffer(1));
        printf ("inf_file custom.inf\n");
        printf ("disk_id = \"Reverse Engineered Registry File\"\n");
        printf ("manual_display_mode_count %i\n\n", count);

        for (i = 0; i < count; i++)
        {
            GEODELX_CUSTOM_MODE *pMode = (GEODELX_CUSTOM_MODE *)(filePtr + (sizeof(GEODELX_CUSTOM_MODE) * i));
            CString base;

            base.Format ("manual_mode_%i", i);
            printf ("%s bpp_mask                              0x%X\n", base.GetBuffer (1), pMode->bpp_mask);
            printf ("%s output_path                           0x%X\n", base.GetBuffer (1), pMode->output_path);
            printf ("%s output_space                          0x%X\n", base.GetBuffer (1), pMode->output_space);
            printf ("%s flicker_strength                      0x%X\n", base.GetBuffer (1), pMode->flicker_strength);
            printf ("%s line_double_video_on_interlaced       0x%X\n", base.GetBuffer (1), pMode->line_double_video_on_interlaced);
            printf ("%s border_color                          0x%X\n", base.GetBuffer (1), pMode->border_color);
            printf ("%s premode_smbus_count                   0x%X\n", base.GetBuffer (1), pMode->premode_smbus_count);
            printf ("%s postmode_smbus_count                  0x%X\n", base.GetBuffer (1), pMode->postmode_smbus_count);
            printf ("%s custom_coefficients                   0x%X\n", base.GetBuffer (1), (pMode->hcoeff && pMode->vcoeff) ? 1 : 0);
            printf ("%s configure_vop                         0x%X\n\n", base.GetBuffer (1), pMode->configure_vop);

            base += "\\manual_timings";
            printf ("%s flags                  0x%X\n", base.GetBuffer(1), pMode->manual_timings.flags);
            printf ("%s src_width              0x%X\n", base.GetBuffer(1), pMode->manual_timings.src_width);
            printf ("%s src_height             0x%X\n", base.GetBuffer(1), pMode->manual_timings.src_height);
            printf ("%s mode_width             0x%X\n", base.GetBuffer(1), pMode->manual_timings.mode_width);
            printf ("%s mode_height            0x%X\n", base.GetBuffer(1), pMode->manual_timings.mode_height);
            printf ("%s panel_width            0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_width);
            printf ("%s panel_height           0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_height);
            printf ("%s panel_tim1             0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_tim1);
            printf ("%s panel_tim2             0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_tim2);
            printf ("%s panel_dither_ctl       0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_dither_ctl);
            printf ("%s panel_pad_sel_low      0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_pad_sel_low);
            printf ("%s panel_pad_sel_high     0x%X\n", base.GetBuffer(1), pMode->manual_timings.panel_pad_sel_high);
            printf ("%s hactive                0x%X\n", base.GetBuffer(1), pMode->manual_timings.hactive);
            printf ("%s hblankstart            0x%X\n", base.GetBuffer(1), pMode->manual_timings.hblankstart);
            printf ("%s hsyncstart             0x%X\n", base.GetBuffer(1), pMode->manual_timings.hsyncstart);
            printf ("%s hsyncend               0x%X\n", base.GetBuffer(1), pMode->manual_timings.hsyncend);
            printf ("%s hblankend              0x%X\n", base.GetBuffer(1), pMode->manual_timings.hblankend);
            printf ("%s htotal                 0x%X\n", base.GetBuffer(1), pMode->manual_timings.htotal);
            printf ("%s vactive                0x%X\n", base.GetBuffer(1), pMode->manual_timings.vactive);
            printf ("%s vblankstart            0x%X\n", base.GetBuffer(1), pMode->manual_timings.vblankstart);
            printf ("%s vsyncstart             0x%X\n", base.GetBuffer(1), pMode->manual_timings.vsyncstart);
            printf ("%s vsyncend               0x%X\n", base.GetBuffer(1), pMode->manual_timings.vsyncend);
            printf ("%s vblankend              0x%X\n", base.GetBuffer(1), pMode->manual_timings.vblankend);
            printf ("%s vtotal                 0x%X\n", base.GetBuffer(1), pMode->manual_timings.vtotal);
            printf ("%s vactive_even           0x%X\n", base.GetBuffer(1), pMode->manual_timings.vactive_even);
            printf ("%s vblankstart_even       0x%X\n", base.GetBuffer(1), pMode->manual_timings.vblankstart_even);
            printf ("%s vsyncstart_even        0x%X\n", base.GetBuffer(1), pMode->manual_timings.vsyncstart_even);
            printf ("%s vsyncend_even          0x%X\n", base.GetBuffer(1), pMode->manual_timings.vsyncend_even);
            printf ("%s vblankend_even         0x%X\n", base.GetBuffer(1), pMode->manual_timings.vblankend_even);
            printf ("%s vtotal_even            0x%X\n", base.GetBuffer(1), pMode->manual_timings.vtotal_even);
            printf ("%s frequency              0x%X\n\n", base.GetBuffer(1), pMode->manual_timings.frequency);

            // Get VOP settings
            //
            if (pMode->configure_vop)
            {
                base.Format ("manual_mode_%i\\vop_settings", i);
                printf ("%s flags                    0x%X\n", base.GetBuffer(1), pMode->vop_config.flags);
                printf ("%s mode                     0x%X\n", base.GetBuffer(1), pMode->vop_config.mode);
                printf ("%s conversion_mode          0x%X\n", base.GetBuffer(1), pMode->vop_config.conversion_mode);
                printf ("%s vsync_out                0x%X\n", base.GetBuffer(1), pMode->vop_config.vsync_out);

                if (pMode->vop_config.mode == VOP_MODE_601)
                {
                    base += "\\vop601";
                    printf ("%s flags             0x%X\n",   base.GetBuffer(1), pMode->vop_config.vop601.flags);
                    printf ("%s vsync_shift       0x%X\n",   base.GetBuffer(1), pMode->vop_config.vop601.vsync_shift);
                    printf ("%s vsync_shift_count 0x%X\n",   base.GetBuffer(1), pMode->vop_config.vop601.vsync_shift_count);
                    printf ("%s output_mode       0x%X\n",   base.GetBuffer(1), pMode->vop_config.vop601.output_mode);
                }

                printf ("\n");
            }

            if (pMode->hcoeff && pMode->vcoeff)
            {
                unsigned long *hcoeff = (unsigned long *)(filePtr + (unsigned long)pMode->hcoeff);
                unsigned long *vcoeff = (unsigned long *)(filePtr + (unsigned long)pMode->vcoeff);

                base.Format ("manual_mode_%i\\custom_hcoeff", i);
                for (x = 0; x < 256; x++)
                {
                    token1.Format ("hcoeff_%i_0              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)(hcoeff[x << 1] & 0x3FF) << 22) >> 22);
                    token1.Format ("hcoeff_%i_1              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)((hcoeff[x << 1] >> 10) & 0x3FF) << 22) >> 22);
                    token1.Format ("hcoeff_%i_2              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)((hcoeff[x << 1] >> 20) & 0x3FF) << 22) >> 22);
                    token1.Format ("hcoeff_%i_3              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)(hcoeff[(x << 1) + 1] & 0x3FF) << 22) >> 22);
                    token1.Format ("hcoeff_%i_4              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)((hcoeff[(x << 1) + 1] >> 10) & 0x3FF) << 22) >> 22);
                }
                base.Format ("manual_mode_%i\\custom_vcoeff", i);
                for (x = 0; x < 256; x++)
                {
                    token1.Format ("vcoeff_%i_0              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)(vcoeff[x] & 0x3FF) << 22) >> 22);
                    token1.Format ("vcoeff_%i_1              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)((vcoeff[x] >> 10) & 0x3FF) << 22) >> 22);
                    token1.Format ("vcoeff_%i_2              ", x);
                    printf ("%s %s%i\n", base.GetBuffer (1), token1.GetBufferSetLength (24),
                        ((int)((vcoeff[x] >> 20) & 0x3FF) << 22) >> 22);
                }
                printf ("\n");
            }

            if (pMode->premode_smbus_count)
            {
                SMBUS_WRITE *pWrite = (SMBUS_WRITE *)(filePtr + (unsigned long)pMode->premode_smbus);

                base.Format ("manual_mode_%i\\premode_smbus", i);
                for (x = 0; x < (int)pMode->premode_smbus_count; x++)
                {
                    token1.Format ("base_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(23), pWrite[x].base);
                    token1.Format ("addr_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(23), pWrite[x].addr);
                    token1.Format ("data_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(23), pWrite[x].data);
                }

                printf ("\n");
            }

            if (pMode->postmode_smbus_count)
            {
                SMBUS_WRITE *pWrite = (SMBUS_WRITE *)(filePtr + (unsigned long)pMode->postmode_smbus);

                base.Format ("manual_mode_%i\\postmode_smbus", i);
                for (x = 0; x < (int)pMode->postmode_smbus_count; x++)
                {
                    token1.Format ("base_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(22), pWrite[x].base);
                    token1.Format ("addr_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(22), pWrite[x].addr);
                    token1.Format ("data_%i                  ", x);
                    printf ("%s %s 0x%2.2X\n", base.GetBuffer(1), token1.GetBufferSetLength(22), pWrite[x].data);
                }

                printf ("\n");
            }
        }
    }
    else
    {
        filetext.ReleaseBuffer();

        // Close the input file as it is no longer required.
        //
        input.Close();

        // The first line of the input file indicates the name of the output file
        //
        if (!NextToken (filetext, token1)) return 0;
        if (!NextToken (filetext, token2)) return 0;

        if (token1 != "binary_file")
        {
            printf ("Invalid file syntax!  First line should be \'binary_file xxxx\'\n");
            return 0;
        }

        bin_file = token2;
        output.Open (bin_file, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive, NULL); 
        if (output.m_hFile == CFile::hFileNull)
        {
            printf ("Unable to open output file: %s\n", bin_file.GetBuffer (1));
            return 0;
        }

        printf ("Binary File specified as %s\n", bin_file.GetBuffer(1));

        // The second line specifies the .inf filename
        //
        if (!NextToken (filetext, token1)) return 0;
        if (!NextToken (filetext, token2)) return 0;

        if (token1 != "inf_file")
        {
            printf ("Invalid file syntax!  Second line should be \'inf_file xxxx\'\n");
            return 0;
        }

        filename = token2;
        inf.Open (filename, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive, NULL); 
        if (inf.m_hFile == CFile::hFileNull)
        {
            printf ("Unable to open output file: %s\n", filename.GetBuffer (1));
            return 0;
        }

        printf (".Inf File specified as %s\n", filename.GetBuffer(1));

        // The third line specifies the install text
        //
        if (!NextToken (filetext, token1)) return 0;
        {
            if (token1 != "disk_id")
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }

            // Extract the id string
            //
            int start = filetext.Find ('"');
            if (start < 0)
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }
            start++;
            int last = filetext.Find ('"', start + 1);
            if (last < 0)
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }

            filetext = filetext.Right (filetext.GetLength() - start);
            disk_id = filetext.Left (last - start);
            filetext = filetext.Right (filetext.GetLength() - (last - start));

            printf ("Disk ID = \"%s\"", disk_id.GetBuffer(1));
        }
        
        if (!NextToken (filetext, token2)) return 0;
        
        // The fourth line specifies the number of manual display modes.
        //
        if (!NextToken (filetext, token1)) return 0;
        if (!NextToken (filetext, token2)) return 0;

        if (token1 != "manual_display_mode_count")
        {
            printf ("Invalid File Syntax!  Fourth line should be \'manual_display_mode_count xxxx\'\n");
            return 0;
        }

        count = ExtractValue (&token2);

        // Allocate enough space for all the custom mode information.  Optional arrays such as
        // coefficients and SMBus writes are allocated later on the fly.  However, when the data is
        // finally written to the file, the pointers that point to these arrays will simply
        // contain offsets into the file.  In this way, no extra memory allocations will be
        // needed in the driver.
        //
        outBuffer  = (unsigned char *) GlobalAlloc (GMEM_FIXED, (sizeof(GEODELX_CUSTOM_MODE) * count));
        extraArray = (unsigned char **)GlobalAlloc (GMEM_FIXED, (sizeof(unsigned char *) * count * 4));
        extraSizes = (unsigned long *) GlobalAlloc (GMEM_FIXED, (sizeof(unsigned long) * count * 4));

        memset (outBuffer,  0, (sizeof(GEODELX_CUSTOM_MODE) * count));
        memset (extraArray, 0, (sizeof(unsigned char *) * count * 4));
        memset (extraSizes, 0, (sizeof(unsigned long) * count * 4));

        if (!outBuffer || !extraArray || !extraSizes)
        {
            printf ("Unable to allocate memory.  Exiting...\n");
            goto cleanUp;
        }

        // Coefficient and SMBus allocations start after last display mode.
        //
        extraOffset = (sizeof(GEODELX_CUSTOM_MODE) * count);

        for (i = 0; i < count; i++)
        {
            GEODELX_CUSTOM_MODE *pMode = (GEODELX_CUSTOM_MODE *)(outBuffer + (i * sizeof(GEODELX_CUSTOM_MODE)));
            CString base;
            CString base2;
            CString setting;
            unsigned long coefficients;

            // Get base settings for the mode.
            //
            base.Format ("manual_mode_%i", i);

            coefficients = 0;
            GET_SETTING (base, "bpp_mask",                        &(pMode->bpp_mask));
            GET_SETTING (base, "output_path",                     &(pMode->output_path));
            GET_SETTING (base, "output_space",                    &(pMode->output_space));
            GET_SETTING (base, "flicker_strength",                &(pMode->flicker_strength));
            GET_SETTING (base, "line_double_video_on_interlaced", &(pMode->line_double_video_on_interlaced));
            GET_SETTING (base, "border_color",                    &(pMode->border_color));
            GET_SETTING (base, "premode_smbus_count",             &(pMode->premode_smbus_count));
            GET_SETTING (base, "postmode_smbus_count",            &(pMode->postmode_smbus_count));
            GET_SETTING (base, "custom_coefficients",             &coefficients);
            GET_SETTING (base, "configure_vop",                   &(pMode->configure_vop));

            base2 = base + "\\manual_timings";
            GET_SETTING (base2, "flags",                &(pMode->manual_timings.flags));
            GET_SETTING (base2, "src_width",            &(pMode->manual_timings.src_width));
            GET_SETTING (base2, "src_height",           &(pMode->manual_timings.src_height));
            GET_SETTING (base2, "mode_width",           &(pMode->manual_timings.mode_width));
            GET_SETTING (base2, "mode_height",          &(pMode->manual_timings.mode_height));
            GET_SETTING (base2, "panel_width",          &(pMode->manual_timings.panel_width));
            GET_SETTING (base2, "panel_height",         &(pMode->manual_timings.panel_height));
            GET_SETTING (base2, "panel_tim1",           &(pMode->manual_timings.panel_tim1));
            GET_SETTING (base2, "panel_tim2",           &(pMode->manual_timings.panel_tim2));
            GET_SETTING (base2, "panel_dither_ctl",     &(pMode->manual_timings.panel_dither_ctl));
            GET_SETTING (base2, "panel_pad_sel_low",    &(pMode->manual_timings.panel_pad_sel_low));
            GET_SETTING (base2, "panel_pad_sel_high",   &(pMode->manual_timings.panel_pad_sel_high));
            GET_SETTING (base2, "hactive",              &(pMode->manual_timings.hactive));
            GET_SETTING (base2, "hblankstart",          &(pMode->manual_timings.hblankstart));
            GET_SETTING (base2, "hsyncstart",           &(pMode->manual_timings.hsyncstart));
            GET_SETTING (base2, "hsyncend",             &(pMode->manual_timings.hsyncend));
            GET_SETTING (base2, "hblankend",            &(pMode->manual_timings.hblankend));
            GET_SETTING (base2, "htotal",               &(pMode->manual_timings.htotal));
            GET_SETTING (base2, "vactive",              &(pMode->manual_timings.vactive));
            GET_SETTING (base2, "vblankstart",          &(pMode->manual_timings.vblankstart));
            GET_SETTING (base2, "vsyncstart",           &(pMode->manual_timings.vsyncstart));
            GET_SETTING (base2, "vsyncend",             &(pMode->manual_timings.vsyncend));
            GET_SETTING (base2, "vblankend",            &(pMode->manual_timings.vblankend));
            GET_SETTING (base2, "vtotal",               &(pMode->manual_timings.vtotal));
            GET_SETTING (base2, "vactive_even",         &(pMode->manual_timings.vactive_even));
            GET_SETTING (base2, "vblankstart_even",     &(pMode->manual_timings.vblankstart_even));
            GET_SETTING (base2, "vsyncstart_even",      &(pMode->manual_timings.vsyncstart_even));
            GET_SETTING (base2, "vsyncend_even",        &(pMode->manual_timings.vsyncend_even));
            GET_SETTING (base2, "vblankend_even",       &(pMode->manual_timings.vblankend_even));
            GET_SETTING (base2, "vtotal_even",          &(pMode->manual_timings.vtotal_even));
            GET_SETTING (base2, "frequency",            &(pMode->manual_timings.frequency));
                    
            // Get VOP settings
            //
            if (pMode->configure_vop)
            {
                base2 = base + "\\vop_settings";
                GET_SETTING (base2, "flags",              &(pMode->vop_config.flags));
                GET_SETTING (base2, "mode",               &(pMode->vop_config.mode));
                GET_SETTING (base2, "conversion_mode",    &(pMode->vop_config.conversion_mode));
                GET_SETTING (base2, "vsync_out",          &(pMode->vop_config.vsync_out));

                if (pMode->vop_config.mode == VOP_MODE_601)
                {
                    base2 += "\\vop601";
                    GET_SETTING (base2, "flags",              &(pMode->vop_config.vop601.flags));
                    GET_SETTING (base2, "vsync_shift",        &(pMode->vop_config.vop601.vsync_shift));
                    GET_SETTING (base2, "vsync_shift_count",  &(pMode->vop_config.vop601.vsync_shift_count));
                    GET_SETTING (base2, "output_mode",        &(pMode->vop_config.vop601.output_mode));
                }
            }

            if (coefficients)
            {
                unsigned long *pCoeff;

                // Allocate space for both horizontal and vertical coefficients
                //
                extraSizes[extraIndex] = 256 * 3 * sizeof(ULONG);
                extraArray[extraIndex] = (unsigned char *)GlobalAlloc (GMEM_FIXED, extraSizes[extraIndex]);
                if (!extraArray[extraIndex])
                {
                    printf ("Error allocating coefficient buffer.  Exiting...\n");
                    goto cleanUp;
                }
                
                pCoeff = (unsigned long *)extraArray[extraIndex++];

                // Set coefficient pointer to be simple offset into a continuous buffer.
                //
                pMode->hcoeff = (unsigned long *)(extraOffset);
                pMode->vcoeff = (unsigned long *)(extraOffset + 2048);
                extraOffset += 3072;

                // Read horizontal coefficients
                //
                base2 = base + "\\custom_hcoeff";
                for (x = 0; x < 256; x++)
                {
                    unsigned long x1, x2, x3, x4, x5;

                    setting.Format ("hcoeff_%i_0", x);
                    GET_SETTING (base2, setting, &x1);
                    if ((long)x1 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("hcoeff_%i_1", x);
                    GET_SETTING (base2, setting, &x2);
                    if ((long)x2 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("hcoeff_%i_2", x);
                    GET_SETTING (base2, setting, &x3);
                    if ((long)x3 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("hcoeff_%i_3", x);
                    GET_SETTING (base2, setting, &x4);
                    if ((long)x4 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("hcoeff_%i_4", x);
                    GET_SETTING (base2, setting, &x5);
                    if ((long)x5 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    if ((x1 + x2 + x3 + x4 + x5) != 512)
                        printf ("Warning: line %i - Coefficients do not add up to 512\n", g_line);

                    pCoeff[x << 1] =  (x1 & 0x3FF)        |
                                    ((x2 & 0x3FF) << 10) |
                                    ((x3 & 0x3FF) << 20);

		            pCoeff[(x << 1) + 1] = (x4 & 0x3FF) | ((x5 & 0x3FF) << 10);
                }

                // Read vertical coefficients
                //
                pCoeff += 512;
                base2 = base + "\\custom_vcoeff";
                for (x = 0; x < 256; x++)
                {
                    unsigned long x1, x2, x3;

                    setting.Format ("vcoeff_%i_0", x);
                    GET_SETTING (base2, setting, &x1);
                    if (x1 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("vcoeff_%i_1", x);
                    GET_SETTING (base2, setting, &x2);
                    if (x2 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("vcoeff_%i_2", x);
                    GET_SETTING (base2, setting, &x3);
                    if (x3 >= 512)
                    {
                        printf ("ERROR: line %i - maximum coefficient size is 511\n", g_line);
                        goto cleanUp;
                    }

                    if ((x1 + x2 + x3) != 512)
                        printf ("Warning: line %i - Coefficients do not add up to 512\n", g_line);

                    pCoeff[x] =  (x1 & 0x3FF)        |
                                ((x2 & 0x3FF) << 10) |
                                ((x3 & 0x3FF) << 20);
                }
            }
            if (pMode->premode_smbus_count)
            {
                SMBUS_WRITE *pSMBus;

                // Allocate space for both horizontal and vertical coefficients
                //
                extraSizes[extraIndex] = (pMode->premode_smbus_count) * sizeof(SMBUS_WRITE);
                extraArray[extraIndex] = (unsigned char *)GlobalAlloc (GMEM_FIXED, extraSizes[extraIndex]);
                if (!extraArray[extraIndex])
                {
                    printf ("Error allocating Pre-mode SMBus buffer.  Exiting...\n");
                    goto cleanUp;
                }
                
                pSMBus = (SMBUS_WRITE *)extraArray[extraIndex++];

                // Set smbus pointer to be simple offset into a continuous buffer.
                //
                pMode->premode_smbus = (SMBUS_WRITE *)(extraOffset);
                extraOffset += sizeof (SMBUS_WRITE) * (pMode->premode_smbus_count);

                base2 = base + "\\premode_smbus";
                for (x = 0; x < (int)pMode->premode_smbus_count; x++)
                {
                    setting.Format ("base_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].base));
                    if (pSMBus[x].base > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus base must be less than 256\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("addr_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].addr));
                    if (pSMBus[x].addr > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus address must be less than 256\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("data_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].data));
                    if (pSMBus[x].data > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus data must be less than 256\n", g_line);
                        goto cleanUp;
                    }
                }
            }
            if (pMode->postmode_smbus_count)
            {
                SMBUS_WRITE *pSMBus;

                // Allocate space for both horizontal and vertical coefficients
                //
                extraSizes[extraIndex] = (pMode->postmode_smbus_count) * sizeof(SMBUS_WRITE);
                extraArray[extraIndex] = (unsigned char *)GlobalAlloc (GMEM_FIXED, extraSizes[extraIndex]);
                if (!extraArray[extraIndex])
                {
                    printf ("Error allocating Post-mode SMBus buffer.  Exiting...\n");
                    goto cleanUp;
                }

                pSMBus = (SMBUS_WRITE *)extraArray[extraIndex++];

                // Set smbus pointer to be simple offset into a continuous buffer.
                //
                pMode->postmode_smbus = (SMBUS_WRITE *)(extraOffset);
                extraOffset += sizeof (SMBUS_WRITE) * (pMode->postmode_smbus_count);

                base2 = base + "\\postmode_smbus";
                for (x = 0; x < (int)pMode->postmode_smbus_count; x++)
                {
                    setting.Format ("base_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].base));
                    if (pSMBus[x].base > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus base must be less than 256\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("addr_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].addr));
                    if (pSMBus[x].addr > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus address must be less than 256\n", g_line);
                        goto cleanUp;
                    }

                    setting.Format ("data_%i", x);
                    GET_SETTING (base2, setting, &(pSMBus[x].data));
                    if (pSMBus[x].data > 0xFF)
                    {
                        printf ("ERROR: line %i - SMBus data must be less than 256\n", g_line);
                        goto cleanUp;
                    }
                }
            }
        }

        // All is well, write the file.  Start with a signature (LXXP_BIN).
        //
        unsigned long signature = LX_MODE_SIGNATURE1;
        output.Write (&signature, sizeof(ULONG));
        signature = LX_MODE_SIGNATURE2;
        output.Write (&signature, sizeof(ULONG));

        // Write the mode count
        //
        output.Write (&count, sizeof(int));

        size = 12;

        // Write the mode data
        //
        output.Write (outBuffer, sizeof(GEODELX_CUSTOM_MODE) * count);
        size += sizeof(GEODELX_CUSTOM_MODE) * count;

        // Write the extra data
        //
        for (i = 0; i < extraIndex; i++)
        {
            output.Write (extraArray[i], extraSizes[i]);
            size += extraSizes[i];
        }

        // Now write the installation .inf
        //
        WriteInfFile (inf, filename, disk_id, bin_file, size);
        inf.Close();

    cleanUp:

        output.Close();
        if (outBuffer)
            GlobalFree (outBuffer);
        if (extraArray)
            GlobalFree (extraArray);
        
	    return 0;
    }

    return 0;
}

int NextToken (CString& input, CString& token)
{
    int next;
    int length;

    if (input.GetLength() == 0)
    {
        printf ("Invalid File! - File ended abruptly on line %i", g_line);
        return 0;
    }

    next = 0;
    length = input.GetLength();

    // TRIM LEADING WHITESPACE
    // The TrimLeft CString routine is not used because we want to detect the end
    // of each line to track the current line number.
    //
    while (input[next] == '\n' || input[next] == '\t' || input[next] == ' ' || input[next] == '\r')
    {
        if (input[next] == '\n')
            g_line++;

        next++;
        if (next == length)
        {
            printf ("Invalid File! - File ended abruptly on line %i", g_line);
            return 0;
        }
    }

    input = input.Right (length - next);

    if (input.GetAt (0) == '(' || input.GetAt(0) == ')' ||
        input.GetAt (0) == '|')
    {
        token = input.GetAt(0);
        input = input.Right (input.GetLength() - 1);
        return 1;
    }

    next = input.FindOneOf (" \t\r\n()|");

    if (next < 0)
    {
        // Last token
        //
        token = input;
        input = "";
    }
    else
    {
        token = input.Left (next);
        input = input.Right (input.GetLength() - next);
    }

    return 1;
}

void WriteInfFile (CFile& inf_file, CString& filename, CString& disk_id, 
    CString& bin_file, ULONG fileSize)
{
    CString oneLine;
    CTime curTime = CTime::GetCurrentTime();

    oneLine.Format ("; %s\r\n", filename.GetBuffer(1));
    inf_file.Write (oneLine.GetBuffer(1), oneLine.GetLength());
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Copyright 2005, Advanced Micro Devices\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Custom Installation inf for the Advanced Micro Devices GeodeLX graphics adapter.\r\n"));
    WRITE_LINE ((";\r\n\r\n"));
    
    WRITE_LINE (("[Version]\r\n"));
    WRITE_LINE (("Signature=\"$WINDOWS NT$\"\r\n"));
    WRITE_LINE (("Provider=%cAMD%c\r\n", '%', '%'));
    WRITE_LINE (("ClassGUID={4D36E968-E325-11CE-BFC1-08002BE10318}\r\n"));
    WRITE_LINE (("Class=Display\r\n"));
    WRITE_LINE (("DriverVer=%2.2i/%2.2i/%4.4i, 3.03.02\r\n", curTime.GetMonth(), curTime.GetDay(), curTime.GetYear()));
    WRITE_LINE (("CatalogFile=lx_win.cat\r\n\r\n"));

    WRITE_LINE (("[DestinationDirs]\r\n"));
    WRITE_LINE (("DefaultDestDir   = 11\r\n"));
    WRITE_LINE (("lx.Miniport  = 12  ; drivers\r\n"));
    WRITE_LINE (("lx.Display   = 11  ; system32\r\n"));
    WRITE_LINE (("lx.Custom    = 11  ; system32\r\n\r\n"));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Driver information\r\n"));
    WRITE_LINE ((";\r\n\r\n"));

    WRITE_LINE (("[Manufacturer]\r\n"));
    WRITE_LINE (("%cAMD%c   = lx.Mfg\r\n\r\n", '%', '%'));

    WRITE_LINE (("[lx.Mfg]\r\n"));
    WRITE_LINE (("%s = LX, PCI\\VEN_1022&DEV_2081\r\n\r\n", disk_id.GetBuffer(1)));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; General installation section\r\n"));
    WRITE_LINE ((";\r\n\r\n"));

    WRITE_LINE (("[LX]\r\n"));
    WRITE_LINE (("CopyFiles=lx.Miniport, lx.Display , lx.Appls, lx.Custom\r\n\r\n"));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; File sections\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("[lx.Miniport]\r\n"));
    WRITE_LINE (("lxmini.sys\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Display]\r\n"));
    WRITE_LINE (("geodelx.dll\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Appls]\r\n"));
    WRITE_LINE (("LXDPApp.dll\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Custom]\r\n"));
    WRITE_LINE (("%s\r\n", bin_file.GetBuffer(1)));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Service Installation\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Services]\r\n"));
    WRITE_LINE (("AddService = LX, 0x00000002, LX_Service_Inst, LX_EventLog_Inst\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_Service_Inst]\r\n"));
    WRITE_LINE (("ServiceType    = 1                  ; SERVICE_KERNEL_DRIVER\r\n"));
    WRITE_LINE (("StartType      = 1                  ; SERVICE_SYSTEM_START\r\n"));
    WRITE_LINE (("ErrorControl   = 0                  ; SERVICE_ERROR_IGNORE\r\n"));
    WRITE_LINE (("LoadOrderGroup = Video\r\n"));
    WRITE_LINE (("ServiceBinary  = %c12%c\\lxmini.sys\r\n", '%', '%'));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_EventLog_Inst]\r\n"));
    WRITE_LINE (("AddReg = LX_EventLog_AddReg\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_EventLog_AddReg]\r\n"));
    WRITE_LINE (("HKR,,EventMessageFile,0x00020000,\"%cSystemRoot%c\\System32\\IoLogMsg.dll;%cSystemRoot%c\\System32\\drivers\\lxmini.sys\"\r\n", '%', '%', '%', '%'));
    WRITE_LINE (("HKR,,TypesSupported,0x00010001,7\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Software Installation\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.SoftwareSettings]\r\n"));
    WRITE_LINE (("AddReg = LX_SoftwareDeviceSettings\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_SoftwareDeviceSettings]\r\n"));
    WRITE_LINE (("HKR,, InstalledDisplayDrivers,     %cREG_MULTI_SZ%c, GeodeLX\r\n", '%', '%'));
    WRITE_LINE (("HKR,, VgaCompatible,               %cREG_DWORD%c,    0\r\n", '%', '%'));
    WRITE_LINE (("HKR,, MultiFunctionSupported,      %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, Compression,                 %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, FBLinear,                    %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, DDCEnabled,                  %cREG_DWORD%c,    0\r\n", '%', '%'));
    WRITE_LINE (("HKR,, CustomFile,                  %cREG_SZ%c,       \"\\??\\C:\\windows\\system32\\%s\"\r\n", '%', '%', bin_file.GetBuffer(1)));
    WRITE_LINE (("HKR,, CustomFileSize,              %cREG_DWORD%c,    %i\r\n", '%', '%', fileSize));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("HKCR,\"CLSID\\{2F5AC606-70CF-461C-BFE1-6063670C3484}\\InProcServer32\",,,\"LXDPApp.dll\"\r\n"));
    WRITE_LINE (("HKCR,\"CLSID\\{2F5AC606-70CF-461C-BFE1-6063670C3484}\\InProcServer32\",ThreadingModel,,\"Apartment\"\r\n"));
    WRITE_LINE (("HKLM,\"Software\\Microsoft\\Windows\\CurrentVersion\\Controls Folder\\Device\\shellex\\PropertySheetHandlers\\GeodeDisplayPage\",,,\"{2F5AC606-70CF-461C-BFE1-6063670C3484}\"\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Source file information\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[SourceDisksNames.x86]\r\n"));
    WRITE_LINE (("1 = %cDiskId%c,,,\"\"\r\n", '%', '%'));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[SourceDisksFiles]\r\n"));
    WRITE_LINE (("lxmini.sys  = 1\r\n"));
    WRITE_LINE (("geodelx.dll = 1\r\n"));
    WRITE_LINE (("LXDPApp.dll = 1\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[Strings]\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Non-Localizable Strings\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("REG_SZ         = 0x00000000\r\n"));
    WRITE_LINE (("REG_MULTI_SZ   = 0x00010000\r\n"));
    WRITE_LINE (("REG_EXPAND_SZ  = 0x00020000\r\n"));
    WRITE_LINE (("REG_BINARY     = 0x00000001\r\n"));
    WRITE_LINE (("REG_DWORD      = 0x00010001\r\n"));
    WRITE_LINE (("SERVICEROOT    = \"System\\CurrentControlSet\\Services\"\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Localizable Strings\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("DiskId       = \"%s\"\r\n", disk_id.GetBuffer(1)));
    WRITE_LINE (("GraphAdap    = \"Graphics Adapter\"\r\n"));
    WRITE_LINE (("AMD          = \"Advanced Micro Devices\"\r\n"));
}
void DisplayUsage (void)
{
    printf ("Usage: lxcustom filename\n");
    printf ("   Process a binary customization file or a registry settings file\n");
    printf ("   and converts it to its inverse.  The type of the file is\n");
    printf ("   determined automatically by the utility.  When processing a\n");
    printf ("   binary file, the registry text is output to standard out.  When\n");
    printf ("   processing a registry text file, the output binary filename is\n");
    printf ("   specified inside the registry file.\n");
}

long ExtractValue (CString *pString)
{
    int index;
    
    index = FindDefine (pString->GetBuffer(1));
    if (index != -1)
        return ValidDefines[index].value;

    if (pString->GetLength() >= 2 && 
        (pString->Left(2) == "0x" || pString->Left(2) == "0X"))
	{
        return (long)(strtoul (pString->GetBuffer(1), NULL, 16));
	}
	else
        return strtol (pString->GetBuffer(1), NULL, 10);
}

int FindDefine (char *var_string)
{
	int high, low, current;
    int compare;

    // As the array is sorted, we can perform a binary search of the array.

    low     = 0;
    high    = g_NumDefines - 1;
    current = (high + low) / 2;

    while (1)
    {
        compare = strcmp (var_string, ValidDefines[current].define_string);

        // A Match ?

        if (compare == 0)
            return current;

        // Reached the end of the search

        else if ((high - low) == 1)
        {
            compare = strcmp (var_string, ValidDefines[high].define_string);
            if (compare == 0)
                return high;
            else
                return -1;
        }
        else
        {
            if (compare < 0)
                high = current;
            else
                low = current;

            current = (high + low) / 2;
        }
    }
}