//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Console application to generate a panel customization file and associated
// .inf install file.
// </DOC_AMD_STD>
//

#include <afx.h>
#include "lxpanel.h"
#include "..\display\lxshared.h"

int g_line = 1;
int g_extraResolutions = 0;
int g_allocatedResolutions = 0;
int simultaneousCRT = 0;
unsigned long *g_resolutionBuffer = NULL;

int main(int argc, char* argv[])
{
    unsigned long length;
    unsigned char *outBuffer;
    unsigned long size;
    CString bin_file;
    CString filename;
    CString filetext;
    CString token1, token2;
    CString disk_id;
    CFile input, output, inf;
    char *filePtr;
            
    if (argc != 2)
    {
        DisplayUsage();
        return 0;
    }

    filename = argv[1];

    input.Open (filename, CFile::modeRead, NULL); 
    if (input.m_hFile == CFile::hFileNull)
    {
        printf ("Unable to open input file: %s\n", filename);
        return 0;
    }

    // Buffer the entire file contents in a CString.  Don't forget
    // the NULL termination.
    //
    length = (long)input.GetLength();
    filePtr = filetext.GetBuffer (length + 1);
    input.Read (filePtr, length);
    filePtr[length] = 0;

    // Check if the first four digits are a binary signature.
    //
    if (*((unsigned long *)filePtr) == LX_MODE_SIGNATURE1 &&
        *((unsigned long *)filePtr + 1) == LX_MODE_SIGNATURE2)
    {
        fprintf (stderr, "Binary Input File: %s\n", filename.GetBuffer (1));

        // Binary file - Reverse engineer the contents and output to standard out.
        //
        filePtr += 8;

        printf ("binary_file %s\n", filename.GetBuffer(1));
        printf ("inf_file custom.inf\n");
        printf ("disk_id = \"Reverse Engineered Registry File\"\n");
        printf ("show_crt_modes %i\n", *((unsigned long *)filePtr));
        printf ("simultaneous_CRT %i\n\n", *((unsigned long *)filePtr + 1));
        
        VG_DISPLAY_MODE *pMode = (VG_DISPLAY_MODE *)(filePtr + 8);
        filePtr += 8;

        printf ("flags                  0x%X\n", pMode->flags);
        printf ("panel_width            0x%X\n", pMode->panel_width);
        printf ("panel_height           0x%X\n", pMode->panel_height);
        printf ("panel_tim1             0x%X\n", pMode->panel_tim1);
        printf ("panel_tim2             0x%X\n", pMode->panel_tim2);
        printf ("panel_dither_ctl       0x%X\n", pMode->panel_dither_ctl);
        printf ("panel_pad_sel_low      0x%X\n", pMode->panel_pad_sel_low);
        printf ("panel_pad_sel_high     0x%X\n", pMode->panel_pad_sel_high);
        printf ("hactive                0x%X\n", pMode->hactive);
        printf ("hblankstart            0x%X\n", pMode->hblankstart);
        printf ("hsyncstart             0x%X\n", pMode->hsyncstart);
        printf ("hsyncend               0x%X\n", pMode->hsyncend);
        printf ("hblankend              0x%X\n", pMode->hblankend);
        printf ("htotal                 0x%X\n", pMode->htotal);
        printf ("vactive                0x%X\n", pMode->vactive);
        printf ("vblankstart            0x%X\n", pMode->vblankstart);
        printf ("vsyncstart             0x%X\n", pMode->vsyncstart);
        printf ("vsyncend               0x%X\n", pMode->vsyncend);
        printf ("vblankend              0x%X\n", pMode->vblankend);
        printf ("vtotal                 0x%X\n", pMode->vtotal);
        printf ("frequency              0x%X\n\n", pMode->frequency);

        filePtr += sizeof(VG_DISPLAY_MODE);
        while (length >= (sizeof(VG_DISPLAY_MODE) + 24))
        {
            printf ("extra_width            0x%X\n", *((ULONG *)filePtr));
            printf ("extra_height           0x%X\n", *((ULONG *)filePtr + 1));
            length -= 8;
            filePtr += 8;
        }
    }
    else
    {
        filetext.ReleaseBuffer();

        // Close the input file as it is no longer required.
        //
        input.Close();

        // The first line of the input file indicates the name of the output file
        //
        if (!NextToken (filetext, token1)) return 0;
        if (!NextToken (filetext, token2)) return 0;

        if (token1 != "binary_file")
        {
            printf ("Invalid file syntax!  First line should be \'binary_file xxxx\'\n");
            return 0;
        }

        bin_file = token2;
        output.Open (bin_file, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive, NULL); 
        if (output.m_hFile == CFile::hFileNull)
        {
            printf ("Unable to open output file: %s\n", bin_file.GetBuffer (1));
            return 0;
        }

        printf ("Binary File specified as %s\n", bin_file.GetBuffer(1));

        // The second line specifies the .inf filename
        //
        if (!NextToken (filetext, token1)) return 0;
        if (!NextToken (filetext, token2)) return 0;

        if (token1 != "inf_file")
        {
            printf ("Invalid file syntax!  Second line should be \'inf_file xxxx\'\n");
            return 0;
        }

        filename = token2;
        inf.Open (filename, CFile::modeCreate | CFile::modeWrite | CFile::shareExclusive, NULL); 
        if (inf.m_hFile == CFile::hFileNull)
        {
            printf ("Unable to open output file: %s\n", filename.GetBuffer (1));
            return 0;
        }

        printf (".Inf File specified as %s\n", filename.GetBuffer(1));

        // The third line specifies the install text
        //
        if (!NextToken (filetext, token1)) return 0;
        {
            if (token1 != "disk_id")
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }

            // Extract the id string
            //
            int start = filetext.Find ('"');
            if (start < 0)
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }
            start++;
            int last = filetext.Find ('"', start + 1);
            if (last < 0)
            {
                printf ("Invalid File Syntax!  Third line should be \'disk_id \"xxxx\"\'\n");
                return 0;
            }

            filetext = filetext.Right (filetext.GetLength() - start);
            disk_id = filetext.Left (last - start);
            filetext = filetext.Right (filetext.GetLength() - (last - start));

            printf ("Disk ID = \"%s\"\n", disk_id.GetBuffer(1));
        }
        
        if (!NextToken (filetext, token2)) return 0;
        
        // Allocate enough space for a single custom mode.
        //
        outBuffer  = (unsigned char *) GlobalAlloc (GMEM_FIXED, (sizeof(VG_DISPLAY_MODE) + 4));
        memset (outBuffer,  0, sizeof(VG_DISPLAY_MODE) + 4);
        
        if (!outBuffer)
        {
            printf ("Unable to allocate memory.  Exiting...\n");
            goto cleanUp;
        }

        VG_DISPLAY_MODE *panelMode;
        panelMode = (VG_DISPLAY_MODE *)(outBuffer + 8);
        
        GET_SETTING ("show_crt_modes",       ((unsigned long *)outBuffer));
        GET_SETTING ("simultaneous_CRT",     ((unsigned long *)(outBuffer + 4)));
        GET_SETTING ("flags",                &(panelMode->flags));
        GET_SETTING ("panel_width",          &(panelMode->panel_width));
        GET_SETTING ("panel_height",         &(panelMode->panel_height));
        GET_SETTING ("panel_tim1",           &(panelMode->panel_tim1));
        GET_SETTING ("panel_tim2",           &(panelMode->panel_tim2));
        GET_SETTING ("panel_dither_ctl",     &(panelMode->panel_dither_ctl));
        GET_SETTING ("panel_pad_sel_low",    &(panelMode->panel_pad_sel_low));
        GET_SETTING ("panel_pad_sel_high",   &(panelMode->panel_pad_sel_high));
        GET_SETTING ("hactive",              &(panelMode->hactive));
        GET_SETTING ("hblankstart",          &(panelMode->hblankstart));
        GET_SETTING ("hsyncstart",           &(panelMode->hsyncstart));
        GET_SETTING ("hsyncend",             &(panelMode->hsyncend));
        GET_SETTING ("hblankend",            &(panelMode->hblankend));
        GET_SETTING ("htotal",               &(panelMode->htotal));
        GET_SETTING ("vactive",              &(panelMode->vactive));
        GET_SETTING ("vblankstart",          &(panelMode->vblankstart));
        GET_SETTING ("vsyncstart",           &(panelMode->vsyncstart));
        GET_SETTING ("vsyncend",             &(panelMode->vsyncend));
        GET_SETTING ("vblankend",            &(panelMode->vblankend));
        GET_SETTING ("vtotal",               &(panelMode->vtotal));
        GET_SETTING ("frequency",            &(panelMode->frequency));
        
        // Program simultaneous CRT in the registry instead of in the
        // binary file.
        //
        simultaneousCRT = *((unsigned long *)(outBuffer + 4));
        
        // Basic error checking
        //
        if (panelMode->hactive != panelMode->panel_width)
            printf ("WARNING: Panel width does not match hactive.\n");

        // Look for user-specified "extra" resolutions
        //
        while (OptionalEntry (filetext))
            ;

        // All is well, write the file.  Start with a signature (LXXP_BIN).
        //
        unsigned long signature = LX_MODE_SIGNATURE1;
        output.Write (&signature, sizeof(ULONG));
        signature = LX_MODE_SIGNATURE2;
        output.Write (&signature, sizeof(ULONG));

        size = 16 + sizeof(VG_DISPLAY_MODE);

        // Write the panel data
        //
        output.Write (outBuffer, size - 8);

        size += g_extraResolutions << 3;
        output.Write (g_resolutionBuffer, g_extraResolutions << 3);
        
        // Now write the installation .inf
        //
        WriteInfFile (inf, filename, disk_id, bin_file, size);
        inf.Close();

    cleanUp:

        output.Close();
        if (outBuffer)
            GlobalFree (outBuffer);
                
	    return 0;
    }

    return 0;
}

int NextToken (CString& input, CString& token, bool quiet)
{
    int next;
    int length;

    if (input.GetLength() == 0)
    {
        if (!quiet)
            printf ("Invalid File! - File ended abruptly on line %i", g_line);
        return 0;
    }

    next = 0;
    length = input.GetLength();

    // TRIM LEADING WHITESPACE
    // The TrimLeft CString routine is not used because we want to detect the end
    // of each line to track the current line number.
    //
    while (input[next] == '\n' || input[next] == '\t' || input[next] == ' ' || input[next] == '\r')
    {
        if (input[next] == '\n')
            g_line++;

        next++;
        if (next == length)
        {
            if (!quiet)
                printf ("Invalid File! - File ended abruptly on line %i", g_line);
            return 0;
        }
    }

    input = input.Right (length - next);
    if (input.GetAt (0) == '(' || input.GetAt(0) == ')' ||
        input.GetAt (0) == '|')
    {
        token = input.GetAt(0);
        input = input.Right (input.GetLength() - 1);
        return 1;
    }

    next = input.FindOneOf (" \t\r\n()|");

    if (next < 0)
    {
        // Last token
        //
        token = input;
        input = "";
    }
    else
    {
        token = input.Left (next);
        input = input.Right (input.GetLength() - next);
    }

    return 1;
}

void WriteInfFile (CFile& inf_file, CString& filename, CString& disk_id, 
    CString& bin_file, ULONG fileSize)
{
    CString oneLine;
    CTime curTime = CTime::GetCurrentTime();

    oneLine.Format ("; %s\r\n", filename.GetBuffer(1));
    inf_file.Write (oneLine.GetBuffer(1), oneLine.GetLength());
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Copyright 2005, Advanced Micro Devices\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Custom Panel Installation inf for the Advanced Micro Devices GeodeLX graphics adapter.\r\n"));
    WRITE_LINE ((";\r\n\r\n"));
    
    WRITE_LINE (("[Version]\r\n"));
    WRITE_LINE (("Signature=\"$WINDOWS NT$\"\r\n"));
    WRITE_LINE (("Provider=%cAMD%c\r\n", '%', '%'));
    WRITE_LINE (("ClassGUID={4D36E968-E325-11CE-BFC1-08002BE10318}\r\n"));
    WRITE_LINE (("Class=Display\r\n"));
    WRITE_LINE (("DriverVer=%2.2i/%2.2i/%4.4i, 3.03.02\r\n", curTime.GetMonth(), curTime.GetDay(), curTime.GetYear()));
    WRITE_LINE (("CatalogFile=lx_win.cat\r\n\r\n"));

    WRITE_LINE (("[DestinationDirs]\r\n"));
    WRITE_LINE (("DefaultDestDir   = 11\r\n"));
    WRITE_LINE (("lx.Miniport  = 12  ; drivers\r\n"));
    WRITE_LINE (("lx.Display   = 11  ; system32\r\n"));
    WRITE_LINE (("lx.Custom    = 11  ; system32\r\n\r\n"));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Driver information\r\n"));
    WRITE_LINE ((";\r\n\r\n"));

    WRITE_LINE (("[Manufacturer]\r\n"));
    WRITE_LINE (("%cAMD%c   = lx.Mfg\r\n\r\n", '%', '%'));

    WRITE_LINE (("[lx.Mfg]\r\n"));
    WRITE_LINE (("%s = LX, PCI\\VEN_1022&DEV_2081\r\n\r\n", disk_id.GetBuffer(1)));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; General installation section\r\n"));
    WRITE_LINE ((";\r\n\r\n"));

    WRITE_LINE (("[LX]\r\n"));
    WRITE_LINE (("CopyFiles=lx.Miniport, lx.Display , lx.Appls, lx.Custom\r\n\r\n"));

    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; File sections\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("[lx.Miniport]\r\n"));
    WRITE_LINE (("lxmini.sys\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Display]\r\n"));
    WRITE_LINE (("geodelx.dll\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Appls]\r\n"));
    WRITE_LINE (("LXDPApp.dll\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Custom]\r\n"));
    WRITE_LINE (("%s\r\n", bin_file.GetBuffer(1)));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Service Installation\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.Services]\r\n"));
    WRITE_LINE (("AddService = LX, 0x00000002, LX_Service_Inst, LX_EventLog_Inst\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_Service_Inst]\r\n"));
    WRITE_LINE (("ServiceType    = 1                  ; SERVICE_KERNEL_DRIVER\r\n"));
    WRITE_LINE (("StartType      = 1                  ; SERVICE_SYSTEM_START\r\n"));
    WRITE_LINE (("ErrorControl   = 0                  ; SERVICE_ERROR_IGNORE\r\n"));
    WRITE_LINE (("LoadOrderGroup = Video\r\n"));
    WRITE_LINE (("ServiceBinary  = %c12%c\\lxmini.sys\r\n", '%', '%'));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_EventLog_Inst]\r\n"));
    WRITE_LINE (("AddReg = LX_EventLog_AddReg\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_EventLog_AddReg]\r\n"));
    WRITE_LINE (("HKR,,EventMessageFile,0x00020000,\"%cSystemRoot%c\\System32\\IoLogMsg.dll;%cSystemRoot%c\\System32\\drivers\\lxmini.sys\"\r\n", '%', '%', '%', '%'));
    WRITE_LINE (("HKR,,TypesSupported,0x00010001,7\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Software Installation\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[lx.SoftwareSettings]\r\n"));
    WRITE_LINE (("AddReg = LX_SoftwareDeviceSettings\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[LX_SoftwareDeviceSettings]\r\n"));
    WRITE_LINE (("HKR,, InstalledDisplayDrivers,     %cREG_MULTI_SZ%c, GeodeLX\r\n", '%', '%'));
    WRITE_LINE (("HKR,, VgaCompatible,               %cREG_DWORD%c,    0\r\n", '%', '%'));
    WRITE_LINE (("HKR,, MultiFunctionSupported,      %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, Compression,                 %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, FBLinear,                    %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, DDCEnabled,                  %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, DDC_GPIOData,                %cREG_DWORD%c,    4\r\n", '%', '%'));
    WRITE_LINE (("HKR,, DDC_GPIOClock,               %cREG_DWORD%c,    3\r\n", '%', '%'));
    WRITE_LINE (("HKR,, EnablePanelScale,            %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, CustomFileSize,              %cREG_DWORD%c,    0\r\n", '%', '%'));
    WRITE_LINE (("HKR,, PanelFile,                   %cREG_SZ%c,       \"\\??\\C:\\windows\\system32\\%s\"\r\n", '%', '%', bin_file.GetBuffer(1)));
    WRITE_LINE (("HKR,, PanelFileSize,               %cREG_DWORD%c,    %i\r\n", '%', '%', fileSize));
    WRITE_LINE (("HKR,, PanelEnable,                 %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, BIOSDetectOutput,            %cREG_DWORD%c,    0\r\n", '%', '%'));
    WRITE_LINE (("HKR,, PanelInSystem,               %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, EnableNon60HzPanel,          %cREG_DWORD%c,    1\r\n", '%', '%'));
    WRITE_LINE (("HKR,, SimultaneousCRT,             %cREG_DWORD%c,    %i\r\n", '%', '%', simultaneousCRT));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("HKCR,\"CLSID\\{2F5AC606-70CF-461C-BFE1-6063670C3484}\\InProcServer32\",,,\"LXDPApp.dll\"\r\n"));
    WRITE_LINE (("HKCR,\"CLSID\\{2F5AC606-70CF-461C-BFE1-6063670C3484}\\InProcServer32\",ThreadingModel,,\"Apartment\"\r\n"));
    WRITE_LINE (("HKLM,\"Software\\Microsoft\\Windows\\CurrentVersion\\Controls Folder\\Device\\shellex\\PropertySheetHandlers\\GeodeDisplayPage\",,,\"{2F5AC606-70CF-461C-BFE1-6063670C3484}\"\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Source file information\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[SourceDisksNames.x86]\r\n"));
    WRITE_LINE (("1 = %cDiskId%c,,,\"\"\r\n", '%', '%'));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[SourceDisksFiles]\r\n"));
    WRITE_LINE (("lxmini.sys  = 1\r\n"));
    WRITE_LINE (("geodelx.dll = 1\r\n"));
    WRITE_LINE (("LXDPApp.dll = 1\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("[Strings]\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Non-Localizable Strings\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("REG_SZ         = 0x00000000\r\n"));
    WRITE_LINE (("REG_MULTI_SZ   = 0x00010000\r\n"));
    WRITE_LINE (("REG_EXPAND_SZ  = 0x00020000\r\n"));
    WRITE_LINE (("REG_BINARY     = 0x00000001\r\n"));
    WRITE_LINE (("REG_DWORD      = 0x00010001\r\n"));
    WRITE_LINE (("SERVICEROOT    = \"System\\CurrentControlSet\\Services\"\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("; Localizable Strings\r\n"));
    WRITE_LINE ((";\r\n"));
    WRITE_LINE (("\r\n"));
    WRITE_LINE (("DiskId       = \"%s\"\r\n", disk_id.GetBuffer(1)));
    WRITE_LINE (("GraphAdap    = \"Graphics Adapter\"\r\n"));
    WRITE_LINE (("AMD          = \"Advanced Micro Devices\"\r\n"));
}
void DisplayUsage (void)
{
    printf ("Usage: lxpanel filename\n");
    printf ("   Process a binary panel customization file or a registry settings file\n");
    printf ("   and converts it to its inverse.  The type of the file is\n");
    printf ("   determined automatically by the utility.  When processing a\n");
    printf ("   binary file, the registry text is output to standard out.  When\n");
    printf ("   processing a registry text file, the output binary filename is\n");
    printf ("   specified inside the registry file.\n");
}

long ExtractValue (CString *pString)
{
    int index;
    
    index = FindDefine (pString->GetBuffer(1));
    if (index != -1)
        return ValidDefines[index].value;

    if (pString->GetLength() >= 2 && 
        (pString->Left(2) == "0x" || pString->Left(2) == "0X"))
	{
        return (long)(strtoul (pString->GetBuffer(1), NULL, 16));
	}
	else
        return strtol (pString->GetBuffer(1), NULL, 10);
}

int FindDefine (char *var_string)
{
	int high, low, current;
    int compare;

    // As the array is sorted, we can perform a binary search of the array.

    low     = 0;
    high    = g_NumDefines - 1;
    current = (high + low) / 2;

    while (1)
    {
        compare = strcmp (var_string, ValidDefines[current].define_string);

        // A Match ?

        if (compare == 0)
            return current;

        // Reached the end of the search

        else if ((high - low) == 1)
        {
            compare = strcmp (var_string, ValidDefines[high].define_string);
            if (compare == 0)
                return high;
            else
                return -1;
        }
        else
        {
            if (compare < 0)
                high = current;
            else
                low = current;

            current = (high + low) / 2;
        }
    }
}

bool OptionalEntry (CString &filetext)
{
    unsigned long width;
    unsigned long height;
    CString token;
    
    // Format for an extra resolution is:
    //  extra_width  number
    //  extra_height number
    //
    // This can be repeated ad nauseum
    //
    if (!NextToken (filetext, token, true)) 
        return false;
    if (token != "extra_width")
        return false;
    if (!NextToken (filetext, token, true)) 
        return false;
    width = (unsigned long)ExtractValue (&token);

    if (!NextToken (filetext, token, true)) 
        return false;
    if (token != "extra_height")
        return false;
    if (!NextToken (filetext, token, true)) 
        return false;
    height = (unsigned long)ExtractValue (&token);

    g_extraResolutions++;
    if (g_extraResolutions > g_allocatedResolutions)
    {
        ULONG *newModes;
        g_allocatedResolutions += 10;
        
        newModes = (ULONG *)malloc (g_allocatedResolutions * 2 * sizeof(ULONG));
        if (!newModes)
        {
            printf ("Failed to allocate space for new modes!\n");
            return false;
        }
        for (int i = 0; i < (g_allocatedResolutions - 10) << 1; i += 2)
        {
            newModes[i] = g_resolutionBuffer[i];
            newModes[i + 1] = g_resolutionBuffer[i + 1];
        }

        if (g_resolutionBuffer)
            delete [] g_resolutionBuffer;

        g_resolutionBuffer = newModes;
    }
    g_resolutionBuffer[(g_extraResolutions - 1) << 1] = width;
    g_resolutionBuffer[((g_extraResolutions - 1) << 1) + 1] = height;

    return true;
}