//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// DXAPI support.
// </DOC_AMD_STD>
//

#include "lxdrv.h"
#define NT_INCLUDED
#include "windef.h"
#include "dxmini.h"
#include "dxapi.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"

extern LX_REGISTRY_SETTINGS g_regSettings;
extern ULONG lxCustomIndex;
DXAPI_INTERFACE LXDXAPI =
{
    sizeof (DXAPI_INTERFACE),
    0,
    NULL,
    NULL,
    NULL,
    LXGetIRQInfo,
    LXEnableIRQ,
    NULL,
    LXBobNextField,
    LXSetState,
    LXLock,
    LXFlipOverlay,
    LXFlipVideoPort,
    LXGetPolarity,
    LXGetCurrentAutoFlip,
    LXGetPreviousAutoFlip,
    NULL,
    NULL
};

unsigned long previousField;
unsigned long skipNextInterrupt;
unsigned long currentFlip;
unsigned long previousFlip;
unsigned long WaitingForField;
unsigned long nextOddY,  nextOddU,  nextOddV;
unsigned long nextEvenY, nextEvenU, nextEvenV;
unsigned long programming = 0;

/////////////////////////////////////////////////////////////////////////
// LXQueryInterface
//
ULONG LXQueryInterface (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PQUERY_INTERFACE        pQueryInterface)
{
    DXAPI_INTERFACE *dxapi;
    ULONG status;
    ULONG size;

    DISPDBG ((3000, "LXQueryInterface\n"));
    if (IsEqualGUID (pQueryInterface->InterfaceType, &GUID_DxApi))
    {
        if (pQueryInterface->Size == sizeof (DXAPI_INTERFACE))
        {
            DISPDBG ((3000, "  Querying DXAPI Interface\n"));

            if (HwDeviceExtension->vipHardware)
            {
                dxapi = (DXAPI_INTERFACE *)pQueryInterface->Interface;

                *dxapi = LXDXAPI;
                dxapi->Version = pQueryInterface->Version;
                dxapi->Context = HwDeviceExtension;
                status = NO_ERROR;
            }
            else
            {
                DISPDBG ((3000, "  VIP Hardware is absent.\n"));
                status = ERROR_INVALID_PARAMETER;
            }
        }
        else
        {
            DISPDBG ((3000, "  Invalid DXAPI Structure Size\n"));
            status = ERROR_INVALID_PARAMETER;
        }
    }
    else
    {
        DISPDBG ((3000, "  Unsupported Interface\n"));
        status = ERROR_INVALID_PARAMETER;
    }

    return status;
}

/////////////////////////////////////////////////////////////////////////
// LXInterrupt
//
BOOLEAN LXInterrupt (PHW_DEVICE_EXTENSION HwDeviceExtension)
{
    VOID (*pIRQCallback)(PDX_IRQDATA) = HwDeviceExtension->IRQCallback;
    PDX_IRQDATA irqData               = HwDeviceExtension->IRQData;
    VG_INTERRUPT_PARAMS intParams;
    BOOLEAN ret = FALSE;
    DWORD   vipState;
    ULONG   dwIrqFlags;
    ULONG   currentField;
    dwIrqFlags = 0;

    if (!HwDeviceExtension->vipHardware)
    {
        // This should never happen...
        //
        return FALSE;
    }

    // VG Vsync interrupt
    //
    if (vg_test_and_clear_interrupt())
    {
        if (HwDeviceExtension->InterlacedOut)
        {
            // Only update during the even VG field.
            //
            if (vg_test_even_field() && nextOddY && nextEvenY && !programming)
            {
                df_set_video_offsets (0, nextOddY,  nextOddU,  nextOddV);
                df_set_video_offsets (1, nextEvenY, nextEvenU, nextEvenV);
            }
        }
        else
        {
            ret = TRUE;
            dwIrqFlags |= DDIRQ_DISPLAY_VSYNC;
        }
    }

    // VIP Interrupts are enabled
    //
    vipState = vip_get_interrupt_state();

    // VIP Error Flags
    //
    if (vipState & (VIP_INT_CLOCK_INPUT | VIP_INT_ACTIVE_PIXELS | VIP_INT_LONGLINE |
        VIP_INT_FIFO_WRAP | VIP_INT_VERTICAL_TIMING))
    {
        DISPDBG ((10000, "VIP Error - 0x%8.8lX\n", vipState));

        if (vipState & VIP_INT_CLOCK_INPUT)
        {
            // CLOCK HAS DISAPPEARED
            // When the VIP clock goes, VIP reset cannot reset everything.  We will thus
            // disable all error interrupts and wait for the next start of field.  We will
            // also disable capture.
            //
            vip_set_interrupt_enable (
                VIP_INT_CLOCK_INPUT     |
                VIP_INT_FIFO_WRAP       |
                VIP_INT_ACTIVE_PIXELS   |
                VIP_INT_VERTICAL_TIMING |
                VIP_INT_LONGLINE, 0);

            // Set capture state to off.
            //
            vip_set_capture_state (VIP_STOPCAPTURE);

            WaitingForField = 1;
        }
        else
        {
            // An error occured.  When VIP writes past its max address, it signals an
            // interrupt and stops capture.  All other errors simply fire an interrupt.
            // We want to generically reset the VIP for all errors.  We also must
            // restart capture when a max address interrupt fires.  Conveniently, Cimarron
            // will reset the VIP whenever starting capture, so we can collapse both
            // problems into one function call.  Any lost fields due to the VIP reset will
            // shake out inside the DPC routines.
            //
            if (HwDeviceExtension->interlaced)
                vip_set_capture_state (VIP_STARTCAPTUREATNEXTFIELD);
            else
                vip_set_capture_state (VIP_STARTCAPTUREATNEXTFRAME);
        }
    }

    else if (vipState & (VIP_INT_START_EVEN | VIP_INT_START_ODD | VIP_INT_END_VBLANK))
    {
        // Re-enable capture if it was previously disabled due to a
        // lost VIP clock.  Reset the VIP as previous VIP resets may have been
        // lost.
        //
        if (WaitingForField)
        {
            WaitingForField = 0;

            // Re-enable VIP errors
            //
            vip_reset_interrupt_state (VIP_ALL_INTERRUPTS);
            vip_set_interrupt_enable (
                VIP_INT_CLOCK_INPUT     |
                VIP_INT_FIFO_WRAP       |
                VIP_INT_ACTIVE_PIXELS   |
                VIP_INT_VERTICAL_TIMING |
                VIP_INT_LONGLINE, 1);

            // Re-enable VIP capture.  Note that updating the capture state will guarantee
            // a VIP reset.
            //
            if (HwDeviceExtension->interlaced)
                vip_set_capture_state (VIP_STARTCAPTUREATNEXTFIELD);
            else
                vip_set_capture_state (VIP_STARTCAPTUREATNEXTFRAME);
        }
        else
        {
            // Check for missed fields.  The DxApi DPC can get confused if we drop
            // fields.  The handler does not query the current field polarity, but
            // rather auto-updates an internal software variable.  If we skip a field,
            // we could start flipping on the wrong field.  Note that DxApi can still
            // get confused if system latency causes the execution of the DPC to be delayed.
            // In this case, the previous DPC might not have been executed before
            // the next is scheduled.  The scheduler will thus ignore the second DPC and
            // DxApi will miss a field.  Note that this only applies when in interlaced
            // modes.
            //
            if (HwDeviceExtension->interlaced)
            {
                currentField = vip_get_current_field();
                if (previousField != currentField && !skipNextInterrupt)
                {
                    dwIrqFlags |= DDIRQ_VPORT0_VSYNC;
                }

                skipNextInterrupt = 0;
                previousField = currentField;
            }
            else
            {
                // Unconditionally set the interrupt flag when
                // running progressive.
                //
                dwIrqFlags |= DDIRQ_VPORT0_VSYNC;
            }
        }
    }

    if (vipState & VIP_INT_LINE_MATCH_TARGET)
        dwIrqFlags |= DDIRQ_VPORT0_LINE;

    if (vipState)
    {
        ret = TRUE;
        vip_reset_interrupt_state (vipState);
    }

    if (HwDeviceExtension->IRQFlags && dwIrqFlags)
    {
        irqData->dwIrqFlags = dwIrqFlags;
        pIRQCallback (irqData);
    }

    return ret;
}

/////////////////////////////////////////////////////////////////////////
// LXGetIRQInfo
//
DWORD LXGetIRQInfo (
    PVOID           HwDeviceExtension,
    PVOID           lpInput,
    PDDGETIRQINFO   GetIRQInfo)
{
    DISPDBG ((200, "LXGetIRQInfo\n"));

    GetIRQInfo->dwFlags = IRQINFO_HANDLED;

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXEnableIRQ
//
DWORD LXEnableIRQ (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PDDENABLEIRQINFO        EnableIRQInfo,
    PVOID                   lpOutput)
{
    VG_INTERRUPT_PARAMS intParams;
    unsigned long enablemask, disablemask;

    DISPDBG ((200, "LXEnableIRQ\n"));

    // Check for invalid caps
    //
    if (EnableIRQInfo->dwIRQSources & ~(DDIRQ_DISPLAY_VSYNC | DDIRQ_VPORT0_LINE | DDIRQ_VPORT0_VSYNC))
    {
        DISPDBG ((200, "  Invalid IRQ Flags!\n"));
        return DXERR_UNSUPPORTED;
    }

    HwDeviceExtension->IRQFlags    = EnableIRQInfo->dwIRQSources;
    HwDeviceExtension->IRQCallback = EnableIRQInfo->IRQCallback;
    HwDeviceExtension->IRQData     = EnableIRQInfo->lpIRQData;

    // Display VSync interrupt
    //
    intParams.flags = 0;
    if (EnableIRQInfo->dwIRQSources & DDIRQ_DISPLAY_VSYNC)
    {
        if (g_regSettings.ManualTimingEnable)
            intParams.line = HwDeviceExtension->customModes[lxCustomIndex].manual_timings.vactive - 1;
        else if (g_regSettings.CustomPanelTimings)
            intParams.line = HwDeviceExtension->panelTimings->vactive - 1;
        else if (g_regSettings.PanelEnable)
            intParams.line = min (g_regSettings.PanelHeight - 1, HwDeviceExtension->ScreenHeight - 1);
        else
            intParams.line = HwDeviceExtension->ScreenHeight - 1;

        intParams.enable = 1;
    }
    else
    {
        intParams.line   = 0;
        intParams.enable = 0;
    }

    vg_configure_line_interrupt (&intParams);

    // VIP line interrupt
    //
    enablemask = disablemask = 0;
    if (EnableIRQInfo->dwIRQSources & DDIRQ_VPORT0_LINE)
        enablemask |= VIP_INT_LINE_MATCH_TARGET;
    else
        disablemask |= VIP_INT_LINE_MATCH_TARGET;

    // VIP VSync interrupt
    //
    if (EnableIRQInfo->dwIRQSources & DDIRQ_VPORT0_VSYNC)
    {
        previousField     = VIP_ODD_FIELD;
        skipNextInterrupt = 0;
        if (HwDeviceExtension->interlaced)
            enablemask |= VIP_INT_START_EVEN | VIP_INT_START_ODD;
        else
            enablemask |= VIP_INT_END_VBLANK;

        WaitingForField   = 0;

        // ENABLE VG VSYNC
        // When displaying interlaced input on an interlaced display,
        // we run into problems when the VIP and VOP refresh rates are
        // not identical.  To address this issue, we enable a VG interrupt
        // and flip the video offsets based on the VG interrupt and the
        // VIP offsets based on the VIP interrupt.  This doubles the
        // interrupt traffic but is the only way to guarantee jitter and
        // tearing-free VIP playback without genlock.
        //
        if (HwDeviceExtension->InterlacedOut)
        {
            intParams.flags = 0;
            intParams.line = 5;
            intParams.enable = 1;

            vg_configure_line_interrupt (&intParams);
        }
    }
    else
        disablemask |= VIP_INT_START_EVEN | VIP_INT_START_ODD | VIP_INT_END_VBLANK;

    // Enable VIP errors
    //
    if (EnableIRQInfo->dwIRQSources & (DDIRQ_VPORT0_VSYNC | DDIRQ_VPORT0_LINE))
    {
        enablemask |= VIP_INT_CLOCK_INPUT | VIP_INT_ACTIVE_PIXELS | VIP_INT_LONGLINE |
            VIP_INT_FIFO_WRAP | VIP_INT_VERTICAL_TIMING;
    }
    else
    {
        disablemask |= VIP_INT_CLOCK_INPUT | VIP_INT_ACTIVE_PIXELS | VIP_INT_LONGLINE |
            VIP_INT_FIFO_WRAP | VIP_INT_VERTICAL_TIMING;
    }

    nextOddY  = nextOddU  = nextOddV  = 0;
    nextEvenY = nextEvenU = nextEvenV = 0;

    // Enable interrupts all at once.
    //
    vip_set_interrupt_enable (disablemask, 0);
    vip_set_interrupt_enable (enablemask,  1);

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXBobNextField
//
DWORD LXBobNextField (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PDDBOBNEXTFIELDINFO     BobNextFieldInfo,
    PVOID                   lpOutput)
{
    if (HwDeviceExtension->InterlacedOut)
    {
        if (vip_get_current_field() != VIP_ODD_FIELD)
            skipNextInterrupt = 1;

        // NO WORK!!!???
        // This routine is called to switch to the even field inside an interleaved
        // surface.  However, we already programmed the odd and even field offsets at
        // the LxFlipOverlay call.  The only work to do here is to check that the VIP
        // is capturing the correct field.
        //
        return DX_OK;
    }
    else
    {
        // CHECK REASON FOR THIS CALL
        // This routine can be called if we are bobbing within one surface (interleaved).  It
        // can also be called if we are bobbing interleaved within a multi-flip chain.  For
        // the latter, we should only be called during the odd field of video port capture.
        //
        if (HwDeviceExtension->numAutoFlip == 1)
        {
            if (vip_get_current_field() == VIP_ODD_FIELD)
            {
                // Odd Field is being captured.  Switch to even field.
                //
                df_set_video_offsets (0,
                    BobNextFieldInfo->lpSurface->dwOverlayOffset + BobNextFieldInfo->lpSurface->lPitch,
                    BobNextFieldInfo->lpSurface->dwDriverReserved1 + (BobNextFieldInfo->lpSurface->lPitch >> 1),
                    BobNextFieldInfo->lpSurface->dwDriverReserved2 + (BobNextFieldInfo->lpSurface->lPitch >> 1));

                WRITE_VID32 (DF_VIDEO_Y_POS, HwDeviceExtension->videoY + 1);
            }
            else
            {
                // Even Field is on.  Switch to odd field.
                //
                df_set_video_offsets (0,
                    BobNextFieldInfo->lpSurface->dwOverlayOffset,
                    BobNextFieldInfo->lpSurface->dwDriverReserved1,
                    BobNextFieldInfo->lpSurface->dwDriverReserved2);

                WRITE_VID32 (DF_VIDEO_Y_POS, HwDeviceExtension->videoY);
            }
        }
        else
        {
            if (vip_get_current_field() != VIP_ODD_FIELD)
                skipNextInterrupt = 1;

            // Program DF to display even field within an interleaved surface.
            //
            df_set_video_offsets (0,
                BobNextFieldInfo->lpSurface->dwOverlayOffset + BobNextFieldInfo->lpSurface->lPitch,
                BobNextFieldInfo->lpSurface->dwDriverReserved1 + (BobNextFieldInfo->lpSurface->lPitch >> 1),
                BobNextFieldInfo->lpSurface->dwDriverReserved2 + (BobNextFieldInfo->lpSurface->lPitch >> 1));

            WRITE_VID32 (DF_VIDEO_Y_POS, HwDeviceExtension->videoY + 1);
        }
    }

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXSetState
//
DWORD LXSetState (
    PVOID               HwDeviceExtension,
    PDDSETSTATEININFO   SetStateInInfo,
    PDDSETSTATEOUTINFO  SetStateOutInfo)
{
    DISPDBG ((200, "LXSetState\n"));

    // This routine is called to transition between Bob and weave.  It is generally
    // used when hardware autoflipping (outside of an interrupt handler).  As we
    // are already using an IRQ based system (software flipping using DxApi), we
    // will simply always say that we need to fall back to software autoflipping
    // if this call is ever made (which it shouldn't be).
    //
    SetStateOutInfo->bSoftwareAutoflip = 1;
    SetStateOutInfo->dwSurfaceIndex    = FindAutoFlipIndex (HwDeviceExtension, currentFlip);
    SetStateOutInfo->dwVBISurfaceIndex = 0;
    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXLock
//
DWORD LXLock (
    PVOID               HwDeviceExtension,
    PDDLOCKININFO       LockInInfo,
    PDDLOCKOUTINFO      LockOutInfo)
{
    DISPDBG ((200, "LXLock\n"));

    // Feed a direct pointer
    //
    LockOutInfo->dwSurfacePtr = (ULONG_PTR)(cim_fb_ptr + LockInInfo->lpSurfaceData->dwSurfaceOffset);
    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXFlipOverlay
//
DWORD LXFlipOverlay (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PDDFLIPOVERLAYINFO      FlipOverlayInfo,
    PVOID                   lpOutput)
{
    ULONG videoY = HwDeviceExtension->videoY + FlipOverlayInfo->lpTargetSurface->dwDriverReserved3;
    static unsigned long storedNextOddY = 0;
    static unsigned long storedNextOddU = 0;
    static unsigned long storedNextOddV = 0;

    if (HwDeviceExtension->InterlacedOut)
    {
        // EVEN FIELD HAS A NON-ZERO VALUE
        //
        if (FlipOverlayInfo->lpTargetSurface->dwDriverReserved3)
        {
            programming = 1;
            nextOddY  = storedNextOddY;
            nextOddU  = storedNextOddU;
            nextOddV  = storedNextOddV;
            nextEvenY = FlipOverlayInfo->lpTargetSurface->dwOverlayOffset;
            nextEvenU = FlipOverlayInfo->lpTargetSurface->dwDriverReserved1;
            nextEvenV = FlipOverlayInfo->lpTargetSurface->dwDriverReserved2;
            programming = 0;
        }

        // ASSUME THE ODD FIELD IS BEING DISPLAYED IF NOT INTERLEAVED
        //
        else if (!(FlipOverlayInfo->lpTargetSurface->dwOverlayFlags & DDOVER_INTERLEAVED))
        {
            storedNextOddY = FlipOverlayInfo->lpTargetSurface->dwOverlayOffset;
            storedNextOddU = FlipOverlayInfo->lpTargetSurface->dwDriverReserved1;
            storedNextOddV = FlipOverlayInfo->lpTargetSurface->dwDriverReserved2;
        }

        // OTHERWISE ASSUME WEAVE
        // We are switching capture to a new frame.
        //
        else
        {
            ULONG vidPitch = FlipOverlayInfo->lpTargetSurface->lPitch;
            programming = 1;
            nextOddY = FlipOverlayInfo->lpTargetSurface->dwOverlayOffset;
            nextOddU = FlipOverlayInfo->lpTargetSurface->dwDriverReserved1;
            nextOddV = FlipOverlayInfo->lpTargetSurface->dwDriverReserved2;

            nextEvenY = FlipOverlayInfo->lpTargetSurface->dwOverlayOffset + vidPitch;
            nextEvenU = FlipOverlayInfo->lpTargetSurface->dwDriverReserved1 + (vidPitch >> 1);
            nextEvenV = FlipOverlayInfo->lpTargetSurface->dwDriverReserved2 + (vidPitch >> 1);
            programming = 0;
        }
    }
    else
    {
        df_set_video_offsets (0,
            FlipOverlayInfo->lpTargetSurface->dwOverlayOffset,
            FlipOverlayInfo->lpTargetSurface->dwDriverReserved1,
            FlipOverlayInfo->lpTargetSurface->dwDriverReserved2);

        WRITE_VID32 (DF_VIDEO_Y_POS, videoY);
    }

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXFlipVideoPort
//
DWORD LXFlipVideoPort (
    PHW_DEVICE_EXTENSION  HwDeviceExtension,
    PDDFLIPVIDEOPORTINFO  FlipVideoPortInfo,
    PVOID                 lpOutput)
{
    VIPINPUTBUFFER vipBuffer;
    VIPINPUTBUFFER_ADDR *taskA;

    // Save the flip configuration
    //
    previousFlip = FlipVideoPortInfo->lpCurrentSurface->dwOverlayOffset;
    currentFlip  = FlipVideoPortInfo->lpTargetSurface->dwOverlayOffset;

    vipBuffer.flags = 0;
    vipBuffer.current_buffer = 0;

    taskA = &(vipBuffer.offsets[VIP_BUFFER_TASK_A]);

    if (FlipVideoPortInfo->lpTargetSurface->dwOverlayFlags & DDOVER_INTERLEAVED)
    {
        // The DxApi interrupt handler is simple in that it does not perform any
        // error recovery for dropped interrupts or for dropped DPCs.  We handle
        // dropped interrupts by a special case in the interrupt handler.  We handle
        // dropped DPCs by checking that all DxFlipVideoPort calls occur during the
        // even field and that all DxBobNextField calls occur during the odd field.
        // If any of these conditions are violated, we cause the next interrupt
        // to be ignored.  This allows us to catch up or to resync.
        //
        if (vip_get_current_field() != VIP_EVEN_FIELD)
            skipNextInterrupt = 1;

        // Set even offset to point to the second line.
        //
        if (HwDeviceExtension->interlaced)
        {
            taskA->odd_base[0] = FlipVideoPortInfo->lpTargetSurface->dwOverlayOffset
                + HwDeviceExtension->PhysLXFBAddr.LowPart;
            taskA->even_base[0] = taskA->odd_base[0] + FlipVideoPortInfo->lpTargetSurface->lPitch;
        }
        else
        {
            taskA->odd_base[0] = taskA->even_base[0] =
                FlipVideoPortInfo->lpTargetSurface->dwOverlayOffset + HwDeviceExtension->PhysLXFBAddr.LowPart;
        }

        // Clear reserved flag (in case of Bob to weave switch)
        //
        FlipVideoPortInfo->lpTargetSurface->dwDriverReserved3 = 0;
    }
    else
    {
        // Progressive or pure Bob deinterlacing.  When doing non-interleaved Bob,
        // the hardware is configured to toggle the video offsets each field,
        // instead of each frame.  Interleaved Bob goes through the code above
        // and the LxBobNextField routine.
        //
        taskA->odd_base[0] = taskA->even_base[0] =
            FlipVideoPortInfo->lpTargetSurface->dwOverlayOffset + HwDeviceExtension->PhysLXFBAddr.LowPart;

        // Set even or odd indicator
        //
        FlipVideoPortInfo->lpTargetSurface->dwDriverReserved3 = 0;
        if (HwDeviceExtension->interlaced && vip_get_current_field() == VIP_ODD_FIELD)
            FlipVideoPortInfo->lpTargetSurface->dwDriverReserved3 = 1;
    }

    vip_toggle_video_offsets (VIP_BUFFER_A, &vipBuffer);

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXGetPolarity
//
DWORD LXGetPolarity (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PDDGETPOLARITYININFO    GetPolarityInInfo,
    PDDGETPOLARITYOUTINFO   GetPolarityOutInfo)
{
    DISPDBG ((200, "LXGetPolarity\n"));

    GetPolarityOutInfo->bPolarity = (vip_get_current_field() == VIP_EVEN_FIELD);

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXGetCurrentAutoFlip
//
DWORD LXGetCurrentAutoFlip (
    PHW_DEVICE_EXTENSION            HwDeviceExtension,
    PDDGETCURRENTAUTOFLIPININFO     GetCurrentAutoFlipInInfo,
    PDDGETCURRENTAUTOFLIPOUTINFO    GetCurrentAutoFlipOutInfo)
{
    unsigned int i;

    DISPDBG ((200, "LXGetCurrentAutoFlip\n"));

    GetCurrentAutoFlipOutInfo->dwSurfaceIndex = FindAutoFlipIndex (HwDeviceExtension, currentFlip);

    // VBI is not yet supported.
    //
    GetCurrentAutoFlipOutInfo->dwVBISurfaceIndex = 0;
    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXGetPreviousAutoFlip
//
DWORD LXGetPreviousAutoFlip (
    PHW_DEVICE_EXTENSION            HwDeviceExtension,
    PDDGETPREVIOUSAUTOFLIPININFO    GetPreviousAutoFlipInInfo,
    PDDGETPREVIOUSAUTOFLIPOUTINFO   GetPreviousAutoFlipOutInfo)
{
    unsigned int i;

    DISPDBG ((200, "LXGetPreviousAutoFlip\n"));

    GetPreviousAutoFlipOutInfo->dwSurfaceIndex = FindAutoFlipIndex (HwDeviceExtension, previousFlip);

    // VBI is not yet supported.
    //
    GetPreviousAutoFlipOutInfo->dwVBISurfaceIndex = 0;

    return DX_OK;
}

/////////////////////////////////////////////////////////////////////////
// LXGetPreviousAutoFlip
//
// Helper routine to find the autoflip index of an overlay offset based on
// the current autoflip chain.
//
ULONG FindAutoFlipIndex (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    ULONG                   OverlayOffset)
{
    ULONG i;

    // Find the current surface in the flip chain.
    //
    for (i = 0; i < HwDeviceExtension->numAutoFlip; i++)
    {
        if (HwDeviceExtension->autoFlipSurfaces[i] == OverlayOffset)
        {
            i++;
            break;
        }
    }

    // Set to last surface if no match is found.  Otherwise use the proper surface.
    //
    return i - 1;
}


