//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Miniport main routines.
// </DOC_AMD_STD>
//

#define WIN2K_DPMS 1

#include "lxdrv.h"
#include "debug.h"
#include "cim_rtns.h"
#include "lxshared.h"
#include "escape.h"
#include "cim_defs.h"
#include "cim_regs.h"

// CHILD DEVICE DEFINITIONS

#define P2_DDC_MONITOR        (0x1357bd00)
#define P2_NONDDC_MONITOR     (0x1357bd01)

#define DRIVER_FLAG_LINEAR_MODE 0x80000000l
#define DRIVER_FLAG_COMPRESSION 0x40000000l

extern void cim_outw (unsigned short port, unsigned short address);
extern unsigned short cim_inw (unsigned short port);
extern int swprintf(wchar_t *str, wchar_t *format, ...);

ULONG *fb_ptr       = NULL;
ULONG heapStart     = 0;
ULONG heapEnd       = 0;
ULONG lxModeIndex   = 0;
ULONG lxCustomIndex = 0;
int from_hibernate  = 0;

char g_ansiBuffer[200];
LX_REGISTRY_SETTINGS g_regSettings;

VIDEO_ACCESS_RANGE LXLegacyResourceList[] =
{
    {0x000C0000, 0x00000000, 0x00010000, 0, 0, 0, 0}, // ROM location
    {0x000A0000, 0x00000000, 0x00020000, 0, 0, 1, 0}, // Frame buffer
    {0x000003B0, 0x00000000, 0x0000000B, 1, 1, 1, 0}, // VGA regs
    {0x000003C0, 0x00000000, 0x00000020, 1, 1, 1, 0}  // VGA regs
};
ULONG LXLegacyResourceEntries = sizeof LXLegacyResourceList / sizeof LXLegacyResourceList[0];

/////////////////////////////////////////////////////////////////////////
// DriverEntry
//
// This is the first miniport routine called by the kernel.
// We aren't supposed to understand Context1 and Context2. we just pass
// them to VideoPortInitialize.
//
// Routine Description:
//
//  Installable driver initialization entry point.
//  This entry point is called directly by the I/O system.
//
// Arguments:
//
//  Context1 - First context value passed by the operating system. This is
//             the value with which the miniport driver calls VideoPortInitialize().
//
//  Context2 - Second context value passed by the operating system. This is
//             the value with which the miniport driver calls VideoPortInitialize().
//
// Return Value:
//
//  Status from VideoPortInitialize()
//
ULONG DriverEntry(PVOID Context1, PVOID Context2)
{
	VIDEO_HW_INITIALIZATION_DATA hwInitData;
	ULONG initStatus;
    ULONG revision, companion;
    int cimStatus;

	DISPDBG((1, "\n\n\nStarting Driver\nDriverEntry ----------\n"));

	VideoPortZeroMemory(&hwInitData, sizeof(VIDEO_HW_INITIALIZATION_DATA));

	initStatus = ERROR_DEV_NOT_EXIST;

    // INITIALIZE CIMARRON
    // We do this here in DriverEntry instead of later in LxFindAdapter because
    // we unhook the driver ISR if VIP is not present in the display device.
    // Unhooking the ISR instead of simply changing the ISR implementation
    // improves system performance and is generally safer.
    //
    cimStatus = init_detect_cpu (&revision, &companion);
	if (cimStatus == CIM_STATUS_CPUNOTFOUND || cimStatus == CIM_STATUS_DISPLAYUNAVAILABLE)
    {
        DISPDBG((99999, "HW Detection Failed!!!!\n"));
    }

	// INITIALIZE THE STRUCTURE
	// At this stage, we fill in an initialization structure with important information,
	// such as pointers to other routines that can be called.

	hwInitData.HwInitDataSize            = sizeof(VIDEO_HW_INITIALIZATION_DATA);
	hwInitData.AdapterInterfaceType      = PCIBus;
    hwInitData.HwFindAdapter             = LXFindAdapter;
	hwInitData.HwInitialize              = LXInitialize;
    hwInitData.HwStartIO                 = LXStartIO;
    hwInitData.HwDeviceExtensionSize     = sizeof(HW_DEVICE_EXTENSION);
    hwInitData.HwResetHw                 = NULL;
    hwInitData.HwTimer                   = NULL;
    hwInitData.HwStartDma                = NULL;
    hwInitData.HwSetPowerState           = LXSetPowerState;
	hwInitData.HwGetPowerState           = LXGetPowerState;
    hwInitData.HwGetVideoChildDescriptor = LXGetChildDescriptor;
    hwInitData.HwQueryInterface          = LXQueryInterface;
    hwInitData.HwLegacyResourceList      = LXLegacyResourceList;
    hwInitData.HwLegacyResourceCount     = LXLegacyResourceEntries;

    if (cimStatus != CIM_STATUS_NOVIPHARDWARE)
        hwInitData.HwInterrupt = LXInterrupt;
    
	initStatus = VideoPortInitialize(Context1, Context2, &hwInitData, NULL);

	DISPDBG((1, "DriverEntry: status from VideoPortInitialize %d\n", initStatus));
	DISPDBG((1, "DriverEntry returning...\n"));

	return initStatus;

}// DriverEntry

/////////////////////////////////////////////////////////////////////////
// CustomFileCallback
//
// This routine is called by the videoport driver to parse a driver
// customization file.
//
VP_STATUS CustomFileCallback(
    PHW_DEVICE_EXTENSION HwDeviceExtension,
    PVOID Context,
    PWSTR ValueName,
    PVOID ValueData,
    ULONG ValueLength)
{
    if (!Context)
        return ERROR_INVALID_PARAMETER;

    // The file size should match the input data length
    //
    if (ValueLength != *((ULONG *)Context))
    {
        DISPDBG ((3000, "ERROR!!!!!! Length Mismatch in Customization File\n"));
        DISPDBG ((3000, "    --- Expected %i, Received %i\n", *((ULONG *)Context), ValueLength));
    }

    // The first two ULONGs of the input data should be a binary signature.
    //
    if ((*((unsigned long *)ValueData)     != LX_MODE_SIGNATURE1) ||
        (*((unsigned long *)ValueData + 1) != LX_MODE_SIGNATURE2))
    {
        DISPDBG ((9999, "Invalid signature in customization file.\n"));
        return ERROR_INVALID_PARAMETER;
    }

    // The third dword is the number of custom modes included in the file.
    //
    HwDeviceExtension->customModeCount = *((unsigned long *)ValueData + 2);

    ValueData = (PVOID)((ULONG)ValueData + 12);

    // Copy the data manually.  We are assuming (with good reason) that the
    // buffer is always DWORD aligned.
    //
    _asm {
        cld
        mov ecx, ValueLength
        sub ecx, 12
        shr ecx, 2
        mov edi, Context
        mov esi, ValueData
        rep movsd
    }

    return NO_ERROR;
}

/////////////////////////////////////////////////////////////////////////
// CustomPanelCallback
//
// This routine is called by the videoport driver to parse a driver
// customization file.
//
VP_STATUS CustomPanelCallback(
    PHW_DEVICE_EXTENSION HwDeviceExtension,
    PVOID Context,
    PWSTR ValueName,
    PVOID ValueData,
    ULONG ValueLength)
{
    ULONG resOffset;
    ULONG *resolutions;
    ULONG vgSize = sizeof(VG_DISPLAY_MODE);

    if (!Context)
        return ERROR_INVALID_PARAMETER;

    // The file size should match the input data length
    //
    if (ValueLength != *((ULONG *)Context))
    {
        DISPDBG ((3000, "ERROR!!!!!! Length Mismatch in Panel File\n"));
        DISPDBG ((3000, "    --- Expected %i, Received %i\n", *((ULONG *)Context), ValueLength));
        return ERROR_INVALID_PARAMETER;
    }

    // The first two ULONGs of the input data should be a binary signature.
    //
    if ((*((unsigned long *)ValueData)     != LX_MODE_SIGNATURE1) ||
        (*((unsigned long *)ValueData + 1) != LX_MODE_SIGNATURE2))
    {
        DISPDBG ((9999, "Invalid signature in panel file.\n"));
        return ERROR_INVALID_PARAMETER;
    }

    // The third dword is 'Show CRT Modes'.
    //
    HwDeviceExtension->showCRTModes = *((unsigned long *)ValueData + 2);

    // The fourth dword used to be Simutaneous CRT, but is now ignored.
    //
    ValueData = (PVOID)((ULONG)ValueData + 16);

    // The next part of the buffer is the display mode timings.  Note that we are 
    // assuming (with good reason) that the buffer is always DWORD aligned.
    //
    _asm {
        cld
        mov ecx, vgSize
        shr ecx, 2
        mov edi, Context
        mov esi, ValueData
        rep movsd
    }

    // The last part of the buffer is one or more optional resolutions to add
    // when displaying on a custom TFT.  Note that we only pull out of the buffer 
    // when 8 bytes or more are specified.
    //
    resOffset = 0;
    resolutions = (ULONG *)((ULONG)ValueData + sizeof(VG_DISPLAY_MODE));
    while (ValueLength >= (sizeof(VG_DISPLAY_MODE) + 24))
    {
        *(HwDeviceExtension->extraTFTResolutions + resOffset++) = *resolutions++;
        *(HwDeviceExtension->extraTFTResolutions + resOffset++) = *resolutions++;
        ValueLength -= 8;
    }
    return NO_ERROR;
}

/////////////////////////////////////////////////////////////////////////
// GeodeRegistryCallback
//
// This routine is called by the videoport driver to parse a value extracted from
// the registry.  The address of this routine is supplied when requesting an extraction.
//
VP_STATUS GeodeRegistryCallback(
    PVOID HwDeviceExtension,
    PVOID Context,
    PWSTR ValueName,
    PVOID ValueData,
    ULONG ValueLength)
{
    if (!Context)
        return ERROR_INVALID_PARAMETER;

    // The first ULONG of context should contain the expected size.
    //
    if (ValueLength != *((ULONG *)Context))
    {
        DISPDBG ((3000, "ERROR!!!!!! Length Mismatch in Registry Key\n"));
        DISPDBG ((3000, "    --- Expected %i, Received %i\n", *((ULONG *)Context), ValueLength));
        return ERROR_INVALID_PARAMETER;
    }

    if (ValueLength > 4)
    {
        ULONG *pContext = (ULONG *)Context;
        ULONG *pValue   = (ULONG *)ValueData;
        UCHAR *pContextByte;
        UCHAR *pValueByte;

        // Fill in by dword increments.
        //
        while (ValueLength > 4)
        {
            *pContext++ = *pValue++;
            ValueLength -= 4;
        }

        // Fill in by byte increments.
        //
        pContextByte = (UCHAR *)pContext;
        pValueByte   = (UCHAR *)pValue;
        while (ValueLength-- )
            *pContextByte++ = *pValueByte++;

        return NO_ERROR;
    }
    else if (ValueLength)
    {
        if (Context)
        {
            *(ULONG *)Context = *(PULONG)ValueData;
        }
        else if (*((PULONG)ValueData) != 0)
        {
            return ERROR_INVALID_PARAMETER;
        }

        return NO_ERROR;
    }
	else
    {
        return ERROR_INVALID_PARAMETER;
    }
}

/////////////////////////////////////////////////////////////////////////
// LXFindAdapter
//
// Make sure we're running a LX and that video memory has been allocated.
//
// Routine Description:
//
//  This routine is called to determine if the adapter for this driver
//  is present in the system.
//  If it is present, the function fills out some information describing
//  the adapter.
//
// Arguments:
//
//  HwDeviceExtension - Supplies the miniport driver's adapter storage. This
//                      storage is initialized to zero before this call.
//
//  HwContext         - Supplies the context value which was passed to
//                      VideoPortInitialize().
//
//  ArgumentString    - Supplies a NULL terminated ASCII string. This string
//                      originates from the user.
//
//  ConfigInfo        - Returns the configuration information structure which is
//                      filled by the miniport driver. This structure is initialized with
//                      any knwon configuration information (such as SystemIoBusNumber) by
//                      the port driver. Where possible, drivers should have one set of
//                      defaults which do not require any supplied configuration information.
//
//  Again             - Indicates if the miniport driver wants the port driver to call
//                      its VIDEO_HW_FIND_ADAPTER function again with a new device extension
//                      and the same config info. This is used by the miniport drivers which
//                      can search for several adapters on a bus.
//
// Return Value:
//
//  This routine must return:
//
//    NO_ERROR                - Indicates a host adapter was found and the
//                              configuration information was successfully determined.
//
//    ERROR_INVALID_PARAMETER - Indicates an adapter was found but there was an
//                              error obtaining the configuration information. If possible an error
//                              should be logged.
//
//    ERROR_DEV_NOT_EXIST     - Indicates no host adapter was found for the
//                              supplied configuration information.
//
VP_STATUS LXFindAdapter(
	PHW_DEVICE_EXTENSION HwDeviceExtension,
	PVOID HwContext,
	PWSTR ArgumentString,
	PVIDEO_PORT_CONFIG_INFO ConfigInfo,
	PUCHAR Again)
{
    INIT_BASE_ADDRESSES baseAddresses;
	ULONG i, ulValue;

	DISPDBG((1, "LXFindAdapter: enter\n"));

	// CLEAR RECURSIVE FLAG
	// We don't want to be called again

	*Again = 0;

    // CLAIM RESOURCES
    // This is very important to do, as it claims any device resources.  If we
    // leave these resources unclaimed, they will not be connected by the PnP manager.
    // In essence, we will lose our IRQ capabilities.
    //
    LXClaimResources (HwDeviceExtension);

    // SET THE CURRENT POWER STATE
    // This prevents us from inadvertently restoring the registers on a
    // failed standby.
    //
    HwDeviceExtension->PreviousPowerState   = HwDeviceExtension->CurrPowerState = VideoPowerOn;
    HwDeviceExtension->PreviousMonitorState = VideoPowerOn;

	// Different for each architecture.  Note that we do not simply use
    // the information available to us from Plug and Play, because the
    // mapping between the BARs and the devices is abstracted inside
    // Cimarron.
    //
    init_read_base_addresses (&baseAddresses);
    HwDeviceExtension->vipHardware = (baseAddresses.vip_register_base != 0) &&
        (baseAddresses.vip_register_base != 0xFFFFFFFF);

	// CLEAR OUT EMULATOR ENTRIES AND STATE SIZE
	// This driver does not support them, as it is not 'VGA-compatible'.

	ConfigInfo->NumEmulatorAccessEntries     = 0;
	ConfigInfo->EmulatorAccessEntries        = NULL;
	ConfigInfo->EmulatorAccessEntriesContext = 0;

	ConfigInfo->HardwareStateSize = 0;

	ConfigInfo->VdmPhysicalVideoMemoryAddress.LowPart  = 0x00000000;
	ConfigInfo->VdmPhysicalVideoMemoryAddress.HighPart = 0x00000000;
	ConfigInfo->VdmPhysicalVideoMemoryLength           = 0x00000000;

	// MAP ADDRESSES IN DEVICE MEMORY
    //
	if (LXMapPhysicalAddress (HwDeviceExtension, baseAddresses.vg_register_base, 0x4000,  FALSE, (PULONG)&cim_vg_ptr))
		return ERROR_DEV_NOT_EXIST;

	if (LXMapPhysicalAddress (HwDeviceExtension, baseAddresses.df_register_base, 0x4000,  FALSE, (PULONG)&cim_vid_ptr))
		return ERROR_DEV_NOT_EXIST;

	if (LXMapPhysicalAddress (HwDeviceExtension, baseAddresses.framebuffer_base,
        baseAddresses.framebuffer_size, FALSE, (PULONG)&cim_fb_ptr))
		return ERROR_DEV_NOT_EXIST;

	if (LXMapPhysicalAddress (HwDeviceExtension, baseAddresses.gp_register_base,  0x4000,  FALSE, (PULONG)&cim_gp_ptr))
		return ERROR_DEV_NOT_EXIST;

    if (HwDeviceExtension->vipHardware)
    {
        if (LXMapPhysicalAddress (HwDeviceExtension, baseAddresses.vip_register_base, 0x4000,  FALSE, (PULONG)&cim_vip_ptr))
		    return ERROR_DEV_NOT_EXIST;
    }
    else
    {
        // NULL out the VIP pointer.  The VIP macros benignly ignore VIP accesses
        // when the pointer is NULL.  NULLing it out here eliminates the possibility
        // of an unintialized variable causing a blue screen.
        //
        cim_vip_ptr = NULL;
    }

    cim_cmd_base_ptr = cim_fb_ptr + baseAddresses.framebuffer_size - LX_CMD_BUFFER_SIZE;

	// STORE PHYSICAL ADDRESSES IN DEVICE MEMORY FOR LATER
    //
	HwDeviceExtension->PhysLXRegAddr.LowPart  = baseAddresses.vg_register_base;
	HwDeviceExtension->PhysLXRegAddr.HighPart = 0;
	HwDeviceExtension->PhysLXVidAddr.LowPart  = baseAddresses.df_register_base;
	HwDeviceExtension->PhysLXVidAddr.HighPart = 0;
    HwDeviceExtension->PhysLXVIPAddr.LowPart  = baseAddresses.vip_register_base;
	HwDeviceExtension->PhysLXVIPAddr.HighPart = 0;
	HwDeviceExtension->PhysLXFBAddr.LowPart   = baseAddresses.framebuffer_base;
	HwDeviceExtension->PhysLXFBAddr.HighPart  = 0;
	HwDeviceExtension->PhysLXGPAddr.LowPart   = baseAddresses.gp_register_base;
	HwDeviceExtension->PhysLXGPAddr.HighPart  = 0;
	HwDeviceExtension->FBSize                 = baseAddresses.framebuffer_size;

	// SET REGISTRY PARAMETERS
    //
	VideoPortSetRegistryParameters(HwDeviceExtension,
								   L"HardwareInformation.ChipType",
								   LX_CHIP_TYPE,
								   sizeof(LX_CHIP_TYPE));

	VideoPortSetRegistryParameters(HwDeviceExtension,
								   L"HardwareInformation.DacType",
								   LX_DAC,
								   sizeof(LX_DAC));

	VideoPortSetRegistryParameters(HwDeviceExtension,
								   L"HardwareInformation.MemorySize",
								   &HwDeviceExtension->FBSize,
								   sizeof(ULONG));

	VideoPortSetRegistryParameters(HwDeviceExtension,
								   L"HardwareInformation.AdapterString",
								   LX_ADAPTER,
								   sizeof(LX_ADAPTER));

	// RETRIEVE USER (OR APPLICATION) SUPPLIED PARAMETERS
    //
    RetrieveRegistryInformation (HwDeviceExtension, &g_regSettings);

    // DETECT PANEL SETTINGS
    //
    UpdateRegistrySettings(HwDeviceExtension, 1);

    // CREATE THE MODE TABLE
    // We dynamically create the mode table based on the available Cimarron modes.
    //
    if (!GenerateModeTable (HwDeviceExtension, 0))
    {
        DISPDBG ((3000, "Unable to allocate mode table!\n"));
        return ERROR_DEV_NOT_EXIST;
    }

	// FILTER VIDEO MODES THAT ARE NOT SUPPORTED BY THE CURRENT HARDWARE CONFIGURATION
    //
	if (!FilterUnsupportedModes (HwDeviceExtension, 0))
		return ERROR_DEV_NOT_EXIST;

	// INDICATE SUCCESSFUL COMPLETION
	//
	return NO_ERROR;

} // LXFindAdapter

/////////////////////////////////////////////////////////////////////////
// GenerateModeTable
//
// Dynamically generate a table based on the current Cimarron modes.  'fromResume'
// indicates that we are regenerating the table from a resume, in which case
// we MUST offer the last programmed mode as the OS will be calling DrvAssertMode
// in the display DLL with the last programmed PDEV.
//
BOOLEAN GenerateModeTable (PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResume)
{
    VG_DISPLAY_MODE displayMode;
    VG_DISPLAY_MODE *pMode;
    int count, modeCount, modesFilled;
    int i, resStart, resStop;
    ULONG currentWidth  = 0;
    ULONG currentHeight = 0;
    ULONG mask;
    ULONG valid;
    ULONG hz;

    HwDeviceExtension->addedExtraModeFromResume = FALSE;
    if (g_regSettings.ManualTimingEnable)
    {        
        modeCount = 0;
        for (i = 0; i < (int)HwDeviceExtension->customModeCount; i++)
        {
            // COUNT THE SUPPORTED MODES
            // Note that 24 and 32BPP are mutually exclusive, as there
            // is no way to distinguish between the two when reporting modes
            // to the OS.  If both are enabled, 32BPP will be chosen.
            //
            mask = HwDeviceExtension->customModes[i].bpp_mask;
            if (mask & 0x1) modeCount++;
            if (mask & 0x2) modeCount++;
            if (mask & 0xC) modeCount++;
        }

        // NO MODES???!!!!
        //
        if (modeCount == 0)
            return FALSE;

        HwDeviceExtension->modeCount = modeCount;
        LxAllocateMemory (HwDeviceExtension, sizeof(VIDEO_MODE_INFORMATION) * modeCount,
            &HwDeviceExtension->pModes);
        if (!HwDeviceExtension->pModes)
            return FALSE;

        // Fill in all Windows mode structures.
        //
        modeCount = 0;
        for (i = 0; i < (int)HwDeviceExtension->customModeCount; i++)
        {
            GEODELX_CUSTOM_MODE *pMode = &(HwDeviceExtension->customModes[i]);
            mask = pMode->bpp_mask;
            
            hz = CalculateRefresh (&pMode->manual_timings);

            // 8BPP
            //
            if (mask & 0x1)
            {
                HwDeviceExtension->pModes[modeCount].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modeCount].ModeIndex                    = (i << 16) | modeCount;
                HwDeviceExtension->pModes[modeCount].VisScreenWidth               = pMode->manual_timings.src_width;
                HwDeviceExtension->pModes[modeCount].VisScreenHeight              = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].ScreenStride                 = POWER_OF2_STRIDE(pMode->manual_timings.src_width);
                HwDeviceExtension->pModes[modeCount].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modeCount].BitsPerPlane                 = 8;
                HwDeviceExtension->pModes[modeCount].XMillimeter                  = pMode->manual_timings.src_width >> 1;
                HwDeviceExtension->pModes[modeCount].YMillimeter                  = pMode->manual_timings.src_height >> 1;
                HwDeviceExtension->pModes[modeCount].NumberRedBits                = 8;
                HwDeviceExtension->pModes[modeCount].NumberGreenBits              = 8;
                HwDeviceExtension->pModes[modeCount].NumberBlueBits               = 8;
                HwDeviceExtension->pModes[modeCount].RedMask                      = 0;
                HwDeviceExtension->pModes[modeCount].GreenMask                    = 0;
                HwDeviceExtension->pModes[modeCount].BlueMask                     = 0;
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(pMode->manual_timings.src_width);
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapHeight      = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modeCount].Frequency                    = hz;
                HwDeviceExtension->pModes[modeCount].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_PALETTE_DRIVEN |
                    VIDEO_MODE_MONO_POINTER | VIDEO_MODE_MANAGED_PALETTE;

                if (pMode->manual_timings.flags & VG_MODEFLAG_INTERLACED)
                    HwDeviceExtension->pModes[modeCount].AttributeFlags |= VIDEO_MODE_INTERLACED;

                modeCount++;
            }

            // 16BPP
            //
            if (mask & 0x2)
            {
                // Fill in the Windows mode structure for 16BPP
                //
                HwDeviceExtension->pModes[modeCount].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modeCount].ModeIndex                    = (i << 16) | modeCount;
                HwDeviceExtension->pModes[modeCount].VisScreenWidth               = pMode->manual_timings.src_width;
                HwDeviceExtension->pModes[modeCount].VisScreenHeight              = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].ScreenStride                 = POWER_OF2_STRIDE(pMode->manual_timings.src_width << 1);
                HwDeviceExtension->pModes[modeCount].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modeCount].BitsPerPlane                 = 16;
                HwDeviceExtension->pModes[modeCount].XMillimeter                  = pMode->manual_timings.src_width >> 1;
                HwDeviceExtension->pModes[modeCount].YMillimeter                  = pMode->manual_timings.src_height >> 1;
                HwDeviceExtension->pModes[modeCount].NumberRedBits                = 6;
                HwDeviceExtension->pModes[modeCount].NumberGreenBits              = 6;
                HwDeviceExtension->pModes[modeCount].NumberBlueBits               = 6;
                HwDeviceExtension->pModes[modeCount].RedMask                      = 0x1F << 11;
                HwDeviceExtension->pModes[modeCount].GreenMask                    = 0x3F << 5;
                HwDeviceExtension->pModes[modeCount].BlueMask                     = 0x1F;
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(pMode->manual_timings.src_width);
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapHeight      = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modeCount].Frequency                    = hz;
                HwDeviceExtension->pModes[modeCount].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

                if (pMode->manual_timings.flags & VG_MODEFLAG_INTERLACED)
                    HwDeviceExtension->pModes[modeCount].AttributeFlags |= VIDEO_MODE_INTERLACED;

                modeCount++;
            }

            // 24/32 BPP
            //
            if (mask & 0xC)
            {
                // Fill in the Windows mode structure for 32BPP
                //
                HwDeviceExtension->pModes[modeCount].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modeCount].ModeIndex                    = (i << 16) | modeCount;
                HwDeviceExtension->pModes[modeCount].VisScreenWidth               = pMode->manual_timings.src_width;
                HwDeviceExtension->pModes[modeCount].VisScreenHeight              = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].ScreenStride                 = POWER_OF2_STRIDE(pMode->manual_timings.src_width << 2);
                HwDeviceExtension->pModes[modeCount].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modeCount].BitsPerPlane                 = 32;
                HwDeviceExtension->pModes[modeCount].XMillimeter                  = pMode->manual_timings.src_width >> 1;
                HwDeviceExtension->pModes[modeCount].YMillimeter                  = pMode->manual_timings.src_height >> 1;
                HwDeviceExtension->pModes[modeCount].NumberRedBits                = 8;
                HwDeviceExtension->pModes[modeCount].NumberGreenBits              = 8;
                HwDeviceExtension->pModes[modeCount].NumberBlueBits               = 8;
                HwDeviceExtension->pModes[modeCount].RedMask                      = 0xFF << 16;
                HwDeviceExtension->pModes[modeCount].GreenMask                    = 0xFF << 8;
                HwDeviceExtension->pModes[modeCount].BlueMask                     = 0xFF;
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(pMode->manual_timings.src_width);
                HwDeviceExtension->pModes[modeCount].VideoMemoryBitmapHeight      = pMode->manual_timings.src_height;
                HwDeviceExtension->pModes[modeCount].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modeCount].Frequency                    = hz;
                HwDeviceExtension->pModes[modeCount].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

                // Display mode is either 24BPP or 32BPP.  The memory organization is the
                // same.  However, 32BPP enables per-pixel alpha blending at the expense
                // of color keying.
                //
                if (mask & 8)
                    HwDeviceExtension->pModes[modeCount].DriverSpecificAttributeFlags = DRIVER_FLAG_32BPP;

                modeCount++;
            }
        }
    }
    else
    {
        unsigned long addedModeWidth = 0;
        unsigned long addedModeHeight = 0;
        int parseCRTModes = 1;
        
        modeCount = 0;
        
        if (g_regSettings.CustomPanelTimings && g_regSettings.PanelEnable)
        {
            // If custom panel timings are specified, we will expose
            // the native panel timings, the user specified modes, and all 
            // Cimarron CRT modes.  The latter are only exposed if requested by
            // the user.
            //
            modeCount = (HwDeviceExtension->extraTFTResolutionCount + 1) * 3;

            // When resuming, the only thing that can change is that the panel can be
            // forced on when the user had only the CRT enabled.  In that case,
            // the mode programmed before the suspend will be in Cimarron's mode tables.
            // This mode may therefore get added anyway and it will certainly pass the 
            // mode filter.  The only check thus required here is to add the mode if the
            // CRT mode list will not be included anyway.
            //
            if (!HwDeviceExtension->showCRTModes)
            {
                parseCRTModes = 0;
                if (fromResume)
                {                    
                    if (g_regSettings.RotationAngle == 90 ||
                        g_regSettings.RotationAngle == 270)
                    {
                        addedModeWidth = HwDeviceExtension->ScreenHeight;
                        addedModeHeight = HwDeviceExtension->ScreenWidth;
                    }
                    else
                    {
                        addedModeWidth = HwDeviceExtension->ScreenWidth;
                        addedModeHeight = HwDeviceExtension->ScreenHeight;
                    }

                    HwDeviceExtension->addedExtraModeFromResume = TRUE;
                    if (g_regSettings.PanelWidth  == addedModeWidth  &&
                        g_regSettings.PanelHeight == addedModeHeight &&
                        HwDeviceExtension->VRefreshRate == g_regSettings.PanelRefresh)
                    {
                        HwDeviceExtension->addedExtraModeFromResume = FALSE;
                    }
                    else
                    {
                        for (i = 0; i < (int)HwDeviceExtension->extraTFTResolutionCount; i++)
                        {
                            if (*(HwDeviceExtension->extraTFTResolutions + (i << 1))     == addedModeWidth  &&
                                *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1) == addedModeHeight &&
                                  HwDeviceExtension->VRefreshRate == g_regSettings.PanelRefresh)
                            {
                                HwDeviceExtension->addedExtraModeFromResume = FALSE;
                                break;
                            }
                        }
                    }
                    if (HwDeviceExtension->addedExtraModeFromResume)
                    {
                        DISPDBG ((3000, "Hacking in an extra mode on resume\n"));
                        modeCount += 3;
                    }
                }
            }
        }

        // DYNAMIC MODE TABLE GENERATION
        // Display modes are generated based on the current Cimarron
        // mode tables.
        //
        if (parseCRTModes)
        {
            count = vg_get_display_mode_count();

            // GENERATE MODE COUNT
            // We end up searching the Cimarron array twice, but we allocate less
            // memory from the pool this way.
            //
            for (i = 0; i < count; i++)
            {
                vg_get_display_mode_information (i, &displayMode);

                // FIND ONLY CRT MODES
                // We are assuming that the panel resolutions are a subset of the CRT
                // resolutions.
                //
                if (!(displayMode.flags & (VG_MODEFLAG_PANELOUT | VG_MODEFLAG_TVOUT)))
                {
                    // Ignore 56Hz modes and modes smaller than 640x480
                    //
                    if (displayMode.vactive < 480 ||
                        displayMode.hactive < 640 ||
                        displayMode.internal_flags & VG_SUPPORTFLAG_56HZ)
                    {
                        continue;
                    }

                    // Assume that all BPP are supported for all modes.
                    //
                    modeCount += 3;
                }
            }
        }

        HwDeviceExtension->modeCount = modeCount;
        LxAllocateMemory (HwDeviceExtension, sizeof(VIDEO_MODE_INFORMATION) * modeCount,
            &HwDeviceExtension->pModes);
        if (!HwDeviceExtension->pModes)
            return FALSE;

        // FILL IN THE DISPLAY MODES
        // Note that we do extra work to enter the modes in the table in order.
        // To match other drivers (and simply for neatness) we group all the refresh rates
        // for a particular BPP and resolution.  Note that the Cimarron table groups
        // first by refresh rate.  To speed up the process, we make the assumption
        // that all modes support all BPP.
        //
        modesFilled = 0;
        resStart    = 0;

        if (g_regSettings.CustomPanelTimings && g_regSettings.PanelEnable)
        {
            modesFilled += 3;

            HwDeviceExtension->pModes[0].Length                       = sizeof(VIDEO_MODE_INFORMATION);
            HwDeviceExtension->pModes[0].ModeIndex                    = 0;
            HwDeviceExtension->pModes[0].VisScreenWidth               = HwDeviceExtension->panelTimings->hactive;
            HwDeviceExtension->pModes[0].VisScreenHeight              = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[0].ScreenStride                 = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive);
            HwDeviceExtension->pModes[0].NumberOfPlanes               = 1;
            HwDeviceExtension->pModes[0].BitsPerPlane                 = 8;
            HwDeviceExtension->pModes[0].Frequency                    = g_regSettings.PanelRefresh;
            HwDeviceExtension->pModes[0].XMillimeter                  = (HwDeviceExtension->panelTimings->hactive) >> 1;
            HwDeviceExtension->pModes[0].YMillimeter                  = (HwDeviceExtension->panelTimings->vactive) >> 1;
            HwDeviceExtension->pModes[0].NumberRedBits                = 8;
            HwDeviceExtension->pModes[0].NumberGreenBits              = 8;
            HwDeviceExtension->pModes[0].NumberBlueBits               = 8;
            HwDeviceExtension->pModes[0].RedMask                      = 0;
            HwDeviceExtension->pModes[0].GreenMask                    = 0;
            HwDeviceExtension->pModes[0].BlueMask                     = 0;
            HwDeviceExtension->pModes[0].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive);
            HwDeviceExtension->pModes[0].VideoMemoryBitmapHeight      = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[0].DriverSpecificAttributeFlags = 0;
            HwDeviceExtension->pModes[0].AttributeFlags               =
                VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_PALETTE_DRIVEN |
                VIDEO_MODE_MONO_POINTER | VIDEO_MODE_MANAGED_PALETTE;

            HwDeviceExtension->pModes[1].Length                       = sizeof(VIDEO_MODE_INFORMATION);
            HwDeviceExtension->pModes[1].ModeIndex                    = 0;
            HwDeviceExtension->pModes[1].VisScreenWidth               = HwDeviceExtension->panelTimings->hactive;
            HwDeviceExtension->pModes[1].VisScreenHeight              = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[1].ScreenStride                 = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive << 1);
            HwDeviceExtension->pModes[1].NumberOfPlanes               = 1;
            HwDeviceExtension->pModes[1].BitsPerPlane                 = 16;
            HwDeviceExtension->pModes[1].Frequency                    = g_regSettings.PanelRefresh;
            HwDeviceExtension->pModes[1].XMillimeter                  = HwDeviceExtension->panelTimings->hactive >> 1;
            HwDeviceExtension->pModes[1].YMillimeter                  = HwDeviceExtension->panelTimings->vactive >> 1;
            HwDeviceExtension->pModes[1].NumberRedBits                = 6;
            HwDeviceExtension->pModes[1].NumberGreenBits              = 6;
            HwDeviceExtension->pModes[1].NumberBlueBits               = 6;
            HwDeviceExtension->pModes[1].RedMask                      = 0x1F << 11;
            HwDeviceExtension->pModes[1].GreenMask                    = 0x3F << 5;
            HwDeviceExtension->pModes[1].BlueMask                     = 0x1F;
            HwDeviceExtension->pModes[1].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive);
            HwDeviceExtension->pModes[1].VideoMemoryBitmapHeight      = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[1].DriverSpecificAttributeFlags = 0;
            HwDeviceExtension->pModes[1].AttributeFlags               =
                VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

            HwDeviceExtension->pModes[2].Length                       = sizeof(VIDEO_MODE_INFORMATION);
            HwDeviceExtension->pModes[2].ModeIndex                    = 0;
            HwDeviceExtension->pModes[2].VisScreenWidth               = HwDeviceExtension->panelTimings->hactive;
            HwDeviceExtension->pModes[2].VisScreenHeight              = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[2].ScreenStride                 = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive << 2);
            HwDeviceExtension->pModes[2].NumberOfPlanes               = 1;
            HwDeviceExtension->pModes[2].BitsPerPlane                 = 32;
            HwDeviceExtension->pModes[2].Frequency                    = g_regSettings.PanelRefresh;
            HwDeviceExtension->pModes[2].XMillimeter                  = HwDeviceExtension->panelTimings->hactive >> 1;
            HwDeviceExtension->pModes[2].YMillimeter                  = HwDeviceExtension->panelTimings->vactive >> 1;
            HwDeviceExtension->pModes[2].NumberRedBits                = 8;
            HwDeviceExtension->pModes[2].NumberGreenBits              = 8;
            HwDeviceExtension->pModes[2].NumberBlueBits               = 8;
            HwDeviceExtension->pModes[2].RedMask                      = 0xFF << 16;
            HwDeviceExtension->pModes[2].GreenMask                    = 0xFF << 8;
            HwDeviceExtension->pModes[2].BlueMask                     = 0xFF;
            HwDeviceExtension->pModes[2].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(HwDeviceExtension->panelTimings->hactive);
            HwDeviceExtension->pModes[2].VideoMemoryBitmapHeight      = HwDeviceExtension->panelTimings->vactive;
            HwDeviceExtension->pModes[2].DriverSpecificAttributeFlags = 0;
            HwDeviceExtension->pModes[2].AttributeFlags               =
                VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

            // Add user-requested resolutions in addition to the custom panel size.
            //
            for (i = 0; i < (int)HwDeviceExtension->extraTFTResolutionCount; i++)
            {
                HwDeviceExtension->pModes[modesFilled + 0] = HwDeviceExtension->pModes[0];
                HwDeviceExtension->pModes[modesFilled + 1] = HwDeviceExtension->pModes[1];
                HwDeviceExtension->pModes[modesFilled + 2] = HwDeviceExtension->pModes[2];

                HwDeviceExtension->pModes[modesFilled + 0].VisScreenWidth          = *(HwDeviceExtension->extraTFTResolutions + (i << 1));
                HwDeviceExtension->pModes[modesFilled + 0].VisScreenHeight         = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);
                HwDeviceExtension->pModes[modesFilled + 0].ScreenStride            = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)));
                HwDeviceExtension->pModes[modesFilled + 0].XMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1)) >> 1;
                HwDeviceExtension->pModes[modesFilled + 0].YMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1) >> 1;
                HwDeviceExtension->pModes[modesFilled + 0].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)));
                HwDeviceExtension->pModes[modesFilled + 0].VideoMemoryBitmapHeight = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);

                HwDeviceExtension->pModes[modesFilled + 1].VisScreenWidth          = *(HwDeviceExtension->extraTFTResolutions + (i << 1));
                HwDeviceExtension->pModes[modesFilled + 1].VisScreenHeight         = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);
                HwDeviceExtension->pModes[modesFilled + 1].ScreenStride            = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)) << 1);
                HwDeviceExtension->pModes[modesFilled + 1].XMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1)) >> 1;
                HwDeviceExtension->pModes[modesFilled + 1].YMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1) >> 1;
                HwDeviceExtension->pModes[modesFilled + 1].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)));
                HwDeviceExtension->pModes[modesFilled + 1].VideoMemoryBitmapHeight = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);

                HwDeviceExtension->pModes[modesFilled + 2].VisScreenWidth          = *(HwDeviceExtension->extraTFTResolutions + (i << 1));
                HwDeviceExtension->pModes[modesFilled + 2].VisScreenHeight         = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);
                HwDeviceExtension->pModes[modesFilled + 2].ScreenStride            = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)) << 2);
                HwDeviceExtension->pModes[modesFilled + 2].XMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1)) >> 1;
                HwDeviceExtension->pModes[modesFilled + 2].YMillimeter             = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1) >> 1;
                HwDeviceExtension->pModes[modesFilled + 2].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(*(HwDeviceExtension->extraTFTResolutions + (i << 1)));
                HwDeviceExtension->pModes[modesFilled + 2].VideoMemoryBitmapHeight = *(HwDeviceExtension->extraTFTResolutions + (i << 1) + 1);

                modesFilled += 3;
            }
        }

        if (HwDeviceExtension->addedExtraModeFromResume)
        {
            HwDeviceExtension->pModes[modesFilled + 0] = HwDeviceExtension->pModes[0];
            HwDeviceExtension->pModes[modesFilled + 1] = HwDeviceExtension->pModes[1];
            HwDeviceExtension->pModes[modesFilled + 2] = HwDeviceExtension->pModes[2];

            HwDeviceExtension->pModes[modesFilled + 0].VisScreenWidth          = addedModeWidth;
            HwDeviceExtension->pModes[modesFilled + 0].VisScreenHeight         = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 0].ScreenStride            = POWER_OF2_STRIDE(addedModeWidth);
            HwDeviceExtension->pModes[modesFilled + 0].XMillimeter             = addedModeWidth >> 1;
            HwDeviceExtension->pModes[modesFilled + 0].YMillimeter             = addedModeHeight >> 1;
            HwDeviceExtension->pModes[modesFilled + 0].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(addedModeWidth);
            HwDeviceExtension->pModes[modesFilled + 0].VideoMemoryBitmapHeight = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 0].Frequency               = HwDeviceExtension->VRefreshRate;

            HwDeviceExtension->pModes[modesFilled + 1].VisScreenWidth          = addedModeWidth;
            HwDeviceExtension->pModes[modesFilled + 1].VisScreenHeight         = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 1].ScreenStride            = POWER_OF2_STRIDE(addedModeWidth << 1);
            HwDeviceExtension->pModes[modesFilled + 1].XMillimeter             = addedModeWidth >> 1;
            HwDeviceExtension->pModes[modesFilled + 1].YMillimeter             = addedModeHeight >> 1;
            HwDeviceExtension->pModes[modesFilled + 1].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(addedModeWidth);
            HwDeviceExtension->pModes[modesFilled + 1].VideoMemoryBitmapHeight = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 1].Frequency               = HwDeviceExtension->VRefreshRate;

            HwDeviceExtension->pModes[modesFilled + 2].VisScreenWidth          = addedModeWidth;
            HwDeviceExtension->pModes[modesFilled + 2].VisScreenHeight         = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 2].ScreenStride            = POWER_OF2_STRIDE(addedModeWidth << 2);
            HwDeviceExtension->pModes[modesFilled + 2].XMillimeter             = addedModeWidth >> 1;
            HwDeviceExtension->pModes[modesFilled + 2].YMillimeter             = addedModeHeight >> 1;
            HwDeviceExtension->pModes[modesFilled + 2].VideoMemoryBitmapWidth  = POWER_OF2_STRIDE(addedModeWidth);
            HwDeviceExtension->pModes[modesFilled + 2].VideoMemoryBitmapHeight = addedModeHeight;
            HwDeviceExtension->pModes[modesFilled + 2].Frequency               = HwDeviceExtension->VRefreshRate;

            modesFilled += 3;
        }

        while (modesFilled != modeCount)
        {
            vg_get_display_mode_information (resStart, &displayMode);

            if ((displayMode.flags & (VG_MODEFLAG_PANELOUT | VG_MODEFLAG_TVOUT)) ||
                displayMode.vactive < 480 ||
                displayMode.hactive < 640 ||
                displayMode.internal_flags & VG_SUPPORTFLAG_56HZ)
            {
                resStart++;
                continue;
            }

            currentWidth  = displayMode.hactive;
            currentHeight = displayMode.vactive;

            valid = 1;
            resStop = resStart + 1;
            while (1)
            {
                // Break on the last mode.
                //
                if (resStop == count)
                    break;

                vg_get_display_mode_information (resStop, &displayMode);

                if ((displayMode.flags & (VG_MODEFLAG_PANELOUT | VG_MODEFLAG_TVOUT))   ||
                     displayMode.vactive < 480                                         ||
                     displayMode.hactive < 640                                         ||
                     displayMode.internal_flags & VG_SUPPORTFLAG_56HZ)
                {
                    resStop++;
                    continue;
                }

                if (displayMode.hactive == currentWidth && displayMode.vactive == currentHeight)
                {
                    valid++;
                    resStop++;
                    continue;
                }

                break;
            }

            // Now that we have the mode range, add all the modes...
            // Remember the assumption is that all BPP are supported.
            // We can thus fill in all 8, 16 and 32 bit modes at the same
            // time by indexing into the table as an array.
            //
            for (i = resStart; i < resStop; i++)
            {
                vg_get_display_mode_information (i, &displayMode);

                switch (displayMode.internal_flags & VG_SUPPORTFLAG_HZMASK)
                {
                    case VG_SUPPORTFLAG_100HZ: hz = 100; break;
                    case VG_SUPPORTFLAG_90HZ:  hz = 90;  break;
                    case VG_SUPPORTFLAG_85HZ:  hz = 85;  break;
                    case VG_SUPPORTFLAG_75HZ:  hz = 75;  break;
                    case VG_SUPPORTFLAG_72HZ:  hz = 72;  break;
                    case VG_SUPPORTFLAG_70HZ:  hz = 70;  break;
                    case VG_SUPPORTFLAG_60HZ:
                    default :                  hz = 60;  break;
                }

                if ((displayMode.flags & (VG_MODEFLAG_PANELOUT | VG_MODEFLAG_TVOUT))   ||
                    displayMode.vactive < 480                                         ||
                    displayMode.hactive < 640                                         ||
                    displayMode.internal_flags & VG_SUPPORTFLAG_56HZ)

                {
                    continue;
                }

                HwDeviceExtension->pModes[modesFilled].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modesFilled].ModeIndex                    = 0;
                HwDeviceExtension->pModes[modesFilled].VisScreenWidth               = displayMode.hactive;
                HwDeviceExtension->pModes[modesFilled].VisScreenHeight              = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled].ScreenStride                 = POWER_OF2_STRIDE(displayMode.hactive);
                HwDeviceExtension->pModes[modesFilled].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modesFilled].BitsPerPlane                 = 8;
                HwDeviceExtension->pModes[modesFilled].Frequency                    = hz;
                HwDeviceExtension->pModes[modesFilled].XMillimeter                  = displayMode.hactive >> 1;
                HwDeviceExtension->pModes[modesFilled].YMillimeter                  = displayMode.vactive >> 1;
                HwDeviceExtension->pModes[modesFilled].NumberRedBits                = 8;
                HwDeviceExtension->pModes[modesFilled].NumberGreenBits              = 8;
                HwDeviceExtension->pModes[modesFilled].NumberBlueBits               = 8;
                HwDeviceExtension->pModes[modesFilled].RedMask                      = 0;
                HwDeviceExtension->pModes[modesFilled].GreenMask                    = 0;
                HwDeviceExtension->pModes[modesFilled].BlueMask                     = 0;
                HwDeviceExtension->pModes[modesFilled].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(displayMode.hactive);
                HwDeviceExtension->pModes[modesFilled].VideoMemoryBitmapHeight      = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modesFilled].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_PALETTE_DRIVEN |
                    VIDEO_MODE_MONO_POINTER | VIDEO_MODE_MANAGED_PALETTE;

                HwDeviceExtension->pModes[modesFilled + valid].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modesFilled + valid].ModeIndex                    = 0;
                HwDeviceExtension->pModes[modesFilled + valid].VisScreenWidth               = displayMode.hactive;
                HwDeviceExtension->pModes[modesFilled + valid].VisScreenHeight              = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled + valid].ScreenStride                 = POWER_OF2_STRIDE(displayMode.hactive << 1);
                HwDeviceExtension->pModes[modesFilled + valid].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modesFilled + valid].BitsPerPlane                 = 16;
                HwDeviceExtension->pModes[modesFilled + valid].Frequency                    = hz;
                HwDeviceExtension->pModes[modesFilled + valid].XMillimeter                  = displayMode.hactive >> 1;
                HwDeviceExtension->pModes[modesFilled + valid].YMillimeter                  = displayMode.vactive >> 1;
                HwDeviceExtension->pModes[modesFilled + valid].NumberRedBits                = 6;
                HwDeviceExtension->pModes[modesFilled + valid].NumberGreenBits              = 6;
                HwDeviceExtension->pModes[modesFilled + valid].NumberBlueBits               = 6;
                HwDeviceExtension->pModes[modesFilled + valid].RedMask                      = 0x1F << 11;
                HwDeviceExtension->pModes[modesFilled + valid].GreenMask                    = 0x3F << 5;
                HwDeviceExtension->pModes[modesFilled + valid].BlueMask                     = 0x1F;
                HwDeviceExtension->pModes[modesFilled + valid].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(displayMode.hactive);
                HwDeviceExtension->pModes[modesFilled + valid].VideoMemoryBitmapHeight      = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled + valid].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modesFilled + valid].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

                HwDeviceExtension->pModes[modesFilled + (valid << 1)].Length                       = sizeof(VIDEO_MODE_INFORMATION);
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].ModeIndex                    = 0;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].VisScreenWidth               = displayMode.hactive;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].VisScreenHeight              = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].ScreenStride                 = POWER_OF2_STRIDE(displayMode.hactive << 2);
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].NumberOfPlanes               = 1;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].BitsPerPlane                 = 32;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].Frequency                    = hz;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].XMillimeter                  = displayMode.hactive >> 1;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].YMillimeter                  = displayMode.vactive >> 1;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].NumberRedBits                = 8;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].NumberGreenBits              = 8;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].NumberBlueBits               = 8;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].RedMask                      = 0xFF << 16;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].GreenMask                    = 0xFF << 8;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].BlueMask                     = 0xFF;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].VideoMemoryBitmapWidth       = POWER_OF2_STRIDE(displayMode.hactive);
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].VideoMemoryBitmapHeight      = displayMode.vactive;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].DriverSpecificAttributeFlags = 0;
                HwDeviceExtension->pModes[modesFilled + (valid << 1)].AttributeFlags               =
                    VIDEO_MODE_COLOR | VIDEO_MODE_GRAPHICS | VIDEO_MODE_MONO_POINTER;

                modesFilled++;
            }

            // Skip over the 16 and 32 bit modes that were added
            //
            modesFilled += valid << 1;

            resStart = resStop;
        }
    }

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// FilterUnsupportedModes
//
// Figure out which modes are supported by the current framebuffer model
// and remove those we can't support.
//
BOOLEAN FilterUnsupportedModes (PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResume)
{
	ULONG curModeIndex;
	ULONG maxStride, minStride;
	ULONG modeSize, compSize;
	ULONG mode, fbmax;
	BOOLEAN reject;
    BOOLEAN exempt;
	int supported = 0;
	int i;
    int extraModes = HwDeviceExtension->addedExtraModeFromResume ? 1 : 0;

    // FILTER CUSTOM MODES
    // Note that we do not check if timings are valid.  Custom modes
    // are an 'advanced' feature, implying the user knows exactly what
    // they are doing.
    //
    if (g_regSettings.ManualTimingEnable)
    {
	    // CALCULATE END OF FRAME BUFFER
	    // We always leave room for the cursor.
        //
	    fbmax = HwDeviceExtension->FBSize - LX_RESERVED_SPACE - 0xA000;

        curModeIndex = 0;
	    for (i = 0; i < (int)HwDeviceExtension->modeCount; i++)
	    {
            // Extract the index for the custom timings.
            //
            mode = HwDeviceExtension->pModes[i].ModeIndex >> 16;

            // PROGRAM THE STRIDE BASED ON THE USER FLAGS
            // We are ignoring the registry settings that indicate pitch and
            // using the Cimarron flags instead.  Note that the mode already
            // contains the power of 2 pitch.
            //
            minStride = HwDeviceExtension->pModes[i].VisScreenWidth << (HwDeviceExtension->pModes[i].BitsPerPlane >> 4);
            minStride = (minStride + 31) & ~31ul;

            if (HwDeviceExtension->customModes[mode].manual_timings.flags & VG_MODEFLAG_LINEARPITCH)
                HwDeviceExtension->pModes[i].ScreenStride = minStride;

            modeSize = HwDeviceExtension->pModes[i].ScreenStride * HwDeviceExtension->pModes[i].VisScreenHeight;

		    // ADD COMPRESSION
		    // Unless compression is requested to be always on, the minimum
		    // mode footprint will have compression disabled.

		    if (g_regSettings.Compression != 0)
		    {
                // Panning and centering are currently disallowed in a custom mode.
                //
                compSize = 544 * HwDeviceExtension->pModes[i].VisScreenHeight;

                // ENABLE COMPRESSION AT THE END OF THE LINE WHENEVER POSSIBLE

                if ((HwDeviceExtension->pModes[i].ScreenStride - minStride) >= 544)
                    compSize = 0;

			    if (g_regSettings.Compression == 2 ||
                   (g_regSettings.Compression == 1 &&
				    ((modeSize + compSize) < fbmax)))
			    {
				    modeSize += compSize;
				    HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= DRIVER_FLAG_COMPRESSION;
			    }
		    }

            // CHECK ROTATION
            // The mode is almost about to be accepted.  At this point, check to see if
            // rotation is enabled.  If so, an extra buffer will be consumed by the driver
            // for the rotated mode.  We reject any modes that cannot fit such a buffer.
            // Also, we take this instance to swizzle the width and height for the modes.
            //
            if (g_regSettings.RotationAngle != 0)
            {
                unsigned long rot_width, rot_height;
                unsigned long rot_stride;

                // Reject all 8BPP modes while rotating.  8BPP rotation is not implemented
                // properly, and is a non-target.
                //
                if (HwDeviceExtension->pModes[i].BitsPerPlane == 8)
                    continue;

                if (g_regSettings.RotationAngle == 90 || g_regSettings.RotationAngle == 270)
                {
                    rot_width  = HwDeviceExtension->pModes[i].VisScreenHeight;
                    rot_height = HwDeviceExtension->pModes[i].VisScreenWidth;
                }
                else
                {
                    rot_width  = HwDeviceExtension->pModes[i].VisScreenWidth;
                    rot_height = HwDeviceExtension->pModes[i].VisScreenHeight;
                }

                rot_stride = rot_width << (HwDeviceExtension->pModes[i].BitsPerPlane >> 4);
                rot_stride = (rot_stride + 31ul) & ~31ul;
                modeSize += (rot_stride * rot_height);

                HwDeviceExtension->pModes[i].VisScreenWidth  = rot_width;
                HwDeviceExtension->pModes[i].VisScreenHeight = rot_height;
                HwDeviceExtension->pModes[i].VideoMemoryBitmapWidth  = rot_width;
                HwDeviceExtension->pModes[i].VideoMemoryBitmapHeight = rot_height;
                HwDeviceExtension->pModes[i].XMillimeter = rot_width >> 1;
                HwDeviceExtension->pModes[i].YMillimeter = rot_height >> 1;
            }

            // SAVE THE OFFSCREEN OFFSET TO AVOID EXTRA WORK IN THE DISPLAY DRIVER
			//
            HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= modeSize;

		    if (modeSize > fbmax)
			    continue;

            // COPY MODE INFORMATION DOWN TO KEEP LIST PACKED
            //
			if((unsigned int)i != curModeIndex)
			{
				HwDeviceExtension->pModes[curModeIndex] = HwDeviceExtension->pModes[i];
			}

			// SET THE INDEX OF THE MODE
			// Set the index of this mode. It has to be done
			// somewhere and experience proves it's laborious
			// and error prone to do it statically.

			HwDeviceExtension->pModes[curModeIndex].ModeIndex &= 0xFFFF0000;
            HwDeviceExtension->pModes[curModeIndex].ModeIndex |= curModeIndex;

            curModeIndex++;
	    }

	    // SAVE VALID MODE COUNT

	    HwDeviceExtension->modeCount = curModeIndex;

        // CHECK FOR NO VALID MODES
        //
        if (curModeIndex == 0)
        {
            DISPDBG ((10000, "No modes available after filtering!\n"));
            return FALSE;
        }
    }

    else
    {
	    // CLEAR INCREMENTAL VARIABLES

	    curModeIndex = 0;

	    // CALCULATE END OF FRAME BUFFER
	    // We always leave room for the cursor

	    fbmax = HwDeviceExtension->FBSize - LX_RESERVED_SPACE - 0xA000;

	    for (i = 0; i < (int)HwDeviceExtension->modeCount; i++)
	    {
		    // ASSUME WE ACCEPT THE MODE

		    reject = FALSE;

            // LAST MODE SET BEFORE A SUSPEND IS EXEMPT

            exempt = fromResume &&
                (HwDeviceExtension->ScreenWidth  == HwDeviceExtension->pModes[i].VisScreenWidth)  &&
                (HwDeviceExtension->ScreenHeight == HwDeviceExtension->pModes[i].VisScreenHeight) &&
                (HwDeviceExtension->VRefreshRate == HwDeviceExtension->pModes[i].Frequency)       &&
                (HwDeviceExtension->BitsPerPixel == HwDeviceExtension->pModes[i].BitsPerPlane);

		    // FILTER NON-60HZ TIMINGS IF REQUESTED
            // We no longer filter modes through Cimarron as the modes are generated from 
            // Cimarron...
            //
            if (!exempt && !g_regSettings.EnableNon60HzPanel && g_regSettings.PanelEnable)
            {
                if (HwDeviceExtension->pModes[i].Frequency != 60 && 
                    (!g_regSettings.CustomPanelTimings || 
                        i >= (int)((HwDeviceExtension->extraTFTResolutionCount + 1 + extraModes) * 3)))
                {
                    reject = TRUE;
                }
            }

		    // CALCULATE THE LINEAR AND RECTANGULAR STRIDES
		    // The rectangular stride is already in the video structure.

		    maxStride = HwDeviceExtension->pModes[i].ScreenStride;
    		minStride = HwDeviceExtension->pModes[i].VisScreenWidth << (HwDeviceExtension->pModes[i].BitsPerPlane >> 4);
            minStride = (minStride + 31) & ~31ul;

		    // CALCULATE BASE MODE SIZE
		    // If a linear frame buffer is disabled, we will always use
		    // the rectangular stride.  For the other two options, the
		    // linear frame buffer will be available, thus reducing the
		    // minimum mode footprint.

		    if (g_regSettings.FBLinear == 0)
		    {
			    modeSize = maxStride * HwDeviceExtension->pModes[i].VisScreenHeight;
		    }
		    else if (g_regSettings.FBLinear == 1)
		    {
			    modeSize = maxStride * HwDeviceExtension->pModes[i].VisScreenHeight;

			    // RECTANGULAR IS TOO BIG - USE LINEAR

			    if (modeSize > fbmax)
			    {
				    modeSize = minStride * HwDeviceExtension->pModes[i].VisScreenHeight;
				    HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= DRIVER_FLAG_LINEAR_MODE;
				    HwDeviceExtension->pModes[i].ScreenStride = minStride;
			    }
		    }
		    else
		    {
			    modeSize = minStride * HwDeviceExtension->pModes[i].VisScreenHeight;
			    HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= DRIVER_FLAG_LINEAR_MODE;
			    HwDeviceExtension->pModes[i].ScreenStride = minStride;
		    }

		    // ADD COMPRESSION
		    // Unless compression is requested to be always on, the minimum
		    // mode footprint will have compression disabled.

		    if (g_regSettings.Compression != 0)
		    {
			    compSize = 544 * HwDeviceExtension->pModes[i].VisScreenHeight;

                // ENABLE COMPRESSION AT THE END OF THE LINE WHENEVER POSSIBLE

                if ((HwDeviceExtension->pModes[i].ScreenStride - minStride) >= 544)
                {
                    compSize = 0;
                }

			    if (g_regSettings.Compression == 2 ||
                    (g_regSettings.Compression == 1 &&
				    ((modeSize + compSize) < fbmax)))
			    {
				    modeSize += compSize;
				    HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= DRIVER_FLAG_COMPRESSION;
			    }
		    }

            // CHECK ROTATION
            // The mode is almost about to be accepted.  At this point, check to see if
            // rotation is enabled.  If so, an extra buffer will be consumed by the driver
            // for the rotated mode.  We reject any modes that cannot fit such a buffer.
            // Also, we take this instance to swizzle the width and height for the modes.
            //
            if (g_regSettings.RotationAngle != 0)
            {
                unsigned long rot_width, rot_height;
                unsigned long rot_stride;

                // Reject all 8BPP modes while rotating.  8BPP rotation is not implemented
                // properly, and is a non-target.
                //
                if (HwDeviceExtension->pModes[i].BitsPerPlane == 8)
                    continue;

                if (g_regSettings.RotationAngle == 90 || g_regSettings.RotationAngle == 270)
                {
                    rot_width  = HwDeviceExtension->pModes[i].VisScreenHeight;
                    rot_height = HwDeviceExtension->pModes[i].VisScreenWidth;
                }
                else
                {
                    rot_width  = HwDeviceExtension->pModes[i].VisScreenWidth;
                    rot_height = HwDeviceExtension->pModes[i].VisScreenHeight;
                }

                rot_stride = rot_width << (HwDeviceExtension->pModes[i].BitsPerPlane >> 4);
                rot_stride = (rot_stride + 31ul) & ~31ul;
                modeSize += (rot_stride * rot_height);

                HwDeviceExtension->pModes[i].VisScreenWidth  = rot_width;
                HwDeviceExtension->pModes[i].VisScreenHeight = rot_height;
                HwDeviceExtension->pModes[i].VideoMemoryBitmapWidth  = rot_width;
                HwDeviceExtension->pModes[i].VideoMemoryBitmapHeight = rot_height;
                HwDeviceExtension->pModes[i].XMillimeter = rot_width >> 1;
                HwDeviceExtension->pModes[i].YMillimeter = rot_height >> 1;
            }

		    if (modeSize > fbmax)
			    reject = TRUE;

		    // SAVE THE OFFSCREEN OFFSET TO AVOID EXTRA WORK IN THE DISPLAY DRIVER

		    else
			    HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags |= modeSize;

		    if(reject)
		    {
			    DISPDBG((100, "mode %d, %d bpp: (%d x %d @ %d) = %d bytes - rejected",
				    i, HwDeviceExtension->pModes[i].BitsPerPlane,
				    HwDeviceExtension->pModes[i].VisScreenWidth,
				    HwDeviceExtension->pModes[i].VisScreenHeight,
				    HwDeviceExtension->pModes[i].Frequency,
				    modeSize));
		    }
		    else
		    {
			    // COPY MODE INFORMATION DOWN TO KEEP LIST PACKED

			    if((unsigned int)i != curModeIndex)
			    {
				    HwDeviceExtension->pModes[curModeIndex] = HwDeviceExtension->pModes[i];
			    }

			    // SET THE INDEX OF THE MODE
			    // Set the index of this mode. It has to be done
			    // somewhere and experience proves it's laborious
			    // and error prone to do it statically.

			    HwDeviceExtension->pModes[curModeIndex].ModeIndex = curModeIndex;

			    curModeIndex++;

			    DISPDBG((100, "mode %d, %d bpp: (%d x %d @ %d) = %d bytes - accepted",
				    i, HwDeviceExtension->pModes[i].BitsPerPlane,
				    HwDeviceExtension->pModes[i].VisScreenWidth,
				    HwDeviceExtension->pModes[i].VisScreenHeight,
				    HwDeviceExtension->pModes[i].Frequency));

			    DISPDBG((100, "Linear %d, Compression %d",
				    (HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags & DRIVER_FLAG_LINEAR_MODE) ? 1 : 0,
				    (HwDeviceExtension->pModes[i].DriverSpecificAttributeFlags & DRIVER_FLAG_COMPRESSION) ? 1 : 0));

		    }
	    }

	    // SAVE VALID MODE COUNT

	    HwDeviceExtension->modeCount = curModeIndex;
    }

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// LXStartIO
//
// This routine allows the VDD and NT to communicate with the miniport
// driver.	It fields a variety of IOCTL requests and performs the requested
// action or returns the desired information.
//
BOOLEAN LXStartIO (PHW_DEVICE_EXTENSION HwDeviceExtension, PVIDEO_REQUEST_PACKET RequestPacket)
{
	VP_STATUS status;
	ULONG inIoSpace;
	PVIDEO_CLUT clutBuffer;
	PVOID virtualAddr;
	ULONG ulValue;

	DISPDBG((0, "StartIO, ioControl 0x%x,  DevExt 0x%x\n",
		RequestPacket->IoControlCode, HwDeviceExtension));

	// SWITCH ON THE I/O CONTROL CODE

	switch (RequestPacket->IoControlCode)
	{
		// DPMS CONTROLS

        case IOCTL_VIDEO_GET_POWER_MANAGEMENT:
		{
		    VIDEO_POWER_MANAGEMENT *powerState;
			powerState = RequestPacket->OutputBuffer;
			if (RequestPacket->OutputBufferLength < sizeof(VIDEO_POWER_MANAGEMENT))
			{
			    RequestPacket->StatusBlock->Information = 0;
			    status = ERROR_INSUFFICIENT_BUFFER;
			    break;
			}
			powerState->PowerState  = HwDeviceExtension->CurrPowerState;
			powerState->Length      = HwDeviceExtension->CurrLength;
			powerState->DPMSVersion = HwDeviceExtension->CurrDPMSVersion;

			// NUMBER OF RETURNED BYTES

			RequestPacket->StatusBlock->Information = sizeof(VIDEO_POWER_MANAGEMENT);
			status = NO_ERROR;
			break;
        }

		case IOCTL_VIDEO_SET_POWER_MANAGEMENT:
		{
			VIDEO_POWER_MANAGEMENT *powerState;
			powerState = RequestPacket->InputBuffer;

			df_set_crt_enable ((int)powerState->PowerState);

			HwDeviceExtension->CurrPowerState  = powerState->PowerState;
			HwDeviceExtension->CurrLength      = powerState->Length;
			HwDeviceExtension->CurrDPMSVersion = powerState->DPMSVersion;
			break;
		}

		case IOCTL_VIDEO_QUERY_PUBLIC_ACCESS_RANGES:
		{
			ULONG LXRegLength;
			VIDEO_PUBLIC_ACCESS_RANGES *publicAccessRange;

			DISPDBG((1, "LXStartIO - IOCTL_VIDEO_QUERY_PUBLIC_ACCESS_RANGES\n"));

			if (RequestPacket->OutputBufferLength <
					(RequestPacket->StatusBlock->Information =
					5*sizeof(VIDEO_PUBLIC_ACCESS_RANGES)))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			publicAccessRange = (VIDEO_PUBLIC_ACCESS_RANGES *)
					RequestPacket->OutputBuffer;

			// REMAP ALL ADDRESSES
			// Even though we have already generated logical addresses for all
			// device memory, we must create a mapping that is compatible with
			// the display driver and not just the kernel.

			// MAP REGISTER POINTER

			status = AddMappedAddressToArray (HwDeviceExtension->PhysLXRegAddr, 0x4000,
				0, (PVOID)publicAccessRange, HwDeviceExtension);

			publicAccessRange++;

			// MAP VIDEO POINTER

			if (status == NO_ERROR)
			{
				status = AddMappedAddressToArray (HwDeviceExtension->PhysLXVidAddr, 0x4000,
					0, (PVOID)publicAccessRange, HwDeviceExtension);
			}
			publicAccessRange++;

			// MAP GP POINTER

			if (status == NO_ERROR)
			{
				status = AddMappedAddressToArray (HwDeviceExtension->PhysLXGPAddr, 0x4000,
					0, (PVOID)publicAccessRange, HwDeviceExtension);
			}
            publicAccessRange++;

            // MAP VIP POINTER

			if (status == NO_ERROR && HwDeviceExtension->vipHardware)
			{
				status = AddMappedAddressToArray (HwDeviceExtension->PhysLXVIPAddr, 0x4000,
					0, (PVOID)publicAccessRange, HwDeviceExtension);
			}
			publicAccessRange++;

			// SET CPU SPEED

			init_read_cpu_frequency ((ULONG *)(&publicAccessRange->VirtualAddress));
			break;
		}

		case IOCTL_VIDEO_MAP_VIDEO_MEMORY:
		{
			VIDEO_MEMORY_INFORMATION *memoryInfo;
			ULONG frameLength;

			DISPDBG((1, "LXStartIO - MapVideoMemory\n"));

			if ((RequestPacket->OutputBufferLength <
					(RequestPacket->StatusBlock->Information =
					sizeof(VIDEO_MEMORY_INFORMATION))) ||
					(RequestPacket->InputBufferLength < sizeof(VIDEO_MEMORY)))
			{
				DISPDBG((1, "output buffer too small\n"));
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
			memoryInfo = RequestPacket->OutputBuffer;

			DISPDBG((1, "Output ptr 0x%x\n", memoryInfo));

			inIoSpace   = 0;
			virtualAddr = NULL;
			frameLength = HwDeviceExtension->FBSize;

			status = VideoPortMapMemory(HwDeviceExtension,
						HwDeviceExtension->PhysLXFBAddr,
						&frameLength,
						&inIoSpace,
						&virtualAddr);

			memoryInfo->VideoRamBase   = virtualAddr;
			memoryInfo->VideoRamLength = HwDeviceExtension->FBSize;
			DISPDBG((1, "status 0x%x, virtual address 0x%x\n",
				status, virtualAddr));

			// FRAME BUFFER AND VIDEO RAM ARE EQUIVALENT

			memoryInfo->FrameBufferBase   = memoryInfo->VideoRamBase;
			memoryInfo->FrameBufferLength = memoryInfo->VideoRamLength;

			DISPDBG((0, "LXStartIO - MapVideoMemory returns.\n"));
			break;
		}

		case IOCTL_VIDEO_UNMAP_VIDEO_MEMORY:

			DISPDBG((0, "LXStartIO - UnMapVideoMemory\n"));

			if (RequestPacket->InputBufferLength < sizeof(VIDEO_MEMORY))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
			status = VideoPortUnmapMemory(HwDeviceExtension,
										  ((PVIDEO_MEMORY)
										   (RequestPacket->InputBuffer))->
										  RequestedVirtualAddress,
										  0);

			break;

		case IOCTL_VIDEO_QUERY_CURRENT_MODE:
		{
			PVIDEO_MODE_INFORMATION modeInfo;

			DISPDBG((0, "LXStartIO - QueryCurrentMode, index %d\n", lxModeIndex));

			modeInfo = RequestPacket->OutputBuffer;

			if (RequestPacket->OutputBufferLength <
					 sizeof(VIDEO_MODE_INFORMATION))
			{
				RequestPacket->StatusBlock->Information = 0;
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
				// STRUCTURE ASSIGNMENT (COPY)

				RequestPacket->StatusBlock->Information =
					 sizeof(VIDEO_MODE_INFORMATION);

				*modeInfo = HwDeviceExtension->pModes[lxModeIndex];

				status = NO_ERROR;
			}

			break;
		}

		case IOCTL_VIDEO_QUERY_AVAIL_MODES:
		{
			VIDEO_MODE_INFORMATION *modeInfo;
			ULONG i;

			DISPDBG((1, "LXStartIO - QueryAvailableModes\n"));

			RequestPacket->StatusBlock->Information =
					HwDeviceExtension->modeCount * sizeof(VIDEO_MODE_INFORMATION);

			DISPDBG((1, "modePtr %x, modes %d, buf size %d, need %d\n",
				RequestPacket->OutputBuffer,
				HwDeviceExtension->modeCount,
				RequestPacket->OutputBufferLength,
				HwDeviceExtension->modeCount * sizeof(VIDEO_MODE_INFORMATION)));

			if (RequestPacket->OutputBufferLength <
						(HwDeviceExtension->modeCount * sizeof(VIDEO_MODE_INFORMATION)))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
				// COPY DATA

				modeInfo = RequestPacket->OutputBuffer;
				for (i = 0; i < HwDeviceExtension->modeCount; i++)
					modeInfo[i] = HwDeviceExtension->pModes[i];

				status = NO_ERROR;
			}
			break;
		}

		case IOCTL_VIDEO_QUERY_NUM_AVAIL_MODES:
		{
			VIDEO_NUM_MODES *videoNumModePtr;

			DISPDBG((0, "LXStartIO - QueryNumAvailableModes\n"));
			DISPDBG((0, "mode count %d\n", HwDeviceExtension->modeCount));

			if (RequestPacket->OutputBufferLength < sizeof(VIDEO_NUM_MODES))
			{
				RequestPacket->StatusBlock->Information = 0;
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
				RequestPacket->StatusBlock->Information = sizeof(VIDEO_NUM_MODES);
				videoNumModePtr = (VIDEO_NUM_MODES *) RequestPacket->OutputBuffer;

				videoNumModePtr->NumModes = HwDeviceExtension->modeCount;
				videoNumModePtr->ModeInformationLength = sizeof(VIDEO_MODE_INFORMATION) ;
				status = NO_ERROR;
			}

			break;
		}

		case IOCTL_VIDEO_SET_CURRENT_MODE:
		{
			ULONG modeIndex;

			DISPDBG((1, "LXStartIO - SetCurrentMode\n"));

			// GRAB MODE INDEX

			modeIndex = ((PVIDEO_MODE) (RequestPacket->InputBuffer))->RequestedMode & 0xFFFF;
			DISPDBG((1, "mode index %d\n", modeIndex));

			if (modeIndex < HwDeviceExtension->modeCount)
			{
				// STORE MODE PARAMETERS

				lxModeIndex   = modeIndex;
				lxCustomIndex = ((PVIDEO_MODE)(RequestPacket->InputBuffer))->RequestedMode >> 16;
                HwDeviceExtension->BitsPerPixel = HwDeviceExtension->pModes[lxModeIndex].BitsPerPlane;
				HwDeviceExtension->ScreenWidth  = HwDeviceExtension->pModes[lxModeIndex].VisScreenWidth;
				HwDeviceExtension->ScreenHeight = HwDeviceExtension->pModes[lxModeIndex].VisScreenHeight;
				HwDeviceExtension->VRefreshRate = HwDeviceExtension->pModes[lxModeIndex].Frequency;
				HwDeviceExtension->ScreenStride = HwDeviceExtension->pModes[lxModeIndex].ScreenStride;
				HwDeviceExtension->ModeFlags    = HwDeviceExtension->pModes[lxModeIndex].DriverSpecificAttributeFlags;

                if (g_regSettings.ManualTimingEnable)
                {
                    HwDeviceExtension->InterlacedOut =
                        HwDeviceExtension->customModes[lxCustomIndex].manual_timings.flags & VG_MODEFLAG_INTERLACED;
                }
                else
                    HwDeviceExtension->InterlacedOut = 0;

				DISPDBG((0, "BitsPerPixel %d\n", HwDeviceExtension->BitsPerPixel));
				DISPDBG((0, "ScreenWidth  %d\n", HwDeviceExtension->ScreenWidth));
				DISPDBG((0, "ScreenHeight %d\n", HwDeviceExtension->ScreenHeight));
				DISPDBG((0, "VRefreshRate %d\n", HwDeviceExtension->VRefreshRate));

				// SET THE MODE ON THE HARDWARE
                // We only do this if this IOCTL was not made by our display driver.  Our
                // display driver will set the mode itself to avoid complications with
                // Cimarron.
                //
                status = NO_ERROR;
                if (RequestPacket->InputBufferLength == sizeof(VIDEO_MODE))
                {
                    DISPDBG ((5000, "Mode set outside of display driver!!!\n"));
				    LXSetDisplayMode(HwDeviceExtension, (int)lxModeIndex);
                }
                else if (RequestPacket->InputBufferLength == (2 * sizeof(VIDEO_MODE)))
                {
                    DISPDBG ((5000, "Mode set by display driver!\n"));
                    
                    // Save the last programmed mode in the registry.
                    //
                    if (g_regSettings.PanelEnable)
                    {
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetTFTModeWidth",
                            &HwDeviceExtension->ScreenWidth,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetTFTModeHeight",
                            &HwDeviceExtension->ScreenHeight,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetTFTRefresh",
                            &HwDeviceExtension->VRefreshRate,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetTFTRotation",
                            &g_regSettings.RotationAngle,
                            sizeof(ULONG));
                    }
                    else
                    {
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetCRTModeWidth",
                            &HwDeviceExtension->ScreenWidth,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetCRTModeHeight",
                            &HwDeviceExtension->ScreenHeight,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetCRTRefresh",
                            &HwDeviceExtension->VRefreshRate,
                            sizeof(ULONG));
                        VideoPortSetRegistryParameters (
                            HwDeviceExtension,
                            L"LastSetCRTRotation",
                            &g_regSettings.RotationAngle,
                            sizeof(ULONG));
                    }
                }
                else
                {
                    DISPDBG ((5000, "Unknown source for mode set!\n"));
                    status = ERROR_INVALID_PARAMETER;
                }
			}
			else
			{
				status = ERROR_INVALID_PARAMETER;
			}

			break;
		}

		case IOCTL_VIDEO_SET_COLOR_REGISTERS:

			DISPDBG((0, "\n\nLXStartIO - SetColorRegs\n"));

			clutBuffer = RequestPacket->InputBuffer;

			DISPDBG((0, "InputBufferLength %d\n", RequestPacket->InputBufferLength));
			DISPDBG((0, "NumEntries %d\n", clutBuffer->NumEntries));

			// CHECK FOR ROOM IN THE INPUT BUFFER

			if ((RequestPacket->InputBufferLength < sizeof(VIDEO_CLUT) -
				 sizeof(ULONG)) ||
				(RequestPacket->InputBufferLength < sizeof(VIDEO_CLUT) +
				 (sizeof(ULONG) * (clutBuffer->NumEntries - 1))))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				DISPDBG((300, "Set color registers, bad buffer\n"));
				break;
			}

			if (HwDeviceExtension->BitsPerPixel == 8)
			{
				LXSetPalette(HwDeviceExtension,
						(PULONG) clutBuffer->LookupTable,
						clutBuffer->FirstEntry,
						clutBuffer->NumEntries);

				status = NO_ERROR;
			}
			else
			{
				status = ERROR_INVALID_PARAMETER;
			}
			break;

		case IOCTL_VIDEO_RESET_DEVICE:

			DISPDBG((0, "LXStartIO - RESET_DEVICE\n"));

			status = NO_ERROR;
			break;

		case IOCTL_VIDEO_QUERY_COLOR_CAPABILITIES:

			// UNSUPPORTED
			status = ERROR_INVALID_FUNCTION;
			break;

		case IOCTL_VIDEO_SHARE_VIDEO_MEMORY:
		{
			PVIDEO_SHARE_MEMORY pShareMemory;
			PVIDEO_SHARE_MEMORY_INFORMATION pShareMemoryInformation;
			PHYSICAL_ADDRESS shareAddress;
			PVOID virtualAddress;
			ULONG sharedViewSize;
			ULONG inIoSpace;

			DISPDBG((1, "LXStartIO - IOCTL_VIDEO_SHARE_VIDEO_MEMORY\n"));

			if ( (RequestPacket->OutputBufferLength < sizeof(VIDEO_SHARE_MEMORY_INFORMATION)) ||
				 (RequestPacket->InputBufferLength < sizeof(VIDEO_MEMORY)) )
			{

				DISPDBG((1, "IOCTL_VIDEO_SHARE_VIDEO_MEMORY - insufficient buffer\n"));
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			pShareMemory = RequestPacket->InputBuffer;

			if ((pShareMemory->ViewOffset > HwDeviceExtension->FBSize) ||
			   ((pShareMemory->ViewOffset + pShareMemory->ViewSize) >
					  HwDeviceExtension->FBSize) )
			{
				DISPDBG((0, "IOCTL_VIDEO_SHARE_VIDEO_MEMORY - invalid parameter\n"));
				status = ERROR_INVALID_PARAMETER;
				break;
			}

			RequestPacket->StatusBlock->Information = sizeof(VIDEO_SHARE_MEMORY_INFORMATION);

			// BEWARE
			// The input buffer and the output buffer are the same buffer, and
			// therefore data should not be copied from one to the other.

			virtualAddress = pShareMemory->ProcessHandle;
			sharedViewSize = HwDeviceExtension->FBSize;
			inIoSpace      = 0;

			shareAddress.QuadPart = HwDeviceExtension->PhysLXFBAddr.QuadPart;

			DISPDBG((1, "share addr (%x,%x), io %d, virtual %x, size %d\n",
					shareAddress.HighPart, shareAddress.LowPart,
					inIoSpace, virtualAddress, sharedViewSize));

			status = VideoPortMapMemory(HwDeviceExtension, shareAddress, &sharedViewSize,
										&inIoSpace, &virtualAddress);

			DISPDBG((1, "status %x, sharedviewsize %x, virtual %x, viewOffset %x\n",
					status, sharedViewSize, virtualAddress, pShareMemory->ViewOffset));

			pShareMemoryInformation = RequestPacket->OutputBuffer;

			pShareMemoryInformation->SharedViewOffset = pShareMemory->ViewOffset;
			pShareMemoryInformation->VirtualAddress   = virtualAddress;
			pShareMemoryInformation->SharedViewSize   = sharedViewSize;
			break;
		}

		case IOCTL_VIDEO_UNSHARE_VIDEO_MEMORY:
		{
			PVIDEO_SHARE_MEMORY pShareMemory;
			DISPDBG((0, "LXStartIO - IOCTL_VIDEO_UNSHARE_VIDEO_MEMORY\n"));

			if (RequestPacket->InputBufferLength < sizeof(VIDEO_SHARE_MEMORY))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			pShareMemory = RequestPacket->InputBuffer;

			// UNMAP LX REGISTER SPACE

			status = VideoPortUnmapMemory(HwDeviceExtension,
										  pShareMemory->RequestedVirtualAddress,
										  pShareMemory->ProcessHandle);

			DISPDBG((1, "Unshared addr %x, process %x\n",
					pShareMemory->RequestedVirtualAddress,
					pShareMemory->ProcessHandle));

			break;
		}

        case IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS:
        {
            VIDEOPARAMETERS *parameters;
			DISPDBG((0, "LXStartIO - IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS\n"));

			if (RequestPacket->InputBufferLength  < sizeof(VIDEOPARAMETERS))
			{
                DISPDBG((3000, "LXStartIO - IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS - Insufficient Input buffer\n"));
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

            // GUID CHECKING IS CURRENTLY SKIPPED
            //
            parameters = RequestPacket->OutputBuffer;
            if (parameters->dwCommand == VP_COMMAND_GET)
            {
			    if (RequestPacket->OutputBufferLength  < sizeof(VIDEOPARAMETERS))
                {
                    DISPDBG((3000, "LXStartIO - IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS - Insufficient Output buffer\n"));
				    status = ERROR_INSUFFICIENT_BUFFER;
				    break;
                }


                *parameters = *((VIDEOPARAMETERS *)RequestPacket->InputBuffer);
                parameters->dwFlags = 0;
                parameters->dwTVStandard = VP_TV_STANDARD_WIN_VGA;

                RequestPacket->StatusBlock->Information = sizeof(VIDEOPARAMETERS);
            }
            else
                RequestPacket->StatusBlock->Information = 0;

			status = NO_ERROR;
			break;
        }

        case IOCTL_GET_CUSTOM_MODE_SIZE:
        {
            ULONG modeIndex;

            if (RequestPacket->OutputBufferLength != 8)
            {
                status = ERROR_INSUFFICIENT_BUFFER;
				break;
            }

            *((ULONG *)(RequestPacket->OutputBuffer)) = HwDeviceExtension->modeBufferSize;
            *((ULONG *)RequestPacket->OutputBuffer + 1) = HwDeviceExtension->customModeCount;

            RequestPacket->StatusBlock->Information = 8;
			status = NO_ERROR;

            break;
        }

        case IOCTL_GET_CUSTOM_MODES:
        {
            ULONG *pOut;
            ULONG *pIn;
            ULONG dwords;

            if (RequestPacket->OutputBufferLength != HwDeviceExtension->modeBufferSize)
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

            pOut = (ULONG *)(RequestPacket->OutputBuffer);
            pIn = (ULONG *)(HwDeviceExtension->customModes);
            dwords = HwDeviceExtension->modeBufferSize >> 2;

            _asm {
                cld
                mov edi, pOut
                mov esi, pIn
                mov ecx, dwords
                rep movsd
            }

            RequestPacket->StatusBlock->Information = HwDeviceExtension->modeBufferSize;
			status = NO_ERROR;

            break;
        }

        case IOCTL_GET_PANEL_TIMINGS:
        {
            ULONG *pOut;
            ULONG *pIn;
            ULONG dwords;

            if (RequestPacket->OutputBufferLength != sizeof(VG_DISPLAY_MODE))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

            pOut = (ULONG *)(RequestPacket->OutputBuffer);
            pIn = (ULONG *)(HwDeviceExtension->panelTimings);
            dwords = (sizeof (VG_DISPLAY_MODE)) >> 2;

            _asm {
                cld
                mov edi, pOut
                mov esi, pIn
                mov ecx, dwords
                rep movsd
            }

            RequestPacket->StatusBlock->Information = sizeof(VG_DISPLAY_MODE);
			status = NO_ERROR;

            break;
        }

        case IOCTL_GET_LX_REGISTRY:
        {
            LX_REGISTRY_SETTINGS *settings;
			DISPDBG((0, "LXStartIO - IOCTL_GET_LX_REGISTRY\n"));

			if (RequestPacket->OutputBufferLength < sizeof(LX_REGISTRY_SETTINGS))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			settings = RequestPacket->OutputBuffer;
            RetrieveRegistryInformation (HwDeviceExtension, settings);

            RequestPacket->StatusBlock->Information = sizeof(LX_REGISTRY_SETTINGS);
			status = NO_ERROR;

			break;
        }

        case IOCTL_GET_MINIPORT_REGISTRY:
        {
            LX_REGISTRY_SETTINGS *settings;
			DISPDBG((0, "LXStartIO - IOCTL_GET_MINIPORT_REGISTRY\n"));

			if (RequestPacket->OutputBufferLength < sizeof(LX_REGISTRY_SETTINGS))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			settings = RequestPacket->OutputBuffer;
            CopyCurrentRegistryInformation (HwDeviceExtension, settings);

            RequestPacket->StatusBlock->Information = sizeof(LX_REGISTRY_SETTINGS);
			status = NO_ERROR;

			break;
        }

        case IOCTL_SET_LX_REGISTRY:
        {
            LX_REGISTRY_SETTINGS *settings;
			DISPDBG((0, "LXStartIO - IOCTL_SET_LX_REGISTRY\n"));

			if (RequestPacket->InputBufferLength < sizeof(LX_REGISTRY_SETTINGS))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}

			settings = RequestPacket->OutputBuffer;
            SetRegistryInformation (HwDeviceExtension, settings);

            RequestPacket->StatusBlock->Information = sizeof(LX_REGISTRY_SETTINGS);
			status = NO_ERROR;

			break;
        }

		case IOCTL_SET_HEAP_INFORMATION:
		{
            ULONG *heapRange;
			DISPDBG((1, "LXStartIO - IOCTL_SET_HEAP_CALLBACK\n"));

			status = NO_ERROR;

			if (RequestPacket->InputBufferLength < 8)
			{
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
                heapRange = RequestPacket->OutputBuffer;
                heapStart = heapRange[0];
                heapEnd   = heapRange[1];

                RequestPacket->StatusBlock->Information = 0;

                DISPDBG ((1000, "Heap Range 0x%8.8lX - 0x%8.8lX\n", heapStart, heapEnd));
            }
		}
		break;

        case IOCTL_SET_GP_RESTORE_INFO:
        {
            GP_SAVE_RESTORE *gpState;
			DISPDBG((1, "LXStartIO - IOCTL_SET_HEAP_CALLBACK\n"));

			status = NO_ERROR;

			if (RequestPacket->InputBufferLength < sizeof (GP_SAVE_RESTORE))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
                gpState = RequestPacket->InputBuffer;
                HwDeviceExtension->gpState = *gpState;

                RequestPacket->StatusBlock->Information = 0;

                DISPDBG ((1000, "Recording saved GP state\n"));
            }
		}
		break;

        case IOCTL_SET_AUTOFLIP_SURFACES:
        {
            ULONG   *autoFlipSurfaces;
            ULONG    numAutoFlip, interlaced;
            ULONG    i;

			DISPDBG((1, "LXStartIO - IOCTL_SET_AUTOFLIP_SURFACES\n"));

			status = NO_ERROR;

			if (RequestPacket->InputBufferLength < 12)
			{
				status = ERROR_INSUFFICIENT_BUFFER;
			}
			else
			{
                autoFlipSurfaces = RequestPacket->InputBuffer;
                interlaced       = *autoFlipSurfaces++;
                numAutoFlip      = *autoFlipSurfaces++;

                if (RequestPacket->InputBufferLength < ((numAutoFlip << 2) + 8))
                    status = ERROR_INSUFFICIENT_BUFFER;
                else
                {
                    HwDeviceExtension->numAutoFlip = numAutoFlip;
                    HwDeviceExtension->interlaced  = interlaced;
                    for (i = 0; i < numAutoFlip; i++)
                    {
                        HwDeviceExtension->autoFlipSurfaces[i] = autoFlipSurfaces[i];
                    }

                    RequestPacket->StatusBlock->Information = 0;

                    DISPDBG ((1000, "Saved autoflipping surfaces\n"));
                }
            }
		}
		break;

        case IOCTL_SAVE_VIDEO_Y:
        {
			DISPDBG((1, "LXStartIO - IOCTL_SAVE_VIDEO_Y\n"));

			status = NO_ERROR;
            RequestPacket->StatusBlock->Information = 0;

			HwDeviceExtension->videoY = READ_VID32 (DF_VIDEO_Y_POS);
        }
        break;

        case IOCTL_GET_LAST_MODE:
        {
            ULONG *lastMode;
            if (RequestPacket->OutputBufferLength < (sizeof(ULONG) * 4))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
            lastMode = RequestPacket->OutputBuffer;

            if (g_regSettings.PanelEnable)
            {
                lastMode[0] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetTFTModeWidth",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[0]) != NO_ERROR)
                {
                    lastMode[0] = g_regSettings.PanelWidth;
                }
                lastMode[1] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetTFTModeHeight",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[1]) != NO_ERROR)
                {
                    lastMode[1] = g_regSettings.PanelHeight;
                }
                lastMode[2] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetTFTRefresh",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[2]) != NO_ERROR)
                {
                    lastMode[2] = 60;
                }
                lastMode[3] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetTFTRotation",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[3]) != NO_ERROR)
                {
                    lastMode[3] = 0;
                }
            }
            else
            {
                lastMode[0] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetCRTModeWidth",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[0]) != NO_ERROR)
                {
                    lastMode[0] = 800;
                }
                lastMode[1] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetCRTModeHeight",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[1]) != NO_ERROR)
                {
                    lastMode[1] = 600;
                }
                lastMode[2] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetCRTRefresh",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[2]) != NO_ERROR)
                {
                    lastMode[2] = 60;
                }
                lastMode[3] = 4;
                if (VideoPortGetRegistryParameters (
                        HwDeviceExtension,
                        L"LastSetCRTRotation",
                        FALSE,
                        GeodeRegistryCallback,
                        &lastMode[3]) != NO_ERROR)
                {
                    lastMode[3] = 0;
                }
            }

            RequestPacket->StatusBlock->Information = sizeof(ULONG)*4;
            status = NO_ERROR;
            break;
        }

        case IOCTL_SET_DISPLAY_GAMMA:
        {
            if (RequestPacket->InputBufferLength < (sizeof(ULONG) * 9))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
            VideoPortSetRegistryParameters (
                HwDeviceExtension,
                L"DisplayGamma",
                RequestPacket->InputBuffer,
                sizeof(ULONG) * 9);

            RequestPacket->StatusBlock->Information = 0;
            status = NO_ERROR;
            break;
        }

        case IOCTL_SET_OVERLAY_GAMMA:
        {
            if (RequestPacket->InputBufferLength < (sizeof(ULONG) * 9))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
            VideoPortSetRegistryParameters (
                HwDeviceExtension,
                L"OverlayGamma",
                RequestPacket->InputBuffer,
                sizeof(ULONG) * 9);

            RequestPacket->StatusBlock->Information = 0;
            status = NO_ERROR;
            break;
        }

        case IOCTL_GET_DISPLAY_GAMMA:
        {
            if (RequestPacket->OutputBufferLength < (sizeof(ULONG) * 9))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
            
            *((ULONG *)RequestPacket->OutputBuffer) = sizeof(ULONG) * 9;
            if (VideoPortGetRegistryParameters (
                HwDeviceExtension,
                L"DisplayGamma",
                FALSE,
                GeodeRegistryCallback,
                RequestPacket->OutputBuffer) != NO_ERROR)
            {
                ULONG *gamma = RequestPacket->OutputBuffer;
                int i;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_GAMMA * 3) + i] = LX_GAMMA_DEFAULT;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_BRIGHT * 3) + i] = LX_BRIGHT_DEFAULT;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_CONTRAST * 3) + i] = LX_CONTRAST_DEFAULT;
            }

            RequestPacket->StatusBlock->Information = sizeof(ULONG) * 9;
            status = NO_ERROR;
            break;
        }

        case IOCTL_GET_OVERLAY_GAMMA:
        {
            if (RequestPacket->OutputBufferLength < (sizeof(ULONG) * 9))
			{
				status = ERROR_INSUFFICIENT_BUFFER;
				break;
			}
            *((ULONG *)RequestPacket->OutputBuffer) = sizeof(ULONG) * 9;
            if (VideoPortGetRegistryParameters (
                HwDeviceExtension,
                L"OverlayGamma",
                FALSE,
                GeodeRegistryCallback,
                RequestPacket->OutputBuffer) != NO_ERROR)
            {
                ULONG *gamma = RequestPacket->OutputBuffer;
                int i;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_GAMMA * 3) + i] = LX_GAMMA_DEFAULT;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_BRIGHT * 3) + i] = LX_BRIGHT_DEFAULT;
                for (i = 0; i < 3; i++)
                    gamma[(LX_MODE_CONTRAST * 3) + i] = LX_CONTRAST_DEFAULT;
            }

            RequestPacket->StatusBlock->Information = sizeof(ULONG) * 9;
            status = NO_ERROR;
            break;
        }

		// INVALID IO CONTROL CODE

		default:
			DISPDBG((3000, "Fell through LX startIO routine - invalid command, %d=%s\n",
				RequestPacket->IoControlCode, FindIoctl(RequestPacket->IoControlCode)));
			status = ERROR_INVALID_FUNCTION;
			break;

	}

	RequestPacket->StatusBlock->Status = status;

	DISPDBG((0, "LXStartIO returning...\n"));
	return TRUE;

} // LXStartIO

/////////////////////////////////////////////////////////////////////////
// LXClaimResources
//
// This is a fake routine to claim the PnP resources (including IRQ) for the
// driver.
//
void LXClaimResources (PHW_DEVICE_EXTENSION HwDeviceExtension)
{
    VIDEO_ACCESS_RANGE  AccessRanges[5];
    VP_STATUS status;
    ULONG slot;

    VideoPortZeroMemory(&AccessRanges[0], 5*sizeof(VIDEO_ACCESS_RANGE));

    status = VideoPortGetAccessRanges (HwDeviceExtension,
                                      0,
                                      NULL,
                                      5,
                                      AccessRanges,
                                      NULL,
                                      NULL,
                                      &slot);

    if (status != NO_ERROR)
        DISPDBG ((3000, "Failed VideoPortGetAccessRanges!\n"));
    else
        DISPDBG ((3000, "VideoPortGetAccessRanges Succeeded\n"));

}

/////////////////////////////////////////////////////////////////////////
// LXInitialize
//
// Stubbed routine.
//
BOOLEAN LXInitialize (PHW_DEVICE_EXTENSION HwDeviceExtension)
{
	DISPDBG((0, "LXInitialize: HwDeviceExtension 0x%x\n", HwDeviceExtension));
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////
// LXSetPalette
//
// Initializes the palette in an 8BPP mode.
//
void LXSetPalette(
	HW_DEVICE_EXTENSION *HwDeviceExtension,
	ULONG *pPal,
	ULONG startIndex,
	ULONG count)
{
    ULONG    i;
    ULONG    BGR;

	for(i = startIndex; i < (startIndex + count); i++)
	{
		// SWIZZLE BGR DATA TO RGB

        BGR = pPal[i];

        vg_set_display_palette_entry (i,
            ((BGR >> 16) & 0x0000FF) |
            ((BGR << 16) & 0xFF0000) |
             (BGR & 0x00FF00));
	}

	DISPDBG((0, "LXSetPalette returns\n"));
}

/////////////////////////////////////////////////////////////////////////
// SetPanelEnable
//
void SetPanelEnable (HW_DEVICE_EXTENSION* HwDeviceExtension, int enable)
{
	unsigned long power;

	// NO WORK IF NO PANEL PRESENT

	if (g_regSettings.PanelEnable == 0)
		return;

	power = READ_VID32 (DF_POWER_MANAGEMENT);

	if (enable) power |=  DF_PM_PANEL_ON;
	else        power &= ~DF_PM_PANEL_ON;

	WRITE_VID32 (DF_POWER_MANAGEMENT, power);

	// WAIT FOR THE PANEL TO BE FULLY POWERED OFF

	if (!enable)
	{
		while (!(READ_VID32 (DF_POWER_MANAGEMENT) & 2))
			;
	}
}

/////////////////////////////////////////////////////////////////////////
// LXI2CWriteClockLine
//
// DDC Support routine
//
VOID LXI2CWriteClockLine (PVOID HwDeviceExtension, UCHAR data)
{
    HW_DEVICE_EXTENSION *pDev = (HW_DEVICE_EXTENSION *)HwDeviceExtension;
    unsigned short address;
    unsigned long value;

    value = data ? DDC_CLK_HIGH : DDC_CLK_LOW;

    // WRITE CLOCK DATA

    address = (unsigned short)pDev->DDCBase + DDC_CLOCK_OUT;

    _asm {
        mov    eax, value
        mov    dx, address
        out    dx, eax
    }
}

/////////////////////////////////////////////////////////////////////////
// LXI2CWriteDataLine
//
// DDC Support routine
//
VOID LXI2CWriteDataLine (PVOID HwDeviceExtension, UCHAR data)
{
    HW_DEVICE_EXTENSION *pDev = (HW_DEVICE_EXTENSION *)HwDeviceExtension;
    unsigned short address;
    unsigned long value;

    value = data ? DDC_DATA_HIGH : DDC_DATA_LOW;

    // WRITE DATA

    address = (unsigned short)pDev->DDCBase + DDC_DATA_OUT;

    _asm {
        mov    eax, value
        mov    dx, address
        out    dx, eax
    }
}

/////////////////////////////////////////////////////////////////////////
// LXI2CReadClockLine
//
// DDC Support routine
//
BOOLEAN LXI2CReadClockLine (PVOID HwDeviceExtension)
{
    HW_DEVICE_EXTENSION *pDev = (HW_DEVICE_EXTENSION *)HwDeviceExtension;
    unsigned short address;
    unsigned long value;

    // READ CLOCK

    address = (unsigned short)pDev->DDCBase + DDC_CLOCK_IN;

    _asm {

        mov    dx, address
        in     eax, dx
        mov    value, eax
    }

    if (value & DDC_CLK_HIGH)
        return TRUE;

    return FALSE;
}

/////////////////////////////////////////////////////////////////////////
// LXI2CReadDataLine
//
// DDC Support routine
//
BOOLEAN LXI2CReadDataLine (PVOID HwDeviceExtension)
{
    HW_DEVICE_EXTENSION *pDev = (HW_DEVICE_EXTENSION *)HwDeviceExtension;
    unsigned short address;
    unsigned long value;

    // READ DATA

    address = (unsigned short)pDev->DDCBase + DDC_DATA_IN;

    _asm {

        mov    dx, address
        in     eax, dx
        mov    value, eax
    }

    if (value & DDC_DATA_HIGH)
        return TRUE;

    return FALSE;
}

/////////////////////////////////////////////////////////////////////////
// LXGetChildDescriptor
//
ULONG LXGetChildDescriptor (
	HW_DEVICE_EXTENSION    *HwDeviceExtension,
	PVIDEO_CHILD_ENUM_INFO  ChildEnumInfo,
	PVIDEO_CHILD_TYPE       pChildType,
	PVOID                   pChildDescriptor,
	PULONG                  pUId,
	PULONG                  pUnused )
{
    switch (ChildEnumInfo->ChildIndex)
    {
		// ACPI
		// Case 0 is used to enumerate devices found by the ACPI firmware.
        // We do not currently support ACPI devices

        case 0:

            return ERROR_NO_MORE_DEVICES;

		// TREAT CASE 1 AS MONITOR
		// Force non-DDC monitor for now.

		case 1:

			*pChildType = Monitor;

            if (LXDDCHelper (HwDeviceExtension, pChildDescriptor,
                ChildEnumInfo->ChildDescriptorSize))
            {
                *pUId = P2_DDC_MONITOR;
            }
            else
                *pUId = P2_NONDDC_MONITOR;

			return ERROR_MORE_DATA;

        default:

            return ERROR_NO_MORE_DEVICES;
    }
}

/////////////////////////////////////////////////////////////////////////
// LXPowerManageDisplay
//
// Updates the display settings for lower power monitor modes.
//
void LXPowerManageDisplay (int powerOn)
{
    static int videoEnable = 0;
    ULONG gcfg, unlock;

    if (powerOn)
    {
        unlock = READ_REG32 (DC3_UNLOCK);
        gcfg   = READ_REG32 (DC3_GENERAL_CFG);
        WRITE_REG32 (DC3_UNLOCK, DC3_UNLOCK_VALUE);
        WRITE_REG32 (DC3_GENERAL_CFG, gcfg | DC3_GCFG_DFLE);
        WRITE_REG32 (DC3_UNLOCK, unlock);

        if (videoEnable)
            df_set_video_enable (1, 0);
    }
    else
    {
        unsigned long flags;
        df_get_video_enable (&videoEnable, &flags);
        df_set_video_enable (0, 0);

        unlock = READ_REG32 (DC3_UNLOCK);
        gcfg   = READ_REG32 (DC3_GENERAL_CFG);
        WRITE_REG32 (DC3_UNLOCK, DC3_UNLOCK_VALUE);
        WRITE_REG32 (DC3_GENERAL_CFG, gcfg & ~DC3_GCFG_DFLE);
        WRITE_REG32 (DC3_UNLOCK, unlock);
    }
}

/////////////////////////////////////////////////////////////////////////
// LXGetPowerState
//
// Returns power state information.
//
VP_STATUS LXGetPowerState (
	HW_DEVICE_EXTENSION     *HwDeviceExtension,
	ULONG                    HwId,
	PVIDEO_POWER_MANAGEMENT  VideoPowerControl)
{
    VP_STATUS status;
    PHW_DEVICE_EXTENSION hwDeviceExtension = HwDeviceExtension;


    DISPDBG((2000, "LXGetPowrState: hwId(%xh) state = %d\n",
                     (int)HwId, (int)VideoPowerControl->PowerState));

    switch((int)HwId)
    {
        case P2_DDC_MONITOR:
        case P2_NONDDC_MONITOR:

            switch (VideoPowerControl->PowerState)
            {
				// WE SUPPORT ALL POWER STATE

                case VideoPowerOn:
                case VideoPowerStandBy:
                case VideoPowerSuspend:
                case VideoPowerOff:
                case VideoPowerHibernate:
                case VideoPowerShutdown:
					status = NO_ERROR;

                    break;

                default:

                    DISPDBG((2000, "LXGetPowerState: Unknown monitor PowerState(%xh)\n",
                                    (int)VideoPowerControl->PowerState));

                    ASSERT(FALSE);
                    status = ERROR_INVALID_PARAMETER;
            }
            break;

        case DISPLAY_ADAPTER_HW_ID:

            // WE SUPPORT ALL POWER STATES

            switch (VideoPowerControl->PowerState)
            {
                case VideoPowerOn:
                case VideoPowerStandBy:
                case VideoPowerSuspend:
                case VideoPowerHibernate:
                case VideoPowerShutdown:
                case VideoPowerOff:
                    status = NO_ERROR;
                    break;

                default:

                    DISPDBG((2000, "LXGetPowerState: Unknown adapter PowerState(%xh)\n",
                                 (int)VideoPowerControl->PowerState));

                    ASSERT(FALSE);

                    status = ERROR_INVALID_PARAMETER;
            }

            break;

        default:

            DISPDBG((2000, "LXGetPowerState: Unknown hwId(%xh)",
                            (int)HwId));
            ASSERT(FALSE);

            status = ERROR_INVALID_PARAMETER;
    }

    DISPDBG((2000, "LXGetPowerState: returning %xh\n", status));

    return status;
}

/////////////////////////////////////////////////////////////////////////
// LXSetPowerState
//
// Sets the power state for a given device.
//
VP_STATUS LXSetPowerState (
	HW_DEVICE_EXTENSION     *HwDeviceExtension,
	ULONG                    HwId,
	PVIDEO_POWER_MANAGEMENT  VideoPowerControl)
{
    PHW_DEVICE_EXTENSION hwDeviceExtension = HwDeviceExtension;
    VP_STATUS status;
    ULONG unlock, gcfg;
    static int videoEnable = 0;

    DISPDBG((2000, "LXSetPowerState: hwId(%xh) state = %d\n",
                   (int)HwId, (int)VideoPowerControl->PowerState));

    switch((int)HwId)
    {

        case P2_DDC_MONITOR:
        case P2_NONDDC_MONITOR:


            switch (VideoPowerControl->PowerState)
            {

                case VideoPowerHibernate:
                case VideoPowerShutdown:

                    DISPDBG((2000, "LXSetPowerState: monitor HIBERNATE\n"));

                    // MONITOR STAYS ON FOR HIBERNATE AND SHUTDOWN

                    status = NO_ERROR;
                    break;

                case VideoPowerOn:

                    DISPDBG((2000, "LXSetPowerState: monitor ON\n"));

					// ENABLE THE MONITOR (AND THE PANEL)
                    //
					df_set_crt_enable (DF_CRT_ENABLE);
                    if ((!g_regSettings.ManualTimingEnable && g_regSettings.PanelEnable) ||
                         (g_regSettings.ManualTimingEnable && 
                           HwDeviceExtension->customModes[lxCustomIndex].manual_timings.flags & VG_MODEFLAG_PANELOUT))
                    {
                         SetPanelEnable (HwDeviceExtension, 1);
                    }

                    // ENABLE DATA FETCHES
                    //
                    if (HwDeviceExtension->PreviousMonitorState != VideoPowerOn)
                        LXPowerManageDisplay (1);

                    status = NO_ERROR;
                    break;

                case VideoPowerStandBy:

                    DISPDBG((2000, "LXSetPowerState: monitor STANDBY\n"));

                    // HSYNC OFF, VSYNC ON
					// As panels have very little latency, we also disable
					// the panel.

                    if ((!g_regSettings.ManualTimingEnable && g_regSettings.PanelEnable) ||
                         (g_regSettings.ManualTimingEnable && 
                           HwDeviceExtension->customModes[lxCustomIndex].manual_timings.flags & VG_MODEFLAG_PANELOUT))
                    {
					    SetPanelEnable(HwDeviceExtension, 0);
                    }
					df_set_crt_enable (DF_CRT_STANDBY);

                    // DISABLE DATA FETCHES
                    //
                    if (HwDeviceExtension->PreviousMonitorState == VideoPowerOn)
                        LXPowerManageDisplay (1);

                    status = NO_ERROR;
                    break;

                case VideoPowerSuspend:

                    DISPDBG((2000, "LXSetPowerState: monitor SUSPEND\n"));

                    // HSYNC ON, VSYNC OFF
					// As panels have very little latency, we also disable
					// the panel.

                    if ((!g_regSettings.ManualTimingEnable && g_regSettings.PanelEnable) ||
                         (g_regSettings.ManualTimingEnable && 
                           HwDeviceExtension->customModes[lxCustomIndex].manual_timings.flags & VG_MODEFLAG_PANELOUT))
                    {
					    SetPanelEnable (HwDeviceExtension, 0);
                    }
					df_set_crt_enable (DF_CRT_STANDBY);

                    // DISABLE DATA FETCHES
                    //
                    if (HwDeviceExtension->PreviousMonitorState == VideoPowerOn)
                        LXPowerManageDisplay (0);

                    status = NO_ERROR;
                    break;

                case VideoPowerOff:

                    DISPDBG((2000, "LXSetPowerState: monitor OFF\n"));

					// DISABLE THE CRT (AND THE PANEL)

                    if ((!g_regSettings.ManualTimingEnable && g_regSettings.PanelEnable) ||
                         (g_regSettings.ManualTimingEnable && 
                           HwDeviceExtension->customModes[lxCustomIndex].manual_timings.flags & VG_MODEFLAG_PANELOUT))
                    {
					    SetPanelEnable (HwDeviceExtension, 0);
                    }
                    df_set_crt_enable (DF_CRT_DISABLE);

                    // DISABLE DATA FETCHES
                    //
                    if (HwDeviceExtension->PreviousMonitorState == VideoPowerOn)
                        LXPowerManageDisplay (0);

                    status = NO_ERROR;
                    break;

                default:

                    DISPDBG((2000, "LXGetPowerState: Unknown monitor PowerState(%xh)\n",
                                     (int)VideoPowerControl->PowerState));

                    ASSERT(FALSE);
                    status = ERROR_INVALID_PARAMETER;
            }

            // ALL WORK IS ACCOMPLISHED BY THIS POINT

            break;

        case DISPLAY_ADAPTER_HW_ID:

			// POWER STATES FOR INTERNAL VIDEO
			// As we have no real control over the power state of the internal
			// graphics controller, we will do nothing for these states and
			// depend on the BIOS to disable the hardware.  Also, when leaving
			// a save-to-RAM situation, Windows passes a DrvAssertMode call
			// that we use to set up all the hardware configuration that was
			// lost.  So, we have no need to save any registers.  (for now)

            switch (VideoPowerControl->PowerState)
            {
                case VideoPowerHibernate:
                case VideoPowerShutdown:

                    DISPDBG((2000, "LXSetPowerState: adapter HIBERNATE\n"));

                    LXSaveHardwareState (HwDeviceExtension);

					// SAVE FRAME BUFFER DATA
					// Windows expects all memory to be preserved when hibernating
					// including offscreen bitmaps.  The display driver is responsible
                    // for specifying the range of valid memory before the system
                    // hibernates.
                    //
                    if ((long)(heapEnd - heapStart) > 0)
                    {
                        LxAllocateMemory (HwDeviceExtension, heapEnd - heapStart,
                            &fb_ptr);

                        if (fb_ptr != NULL)
                        {
                            from_hibernate = 1;

                            memcpy (fb_ptr, cim_fb_ptr + heapStart, heapEnd - heapStart);
                        }
                        else
                        {
                            DISPDBG ((3000, "Unable to allocate space for video memory\n"));
                        }
                    }

					status = NO_ERROR;
                    break;

                case VideoPowerOn:

                    DISPDBG((2000, "LXSetPowerState: adapter ON\n"));

					// RESTORE FRAME BUFFER DATA IF WE SUCCESSFULLY SAVED THE ENTIRE THING

					if (from_hibernate)
					{
						from_hibernate = 0;
						memcpy (cim_fb_ptr + heapStart, fb_ptr, heapEnd - heapStart);

						LxFreeMemory (HwDeviceExtension, fb_ptr);
						fb_ptr = NULL;
					}

                    if (HwDeviceExtension->PreviousPowerState != VideoPowerOn)
                    {
                        DISPDBG((2000, "  Restoring State\n"));
                        LXRestoreHardwareState (HwDeviceExtension);

                        // Reinitialize registry settings in case of HW change before
                        // resume.
                        //
                        UpdateRegistrySettings (HwDeviceExtension, 1);

                        // Regenerate the HW mode table, as the TFT/CRT configuration may have
                        // changed.
                        //
                        if (HwDeviceExtension->pModes)
                        {
                            DISPDBG ((3000, "Deleting old mode list and regenerating a new one.\n"));
                            LxFreeMemory (HwDeviceExtension, HwDeviceExtension->pModes);
                            HwDeviceExtension->pModes = NULL;
                        }
                        if (!GenerateModeTable (HwDeviceExtension, 1))
                        {
                            DISPDBG ((3000, "Unable to allocate mode table!\n"));
                        }
                        if (!FilterUnsupportedModes (HwDeviceExtension, 1))
                        {
                            DISPDBG ((3000, "Mode Filter Failed!\n"));
                        }
                    }

                    status = NO_ERROR;
                    break;

                case VideoPowerStandBy:

                    DISPDBG((2000, "LXSetPowerState: adapter STANDBY\n"));
                    LXSaveHardwareState (HwDeviceExtension);
					status = NO_ERROR;
                    break;

                case VideoPowerSuspend:

                    DISPDBG((2000, "LXSetPowerState: adapter SUSPEND\n"));
                    LXSaveHardwareState (HwDeviceExtension);
					status = NO_ERROR;
                    break;

                case VideoPowerOff:

                    DISPDBG((2000, "LXSetPowerState: adapter OFF\n"));
                    LXSaveHardwareState (HwDeviceExtension);
					status = NO_ERROR;
                    break;

                default:

                    DISPDBG((2000, "LXGetPowerState: Unknown adapter PowerState(%xh)\n",
                                     (int)VideoPowerControl->PowerState));

                    ASSERT(FALSE);
                    status = ERROR_INVALID_PARAMETER;
            }

            break;

        default:

            DISPDBG((2000, "LXSetPowerState: Unknown hwId(%xh)\n",
                             (int)HwId));

            ASSERT(FALSE);
            status = ERROR_INVALID_PARAMETER;
    }

    if (status == NO_ERROR && HwId == DISPLAY_ADAPTER_HW_ID)
	{
        // SAVE CURRENT POWER STATE
        // We will only restore the hardware registers when they have been previously
        // restored.
        //
		HwDeviceExtension->PreviousPowerState = VideoPowerControl->PowerState;
		HwDeviceExtension->CurrPowerState     = VideoPowerControl->PowerState;
		HwDeviceExtension->CurrLength         = VideoPowerControl->Length;
		HwDeviceExtension->CurrDPMSVersion    = VideoPowerControl->DPMSVersion;
	}
    else if (status == NO_ERROR)
    {
        HwDeviceExtension->PreviousMonitorState = VideoPowerControl->PowerState;
    }

    DISPDBG((2000, "LXSetPowerState: returning %xh\n", status));

    return status;
}

/////////////////////////////////////////////////////////////////////////
// LXSaveHardwareState
//
// Support routine to save a minimal amount of hardware configuration.
//
void LXSaveHardwareState (PHW_DEVICE_EXTENSION  HwDeviceExtension)
{
    unsigned short vgConfig;

    if (HwDeviceExtension->vipHardware)
        vip_save_state (&HwDeviceExtension->vipState);
    df_save_state  (&HwDeviceExtension->dfState);
    vg_save_state  (&HwDeviceExtension->vgState);

    // Disable all BIOS power management code.  Setting this bit
    // indicates to the BIOS that the driver should control all HW
    // initialization.  The BIOS will thus only perform a minimal set
    // of services, such as routing the frame buffer.
    //
    cim_outw (VR_INDEX, VR_UNLOCK);
	cim_outw (VR_INDEX, (VRC_VG << 8) | VG_MEM_SIZE);
    vgConfig = cim_inw (VR_DATA);
    vgConfig |= GRAPHICS_CONTROL;

    cim_outw (VR_INDEX, VR_UNLOCK);
	cim_outw (VR_INDEX, (VRC_VG << 8) | VG_MEM_SIZE);
    cim_outw (VR_DATA, vgConfig);
}

/////////////////////////////////////////////////////////////////////////
// LXRestoreHardwareState
//
// Support routine to restore a minimal amount of hardware configuration.
//
void LXRestoreHardwareState (PHW_DEVICE_EXTENSION  HwDeviceExtension)
{
    unsigned long i;
    unsigned char *fbptr = cim_fb_ptr;

    // Clear the frame buffer to black.
    //
    for (i = 0; i < HwDeviceExtension->ScreenHeight; i++)
    {
        memset (fbptr, 0, HwDeviceExtension->ScreenWidth << ((HwDeviceExtension->BitsPerPixel + 7) >> 4));
        fbptr += HwDeviceExtension->ScreenStride;
    }

    vg_restore_state  (&HwDeviceExtension->vgState);
	df_restore_state  (&HwDeviceExtension->dfState);
    if (HwDeviceExtension->vipHardware)
        vip_restore_state (&HwDeviceExtension->vipState);
	gp_restore_state  (&HwDeviceExtension->gpState);
}

/////////////////////////////////////////////////////////////////////////
// LXMapPhysicalAddress
//
// Support routine to map a physical address into a usable virtual
// address.
//
int LXMapPhysicalAddress (PVOID HwDeviceExtension, ULONG address,
						  ULONG size, UCHAR IO, PULONG ptr_out)
{
	PHYSICAL_ADDRESS temp_address;

	// CHECK FOR NULL ADDRESS
	// Address == 0 implies that the address should not be mapped

	if (address == 0 || ptr_out == NULL || size == 0)
		return 0;

	DISPDBG((1, "LXFindAdapter: Mapping register address 0x%x, length 0x%x\n",
			address, size));

	temp_address.LowPart  = address;
	temp_address.HighPart = 0;

	*ptr_out = (ULONG) VideoPortGetDeviceBase(HwDeviceExtension, temp_address, size, IO);

	if (*ptr_out == 0)
	{
		DISPDBG((1, "Failed to map address\n"));
		return 1;
	}

	return 0;
}

/////////////////////////////////////////////////////////////////////////
// AddMappedAddressToArray
//
// Support routine to map a physical address space into a pointer usable by
// the calling process.
//
VP_STATUS AddMappedAddressToArray (PHYSICAL_ADDRESS address, ULONG length, ULONG IO,
								   PVOID array, PVOID HwDeviceExtension)
{
	PVOID virtualAddr = NULL;
	VP_STATUS status = NO_ERROR;
	VIDEO_PUBLIC_ACCESS_RANGES *accessRanges = (VIDEO_PUBLIC_ACCESS_RANGES *)array;

	if (address.LowPart != 0)
	{
		DISPDBG((1, "Phys Addr 0x%x,0x%x, length 0x%x\n",
				(address).HighPart,
				(address).LowPart,
				length));

		status = VideoPortMapMemory (HwDeviceExtension, address, &length, &IO, &virtualAddr);

		accessRanges->VirtualAddress  = (void *) virtualAddr;
		accessRanges->InIoSpace       = IO;
		accessRanges->MappedInIoSpace = IO;

		DISPDBG((1, "VideoPortMapMemory status 0x%x, addr 0x%x\n",
					status, virtualAddr));
	}

	return status;
}

/////////////////////////////////////////////////////////////////////////
// RetrieveRegistryInformation
//
// Support routine to extract all driver configuration information from the
// registry.
//
int RetrieveRegistryInformation (PHW_DEVICE_EXTENSION HwDeviceExtension, LX_REGISTRY_SETTINGS *pSettings)
{
    unsigned long i;

    LX_REGISTRY_PAIR constantRegKeys[] = {
        { L"Compression",       &pSettings->Compression         },
        { L"FBLinear",          &pSettings->FBLinear            },
        { L"EnablePanelScale",  &pSettings->EnablePanelScale    },
        { L"RotationAngle",     &pSettings->RotationAngle       },
    };

    LX_REGISTRY_PAIR optionalRegKeys[] = {
        { L"BIOSDetectOutput",  &pSettings->BIOSDetectOutput    },
        { L"PanelEnable",       &pSettings->PanelEnable         },
        { L"SimultaneousCRT",   &pSettings->SimultaneousCRT     },
        { L"PanelWidth",        &pSettings->PanelWidth          },
        { L"PanelHeight",       &pSettings->PanelHeight         },
        { L"XVGAPanel",         &pSettings->XVGAPanel           },
        { L"DDCEnabled",        &pSettings->DDCEnabled          },
        { L"DDC_GPIOData",      &pSettings->DDC_GPIOData        },
        { L"DDC_GPIOClock",     &pSettings->DDC_GPIOClock       },
        { L"EnableNon60HzPanel",&pSettings->EnableNon60HzPanel  },
        { L"PanelInSystem",     &pSettings->PanelInSystem       },
    };

    pSettings->CustomPanelTimings = 0;
    pSettings->ManualTimingEnable = 0;

    // Test if we will be parsing a customization file.
    //
    HwDeviceExtension->modeBufferSize = 4;
    if (VideoPortGetRegistryParameters (
            HwDeviceExtension,
            L"CustomFileSize",
            FALSE,
            GeodeRegistryCallback,
            &(HwDeviceExtension->modeBufferSize)) != NO_ERROR)
    {
        HwDeviceExtension->modeBufferSize = 0;
    }

    // The first three DWORDs of the customization file are a binary signature and a
    // mode count.
    //
    if (HwDeviceExtension->modeBufferSize >= 12)
    {
        DISPDBG((9000, "Custom File Size is %i\n", HwDeviceExtension->modeBufferSize));
        LxAllocateMemory (HwDeviceExtension, HwDeviceExtension->modeBufferSize - 12,
            &HwDeviceExtension->customModes);

        if (!HwDeviceExtension->customModes)
        {
            DISPDBG((9000, "Unable to allocate space for custom modes\n"));
            return 1;
        }

        *((unsigned long *)(HwDeviceExtension->customModes)) = HwDeviceExtension->modeBufferSize;

        if (VideoPortGetRegistryParameters (
            HwDeviceExtension,
            L"CustomFile",
            TRUE,
            CustomFileCallback,
            HwDeviceExtension->customModes) != NO_ERROR)
        {
            DISPDBG ((9999, "Error reading customization file.\n"));
            LxFreeMemory (HwDeviceExtension, HwDeviceExtension->customModes);
            HwDeviceExtension->customModes = NULL;
        }

        // Actual size of buffer will not include the first three DWORDs.
        //
        HwDeviceExtension->modeBufferSize -= 12;

        pSettings->ManualTimingEnable = 1;
    }
    else
    {
        unsigned long size = 4;

        // In addition to manually specifying all available modes, the
        // user can instead simply specify panel timings.  Under these
        // circumstances, all CRT modes will be presented to the user.
        // When displaying on the TFT, centering, panning or scaling will
        // be used.  When the display is switched to CRT only, the full
        // CRT timings will be used.  The user can optionally only export
        // the native panel mode or add slaved panel resolutions.
        //
        if (VideoPortGetRegistryParameters (
                HwDeviceExtension,
                L"PanelFileSize",
                FALSE,
                GeodeRegistryCallback,
                &size) != NO_ERROR)
        {
            size = 0;
        }

        if (size >= (sizeof(VG_DISPLAY_MODE) + 16))
        {
            DISPDBG ((9999, "Parsing panel customization file\n"));

            LxAllocateMemory (HwDeviceExtension, sizeof(VG_DISPLAY_MODE),
                &HwDeviceExtension->panelTimings);
            if (!HwDeviceExtension->panelTimings)
            {
                DISPDBG ((9999, "Unable to allocate space for panel customization.\n"));
                return 1;
            }

            // Requires 8 bytes per resolution (ULONG width and height)
            //
            HwDeviceExtension->extraTFTResolutionCount = 
                (size - (sizeof(VG_DISPLAY_MODE) + 16)) >> 3;
            if (HwDeviceExtension->extraTFTResolutionCount)
            {
                LxAllocateMemory (HwDeviceExtension, size - (sizeof(VG_DISPLAY_MODE) + 16),
                    &HwDeviceExtension->extraTFTResolutions);
                if (!HwDeviceExtension->extraTFTResolutions)
                {
                    DISPDBG ((9999, "Unable to allocate space for extra resolutions.\n"));
                    return 1;
                }
            }

            *((unsigned long *)(HwDeviceExtension->panelTimings)) = size;

            if (VideoPortGetRegistryParameters (
                HwDeviceExtension,
                L"PanelFile",
                TRUE,
                CustomPanelCallback,
                HwDeviceExtension->panelTimings) != NO_ERROR)
            {
                DISPDBG ((9999, "Error reading panel file.\n"));
                LxFreeMemory (HwDeviceExtension, HwDeviceExtension->panelTimings);
                HwDeviceExtension->panelTimings = NULL;
            }
            else
            {
                pSettings->CustomPanelTimings = 1;            

                // Hardcode the flags to use PANELOUT
                //
                HwDeviceExtension->panelTimings->flags &= ~VG_MODEFLAG_TVOUT;
                HwDeviceExtension->panelTimings->flags |=  VG_MODEFLAG_PANELOUT;
            }
        }
    }

    // Read the standard (PC-like) registry keys.
    //
    for (i = 0; i < (sizeof(constantRegKeys) / sizeof(LX_REGISTRY_PAIR)); i++)
    {
        *((unsigned long *)constantRegKeys[i].data_element) = 4;
        if (VideoPortGetRegistryParameters (
            HwDeviceExtension,
            constantRegKeys[i].reg_string,
            FALSE,
            GeodeRegistryCallback,
            constantRegKeys[i].data_element) != NO_ERROR)
        {
            *((unsigned long *)constantRegKeys[i].data_element) = 0;
        }
#if DBG
        UnicodeToAnsi (g_ansiBuffer, (unsigned short *)constantRegKeys[i].reg_string);
        DISPDBG((1000, "Registry -- %s - %i\n", g_ansiBuffer, *((unsigned long *)constantRegKeys[i].data_element)));
#endif
    }

    // Read the optional registry keys only if manual display modes are not
    // enabled.
    //
    if (!HwDeviceExtension->customModes)
    {
        for (i = 0; i < (sizeof(optionalRegKeys) / sizeof(LX_REGISTRY_PAIR)); i++)
        {
            *((unsigned long *)optionalRegKeys[i].data_element) = 4;
            if (VideoPortGetRegistryParameters (
                HwDeviceExtension,
                optionalRegKeys[i].reg_string,
                FALSE,
                GeodeRegistryCallback,
                optionalRegKeys[i].data_element) != NO_ERROR)
            {
                *((unsigned long *)optionalRegKeys[i].data_element) = 0;
            }
#if DBG
            UnicodeToAnsi (g_ansiBuffer, (unsigned short *)optionalRegKeys[i].reg_string);
            DISPDBG((1000, "Registry -- %s - %i\n", g_ansiBuffer, *((unsigned long *)optionalRegKeys[i].data_element)));
#endif
        }
    }
    else
    {
        for (i = 0; i < (sizeof(optionalRegKeys) / sizeof(LX_REGISTRY_PAIR)); i++)
            *((unsigned long *)optionalRegKeys[i].data_element) = 0;
    }

    return 0;
}

/////////////////////////////////////////////////////////////////////////
// CopyCurrentRegistryInformation
//
// Support routine to update all driver configuration information inside the
// registry.
//
void CopyCurrentRegistryInformation (PVOID HwDeviceExtension, LX_REGISTRY_SETTINGS *pSettings)
{
    unsigned long i;
    LX_REGISTRY_PAIR regKeys[] = {
        { L"Compression",        &pSettings->Compression         },
        { L"FBLinear",           &pSettings->FBLinear            },
        { L"DDCEnabled",         &pSettings->DDCEnabled          },
        { L"BIOSDetectOutput",   &pSettings->BIOSDetectOutput    },
        { L"PanelEnable",        &pSettings->PanelEnable         },
        { L"SimultaneousCRT",    &pSettings->SimultaneousCRT     },
        { L"PanelWidth",         &pSettings->PanelWidth          },
        { L"PanelHeight",        &pSettings->PanelHeight         },
        { L"XVGAPanel",          &pSettings->XVGAPanel           },
        { L"DDC_GPIOData",       &pSettings->DDC_GPIOData        },
        { L"DDC_GPIOClock",      &pSettings->DDC_GPIOClock       },
        { L"ManualTimingEnable", &pSettings->ManualTimingEnable  },
        { L"EnablePanelScale",   &pSettings->EnablePanelScale  },
        { L"EnableNon60HzPanel", &pSettings->EnableNon60HzPanel  },
        { L"RotationAngle",      &pSettings->RotationAngle       },
        { L"PanelInSystem",      &pSettings->PanelInSystem       },
    };

    *pSettings = g_regSettings;

#if DBG
    for (i = 0; i < (sizeof(regKeys) / sizeof(LX_REGISTRY_PAIR)); i++)
    {
        UnicodeToAnsi (g_ansiBuffer, (unsigned short *)regKeys[i].reg_string);
        DISPDBG((1000, "Registry -- %s - %i\n", g_ansiBuffer, *((unsigned long *)regKeys[i].data_element)));
    }
#endif
}

/////////////////////////////////////////////////////////////////////////
// SetRegistryInformation
//
// Support routine to update all driver configuration information inside the
// registry.
//
void SetRegistryInformation (PHW_DEVICE_EXTENSION HwDeviceExtension, LX_REGISTRY_SETTINGS *pSettings)
{
    unsigned long i;

    LX_REGISTRY_PAIR regKeys[] = {
        { L"Compression",           &pSettings->Compression         },
        { L"FBLinear",              &pSettings->FBLinear            },
        { L"DDCEnabled",            &pSettings->DDCEnabled          },
        { L"BIOSDetectOutput",      &pSettings->BIOSDetectOutput    },
        { L"PanelEnable",           &pSettings->PanelEnable         },
        { L"SimultaneousCRT",       &pSettings->SimultaneousCRT     },
        { L"PanelWidth",            &pSettings->PanelWidth          },
        { L"PanelHeight",           &pSettings->PanelHeight         },
        { L"XVGAPanel",             &pSettings->XVGAPanel           },
        { L"DDC_GPIOData",          &pSettings->DDC_GPIOData        },
        { L"DDC_GPIOClock",         &pSettings->DDC_GPIOClock       },
        { L"EnablePanelScale",      &pSettings->EnablePanelScale  },
        { L"EnableNon60HzPanel",    &pSettings->EnableNon60HzPanel  },
        { L"RotationAngle",         &pSettings->RotationAngle       },
    };

    for (i = 0; i < (sizeof(regKeys) / sizeof(LX_REGISTRY_PAIR)); i++)
    {
        VideoPortSetRegistryParameters (
            HwDeviceExtension,
            regKeys[i].reg_string,
            regKeys[i].data_element,
            sizeof(ULONG));

#if DBG
        UnicodeToAnsi (g_ansiBuffer, (unsigned short *)regKeys[i].reg_string);
        DISPDBG((1000, "Registry -- %s - %i\n", g_ansiBuffer, *((unsigned long *)regKeys[i].data_element)));
#endif
    }

    // Store the registry settings in the current global settings.  This allows
    // updates to take effect on the next mode set.  Some registry settings are
    // ignored until a reboot.
    //
    g_regSettings.Compression         = pSettings->Compression;
    g_regSettings.FBLinear            = pSettings->FBLinear;
    g_regSettings.DDCEnabled          = pSettings->DDCEnabled;
    g_regSettings.BIOSDetectOutput    = pSettings->BIOSDetectOutput;
    g_regSettings.PanelEnable         = pSettings->PanelEnable;
    g_regSettings.SimultaneousCRT     = pSettings->SimultaneousCRT;
    g_regSettings.PanelWidth          = pSettings->PanelWidth;
    g_regSettings.PanelHeight         = pSettings->PanelHeight;
    g_regSettings.XVGAPanel           = pSettings->XVGAPanel;
    g_regSettings.DDC_GPIOData        = pSettings->DDC_GPIOData;
    g_regSettings.DDC_GPIOClock       = pSettings->DDC_GPIOClock;
    g_regSettings.EnablePanelScale    = pSettings->EnablePanelScale;
    g_regSettings.EnableNon60HzPanel  = pSettings->EnableNon60HzPanel;
    g_regSettings.RotationAngle       = pSettings->RotationAngle;

    // Re-update reg settings to override panel settings with BIOS settings if
    // necessary.
    //
    UpdateRegistrySettings(HwDeviceExtension, 0);

    // Regenerate the HW mode table.  Settings like compression and linear pitch
    // affect the way that timings are generated.
    //
    if (HwDeviceExtension->pModes)
    {
        DISPDBG ((3000, "Deleting old mode list and regenerating a new one.\n"));
        LxFreeMemory (HwDeviceExtension, HwDeviceExtension->pModes);
        HwDeviceExtension->pModes = NULL;
    }
    if (!GenerateModeTable (HwDeviceExtension, 0))
    {
        DISPDBG ((3000, "Unable to allocate mode table!\n"));
    }
    if (!FilterUnsupportedModes (HwDeviceExtension, 0))
    {
        DISPDBG ((3000, "Mode Filter Failed!\n"));
    }
}

/////////////////////////////////////////////////////////////////////////
// UpdateRegistrySettings
// 
// Helper routine to calculate the current output setting when BIOSDetectOutput
// is set.  The fromResetSuspend variable is set when entering this routine from a reset
// or a resume, in which case we need to redetect hardware and possibly adjust settings.
//
void UpdateRegistrySettings (PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResetSuspend)
{
    // DETECT PANEL SETTINGS
    // If the user requested that the panel be detected from the BIOS, then we
    // will fill in the panel width, height and enable in the registry structure
    // with the BIOS values.  These will be used later by the display driver when
    // setting the display mode.
    //
    if (g_regSettings.BIOSDetectOutput && !g_regSettings.ManualTimingEnable &&
        !g_regSettings.CustomPanelTimings)
    {
        unsigned short vgConfig, vgPanel;
        static int firstBoot = 1;

        // Only program the settings at first boot.  This allows the
        // user to enable/disable the panel as they see fit during normal
        // operation.
        //
        if (firstBoot)
        {
            firstBoot = 0;
	        cim_outw (VR_INDEX, VR_UNLOCK);
	        cim_outw (VR_INDEX, (VRC_VG << 8) | VG_MEM_SIZE);
            vgConfig = cim_inw (VR_DATA);

            if (vgConfig & FP_DETECT_MASK)
            {
                // The BIOS is configured for a panel.  Now detect the resolution.
                //
                g_regSettings.PanelInSystem = 1;
                g_regSettings.PanelEnable = 1;
                
                if (vgConfig & FP_SIMULTANEOUS_CRT)
                    g_regSettings.SimultaneousCRT = 1;
                else
                    g_regSettings.SimultaneousCRT = 0;

                // Read the current panel size.  The panel dimensions are read
                // from the hactive and vactive virtual registers.  This is to
                // allow support for arbitrary sized panels in the future.
                //
                cim_outw (VR_INDEX, VR_UNLOCK);
		        cim_outw (VR_INDEX, (VRC_VG << 8) | VG_FP_HACTIVE);
                g_regSettings.PanelWidth = cim_inw (VR_DATA) + 1;

                cim_outw (VR_INDEX, VR_UNLOCK);
		        cim_outw (VR_INDEX, (VRC_VG << 8) | VG_FP_VACTIVE);
                g_regSettings.PanelHeight = cim_inw (VR_DATA) + 1;
            }
            else
            {
                g_regSettings.PanelInSystem = 0;
                g_regSettings.PanelEnable = 0;
            }
        }
        else if (!g_regSettings.PanelInSystem)
        {
            // Panel cannot be enabled when the BIOS does not detect a panel
            //
            g_regSettings.PanelEnable = 0;
        }
    }

    if (g_regSettings.CustomPanelTimings)
    {
        // Override the panel dimensions
        //
        g_regSettings.PanelWidth  = HwDeviceExtension->panelTimings->panel_width;
        g_regSettings.PanelHeight = HwDeviceExtension->panelTimings->panel_height;

        // Calculate the hz value for the custom panel.
        //
        g_regSettings.PanelRefresh = CalculateRefresh (HwDeviceExtension->panelTimings);
    }
    else
    {
        // Only supporting 60Hz in our internal tables.
        //
        g_regSettings.PanelRefresh = 60;
    }

    // Verify a valid rotation index
    //
    if (g_regSettings.RotationAngle != 0)
    {
        if (g_regSettings.RotationAngle != 90 && g_regSettings.RotationAngle != 180 && 
            g_regSettings.RotationAngle != 270)
        {
            DISPDBG ((3000, "Invalid Rotation Angle Passed in to driver: %i", (int)g_regSettings.RotationAngle));
            g_regSettings.RotationAngle = 0;
        }
    }

    // Look for a CRT if the user is requesting a CRT-only mode when running 
    // on a TFT-enabled system.  We use DDC to look for a connected monitor.
    // This code only executes on a resume or a fresh boot.
    //
    if (fromResetSuspend && g_regSettings.PanelInSystem && !g_regSettings.PanelEnable)
    {
        // Only perform a DDC operation if DDC is enabled by the user.
        //
        UCHAR Edid[256];
        BOOLEAN CRTAttached = LXDDCHelper (
            HwDeviceExtension, &Edid[0], sizeof(Edid));        

        if (!CRTAttached)
        {
            DISPDBG ((3000, "CRT Not found.  Resorting to TFT timings."));

            g_regSettings.PanelEnable = 1;
            g_regSettings.SimultaneousCRT = 0;
        }
    }

    // Force re-enumeration of child devices
    //
    if (HwDeviceExtension->panelWasEnabled != g_regSettings.PanelEnable)
    {
        DISPDBG ((3000, "Trying to re-enumerate children."));
        HwDeviceExtension->panelWasEnabled = g_regSettings.PanelEnable;
        VideoPortEnumerateChildren (HwDeviceExtension, NULL);
    }
}

/////////////////////////////////////////////////////////////////////////
// CalculateRefresh
//
// Calculates the refresh rate for any custom mode.
//
ULONG CalculateRefresh (VG_DISPLAY_MODE *pMode)
{
    ULONG hz, pixels;
    ULONG pll, vtotal;

    if (pMode->flags & VG_MODEFLAG_MANUAL_FREQUENCY)
    {
        pll  = pMode->frequency;
        pll  = (48 * (((pll >> 4) & 0xFF) + 1) * 100000);
        pll /= (((pMode->frequency >> 12) & 0x7) + 1);
        pll *= 10;
        pll /= ((pMode->frequency & 0xF) + 1);
        pll <<= 3;
    }
    else
    {
        pll = (pMode->frequency >> 13) * 1000000;
    }

    if (pMode->flags & VG_MODEFLAG_QVGA)
        pll >>= 2;

    // Calculate approximate refresh.  We use an approximated 29.3 fixed
    // point calculation with a round at the end.  Note that Windows
    // automatically assumes that modes less than or equal to 50Hz are
    // interlaced.  The only way to denote an interlaced mode on the
    // display mode settings dialog is to use the frames/second as the
    // refresh.
    //
    vtotal = pMode->vtotal;
    if (pMode->flags & VG_MODEFLAG_INTERLACED)
    {
        if (pMode->flags & VG_MODEFLAG_HALFCLOCK)
            vtotal <<= 2;
        else
            vtotal <<= 1;
    }
    pixels = pMode->htotal * vtotal;
    hz = pll / pixels;

    // Add .5 to round up
    //
    hz = (hz + 4) >> 3;

    return hz;
}

/////////////////////////////////////////////////////////////////////////
// LXDDCHelper
//
// Attempts to gather an Edid from an attached monitor.
//
BOOLEAN LXDDCHelper (
	HW_DEVICE_EXTENSION    *HwDeviceExtension,
	PVOID                   ddcBuffer,
	ULONG                   size)
{
    DDC_CONTROL ddcControl;
    unsigned long base;
    unsigned long dataHigh = DDC_DATA_HIGH;
    unsigned long clkHigh  = DDC_CLK_HIGH;
    BOOLEAN bGotEdid = FALSE;

    if (g_regSettings.DDCEnabled && !g_regSettings.PanelEnable)
    {
        // READ THE IO BASE FOR GPIOS
        //
        Q_WORD msr_value;
        msr_read64 (MSR_DEVICE_5535_MDD, 0xC, &msr_value);
        base = msr_value.low & 0xFF00;
        DISPDBG((5000, "DDC Base = 0x%X\n", base));

        // Enable bit for LBAR is bit 32.
        //
        if (base != 0 && (msr_value.high & 0x1))
        {
            // ENABLE CLOCK AND DATA INPUTS AND OUTPUTS
            //
            _asm {
                mov   edx, base
                add   dx,  DDC_OUT_ENABLE
                mov   eax, dataHigh
                or    eax, clkHigh
                out   dx,  eax
                mov   edx, base
                add   dx,  DDC_IN_ENABLE
                out   dx,  eax
            }

            HwDeviceExtension->DDCBase = base;
            HwDeviceExtension->DDCClockMode = DDC_MODE_UNDEFINED;
            HwDeviceExtension->DDCDataMode  = DDC_MODE_UNDEFINED;

            ddcControl.Size = sizeof(DDC_CONTROL);
            ddcControl.I2CCallbacks.WriteClockLine = LXI2CWriteClockLine;
            ddcControl.I2CCallbacks.WriteDataLine  = LXI2CWriteDataLine;
            ddcControl.I2CCallbacks.ReadClockLine  = LXI2CReadClockLine;
            ddcControl.I2CCallbacks.ReadDataLine   = LXI2CReadDataLine;
            ddcControl.EdidSegment = 0;

            bGotEdid = VideoPortDDCMonitorHelper (
                HwDeviceExtension,
                &ddcControl,
                ddcBuffer,
                size);

            DISPDBG((5000, "DDC Helper returned %s\n", bGotEdid ? "TRUE" : "FALSE"));
        }
    }

    return bGotEdid;
}

/////////////////////////////////////////////////////////////////////////
// Memory Allocation
// 
// Support routine to allocate a linear buffer of memory.  For XP, this
// routine uses the modern VideoPortAllocatePool, which allows tracking
// the memory with a tag.  For Win2K, an older deprecated routine is used.
//
void LxAllocateMemory (PVOID HwDeviceExtension, unsigned long size, PVOID *buf_ptr)
{
#if _WIN32_WINNT >= 0x0501
    // XP STYLE ALLOCATION
    // Allows tracking with a tag and is not marked as obsolete.
    //
    *buf_ptr = (PVOID)VideoPortAllocatePool (HwDeviceExtension, VpPagedPool, 
	    size, 0x20786744);
#else
    // WIN2K STYLE ALLOCATION
    // Simple buffer allocation.
    //
    VP_STATUS bufOk = VideoPortAllocateBuffer (HwDeviceExtension, 
        size, buf_ptr);

    if (bufOk != NO_ERROR)
        *buf_ptr = NULL;
#endif
}

/////////////////////////////////////////////////////////////////////////
// Memory Free
// 
// Complement routine to LxAllocateMemory
//
void LxFreeMemory (PVOID HwDeviceExtension, PVOID buf_ptr)
{
#if _WIN32_WINNT >= 0x0501
    VideoPortFreePool (HwDeviceExtension, buf_ptr);
#else
    VideoPortReleaseBuffer (HwDeviceExtension, buf_ptr);
#endif
}

/////////////////////////////////////////////////////////////////////////
// UnicodeToAnsi
//
// Support routine to translate a unicode string into 8-bit format.
//
void UnicodeToAnsi (char *output, unsigned short *input)
{
    int index = 0;

    while (input[index])
    {
        output[index] = (char)input[index];
        index++;
    }

    output[index] = 0;
}

#if DBG

typedef struct tagVideoString
{
    char  *string;
    ULONG  code;

} VIDEO_STRING;

VIDEO_STRING allStrings[] =
{
    { "IOCTL_VIDEO_ENABLE_VDM",                         IOCTL_VIDEO_ENABLE_VDM },
    { "IOCTL_VIDEO_DISABLE_VDM",                        IOCTL_VIDEO_DISABLE_VDM },
    { "IOCTL_VIDEO_REGISTER_VDM",                       IOCTL_VIDEO_REGISTER_VDM },
    { "IOCTL_VIDEO_SET_OUTPUT_DEVICE_POWER_STATE",      IOCTL_VIDEO_SET_OUTPUT_DEVICE_POWER_STATE },
    { "IOCTL_VIDEO_GET_OUTPUT_DEVICE_POWER_STATE",      IOCTL_VIDEO_GET_OUTPUT_DEVICE_POWER_STATE },
    { "IOCTL_VIDEO_MONITOR_DEVICE",                     IOCTL_VIDEO_MONITOR_DEVICE },
    { "IOCTL_VIDEO_ENUM_MONITOR_PDO",                   IOCTL_VIDEO_ENUM_MONITOR_PDO },
    { "IOCTL_VIDEO_INIT_WIN32K_CALLBACKS",              IOCTL_VIDEO_INIT_WIN32K_CALLBACKS },
    { "IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS",             IOCTL_VIDEO_HANDLE_VIDEOPARAMETERS },
    { "IOCTL_VIDEO_IS_VGA_DEVICE",                      IOCTL_VIDEO_IS_VGA_DEVICE },
    { "IOCTL_VIDEO_SAVE_HARDWARE_STATE",                IOCTL_VIDEO_SAVE_HARDWARE_STATE },
    { "IOCTL_VIDEO_RESTORE_HARDWARE_STATE",             IOCTL_VIDEO_RESTORE_HARDWARE_STATE },
    { "IOCTL_VIDEO_QUERY_AVAIL_MODES",                  IOCTL_VIDEO_QUERY_AVAIL_MODES     },
    { "IOCTL_VIDEO_QUERY_NUM_AVAIL_MODES",              IOCTL_VIDEO_QUERY_NUM_AVAIL_MODES     },
    { "IOCTL_VIDEO_QUERY_CURRENT_MODE",                 IOCTL_VIDEO_QUERY_CURRENT_MODE },
    { "IOCTL_VIDEO_SET_CURRENT_MODE",                   IOCTL_VIDEO_SET_CURRENT_MODE },
    { "IOCTL_VIDEO_RESET_DEVICE",                       IOCTL_VIDEO_RESET_DEVICE },
    { "IOCTL_VIDEO_LOAD_AND_SET_FONT",                  IOCTL_VIDEO_LOAD_AND_SET_FONT },
    { "IOCTL_VIDEO_SET_PALETTE_REGISTERS",              IOCTL_VIDEO_SET_PALETTE_REGISTERS },
    { "IOCTL_VIDEO_SET_COLOR_REGISTERS",                IOCTL_VIDEO_SET_COLOR_REGISTERS },
    { "IOCTL_VIDEO_ENABLE_CURSOR",                      IOCTL_VIDEO_ENABLE_CURSOR },
    { "IOCTL_VIDEO_DISABLE_CURSOR",                     IOCTL_VIDEO_DISABLE_CURSOR },
    { "IOCTL_VIDEO_SET_CURSOR_ATTR",                    IOCTL_VIDEO_SET_CURSOR_ATTR },
    { "IOCTL_VIDEO_QUERY_CURSOR_ATTR",                  IOCTL_VIDEO_QUERY_CURSOR_ATTR },
    { "IOCTL_VIDEO_SET_CURSOR_POSITION",                IOCTL_VIDEO_SET_CURSOR_POSITION },
    { "IOCTL_VIDEO_QUERY_CURSOR_POSITION",              IOCTL_VIDEO_QUERY_CURSOR_POSITION },
    { "IOCTL_VIDEO_ENABLE_POINTER",                     IOCTL_VIDEO_ENABLE_POINTER },
    { "IOCTL_VIDEO_DISABLE_POINTER",                    IOCTL_VIDEO_DISABLE_POINTER },
    { "IOCTL_VIDEO_SET_POINTER_ATTR",                   IOCTL_VIDEO_SET_POINTER_ATTR },
    { "IOCTL_VIDEO_QUERY_POINTER_ATTR",                 IOCTL_VIDEO_QUERY_POINTER_ATTR },
    { "IOCTL_VIDEO_SET_POINTER_POSITION",               IOCTL_VIDEO_SET_POINTER_POSITION },
    { "IOCTL_VIDEO_QUERY_POINTER_POSITION",             IOCTL_VIDEO_QUERY_POINTER_POSITION },
    { "IOCTL_VIDEO_QUERY_POINTER_CAPABILITIES",         IOCTL_VIDEO_QUERY_POINTER_CAPABILITIES },
    { "IOCTL_VIDEO_GET_BANK_SELECT_CODE",               IOCTL_VIDEO_GET_BANK_SELECT_CODE },
    { "IOCTL_VIDEO_MAP_VIDEO_MEMORY",                   IOCTL_VIDEO_MAP_VIDEO_MEMORY },
    { "IOCTL_VIDEO_UNMAP_VIDEO_MEMORY",                 IOCTL_VIDEO_UNMAP_VIDEO_MEMORY },
    { "IOCTL_VIDEO_QUERY_PUBLIC_ACCESS_RANGES",         IOCTL_VIDEO_QUERY_PUBLIC_ACCESS_RANGES },
    { "IOCTL_VIDEO_FREE_PUBLIC_ACCESS_RANGES",          IOCTL_VIDEO_FREE_PUBLIC_ACCESS_RANGES },
    { "IOCTL_VIDEO_QUERY_COLOR_CAPABILITIES",           IOCTL_VIDEO_QUERY_COLOR_CAPABILITIES },
    { "IOCTL_VIDEO_SET_POWER_MANAGEMENT",               IOCTL_VIDEO_SET_POWER_MANAGEMENT },
    { "IOCTL_VIDEO_GET_POWER_MANAGEMENT",               IOCTL_VIDEO_GET_POWER_MANAGEMENT },
    { "IOCTL_VIDEO_SHARE_VIDEO_MEMORY",                 IOCTL_VIDEO_SHARE_VIDEO_MEMORY },
    { "IOCTL_VIDEO_UNSHARE_VIDEO_MEMORY",               IOCTL_VIDEO_UNSHARE_VIDEO_MEMORY },
    { "IOCTL_VIDEO_SET_COLOR_LUT_DATA",                 IOCTL_VIDEO_SET_COLOR_LUT_DATA },
    { "IOCTL_VIDEO_GET_CHILD_STATE",                    IOCTL_VIDEO_GET_CHILD_STATE },
    { "IOCTL_VIDEO_VALIDATE_CHILD_STATE_CONFIGURATION", IOCTL_VIDEO_VALIDATE_CHILD_STATE_CONFIGURATION },
    { "IOCTL_VIDEO_SET_CHILD_STATE_CONFIGURATION",      IOCTL_VIDEO_SET_CHILD_STATE_CONFIGURATION },
    { "IOCTL_FSVIDEO_COPY_FRAME_BUFFER",                IOCTL_FSVIDEO_COPY_FRAME_BUFFER },
    { "IOCTL_FSVIDEO_WRITE_TO_FRAME_BUFFER",            IOCTL_FSVIDEO_WRITE_TO_FRAME_BUFFER },
    { "IOCTL_FSVIDEO_REVERSE_MOUSE_POINTER",            IOCTL_FSVIDEO_REVERSE_MOUSE_POINTER },
    { "IOCTL_FSVIDEO_SET_CURRENT_MODE",                 IOCTL_FSVIDEO_SET_CURRENT_MODE },
    { "IOCTL_FSVIDEO_SET_SCREEN_INFORMATION",           IOCTL_FSVIDEO_SET_SCREEN_INFORMATION },
    { "IOCTL_FSVIDEO_SET_CURSOR_POSITION",              IOCTL_FSVIDEO_SET_CURSOR_POSITION },
    { "UNKNOWN IOCTL",                                  0 },

};

#define NUM_IOCTLS (sizeof(allStrings)/sizeof(VIDEO_STRING))

char *FindIoctl (ULONG code)
{
    unsigned int i;

    for (i = 0; i < NUM_IOCTLS; i++)
    {
        if (allStrings[i].code == code)
            return allStrings[i].string;

    }

    return allStrings[NUM_IOCTLS - 1].string;
}
#endif

