//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Miniport routine definitions.
// </DOC_AMD_STD>
//

#include "dderror.h"
#include "devioctl.h"
#include "miniport.h"
#include "ntddvdeo.h"
#include "video.h"
#include "debug.h"
#include "lxshared.h"
#include "cim_parm.h"

#define LX_CHIP_TYPE	L"GeodeLX"
#define LX_DAC			L"Internal DAC"
#define LX_ADAPTER		L" Advanced Micro Devices"

// SOFTVG VIRTUAL REGISTER DEFINITIONS

#define VR_INDEX                0xAC1C
#define VR_DATA                 0xAC1E
#define VR_UNLOCK               0xFC53
#define VRC_VG                  0x0002
#define VG_MEM_SIZE             0x0000
#define VG_FP_TYPE              0x0002
#define VG_FP_HACTIVE           0x0011
#define VG_FP_VACTIVE           0x0017
#define GRAPHICS_CONTROL        0x2000
#define FP_DETECT_MASK          0x0100
#define TV_DETECT_MASK          0x0200
#define FP_SIMULTANEOUS_CRT     0x0400
#define FP_RESOLUTION_MASK      0x0038
#define FP_RES_640_480          0x0008
#define FP_RES_800_600          0x0010
#define FP_RES_1024_768         0x0018
#define FP_RES_1152_864         0x0020
#define FP_RES_1280_1024        0x0028
#define FP_RES_1600_1200        0x0030
#define TV_RESOLUTION_MASK      0x000E

// 5535/5536 DDC DEFINES

#define DDC_DATA_OUT            0x0000
#define DDC_DATA_IN             0x0030
#define DDC_CLOCK_OUT           0x0000
#define DDC_CLOCK_IN            0x0030
#define DDC_OUT_ENABLE          0x0004
#define DDC_IN_ENABLE           0x0020
#define DDC_CLK_BIT             (1 << g_regSettings.DDC_GPIOClock)
#define DDC_DATA_BIT            (1 << g_regSettings.DDC_GPIOData)
#define DDC_DATA_HIGH           DDC_DATA_BIT
#define DDC_DATA_LOW            (DDC_DATA_BIT << 16)
#define DDC_CLK_HIGH            DDC_CLK_BIT
#define DDC_CLK_LOW             (DDC_CLK_BIT << 16)

#define DDC_MODE_UNDEFINED      0
#define DDC_MODE_OUTPUT         1
#define DDC_MODE_INPUT          2

// STRIDE CALCULATOR
//
#define POWER_OF2_STRIDE(width) \
    (((width) > 4096) ? 8192 : (((width) > 2048) ? 4096 : (((width) > 1024) ? 2048 : 1024)))

// GLOBAL REGISTRY LOOKUP STRUCTURE
// The following structure facilitates reading all registry
// parameters.
//
typedef struct tagRegistryPair
{
    PWSTR reg_string;
    void *data_element;
} LX_REGISTRY_PAIR;

// GLOBAL DEVICE STRUCTURE
//
typedef struct _HW_DEVICE_EXTENSION
{
	// HARDWARE PHYSICAL ADDRESSES
	// Not all addresses will be filled in for all chips

	PHYSICAL_ADDRESS PhysLXRegAddr;
	PHYSICAL_ADDRESS PhysLXGPAddr;
	PHYSICAL_ADDRESS PhysLXVidAddr;
	PHYSICAL_ADDRESS PhysLXVIPAddr;
	PHYSICAL_ADDRESS PhysLXFBAddr;
	
	// REGION LENGTHS

	ULONG       FBSize;

	ULONG		BitsPerPixel;		// Current depth of the framebuffer
	ULONG		ScreenWidth;		// width of the screen
	ULONG		ScreenHeight;		// height of the screen
		
	ULONG       ModeFlags;
 	ULONG		VRefreshRate;		// vertical refresh rate
	ULONG       ScreenStride;
    ULONG       InterlacedOut;

    ULONG       DDCBase;
    int         DDCClockMode;
    int         DDCDataMode;

	// DPMS VARIABLES

	ULONG       CurrPowerState;     //the current power state for DPMS
	ULONG		CurrDPMSVersion;    // the current DPMS version
	ULONG		CurrLength;
	ULONG       PreviousPowerState; //the Pervious power state for DPMS
    ULONG       PreviousMonitorState;

	ULONG		savBitsPerPixel;	// Current depth of the framebuffer
	ULONG		savScreenWidth;		// width of the screen
	ULONG		savScreenHeight;	// height of the screen
	ULONG		savVRefreshRate;	// vertical refresh rate
		
    // Mode Information

    VIDEO_MODE_INFORMATION *pModes;
    ULONG                   modeCount;

    // Save/Restore structures

    GP_SAVE_RESTORE gpState;
    VG_SAVE_RESTORE vgState;
    DF_SAVE_RESTORE dfState;
    VIPSTATEBUFFER  vipState;

    // Interrupt Variables
    //
    PVOID       IRQCallback;
    PVOID       IRQData;
    ULONG       IRQFlags;
    ULONG       numAutoFlip;
    ULONG       interlaced;
    ULONG       autoFlipSurfaces[VIP_MAX_BUFFERS];
    ULONG       videoY;

    // VIP-less Solution?
    //
    BOOLEAN     vipHardware;

    // Custom Mode Information
    //
    GEODELX_CUSTOM_MODE *customModes;
    ULONG modeBufferSize;
    ULONG customModeCount;

    // Custom Panel Information
    //
    VG_DISPLAY_MODE *panelTimings;
    unsigned long showCRTModes;
    unsigned long SimultaneousCRT;
    unsigned long *extraTFTResolutions;
    unsigned long extraTFTResolutionCount;
    BOOLEAN addedExtraModeFromResume;
    unsigned long panelWasEnabled;

} HW_DEVICE_EXTENSION, *PHW_DEVICE_EXTENSION;

#define READ_REG_KEY(string,ptr)            \
{                                           \
    *(ptr) = 4;                             \
    if (VideoPortGetRegistryParameters (    \
        HwDeviceExtension,                  \
        string,                             \
        FALSE,                              \
        GeodeRegistryCallback,              \
        ptr) != NO_ERROR)                   \
        {                                   \
            *(ptr) = 0;                     \
        }                                   \
}

typedef LONG VP_STATUS;

#define SIZE_NT4_VIDEO_PORT_CONFIG_INFO           0x42
#define SIZE_NT4_VIDEO_HW_INITIALIZATION_DATA     0x28

// GLOBAL DRIVER ROUTINES
//

// DRIVER ROUTINE DEFINITIONS

VP_STATUS LXFindAdapter(HW_DEVICE_EXTENSION *HwDeviceExtension,
	VOID *HwContext, PWSTR ArgumentString,
	PVIDEO_PORT_CONFIG_INFO ConfigInfo,
	UCHAR *Again);

BOOLEAN LXInitialize(HW_DEVICE_EXTENSION *HwDeviceExtension);

BOOLEAN LXStartIO(HW_DEVICE_EXTENSION *HwDeviceExtension,
	VIDEO_REQUEST_PACKET *RequestPacket);

void LXClaimResources (PHW_DEVICE_EXTENSION HwDeviceExtension);

BOOLEAN GenerateModeTable (PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResume);
BOOLEAN FilterUnsupportedModes(PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResume);

void LXSetPalette(HW_DEVICE_EXTENSION *HwDeviceExtension,ULONG *pPal, ULONG StartIndex, ULONG Count);
int  LXMapPhysicalAddress (PVOID HwDeviceExtension, ULONG address, ULONG size, UCHAR IO, PULONG ptr_out);
int RetrieveRegistryInformation (PVOID HwDeviceExtension, LX_REGISTRY_SETTINGS *pSettings);
void SetRegistryInformation (PHW_DEVICE_EXTENSION HwDeviceExtension, LX_REGISTRY_SETTINGS *pSettings);
void CopyCurrentRegistryInformation (PVOID HwDeviceExtension, LX_REGISTRY_SETTINGS *settings);
void UpdateRegistrySettings (PHW_DEVICE_EXTENSION HwDeviceExtension, int fromResetSuspend);
void LxAllocateMemory (PVOID HwDeviceExtension, unsigned long size, PVOID *buf_ptr);
void LxFreeMemory (PVOID HwDeviceExtension, PVOID buf_ptr);
void UnicodeToAnsi (char *output, unsigned short *input);
VP_STATUS AddMappedAddressToArray (PHYSICAL_ADDRESS address, ULONG length, ULONG IO,
								   PVOID array, PVOID HwDeviceExtension);
void LXSaveHardwareState (PHW_DEVICE_EXTENSION  HwDeviceExtension);
void LXRestoreHardwareState (PHW_DEVICE_EXTENSION  HwDeviceExtension);
BOOLEAN LXDDCHelper (HW_DEVICE_EXTENSION *HwDeviceExtension,
    PVOID ddcBuffer, ULONG size);
ULONG CalculateRefresh (VG_DISPLAY_MODE *pMode);

// WINXP DPMS HELPER ROUTINES

VP_STATUS LXGetPowerState(HW_DEVICE_EXTENSION *HwDeviceExtension, ULONG  HwId,
						  PVIDEO_POWER_MANAGEMENT  VideoPowerControl);

VP_STATUS LXSetPowerState(HW_DEVICE_EXTENSION *HwDeviceExtension, ULONG  HwId,
						  PVIDEO_POWER_MANAGEMENT  VideoPowerControl);

ULONG LXQueryInterface (
    PHW_DEVICE_EXTENSION    HwDeviceExtension,
    PQUERY_INTERFACE        pQueryInterface);

BOOLEAN LXInterrupt (PHW_DEVICE_EXTENSION HwDeviceExtension);

ULONG LXGetChildDescriptor(
    HW_DEVICE_EXTENSION *HwDeviceExtension, PVIDEO_CHILD_ENUM_INFO ChildEnumInfo,
    PVIDEO_CHILD_TYPE pChildType, PVOID pChildDescriptor,
    PULONG pUId, PULONG pUnused);

VP_STATUS LXSetDisplayMode(HW_DEVICE_EXTENSION *HwDeviceExtension,
		int display_mode);
VP_STATUS LXSetMode3(HW_DEVICE_EXTENSION *HwDeviceExtension);

#if DBG
char *FindIoctl (ULONG code);
#endif

