//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Miniport support routines.
// </DOC_AMD_STD>
//

#include "dderror.h"
#include "devioctl.h"
#include "miniport.h"
#include "ntddvdeo.h"
#include "video.h"
#include "lxdrv.h"
#include "debug.h"
#include "cim_rtns.h"
#include "cim_regs.h"

extern void cim_outb (unsigned short port, unsigned char data);
extern unsigned char cim_inb (unsigned short port);

extern LX_REGISTRY_SETTINGS g_regSettings;
extern ULONG lxCustomIndex;

/////////////////////////////////////////////////////////////////////////
// LXSetDisplayMode
//
// Set the display mode requested by the display driver. There is a handshake
// on startup where the display driver requests from the miniport a list
// of all supported modes. The display driver picks the one it likes the
// best from the list and requests the miniort to establish that mode.  Note that
// in the GeodeLX driver, the behavior has been modified such that the mode is set in
// the display driver.  To accomplish this, a mode set simply records the mode
// information.  It is then passed to the display driver via a call to
// IOCTL_VIDEO_QUERY_CURRENT_MODE.  The display driver then sets the mode using
// Cimarron.  This is done to keep as much of the hardware abstraction as possible
// in the display driver.  With that said, this routine is only called if something
// other than the display driver uses IOCTL_VIDEO_SET_CURRENT_MODE.
//
// display_mode: index of the chosen mode.
//

VP_STATUS LXSetDisplayMode(
	HW_DEVICE_EXTENSION *HwDeviceExtension,
	int display_mode)
{
    DF_VIDEO_CURSOR_PARAMS cursorKey;

    // SET THE CUSTOM MODE ON THE HARDWARE
    //
    if (g_regSettings.ManualTimingEnable)
    {
        vg_set_custom_mode (&HwDeviceExtension->customModes[lxCustomIndex].manual_timings,
            HwDeviceExtension->BitsPerPixel);
    }

	// SET THE MODE
	
	else if (g_regSettings.PanelEnable)
	{
		DISPDBG((2000, "Panel %d %d \n",
			g_regSettings.PanelWidth,
			g_regSettings.PanelHeight));
			
        vg_set_panel_mode (
            HwDeviceExtension->ScreenWidth,
			HwDeviceExtension->ScreenHeight,
            HwDeviceExtension->ScreenWidth,
			HwDeviceExtension->ScreenHeight,
			g_regSettings.PanelWidth,
			g_regSettings.PanelHeight,			
			HwDeviceExtension->BitsPerPixel,
            0);

		// SET THE PANEL BACKGROUND COLOR
		// When the display mode is smaller than the panel resolution
		// panel centering is enabled.  This effectively creates a
		// solid border around the display mode.  We initialize this
		// color to be black.

		vg_set_border_color (0);
	}
	else
	{	
        vg_set_display_mode (
            HwDeviceExtension->ScreenWidth,
            HwDeviceExtension->ScreenHeight,
            HwDeviceExtension->ScreenWidth,
            HwDeviceExtension->ScreenHeight,
            HwDeviceExtension->BitsPerPixel,
            HwDeviceExtension->VRefreshRate, 0);
	}

    // SET THE DISPLAY PITCH
    //
    vg_set_display_pitch (HwDeviceExtension->ScreenStride);

    // SET THE BPP

    gp_set_bpp (HwDeviceExtension->BitsPerPixel);

	// SET CURSOR COLOR KEY

    cursorKey.key           = 0xFCFEFD;
    cursorKey.mask          = 0xFFFFFF;
    cursorKey.flags         = 0;
    cursorKey.color1        = 0xFCFEFD;
    cursorKey.color2        = 0xFCFEFD;
    cursorKey.select_color2 = 0;
	df_configure_video_cursor_color_key (&cursorKey);
    df_set_video_cursor_color_key_enable (1);
	
    if(g_regSettings.Compression)
    {
        VG_COMPRESSION_DATA compData;

		// CONFIGURE COMPRESSION PARAMETERS
		// Compression will only be enabled once all bitmaps have been deleted
		// from the previous mode to prevent corruption.
		
        compData.compression_offset = vg_get_display_pitch() * HwDeviceExtension->ScreenHeight;
        compData.pitch              = 544;
        compData.size               = 544;
		compData.flags              = 0;
        vg_configure_compression (&compData);
	}

	return(1);
}

/////////////////////////////////////////////////////////////////////////
// LXSetMode3
//
// Set the display mode to mode 3 (VGA 80-column text).
//
VP_STATUS LXSetMode3 (HW_DEVICE_EXTENSION *HwDeviceExtension)
{
	VP_STATUS status;
	ULONG i;
	VIDEO_X86_BIOS_ARGUMENTS biosArguments;

	DISPDBG((1, "LXSetMode3\n"));

	VideoPortZeroMemory (&biosArguments, sizeof(biosArguments));

	// SET MODE 3 (AH = 00h, AL = 03h)
	
	biosArguments.Eax = 0x0003;
	biosArguments.Ebx = 0;
	status = VideoPortInt10(HwDeviceExtension, &biosArguments);

	DISPDBG((1, "Status from set mode %x\n", status));
	return(status);
}

